unit Wincgi;

interface

uses
 SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
 Forms, Dialogs, IniFiles;

type
 TWinCGI = class(TComponent)
 private
 { Private declarations }
 CgiIniFile : TIniFile;
 CgiIniFileName : String;
 OutputFile : TextFile;
 OutputFileName : String;
 WebServer : String;

 procedure SetIniFile( testini : String );
 function GetIniFile: String;
 procedure SetOutputFile ( outfile : String);
 function GetOutputFile: String;
 procedure SetWebServer ( web : String );
 function GetWebServer: String;

 function OpenOutputFile: Boolean;
 function GetGMTTime : String;

 protected
 { Protected declarations }
 public
 { Public declarations }
 REQUEST_METHOD : String;
 SERVER_NAME : String;
 REMOTE_ADDR : String;

 constructor Create(AOwner: TComponent); override;
 function InitWinCGI: Boolean;
 function CloseWinCGI: Boolean;
 function GetLiteral( keyname : String ) : String;
 function Send( dataout : String ) : Boolean;
 function SendHTTPHeader: Boolean;

 published
 { Published declarations }
 property OutputDataFile : String read GetOutputFile write SetOutputFile;
 property CGIDataIniFile : String read GetIniFile write SetIniFile;
 property WebServerName : String read GetWebServer write SetWebServer;
 end;

procedure Register;

implementation

constructor TWinCGI.Create(AOwner: TComponent);
begin
 inherited Create(AOwner);
 WebServer := 'Website 1.X';
end;

function TWinCGI.InitWinCGI : Boolean;
begin
 try
 CgiIniFile := TIniFile.Create(ParamStr(1));
 application.processMessages;

 OutputFileName := CgiIniFile.ReadString('System', 'Output File', '');

 {Retrieve the three data items that you want for this examples:
 request method, server name, and the user's IP address/remote address}

 REQUEST_METHOD := CgiIniFile.ReadString('CGI', 'Request Method', '');
 SERVER_NAME := CgiIniFile.ReadString('CGI', 'Server Name', '');
 REMOTE_ADDR := CgiIniFile.ReadString('CGI', 'Remote Address', '');

 OpenOutputFile;
 Result := True;
 except
 Result := False;
 end;
end;

function TWinCGI.CloseWinCGI : Boolean;
begin
try
 CgiIniFile.Free;
 CloseFile(OutputFile);
 Result := True;
except
 Result := False;
end;
end;

function TWinCGI.GetLiteral( keyname : String) : String;
begin
 try
 Result := CgiIniFile.ReadString('Form Literal', keyname, '');
 except
 Result := 'Unknown';
 end;
end;

procedure TWinCGI.SetOutputFile ( outfile : String);
begin
 OutputFileName := outfile;
end;

function TWinCGI.GetOutputFile : String;
begin
 Result := OutputFileName;
end;

procedure TWinCGI.SetIniFile ( testini : String);
begin
 CGIIniFileName := testini;
end;

function TWinCGI.GetIniFile : String;
begin
 Result := CGIIniFileName;
end;

procedure TWinCGI.SetWebServer ( web : String);
begin
 WebServer := web;
end;

function TWinCGI.GetWebServer : String;
begin
 Result := WebServer;
end;

function TWinCGI.GetGMTTime : String;
begin
 Result := FormatDateTime('ddd, dd mmm yyyy hh:nn:ss "GMT"', Now);
end;

function TWinCGI.OpenOutputFile : Boolean;
begin
 if OutputFileName <> '' then
 begin
 try
 AssignFile( OutputFile, OutputFileName);
 ReWrite(OutputFile);
 Result := True;
 except
 Result := False
 end
 end
 else
 Result := False;
end;

function TWinCGI.Send( dataout : String ) : Boolean;
begin
 try
 Writeln(OutputFile, dataout);
 Result := True;
 except
 Result := False;
 end;
end;

function TWinCGI.SendHTTPHeader : Boolean;
begin
try
 Writeln(OutputFile, 'HTTP/1.0 200 OK');
 Writeln(OutputFile, 'Date: ' + GetGMTTime); 
    {Thursday, 12-Jan-96 16:04:30 GMT');}
 Writeln(OutputFile, 'Server: ' + WebServer);
 Writeln(OutputFile, 'Content-type: text/html');
 Writeln(OutputFile, '');
 Result := True;
except
 Result := False;
end;
end;

procedure Register;
begin
 RegisterComponents('WinCGI', [TWinCGI]);
end;

end.
 
