unit Tftpc;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, TSocketc, NWinSock, StdCtrls, FileCtrl;

{ If Using the Delphi 32 Winsock.pas }
type
     _Socket = TSocket;
     fd_set = TFDSet;
     Pfd_set = PFDSet;
     timeval = TTimeVal;
     hostent = THostEnt;
     netent = TNetEnt;
     servent = TServEnt;
     protoent = TProtoEnt;
     in_addr = TInAddr;
     Pin_addr = PInAddr;
     Sockaddr_in = TSockAddrIn;
     PSockaddr_in = PSockAddrIn;
     WSADATA = TWSAData;
     LPWSADATA = PWSAData;
     sockaddr = TSockAddr;
     sockproto = TSockProto;
     linger = TLinger;
     LPSOCKADDR = ^sockaddr;
     LPSOCKADDR_IN = ^sockaddr_in;
     LPLINGER = ^linger;
     LPIN_ADDR = ^in_addr;
     LPFD_SET = ^fd_set;
     LPHOSTENT = ^hostent;
     LPSERVENT = ^servent;
     LPPROTOENT = ^protoent;
     LPTIMEVAL = ^timeval;
const
     FIONREAD = (IOC_OUT or ((4 and IOCPARM_MASK) SHL 16) or
           (102 SHL 8) or 127);
{ End of Delphi 32 Winsock.pas compatability type definitions }
const
  SKT_NOOP = 0;
  SKT_GETHOSTADDR = 1;
  SKT_CONNECTTOHOST = 2;
  SKT_SENDPACKET = 3;
  SKT_RECVPACKET = 4;
{Constant declarations to represent FTP commands. These commands
keep track of the command last issued, so that the server's responses
can be directed to the appropriate procedure}
  FTP_NOOP = 0;
  FTP_CONNECT = 1;
  FTP_USER = 2;
  FTP_PASS = 3;
  FTP_CWD = 4;
  FTP_CDUP = 5;
  FTP_REIN = 6;
  FTP_QUIT = 7;
  FTP_PORT = 8;
  FTP_PASV = 9;
  FTP_TYPE = 10;
  FTP_STRU = 11;
  FTP_MODE = 12;
  FTP_RETR = 13;
  FTP_STOR = 14;
  FTP_STOU = 15;
  FTP_APPE = 16;
  FTP_REST = 17;
  FTP_RNFR = 18;
  FTP_RNTO = 19;
  FTP_ABOR = 20;
  FTP_DELE = 21;
  FTP_RMD = 22;
  FTP_MKD = 23;
  FTP_PWD = 24;
  FTP_LIST = 25;
  FTP_NLST = 26;
  FTPD_NOOP = 0;
  FTPD_WAITFORCONN = 1;
  FTPD_SEND = 2;
  FTPD_RECV = 3;
  FTPD_RECVDISP = 4;
{User-defined event declarations, for socket events. The first
is used for the primary socket, the second for the data socket,
and the third for address lookup notifications}
  UWM_SOCKETEVENT = WM_USER + 500;
  UWM_DATASOCKETEVENT = WM_USER + 501;
  UWM_ADDRSOCKETEVENT = WM_USER + 502;
type
  TFtp = class(TForm)
    DirectoryListBox1: TDirectoryListBox;
    FileListBox1: TFileListBox;
    Label1: TLabel;
    stLocalPath: TLabel;
    DriveComboBox1: TDriveComboBox;
    Label2: TLabel;
    sleHostName: TEdit;
    cbConnect: TButton;
    lbHostFiles: TListBox;
    Label3: TLabel;
    stHostPath: TLabel;
    Label4: TLabel;
    stStatus: TLabel;
    cbSend: TButton;
    cbGet: TButton;
    cbQuit: TButton;
    Label5: TLabel;
    sleLogin: TEdit;
    Label6: TLabel;
    slePassword: TEdit;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure ClearBuffer(DirtyBuffer: Pchar; BufLength: Integer);
    function CreateDataSocket: boolean;
    procedure OpenReceiveFile;
    procedure OpenSendFile;
    procedure OpenDataConnection(ConnectionAddr: String);
    procedure CloseDataConnection;
    procedure ProcessDataSend;
    procedure AcceptDataConnection;
    procedure ProcessDataReceive;
    procedure ProcessDataConnect;
    procedure ProcessHostLookup;
    procedure ProcessHostConnect;
    procedure ProcessHostWelcome;
    procedure ProcessHostSend;
    procedure ProcessHostAbort;
    procedure ProcessHostNOOP;
    procedure ProcessHostQuit;
    procedure ProcessHostLogin;
    procedure ProcessHostPassword;
    procedure ProcessHostCWD;
    procedure ProcessHostPWD;
    procedure ProcessHostType;
    procedure ProcessHostPort;
    procedure ProcessHostPasv;
    procedure ProcessHostList;
    procedure ProcessHostNList;
    procedure ProcessHostRetr;
    procedure ProcessHostStor;
    procedure ProcessHostReceive;
    procedure DirectoryListBox1Change(Sender: TObject);
    procedure DriveComboBox1Change(Sender: TObject);
    procedure FileListBox1Change(Sender: TObject);
    procedure cbQuitClick(Sender: TObject);
    procedure cbConnectClick(Sender: TObject);
    procedure lbHostFilesDblClick(Sender: TObject);
    procedure FileListBox1DblClick(Sender: TObject);
  private
    { Private declarations }
    {The command and data sockets}
    isConnectionSocket: _SOCKET;
    isDataSocket: _SOCKET;
    {A Boolean flag that indicates whether the connection has been
    established}
    ibConnected: Boolean;
    {The command status flag, used to keep track of the last command
    so that responses can be routed to the correct processing
    procedure}
    iiFtpStatus: Integer;
    {The next variable is a command status flag that indicates which
    command the application is going to be issuing. Use this flag
    when you have to send some commands to the server to prepare for
    the command that this variable holds.}
    iiFtpWorkingToward: Integer;
    {The data connection status flag}
    iiDataConnStatus: Integer;
    {The data type of the data connection}
    isDataConnType: Char;
    {A flag to indicate whether the transfer file is open 
    (for either sending or receiving}
    ibFileOpen: boolean;
    {The name of the file being sent or received}
    isFileName: String;
    {The amount of data in the data transfer buffer}
    iiBufferAmt: Integer;
    {The socket event handlers}
    procedure UWMSocketEvent(var Msg: TMessage); message UWM_SOCKETEVENT;
    procedure UWMDataSocketEvent(var Msg: TMessage); message UWM_DATASOCKETEVENT;
    procedure UWMAddrSocketEvent(var Msg: TMessage); message UWM_ADDRSOCKETEVENT;
  public
    { Public declarations }
    {The TSocket object}
    NetSocket: _TSocket;
    {Your receive and transfer buffers}
    ReceiveBuffer: array[0..8192] of Char;
    TempBuffer: array[0..8192] of Char;
    TransferBuffer: array[0..8192] of Char;
    {The file being transferred}
    TransferFile: File of Byte;
    {The Internet address and port of the server's data socket}
    DataAddress: String;
  end;

var
  Ftp: TFtp;

implementation

{$R *.DFM}

procedure TFtp.FormCreate(Sender: TObject);
begin
   {Initialize the appropriate flags}
   ibConnected := FALSE;
   iiFtpStatus := FTP_NOOP;
   iiDataConnStatus := FTPD_NOOP;
   iiFtpWorkingToward := FTP_NOOP;
   isDataConnType := 'I';
   ibFileOpen := FALSE;
   {Create the _TSocket class}
   NetSocket := _TSocket.Create;
   {Startup the WinSock services}
   if not NetSocket.InitializeSocket then
      Close;
end;

procedure TFtp.FormDestroy(Sender: TObject);
begin
     {Clean up all WinSock services and interface}
     NetSocket.ShutdownSocket;
     NetSocket.Free;
end;

procedure TFtp.ClearBuffer(DirtyBuffer: Pchar; BufLength: Integer);
var
   liCurPos: Integer;
   lcNullChar: Char;
begin
   {Initialize the position pointer to 0}
   liCurPos := 0;
   {Create a null character}
   lcNullChar := Char(0);
   {Loop through the entire buffer passed in, setting each character
   equal to the null character}
   while (liCurPos < BufLength) do
   begin
      DirtyBuffer^ := lcNullChar;
      Inc(liCurPos);
      Inc(DirtyBuffer);
   end;
end;

function TFtp.CreateDataSocket: boolean;
begin
  {Create the data socket}
  isDataSocket := NetSocket.CreateDataSocket(isConnectionSocket, handle, UWM_DATASOCKETEVENT);
  {Did you get a valid socket?}
  if (isDataSocket = INVALID_SOCKET) then
     CreateDataSocket := FALSE
  else
  begin
     {Yes, set the data socket status to show that it is waiting for a 
     connection, then return TRUE}
     iiDataConnStatus := FTPD_WAITFORCONN;
     CreateDataSocket := TRUE;
  end;
end;

procedure TFtp.OpenReceiveFile;
var
   lsReceiveFileName: String;
   lsLocalDir: String;
   lsHostFile: String;
begin
   {Build the name for the incoming file}
   lsLocalDir := DirectoryListBox1.Directory;
   lsHostFile := isFileName;
   lsReceiveFileName := lsLocalDir + '\' + lsHostFile;
   ibFileOpen := TRUE;
   {Open the file for writing}
   AssignFile(TransferFile, lsReceiveFileName);
   Rewrite(TransferFile);
end;

procedure TFtp.OpenSendFile;
begin
   {Open the specified file for reading}
   ibFileOpen := TRUE;
   AssignFile(TransferFile, stLocalPath.Caption);
   Reset(TransferFile);
end;

procedure TFtp.OpenDataConnection(ConnectionAddr: String);
var
   liPort: Integer;
   lsAddress: String;
   lsPortHigh, lsPortLow: String;
   liPosComma: Integer;
begin
   {Convert the server data socket address into a usable address}
   {First, convert the first four numbers into the standard 
    ###.###.###.### notation}
   liPosComma := Pos(',', ConnectionAddr);
   if (liPosComma > 0) then
   begin
      lsAddress := Copy(ConnectionAddr, 1, (liPosComma - 1)) + '.';
      ConnectionAddr := Copy(ConnectionAddr, (liPosComma + 1), Length(ConnectionAddr));
   end;
   liPosComma := Pos(',', ConnectionAddr);
   if (liPosComma > 0) then
   begin
      lsAddress := lsAddress + Copy(ConnectionAddr, 1, (liPosComma - 1)) + '.';
      ConnectionAddr := Copy(ConnectionAddr, (liPosComma + 1), Length(ConnectionAddr));
   end;
   liPosComma := Pos(',', ConnectionAddr);
   if (liPosComma > 0) then
   begin
      lsAddress := lsAddress + Copy(ConnectionAddr, 1, (liPosComma - 1)) + '.';
      ConnectionAddr := Copy(ConnectionAddr, (liPosComma + 1), Length(ConnectionAddr));
   end;
   liPosComma := Pos(',', ConnectionAddr);
   if (liPosComma > 0) then
   begin
      lsAddress := lsAddress + Copy(ConnectionAddr, 1, (liPosComma - 1));
      ConnectionAddr := Copy(ConnectionAddr, (liPosComma + 1), Length(ConnectionAddr));
   end;
   {Now take the last two numbers and combine them into the port number}
   liPosComma := Pos(',', ConnectionAddr);
   if (liPosComma > 0) then
   begin
      {The first number is the high byte, the second is the low}
      lsPortHigh := Copy(ConnectionAddr, 1, (liPosComma - 1));
      lsPortLow := Copy(ConnectionAddr, (liPosComma + 1), Length(ConnectionAddr));
      {Combine the two, shifting the high byte to the left 8 bits}
      liPort := (StrToInt(lsPortHigh) shl 8) + StrToInt(lsPortLow);
   end;
   {Open the data socket connection, using the address and port
    that you just extracted from the message that the server returned}
   NetSocket.OpenDataSocketConnection(handle, UWM_DATASOCKETEVENT, isDataSocket,
                      lsAddress, liPort);
end;

procedure TFtp.CloseDataConnection;
begin
  {Shut down the data socket}
  shutdown(isDataSocket, 2);
  {Close the data socket}
  NetSocket.CloseSocketConnection(isDataSocket);
  {If you are working with an open file, close it}
  if ((iiDataConnStatus = FTPD_SEND) or
     (iiDataConnStatus = FTPD_RECV)) then
  begin
     if ibFileOpen then
     begin
        CloseFile(TransferFile);
        ibFileOpen := FALSE;
     end;
  end;
end;

procedure TFtp.ProcessDataSend;
var
   liCurByteCount: Integer;
   lcCurByte: Byte;
   lpTransferDataPointer: PChar;
begin
   {Is the data file open?}
   if ibFileOpen then
   begin
      {Initialize the counter to keep track of how much you have sent}
      liCurByteCount := 0;
      {Are you at the end of the file?}
      if (not Eof(TransferFile)) then
      begin
         {No, clear the data transfer buffer}
         ClearBuffer(TransferBuffer, 8192);
         {Read the next 4,096 bytes from the file into the transfer 
         buffer}
         while ((liCurByteCount < 4096) and (not Eof(TransferFile))) do
         begin
            Read(TransferFile, lcCurByte);
            TransferBuffer[liCurByteCount] := Char(lcCurByte);
            Inc(liCurByteCount);
         end;
         {Send to the server the data transfer buffer's contents}
         NetSocket.SocketSendBuffer(isDataSocket, handle, UWM_DATASOCKETEVENT,
                  TransferBuffer, liCurByteCount, SKT_SENDPACKET);
      end
      else
        {You are at the end of the file, so close the data connection}
         CloseDataConnection;
   end;
end;

procedure TFtp.AcceptDataConnection;
begin
  isDataSocket := NetSocket.AcceptDataSocket(isDataSocket, handle, UWM_DATASOCKETEVENT);
  case iiFtpStatus of
     FTP_LIST:
     begin
        iiDataConnStatus := FTPD_RECVDISP;
        ClearBuffer(TransferBuffer, 8192);

     end;
     FTP_NLST:
     begin
        iiDataConnStatus := FTPD_RECVDISP;
        ClearBuffer(TransferBuffer, 8192);

     end;
     FTP_RETR:
     begin
        iiDataConnStatus := FTPD_RECV;
        OpenReceiveFile;
        ClearBuffer(TransferBuffer, 8192);

     end;
     FTP_STOR:
     begin
        iiDataConnStatus := FTPD_SEND;
        OpenSendFile;
        ProcessDataSend;
     end;
     FTP_STOU:
     begin
        iiDataConnStatus := FTPD_SEND;
        OpenSendFile;
        ProcessDataSend;
     end;
  end;
end;

procedure TFtp.ProcessDataReceive;
var
   lsTempStr: String;
   liTempBytes: Integer;
   liRecvBytes: Integer;
begin
   {Is your data connection in ASCII mode?}
   if (isDataConnType = 'A') then
   begin
      {Yes, you are probably going to be sending this data to the display
      so that the user can view it}
      {Did you receive any data?}
      liRecvBytes := recv(isDataSocket, TransferBuffer, 8192, 0);
      TransferBuffer[liRecvBytes] := Char(0);
      if (TransferBuffer[0] <> Char(0)) then
      begin
         {Until you reach the end of the received data, extract each line
         from the received data}
         while (NetSocket.GetLine(TempBuffer, TransferBuffer)) do
         begin
            {Convert the line into a Pascal string and add it to the
            list box so that the user can view it}
            lsTempStr := Strpas(TempBuffer);
            if (iiDataConnStatus = FTPD_RECVDISP) then
            begin
               lbHostFiles.Items.Add(lsTempStr);
               lbHostFiles.Update;
            end;
         end;
         {Clear the data transfer buffer}
         ClearBuffer(TransferBuffer, 8192);
      end
      else
      begin
         {You didn't receive anything, so close the connection and send
         the application into idle mode}
         CloseDataConnection;
         iiDataConnStatus := FTPD_NOOP;
         {Send an ABOR command to abort the transfer}
         iiFtpStatus := FTP_ABOR;
         NetSocket.SocketSend(isConnectionSocket, handle, UWM_SOCKETEVENT, 'ABOR', SKT_SENDPACKET);
         {Clear the command transfer buffer}
         ClearBuffer(ReceiveBuffer, 8192);
      end;
   end
   else
   begin
      {You must have your data connection in IMAGE mode.
       Is is the transfer file open?}
      if ibFileOpen then
      begin
         {Set the data socket into synchronous mode}
         WSAAsyncSelect(isDataSocket, handle, 0, 0);
         {Receive the first set of data}
         liRecvBytes := recv(isDataSocket, TransferBuffer, 8192, 0);
         {Did you receive anything?}
         if (liRecvBytes >= 0) then
         begin
            {As long as you continue to receive data, loop}
            while (liRecvBytes > 0) do
            begin
               {Write what you have received to the file}
               for liTempBytes := 0 to (liRecvBytes - 1) do
               begin
                  Write(TransferFile, Byte(TransferBuffer[liTempBytes]));
               end;
               {Clear the data transfer buffer}
               ClearBuffer(TransferBuffer, 8192);
               {Receive the next set of data}
               liRecvBytes := recv(isDataSocket, TransferBuffer, 8192, 0);
            end;
         end;
         {Close the data socket connection}
         CloseDataConnection;
         iiDataConnStatus := FTPD_NOOP;
      end;
   end;
end;

procedure TFtp.ProcessDataConnect;
var
   lsReplyMsg, lsRetnCode: String;
   lbProceed: boolean;
   liAddrStart, liAddrEnd: Integer;
begin
   {What command do you want to send?}
   case iiFtpWorkingToward of
      {You want to send a LIST command}
      FTP_LIST:
      begin
         {Set the command status flag to LIST and send the LIST command
         to the server}
         iiFtpStatus := FTP_LIST;
         NetSocket.SocketSend(isConnectionSocket, handle, UWM_SOCKETEVENT, 'LIST', SKT_SENDPACKET);
      end;
      {You want to send an NLST command}
      FTP_NLST:
      begin
         {Set the command status flag to NLST and send the NLST command
         to the server}
         iiFtpStatus := FTP_NLST;
         NetSocket.SocketSend(isConnectionSocket, handle, UWM_SOCKETEVENT, 'NLST', SKT_SENDPACKET);
      end;
      {You want to send an RETR command}
      FTP_RETR:
      begin
         {Set the command status flag to RETR and send the RETR command
         to the server}
         iiFtpStatus := FTP_RETR;
         NetSocket.SocketSend(isConnectionSocket, handle, UWM_SOCKETEVENT, 'RETR ' + isFileName, SKT_SENDPACKET);
      end;
      {You want to send a STOR command}
      FTP_STOR:
      begin
         {Set the command status flag to STOR and send the STOR command
         to the server}
         iiFtpStatus := FTP_STOR;
         NetSocket.SocketSend(isConnectionSocket, handle, UWM_SOCKETEVENT, 'STOR ' + isFileName, SKT_SENDPACKET);
      end;
   end;
   {Clear the receive buffer}
   ClearBuffer(ReceiveBuffer, 8192);
end;

procedure TFtp.ProcessHostLookup;
begin
   {We have received the host address, now we need to connect to it.}
   stStatus.Caption := 'Connecting To Host';
   stStatus.Update;
   if (NetSocket.ProcessHostAddress = TRUE) then
   begin
      {Connect to the host computer}
      isConnectionSocket := NetSocket.OpenSocketConnection(Handle, UWM_SOCKETEVENT, SKT_CONNECTTOHOST, IPPORT_FTP);
   end
   else
   begin
      stStatus.Caption := 'Cannot Find Host Address';
      stStatus.Update;
   end;
end;

procedure TFtp.ProcessHostConnect;
begin
   {We have connected to the other computer, so update the display}
   stStatus.Caption := 'Receiving Welcome From Host';
   stStatus.Update;
   iiFtpStatus := FTP_CONNECT;
   {Clear the receive buffer}
   ClearBuffer(ReceiveBuffer, 8192);
end;

procedure TFtp.ProcessHostWelcome;
var
   lsWelcomeMsg, lsRetnCode: String;
begin
   {Convert the response into a Pascal string}
   lsWelcomeMsg := StrPas(ReceiveBuffer);
   {Extract the response code}
   lsRetnCode := Copy(lsWelcomeMsg, 1, 3);
   {What response code was received, are you connected?}
   case StrToInt(lsRetnCode) of
      120: ibConnected := FALSE;
      220: ibConnected := TRUE;
      421: ibConnected := FALSE;
   end;
   iiFtpStatus := FTP_NOOP;
   {Display the response message}
   stStatus.Caption := lsWelcomeMsg;
   stStatus.Update;
   {Are you connected?}
   if ibConnected then
   begin
      {Send the USER command with the user ID 
       that the application's user supplied}
      iiFtpStatus := FTP_USER;
      NetSocket.SocketSend(isConnectionSocket, handle, UWM_SOCKETEVENT, 'USER ' + sleLogin.Text, SKT_SENDPACKET);
      {Clear the receive buffer}
      ClearBuffer(ReceiveBuffer, 8192);
   end;
end;

procedure TFtp.ProcessHostSend;
begin
   {Clear the receive buffer}
   if (iiFtpStatus <> FTP_NOOP) then
   begin
      ClearBuffer(ReceiveBuffer, 8192);
   end;
end;

procedure TFtp.ProcessHostAbort;
var
   lsReplyMsg, lsRetnCode: String;
   lbProceed: boolean;
begin
   lbProceed := FALSE;
   {Convert the response into a Pascal string}
   lsReplyMsg := StrPas(ReceiveBuffer);
   {Extract the response code}
   lsRetnCode := Copy(lsReplyMsg, 1, 3);
   {What response code was received? Was the command successful?}
   case StrToInt(lsRetnCode) of
      225: lbProceed := TRUE;
      226: lbProceed := TRUE;
      421: ibConnected := FALSE;
      500:;
      501:;
      502:;
   end;
   {Display the response message}
   stStatus.Caption := lsReplyMsg;
   stStatus.Update;
   {Was the command successful?}
   if not lbProceed then
   begin
      iiFtpStatus := FTP_NOOP;
      NetSocket.SocketSend(isConnectionSocket, handle, UWM_SOCKETEVENT, 'NOOP', SKT_SENDPACKET);
      {Clear the receive buffer}
      ClearBuffer(ReceiveBuffer, 8192);
   end;
   {If you are no longer connected, close the socket}
   if not ibConnected then
      NetSocket.CloseSocketConnection(isConnectionSocket);
end;

procedure TFtp.ProcessHostNOOP;
var
   lsReplyMsg, lsRetnCode: String;
   lbProceed: boolean;
begin
   lbProceed := FALSE;
   {Convert the response into a Pascal string}
   lsReplyMsg := StrPas(ReceiveBuffer);
   {Extract the response code}
   lsRetnCode := Copy(lsReplyMsg, 1, 3);
   {What response code was received? Was the command successful?}
   case StrToInt(lsRetnCode) of
      200: lbProceed := TRUE;
      421: ibConnected := FALSE;
      500:;
   end;
   {Display the response message}
   stStatus.Caption := lsReplyMsg;
   stStatus.Update;
   {Was the command successful?}
   if not lbProceed then
   begin
      iiFtpStatus := FTP_NOOP;
      NetSocket.SocketSend(isConnectionSocket, handle, UWM_SOCKETEVENT, 'NOOP', SKT_SENDPACKET);
      {Clear the receive buffer}
      ClearBuffer(ReceiveBuffer, 8192);
   end;
   {If you are no longer connected, close the socket}
   if not ibConnected then
      NetSocket.CloseSocketConnection(isConnectionSocket);
end;

procedure TFtp.ProcessHostQuit;
var
   lsReplyMsg, lsRetnCode: String;
   lbProceed: boolean;
begin
   lbProceed := FALSE;
   {Convert the response into a Pascal string}
   lsReplyMsg := StrPas(ReceiveBuffer);
   {Extract the response code}
   lsRetnCode := Copy(lsReplyMsg, 1, 3);
   {What response code was received? Was the command successful?}
   case StrToInt(lsRetnCode) of
      221: lbProceed := TRUE;
      500:;
   end;
   iiFtpStatus := FTP_NOOP;
   {Display the response message}
   stStatus.Caption := lsReplyMsg;
   stStatus.Update;
   NetSocket.CloseSocketConnection(isConnectionSocket);
   Close;
end;

procedure TFtp.ProcessHostLogin;
var
   lsReplyMsg, lsRetnCode: String;
   lbProceed: boolean;
begin
   lbProceed := FALSE;
   {Convert the response into a Pascal string}
   lsReplyMsg := StrPas(ReceiveBuffer);
   {Extract the response code}
   lsRetnCode := Copy(lsReplyMsg, 1, 3);
   {What response code was received? Was the command successful?}
   case StrToInt(lsRetnCode) of
      230: lbProceed := TRUE;
      331: lbProceed := TRUE;
      332: ibConnected := FALSE;
      421: ibConnected := FALSE;
      500:;
      501:;
      530:;
   end;
   iiFtpStatus := FTP_NOOP;
   {Display the response message}
   stStatus.Caption := lsReplyMsg;
   stStatus.Update;
   {Was the command successful?}
   if lbProceed then
   begin
      {Send the PASS command with the password 
       that the application's user supplied}
      iiFtpStatus := FTP_PASS;
      NetSocket.SocketSend(isConnectionSocket, handle, UWM_SOCKETEVENT, 'PASS ' + slePassword.Text, SKT_SENDPACKET);
      {Clear the receive buffer}
      ClearBuffer(ReceiveBuffer, 8192);
   end;
   {If you are no longer connected, close the socket}
   if not ibConnected then
      NetSocket.CloseSocketConnection(isConnectionSocket);
end;

procedure TFtp.ProcessHostPassword;
var
   lsReplyMsg, lsRetnCode: String;
   lbProceed: boolean;
begin
   lbProceed := FALSE;
   {Convert the response into a Pascal string}
   lsReplyMsg := StrPas(ReceiveBuffer);
   {Extract the response code}
   lsRetnCode := Copy(lsReplyMsg, 1, 3);
   {What response code was received? Was the command successful?}
   case StrToInt(lsRetnCode) of
      202: lbProceed := TRUE;
      230: lbProceed := TRUE;
      332: ibConnected := FALSE;
      421: ibConnected := FALSE;
      500:;
      501:;
      503:;
      530:;
   end;
   iiFtpStatus := FTP_NOOP;
   {Display the response message}
   stStatus.Caption := lsReplyMsg;
   stStatus.Update;
   {Was the command successful?}
   if lbProceed then
   begin
      {Send the PWD command}
      iiFtpStatus := FTP_PWD;
      NetSocket.SocketSend(isConnectionSocket, handle, UWM_SOCKETEVENT, 'PWD', SKT_SENDPACKET);
      {Clear the receive buffer}
      ClearBuffer(ReceiveBuffer, 8192);
   end;
   {If you are no longer connected, close the socket}
   if not ibConnected then
      NetSocket.CloseSocketConnection(isConnectionSocket);
end;

procedure TFtp.ProcessHostCWD;
var
   lsReplyMsg, lsRetnCode: String;
   lbProceed: boolean;
begin
   lbProceed := FALSE;
   {Convert the response into a Pascal string}
   lsReplyMsg := StrPas(ReceiveBuffer);
   {Extract the response code}
   lsRetnCode := Copy(lsReplyMsg, 1, 3);
   {What response code was received? Was the command successful?}
   case StrToInt(lsRetnCode) of
      250: lbProceed := TRUE;
      421: ibConnected := FALSE;
      500:;
      501:;
      502:;
      530:;
      550:;
   end;
   iiFtpStatus := FTP_NOOP;
   {Display the response message}
   stStatus.Caption := lsReplyMsg;
   stStatus.Update;
   {Was the command successful?}
   if lbProceed then
   begin
      {Send the PWD command}
      iiFtpStatus := FTP_PWD;
      NetSocket.SocketSend(isConnectionSocket, handle, UWM_SOCKETEVENT, 'PWD', SKT_SENDPACKET);
      {Clear the receive buffer}
      ClearBuffer(ReceiveBuffer, 8192);
   end;
   {If you are no longer connected, close the socket}
   if not ibConnected then
      NetSocket.CloseSocketConnection(isConnectionSocket);
end;

procedure TFtp.ProcessHostPWD;
var
   lsReplyMsg, lsRetnCode: String;
   lbProceed: boolean;
   liDirStart, liDirEnd: Integer;
begin
   lbProceed := FALSE;
   {Convert the response into a Pascal string}
   lsReplyMsg := StrPas(ReceiveBuffer);
   {Extract the response code}
   lsRetnCode := Copy(lsReplyMsg, 1, 3);
   {What response code was received? Was the command successful?}
   case StrToInt(lsRetnCode) of
      257: lbProceed := TRUE;
      421: ibConnected := FALSE;
      500:;
      501:;
      502:;
      550:;
   end;
   iiFtpStatus := FTP_NOOP;
   {Display the response message}
   stStatus.Caption := lsReplyMsg;
   stStatus.Update;
   {Was the command successful?}
   if lbProceed then
   begin
      iiFtpStatus := FTP_NOOP;
      {Parse the directory name and path from the response message and
      display it for the user}
      liDirStart := Pos('"', lsReplyMsg);
      if (liDirStart > 0) then
      begin
         lsRetnCode := Copy(lsReplyMsg, (liDirStart + 1), Length(lsReplyMsg));
         liDirEnd := Pos('"', lsRetnCode);
         if (liDirEnd > 0) then
            lsRetnCode := Copy(lsRetnCode, 1, (liDirEnd - 1));
         stHostPath.Caption := lsRetnCode;
      end;
      {Send the TYPE command, specifying ASCII data type}
      iiFtpStatus := FTP_TYPE;
      {You eventually want to send want to send a LIST command}
      iiFtpWorkingToward := FTP_LIST;
      isDataConnType := 'A';
      NetSocket.SocketSend(isConnectionSocket, handle, UWM_SOCKETEVENT, 'TYPE A', SKT_SENDPACKET);
      {Clear the receive buffer}
      ClearBuffer(ReceiveBuffer, 8192);
   end;
   {If you are no longer connected, close the socket}
   if not ibConnected then
      NetSocket.CloseSocketConnection(isConnectionSocket);
end;

procedure TFtp.ProcessHostType;
var
   lsReplyMsg, lsRetnCode: String;
   lbProceed: boolean;
begin
   lbProceed := FALSE;
   {Convert the response into a Pascal string}
   lsReplyMsg := StrPas(ReceiveBuffer);
   {Extract the response code}
   lsRetnCode := Copy(lsReplyMsg, 1, 3);
   {What response code was received? Was the command successful?}
   case StrToInt(lsRetnCode) of
      200: lbProceed := TRUE;
      421: ibConnected := FALSE;
      500:;
      501:;
      504:;
      530:;
   end;
   iiFtpStatus := FTP_NOOP;
   {Display the response message}
   stStatus.Caption := lsReplyMsg;
   stStatus.Update;
   {Was the command successful?}
   if lbProceed then
   begin
      iiFtpStatus := FTP_PASV;
      {Create the data socket}
      if CreateDataSocket then
      begin
         {Get the socket address (this is legacy code from when you were
         trying to enable the server to establish the data connection}
         lsRetnCode := NetSocket.GetSockAddr;
         {Send the PASV command}
         NetSocket.SocketSend(isConnectionSocket, handle, UWM_SOCKETEVENT, 'PASV', SKT_SENDPACKET);
         {Clear the receive buffer}
         ClearBuffer(ReceiveBuffer, 8192);
      end
      else
         {You couldn't create the data socket, so place the 
         application into idle mode}
         iiFtpStatus := FTP_NOOP;
   end;
   {If you are no longer connected, close the socket}
   if not ibConnected then
      NetSocket.CloseSocketConnection(isConnectionSocket);
end;

procedure TFtp.ProcessHostPort;
var
   lsReplyMsg, lsRetnCode: String;
   lbProceed: boolean;
begin
   lbProceed := FALSE;
   {Convert the response into a Pascal string}
   lsReplyMsg := StrPas(ReceiveBuffer);
   {Extract the response code}
   lsRetnCode := Copy(lsReplyMsg, 1, 3);
   {What response code was received? Was the command successful?}
   case StrToInt(lsRetnCode) of
      200: lbProceed := TRUE;
      421: ibConnected := FALSE;
      500:;
      501:;
      530:;
   end;
   iiFtpStatus := FTP_NOOP;
   {Display the response message}
   stStatus.Caption := lsReplyMsg;
   stStatus.Update;
   {Was the command successful?}
   if lbProceed then
   begin
      case iiFtpWorkingToward of
         FTP_LIST:
         begin
            iiFtpStatus := FTP_LIST;
            NetSocket.SocketSend(isConnectionSocket, handle, UWM_SOCKETEVENT, 'LIST', SKT_SENDPACKET);
            {Clear the receive buffer}
            ClearBuffer(ReceiveBuffer, 8192);
         end;
         FTP_NLST:
         begin
            iiFtpStatus := FTP_NLST;
            NetSocket.SocketSend(isConnectionSocket, handle, UWM_SOCKETEVENT, 'NLST', SKT_SENDPACKET);
            {Clear the receive buffer}
            ClearBuffer(ReceiveBuffer, 8192);
         end;
      end;
   end;
   {If you are no longer connected, close the socket}
   if not ibConnected then
      NetSocket.CloseSocketConnection(isConnectionSocket);
end;

procedure TFtp.ProcessHostPasv;
var
   lsReplyMsg, lsRetnCode: String;
   lbProceed: boolean;
   liAddrStart, liAddrEnd: Integer;
begin
   lbProceed := FALSE;
   {Convert the response into a Pascal string}
   lsReplyMsg := StrPas(ReceiveBuffer);
   {Extract the response code}
   lsRetnCode := Copy(lsReplyMsg, 1, 3);
   {What response code was received? Was the command successful?}
   case StrToInt(lsRetnCode) of
      227: lbProceed := TRUE;
      421: ibConnected := FALSE;
      500:;
      501:;
      502:;
      530:;
   end;
   iiFtpStatus := FTP_NOOP;
   {Display the response message}
   stStatus.Caption := lsReplyMsg;
   stStatus.Update;
   {Was the command successful?}
   if lbProceed then
   begin
      {Find the parentheses that enclose the server data socket address}
      liAddrStart := Pos( '(', lsReplyMsg);
      liAddrEnd := Pos( ')', lsReplyMsg);
      {Extract the server data socket address from the response message}
      DataAddress := Copy(lsReplyMsg, (liAddrStart + 1), (liAddrEnd - (liAddrStart + 1)));
      {Open the data socket connection, using the Internet address that you received}
      OpenDataConnection(DataAddress);
   end;
   {If you are no longer connected, close the socket}
   if not ibConnected then
      NetSocket.CloseSocketConnection(isConnectionSocket);
end;

procedure TFtp.ProcessHostList;
var
   lsReplyMsg, lsRetnCode: String;
   lbProceed: boolean;
   lbDoAccept: boolean;
   lbDone: boolean;
begin
   lbProceed := FALSE;
   lbDone := FALSE;
   lbDoAccept := FALSE;
   {Convert the response into a Pascal string}
   lsReplyMsg := StrPas(ReceiveBuffer);
   {Extract the response code}
   lsRetnCode := Copy(lsReplyMsg, 1, 3);
   {What response code was received? Was the command successful?}
   case StrToInt(lsRetnCode) of
      125: lbProceed := TRUE;
      150:
      begin
         lbDoAccept := TRUE;
         lbProceed := TRUE;
      end;
      226: lbDone := TRUE;
      250: lbDone := TRUE;
      421: ibConnected := FALSE;
      425: lbDone := TRUE;
      426: lbDone := TRUE;
      450: lbDone := TRUE;
      451: lbDone := TRUE;
      500:;
      501:;
      502:;
      530:;
   end;
   {Display the response message}
   stStatus.Caption := lsReplyMsg;
   stStatus.Update;
   {Was the command successful?}
   if not lbProceed then
   begin
      iiFtpStatus := FTP_NOOP;
      lbDone := TRUE;
   end
   else
   begin
      {If the data socket is waiting}
      if ((iiDataConnStatus = FTPD_NOOP) or
         (iiDataConnStatus = FTPD_WAITFORCONN)) then
      begin
         {Clear the server directory listing list box}
         lbHostFiles.Clear;
         iiDataConnStatus := FTPD_RECVDISP;
         {Clear the transfer buffer}
         ClearBuffer(TransferBuffer, 8192);
      end;
      {Clear the receive buffer}
      ClearBuffer(ReceiveBuffer, 8192);
   end;
   {Are you done?}
   if lbDone then
   begin
      {Close the data socket and wait for the user to decide what to do next}
      CloseDataConnection;
      iiFtpStatus := FTP_NOOP;
   end;
   {If you are no longer connected, close the socket}
   if not ibConnected then
      NetSocket.CloseSocketConnection(isConnectionSocket);
end;

procedure TFtp.ProcessHostNList;
var
   lsReplyMsg, lsRetnCode: String;
   lbProceed: boolean;
   lbDone: boolean;
begin
   lbProceed := FALSE;
   lbDone := FALSE;
   {Convert the response into a Pascal string}
   lsReplyMsg := StrPas(ReceiveBuffer);
   {Extract the response code}
   lsRetnCode := Copy(lsReplyMsg, 1, 3);
   {What response code was received? Was the command successful?}
   case StrToInt(lsRetnCode) of
      125: lbProceed := TRUE;
      150: lbProceed := TRUE;
      226: lbDone := TRUE;
      250: lbDone := TRUE;
      421: ibConnected := FALSE;
      425:;
      426: lbDone := TRUE;
      450: lbDone := TRUE;
      451: lbDone := TRUE;
      500:;
      501:;
      502:;
      530:;
   end;
   {Display the response message}
   stStatus.Caption := lsReplyMsg;
   stStatus.Update;
   {Was the command successful?}
   if not lbProceed then
   begin
      iiFtpStatus := FTP_NOOP;
      lbDone := TRUE;
   end
   else
   begin
      {If the data socket is waiting}
      if ((iiDataConnStatus = FTPD_NOOP) or
         (iiDataConnStatus = FTPD_WAITFORCONN)) then
      begin
         {Clear the server directory listing list box}
         lbHostFiles.Clear;
         iiDataConnStatus := FTPD_RECVDISP;
         {Clear the transfer buffer}
         ClearBuffer(TransferBuffer, 8192);
      end;
      {Clear the receive buffer}
      ClearBuffer(ReceiveBuffer, 8192);
   end;
   {Are you done?}
   if lbDone then
   begin
      {Close the data socket and wait for the user to decide what to do next}
      CloseDataConnection;
      iiFtpStatus := FTP_NOOP;
   end;
   {If you are no longer connected, close the socket}
   if not ibConnected then
      NetSocket.CloseSocketConnection(isConnectionSocket);
end;

procedure TFtp.ProcessHostRetr;
var
   lsReplyMsg, lsRetnCode: String;
   lbProceed: boolean;
   lbDoAccept: boolean;
   lbDone: boolean;
begin
   lbProceed := FALSE;
   lbDone := FALSE;
   lbDoAccept := FALSE;
   {Convert the response into a Pascal string}
   lsReplyMsg := StrPas(ReceiveBuffer);
   {Extract the response code}
   lsRetnCode := Copy(lsReplyMsg, 1, 3);
   {What response code was received? Was the command successful?}
   case StrToInt(lsRetnCode) of
      110: lbProceed := TRUE;
      125: lbProceed := TRUE;
      150:
      begin
         lbDoAccept := TRUE;
         lbProceed := TRUE;
      end;
      226: lbDone := TRUE;
      250: lbDone := TRUE;
      421: ibConnected := FALSE;
      425: lbDone := TRUE;
      426: lbDone := TRUE;
      450: lbDone := TRUE;
      451: lbDone := TRUE;
      500:;
      501:;
      530:;
      550:;
   end;
   {Display the response message}
   stStatus.Caption := lsReplyMsg;
   stStatus.Update;
   {Was the command successful?}
   if not lbProceed then
   begin
      iiFtpStatus := FTP_NOOP;
      lbDone := TRUE;
   end
   else
   begin
      {Is the data socket waiting?}
      if ((iiDataConnStatus = FTPD_NOOP) or
         (iiDataConnStatus = FTPD_WAITFORCONN)) then
      begin
         iiDataConnStatus := FTPD_RECV;
         {Create and open the file to receive}
         OpenReceiveFile;
         {Clear the receive buffer}
         ClearBuffer(TransferBuffer, 512);
         {Begin receiving the file}
         ProcessDataReceive;
      end;
      {Clear the receive buffer}
      ClearBuffer(ReceiveBuffer, 8192);
   end;
   {Are you done?}
   if lbDone then
   begin
      {You are, so close the data socket and wait for the user to decide
      what to do next}
      CloseDataConnection;
      iiFtpStatus := FTP_NOOP;
   end;
   {If you are no longer connected, close the socket}
   if not ibConnected then
      NetSocket.CloseSocketConnection(isConnectionSocket);
end;

procedure TFtp.ProcessHostStor;
var
   lsReplyMsg, lsRetnCode: String;
   lbProceed: boolean;
   lbDoAccept: boolean;
   lbDone: boolean;
begin
   lbProceed := FALSE;
   lbDone := FALSE;
   lbDoAccept := FALSE;
   {Convert the response into a Pascal string}
   lsReplyMsg := StrPas(ReceiveBuffer);
   {Extract the response code}
   lsRetnCode := Copy(lsReplyMsg, 1, 3);
   {What response code was received? Was the command successful?}
   case StrToInt(lsRetnCode) of
      110: lbProceed := TRUE;
      125: lbProceed := TRUE;
      150:
      begin
         lbDoAccept := TRUE;
         lbProceed := TRUE;
      end;
      226: lbDone := TRUE;
      250: lbDone := TRUE;
      421: ibConnected := FALSE;
      425: lbDone := TRUE;
      426: lbDone := TRUE;
      450: lbDone := TRUE;
      451: lbDone := TRUE;
      452: lbDone := TRUE;
      500:;
      501:;
      530:;
      532:;
      551:;
      552:;
      553:;
   end;
   {Display the response message}
   stStatus.Caption := lsReplyMsg;
   stStatus.Update;
   {Was the command successful?}
   if not lbProceed then
   begin
      iiFtpStatus := FTP_NOOP;
      lbDone := TRUE;
   end
   else
   begin
      {Is the data socket waiting?}
      if ((iiDataConnStatus = FTPD_NOOP) or
         (iiDataConnStatus = FTPD_WAITFORCONN)) then
      begin
         {Set the data socket status to sending}
         iiDataConnStatus := FTPD_SEND;
         {Open the selected file}
         OpenSendFile;
         {Send the file}
         ProcessDataSend;
      end;
      {Clear the receive buffer}
      ClearBuffer(ReceiveBuffer, 8192);
   end;
   {Have you sent the entire file?}
   if lbDone then
   begin
      {Close the data socket and wait for the user to decide what to do next}
      CloseDataConnection;
      iiFtpStatus := FTP_NOOP;
   end;
   {If you are no longer connected, close the socket}
   if not ibConnected then
      NetSocket.CloseSocketConnection(isConnectionSocket);
end;

procedure TFtp.ProcessHostReceive;
var
   lsRetnStr: String;
   liCarRetnPos: Integer;
begin
   try
      {Receive the incoming data}
      liCarRetnPos := recv(isConnectionSocket, ReceiveBuffer, 8192, 0);
      ReceiveBuffer[liCarRetnPos] := char(0);
      {Have you received anything?}
      if (ReceiveBuffer[0] = Char(0)) then
      begin
         {No, be sure and clear the command receive buffer}
         ClearBuffer(ReceiveBuffer, 8192);
      end
      else
      begin
         {What was the FTP command to which this is a response? Call the
         appropriate procedure to process the response}
         case iiFtpStatus of
            FTP_NOOP: ProcessHostNOOP;
            FTP_CONNECT: ProcessHostWelcome;
            FTP_USER: ProcessHostLogin;
            FTP_PASS: ProcessHostPassword;
            FTP_CWD: ProcessHostCWD;
            FTP_CDUP:;
            FTP_REIN:;
            FTP_QUIT: ProcessHostQuit;
            FTP_PORT: ProcessHostPort;
            FTP_PASV: ProcessHostPasv;
            FTP_TYPE: ProcessHostType;
            FTP_STRU:;
            FTP_MODE:;
            FTP_RETR: ProcessHostRetr;
            FTP_STOR: ProcessHostStor;
            FTP_STOU:;
            FTP_APPE:;
            FTP_REST:;
            FTP_RNFR:;
            FTP_RNTO:;
            FTP_ABOR: ProcessHostAbort;
            FTP_DELE:;
            FTP_RMD:;
            FTP_MKD:;
            FTP_PWD: ProcessHostPWD;
            FTP_LIST: ProcessHostList;
            FTP_NLST: ProcessHostNList;
         end;
      end;
   except
      {Clear the receive buffer}
      ClearBuffer(ReceiveBuffer, 8192);
   end;
end;

{This is the command socket event handler}
procedure TFtp.UWMSocketEvent(var Msg: TMessage);
begin
   {What command socket event message did you receive? Call the 
   appropriate procedure to handle the event}
   case Msg.LParamLo of
      FD_READ: ProcessHostReceive;
      FD_WRITE: ProcessHostSend;
      FD_CONNECT: ProcessHostConnect;
      FD_CLOSE: ;
   end;
end;

procedure TFtp.UWMDataSocketEvent(var Msg: TMessage);
begin
   {Which data socket event message did you receive? Call the appropriate
   procedure to handle the event}
   case Msg.LParamLo of
      FD_READ: ProcessDataReceive;
      FD_WRITE: ProcessDataSend;
      FD_OOB: ;
      FD_ACCEPT: AcceptDataConnection;
      FD_CONNECT: ProcessDataConnect;
      FD_CLOSE: CloseDataConnection;
   end;
end;

procedure TFtp.UWMAddrSocketEvent(var Msg: TMessage);
begin
   if (Msg.lParamHi = 0) then
      ProcessHostLookup;
end;

procedure TFtp.DirectoryListBox1Change(Sender: TObject);
begin
   FileListBox1.Directory := DirectoryListBox1.Directory;
   stLocalPath.Caption := DirectoryListBox1.Directory;
end;

procedure TFtp.DriveComboBox1Change(Sender: TObject);
begin
   DirectoryListBox1.Drive := DriveComboBox1.Drive
end;

procedure TFtp.FileListBox1Change(Sender: TObject);
begin
   stLocalPath.Caption := FileListBox1.FileName;
end;

procedure TFtp.cbQuitClick(Sender: TObject);
begin
   if not ibConnected then
      Close
   else
   begin
      stStatus.Caption := 'Sending Logout';
      stStatus.Update;
      iiFtpStatus := FTP_QUIT;
      NetSocket.SocketSend(isConnectionSocket, handle, UWM_SOCKETEVENT, 'QUIT', SKT_SENDPACKET);
         ClearBuffer(ReceiveBuffer, 8192);
   end;
end;

procedure TFtp.cbConnectClick(Sender: TObject);
begin
   if ((Length(sleHostName.Text) > 0) and
      (Length(sleLogin.Text) > 0) and
      (Length(slePassword.Text) > 0)) then
   begin
      stStatus.Caption := 'Looking Up Host Address';
      stStatus.Update;
      iiFtpStatus := FTP_CONNECT;
      NetSocket.GetHostAddress(Handle, UWM_ADDRSOCKETEVENT, sleHostName.text, SKT_GETHOSTADDR);
   end;
end;

procedure TFtp.lbHostFilesDblClick(Sender: TObject);
var
   lsFile: String;
   lbIsDir: Boolean;
   liSpaceChar: Integer;
begin
   {Get the selected line from the list box}
   lsFile := lbHostFiles.Items[lbHostFiles.ItemIndex];
   lbIsDir := FALSE;
   {Is it a directory? This handles UNIX and Windows NT FTP servers
   ONLY. If you want to handle other operating systems, you must 
   extend this to handle the directory listings of those operating
   systems}
   if ((Copy(lsFile, 1, 1) = 'd') or
      (Pos('<DIR>', lsFile) > 0)) then
   begin
      {The user selected a directory}
      lbIsDir := TRUE
   end;
   {Trim everything from around the file name}
   liSpaceChar := Pos(' ', lsFile);
   while (liSpaceChar > 0) do
   begin
      lsFile := Copy(lsFile, (liSpaceChar + 1), Length(lsFile));
      liSpaceChar := Pos(' ', lsFile);
   end;
   {Is it a directory?}
   if lbIsDir then
   begin
      {Tell the user that you are changing the directory on the server}
      stStatus.Caption := 'Changing Working Directory to ' + lsFile;
      stStatus.Update;
      {Send the CWD command, specifying the directory to which to move}
      iiFtpStatus := FTP_CWD;
      NetSocket.SocketSend(isConnectionSocket, handle, UWM_SOCKETEVENT, 'CWD ' + lsFile, SKT_SENDPACKET);
      {Clear the receive buffer}
      ClearBuffer(ReceiveBuffer, 8192);
   end
   else
   begin
      {Tell the user that you are receiving the selected file}
      stStatus.Caption := 'Retrieving File - ' + lsFile;
      stStatus.Update;
      isFileName := lsFile;
      {Send the TYPE command, specifying the IMAGE data type}
      iiFtpStatus := FTP_TYPE;
      {Set the working toward flag so that you have a reminder 
       that you want to send the RETR command after taking 
       care of the preparations}
      iiFtpWorkingToward := FTP_RETR;
      isDataConnType := 'I';
      NetSocket.SocketSend(isConnectionSocket, handle, UWM_SOCKETEVENT, 'TYPE I', SKT_SENDPACKET);
      {Clear the receive buffer}
      ClearBuffer(ReceiveBuffer, 8192);
   end;
end;

procedure TFtp.FileListBox1DblClick(Sender: TObject);
var
   lsFile: String;
   lbIsDir: Boolean;
   liSlashChar: Integer;
begin
   {What is the selected file's name?}
   lsFile := FileListBox1.FileName;
   {Remove all the directory and path information from the file name.
   You do so by looking for each slash character and removing all text
   in front of it (including the slash character)}
   liSlashChar := Pos('\', lsFile);
   while (liSlashChar > 0) do
   begin
      lsFile := Copy(lsFile, (liSlashChar + 1), Length(lsFile));
      liSlashChar := Pos('\', lsFile);
   end;
   {Tell the user that you are sending the selected file}
   stStatus.Caption := 'Sending File - ' + lsFile;
   stStatus.Update;
   isFileName := lsFile;
   {Send the TYPE command, specifying IMAGE data type}
   iiFtpStatus := FTP_TYPE;
   {Set the working toward flag so that you have a reminder 
    that you want to send the STOR command  after taking 
    care of the preparations}
   iiFtpWorkingToward := FTP_STOR;
   isDataConnType := 'I';
   NetSocket.SocketSend(isConnectionSocket, handle, UWM_SOCKETEVENT, 'TYPE I', SKT_SENDPACKET);
   {Clear the receive buffer}
   ClearBuffer(ReceiveBuffer, 8192);
end;

end.
