unit tftps;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  TSocketc, NWinSock, StdCtrls, FileCtrl;

{ If Using the Delphi 32 Winsock.pas }
type
     _Socket = TSocket;
     fd_set = TFDSet;
     Pfd_set = PFDSet;
     timeval = TTimeVal;
     hostent = THostEnt;
     netent = TNetEnt;
     servent = TServEnt;
     protoent = TProtoEnt;
     in_addr = TInAddr;
     Pin_addr = PInAddr;
     Sockaddr_in = TSockAddrIn;
     PSockaddr_in = PSockAddrIn;
     WSADATA = TWSAData;
     LPWSADATA = PWSAData;
     sockaddr = TSockAddr;
     sockproto = TSockProto;
     linger = TLinger;
     LPSOCKADDR = ^sockaddr;
     LPSOCKADDR_IN = ^sockaddr_in;
     LPLINGER = ^linger;
     LPIN_ADDR = ^in_addr;
     LPFD_SET = ^fd_set;
     LPHOSTENT = ^hostent;
     LPSERVENT = ^servent;
     LPPROTOENT = ^protoent;
     LPTIMEVAL = ^timeval;
{ End of Delphi 32 Winsock.pas compatability type definitions }
const
  SKT_NOOP = 0;
  SKT_GETHOSTADDR = 1;
  SKT_CONNECTTOHOST = 2;
  SKT_SENDPACKET = 3;
  SKT_RECVPACKET = 4;
  FTP_NOOP = 0;
  FTP_CONNECT = 1;
  FTP_USER = 2;
  FTP_PASS = 3;
  FTP_CWD = 4;
  FTP_CDUP = 5;
  FTP_REIN = 6;
  FTP_QUIT = 7;
  FTP_PORT = 8;
  FTP_PASV = 9;
  FTP_TYPE = 10;
  FTP_STRU = 11;
  FTP_MODE = 12;
  FTP_RETR = 13;
  FTP_STOR = 14;
  FTP_STOU = 15;
  FTP_APPE = 16;
  FTP_REST = 17;
  FTP_RNFR = 18;
  FTP_RNTO = 19;
  FTP_ABOR = 20;
  FTP_DELE = 21;
  FTP_RMD = 22;
  FTP_MKD = 23;
  FTP_PWD = 24;
  FTP_LIST = 25;
  FTP_NLST = 26;
  FTP_OK = 27;
  FTP_SYST = 28;
  FTP_UNKWN = 29;
  FTPD_NOOP = 0;
  FTPD_WAITFORCONN = 1;
  FTPD_SEND = 2;
  FTPD_RECV = 3;
  FTPD_RECVDISP = 4;
  UWM_SOCKETEVENT = WM_USER + 500;
  UWM_DATASOCKETEVENT = WM_USER + 501;
  UWM_LISTENSOCKETEVENT = WM_USER + 502;
  UWM_REJECTSOCKETEVENT = WM_USER + 503;
type
  TFTPServe = class(TForm)
    Label1: TLabel;
    stStatus: TLabel;
    mleReceived: TMemo;
    mleSent: TMemo;
    cbClose: TButton;
    Label2: TLabel;
    stAddress: TLabel;
    flbCurDir: TFileListBox;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure cbCloseClick(Sender: TObject);
    procedure ClearBuffer(DirtyBuffer: Pchar; BufLength: Integer);
    function CreateDataSocket: boolean;
    procedure OpenReceiveFile;
    procedure OpenSendFile;
    procedure OpenDataConnection(ConnectionAddr: String);
    procedure CloseDataConnection;
    procedure ProcessDataSend;
    procedure AcceptDataConnection;
    procedure ProcessDataReceive;
    procedure ProcessDataConnect;
    procedure ProcessListenAccept;
    procedure SendConnectReject;
    procedure SendMsg(aSendSock:_SOCKET; iMsg, iNewState, aiEvent: Integer; asAdditionalMsg: String);
    procedure SendContextMsg;
    procedure ValidateLogin(asUserName: String);
    procedure ValidatePassword(asPassword: String);
    procedure SendWorkingDirectory;
    procedure CheckModeRequest(asMode: String);
    procedure CheckTypeRequest(asType: String);
    procedure CheckPortRequest(asPort: String);
    procedure SendDirectoryList;
    procedure SendRecvFile(asFileName: String);
    procedure ReceiveStoreFile(asFileName: String);
    procedure CheckCommandMsg(asCommand, asArguments: String);
    procedure ReadCommandMsg;
  private
    { Private declarations }
    {The primary command connection socket}
    isConnectionSocket: _SOCKET;
    {The data connection socket}
    isDataSocket: _SOCKET;
    {The listening socket (waiting for connection requests from clients}
    isListenSocket: _SOCKET;
    {The rejection socket, used to accept connections long enough to send
    a rejection message}
    isRejectSocket: _SOCKET;
    {Are you connected to a client?}
    ibConnected: Boolean;
    {Is the data socket connected to a client?}
    ibDataConnected: Boolean;
    {The FTP command status flag}
    iiFTPStatus: Integer;
    {The holding buffer}
    isHoldBuffer: String;
    {The data port for the client}
    iiAltPort: Integer;
    {The number of current client connections}
    iiNumConnected: Integer;
    {Flag to indicate whether a rejection message has been sent to a client}
    ibRejectSent: Boolean;
    {The message to send to the client}
    isSendMsg: String;
    {The status of the data socket connection}
    iiDataConnStatus: Integer;
    {The type of the data connection}
    isDataConnType: Char;
    {Is the transfer file open?}
    ibFileOpen: boolean;
    {The transfer file name}
    isFileName: String;
    {The amount in the transfer buffer}
    iiBufferAmt: Integer;
    {The client TCP/IP address}
    isClientAddr: String;
{User-defined event declarations, for socket events. The first is 
for the primary command socket, the second for the data socket 
connection, the third for the listening socket, and the last
for the socket connection that is being rejected}
    procedure UWMSocketEvent(var Msg: TMessage); message UWM_SOCKETEVENT;
    procedure UWMDataSocketEvent(var Msg: TMessage); message UWM_DATASOCKETEVENT;
    procedure UWMListenSocketEvent(var Msg: TMessage); message UWM_LISTENSOCKETEVENT;
    procedure UWMRejectSocketEvent(var Msg: TMessage); message UWM_REJECTSOCKETEVENT;
  public
    { Public declarations }
    NetSocket: _TSocket;
    ReceiveBuffer: array[0..8192] of Char;
    TempBuffer: array[0..8192] of Char;
    TransferBuffer: array[0..8192] of Char;
    TransferFile: File of Byte;
    DataAddress: String;
  end;

var
  FTPServe: TFTPServe;

implementation

{$R *.DFM}

procedure TFTPServe.FormCreate(Sender: TObject);
begin
   {Initialize the various variables}
   ibConnected := FALSE;
   ibDataConnected := FALSE;
   ibRejectSent := FALSE;
   iiFTPStatus := FTP_NOOP;
   iiNumConnected := 0;
   {Create the TSocket class object}
   NetSocket := _TSocket.Create;
   {Initialize the WinSock interface}
   if not NetSocket.InitializeSocket then
      Close
   else
   begin
      {Create the listening socket, looking for clients requesting
      a connection}
      isListenSocket := NetSocket.CreateListenSocket(handle,
                        UWM_LISTENSOCKETEVENT, IPPORT_FTP);
      if (isListenSocket = INVALID_SOCKET) then
         Close
      else
      begin
         {Display the socket address for the user}
         stAddress.Caption := NetSocket.GetSockAddr;
         stAddress.Update;
      end;
   end;
end;

procedure TFTPServe.FormDestroy(Sender: TObject);
begin
     {Cleanup the WinSock interface and system}
     NetSocket.ShutdownSocket;
     NetSocket.Free;
end;

procedure TFTPServe.cbCloseClick(Sender: TObject);
begin
   {Close any open sockets}
   NetSocket.CloseSocketConnection(isListenSocket);
   if ibConnected then
      NetSocket.CloseSocketConnection(isConnectionSocket);
   if ibDataConnected then
      NetSocket.CloseSocketConnection(isDataSocket);
   {Close the application}
   Close;
end;

procedure TFTPServe.ClearBuffer(DirtyBuffer: Pchar; BufLength: Integer);
var
   liCurPos: Integer;
   lcNullChar: Char;
begin
   {Initialize the position pointer to 0}
   liCurPos := 0;
   {Create a null character}
   lcNullChar := Char(0);
   {Loop through the entire buffer passed in, setting each character
   equal to the null character}
   while (liCurPos < BufLength) do
   begin
      DirtyBuffer^ := lcNullChar;
      Inc(liCurPos);
      Inc(DirtyBuffer);
   end;
end;

function TFTPServe.CreateDataSocket: boolean;
begin
  {Create the data socket}
  isDataSocket := NetSocket.CreateDataSocket(isConnectionSocket, handle, UWM_DATASOCKETEVENT);
  {Did you get a valid socket?}
  if (isDataSocket = INVALID_SOCKET) then
     {No, return FALSE}
     CreateDataSocket := FALSE
  else
  begin
     {Yes, set the data socket status to show that it is waiting for a 
     connection, then return TRUE}
     iiDataConnStatus := FTPD_WAITFORCONN;
     CreateDataSocket := TRUE;
  end;
end;

procedure TFTPServe.OpenReceiveFile;
var
   lsReceiveFileName: String;
begin
   {Build the file name for the incoming file, then open it for writing}
   lsReceiveFileName := isFileName;
   ibFileOpen := TRUE;
   AssignFile(TransferFile, lsReceiveFileName);
   Rewrite(TransferFile);
end;

procedure TFTPServe.OpenSendFile;
begin
   {Open the requested file for reading}
   ibFileOpen := TRUE;
   AssignFile(TransferFile, isFileName);
   Reset(TransferFile);
end;

procedure TFTPServe.OpenDataConnection(ConnectionAddr: String);
var
   liPort: Integer;
   lsAddress: String;
   lsPortHigh, lsPortLow: String;
   liPosComma: Integer;
begin
   {Convert the server data socket address into a usable address}
   {First, convert the first four numbers into the standard 
    ###.###.###.### notation}
   liPosComma := Pos(',', ConnectionAddr);
   if (liPosComma > 0) then
   begin
      lsAddress := Copy(ConnectionAddr, 1, (liPosComma - 1)) + '.';
      ConnectionAddr := Copy(ConnectionAddr, (liPosComma + 1), Length(ConnectionAddr));
   end;
   liPosComma := Pos(',', ConnectionAddr);
   if (liPosComma > 0) then
   begin
      lsAddress := lsAddress + Copy(ConnectionAddr, 1, (liPosComma - 1)) + '.';
      ConnectionAddr := Copy(ConnectionAddr, (liPosComma + 1), Length(ConnectionAddr));
   end;
   liPosComma := Pos(',', ConnectionAddr);
   if (liPosComma > 0) then
   begin
      lsAddress := lsAddress + Copy(ConnectionAddr, 1, (liPosComma - 1)) + '.';
      ConnectionAddr := Copy(ConnectionAddr, (liPosComma + 1), Length(ConnectionAddr));
   end;
   liPosComma := Pos(',', ConnectionAddr);
   if (liPosComma > 0) then
   begin
      lsAddress := lsAddress + Copy(ConnectionAddr, 1, (liPosComma - 1));
      ConnectionAddr := Copy(ConnectionAddr, (liPosComma + 1), Length(ConnectionAddr));
   end;
   {Now take the last two numbers and combine them into the port number}
   liPosComma := Pos(',', ConnectionAddr);
   if (liPosComma > 0) then
   begin
      {The first number is the high byte, the second is the low}
      lsPortHigh := Copy(ConnectionAddr, 1, (liPosComma - 1));
      lsPortLow := Copy(ConnectionAddr, (liPosComma + 1), Length(ConnectionAddr));
      {Combine the two, shifting the high byte to the left 8 bits}
      liPort := (StrToInt(lsPortHigh) shl 8) + StrToInt(lsPortLow);
   end;
   {Open the data socket connection, using the address and port
    that you just extracted from the message that the server returned}
   NetSocket.OpenDataSocketConnection(handle, UWM_DATASOCKETEVENT, isDataSocket,
                      lsAddress, liPort);
end;

procedure TFTPServe.CloseDataConnection;
begin
  {Shut down the data socket}
  shutdown(isDataSocket, 2);
  {Close the data socket}
  NetSocket.CloseSocketConnection(isDataSocket);
  {If you are working with an open file, close it}
  if ((iiDataConnStatus = FTPD_SEND) or
     (iiDataConnStatus = FTPD_RECV)) then
  begin
     if ibFileOpen then
     begin
        CloseFile(TransferFile);
        ibFileOpen := FALSE;
        SendMsg(isConnectionSocket, 226, FTP_NOOP,UWM_SOCKETEVENT, isSendMsg);
     end;
  end;
end;

procedure TFTPServe.ProcessDataSend;
var
   liCurByteCount: Integer;
   lcCurByte: Byte;
   lpTransferDataPointer: PChar;
begin
   {Is the data file open?}
   if ibFileOpen then
   begin
      {Initialize the counter to keep track of how much you have sent}
      liCurByteCount := 0;
      {Are you at the end of the file?}
      if (not Eof(TransferFile)) then
      begin
         {No, clear the data transfer buffer}
         ClearBuffer(TransferBuffer, 8192);
         {Read the next 4,096 bytes from the file into the transfer 
         buffer}
         while ((liCurByteCount < 4096) and (not Eof(TransferFile))) do
         begin
            Read(TransferFile, lcCurByte);
            TransferBuffer[liCurByteCount] := Char(lcCurByte);
            Inc(liCurByteCount);
         end;
         {Send to the server the data transfer buffer's contents}
         NetSocket.SocketSendBuffer(isDataSocket, handle, UWM_DATASOCKETEVENT,
                  TransferBuffer, liCurByteCount, SKT_SENDPACKET);
      end
      else
         {You are at the end of the file, so close the data connection}
         CloseDataConnection;
   end;
end;

procedure TFTPServe.AcceptDataConnection;
begin
  isDataSocket := NetSocket.AcceptDataSocket(isDataSocket, handle, UWM_DATASOCKETEVENT);
  case iiFtpStatus of
     FTP_LIST:
     begin
        iiDataConnStatus := FTPD_RECVDISP;
        ClearBuffer(TransferBuffer, 8192);
        NetSocket.SocketReceive(isDataSocket, handle, UWM_DATASOCKETEVENT,
                  TransferBuffer, 8192, SKT_RECVPACKET);

     end;
     FTP_NLST:
     begin
        iiDataConnStatus := FTPD_RECVDISP;
        ClearBuffer(TransferBuffer, 8192);
        NetSocket.SocketReceive(isDataSocket, handle, UWM_DATASOCKETEVENT,
                  TransferBuffer, 8192, SKT_RECVPACKET);

     end;
     FTP_RETR:
     begin
        iiDataConnStatus := FTPD_RECV;
        OpenReceiveFile;
        ClearBuffer(TransferBuffer, 8192);
        NetSocket.SocketReceive(isDataSocket, handle, UWM_DATASOCKETEVENT,
                  TransferBuffer, 8192, SKT_RECVPACKET);

     end;
     FTP_STOR:
     begin
        iiDataConnStatus := FTPD_SEND;
        OpenSendFile;
        ProcessDataSend;
     end;
     FTP_STOU:
     begin
        iiDataConnStatus := FTPD_SEND;
        OpenSendFile;
        ProcessDataSend;
     end;
  end;
end;

procedure TFTPServe.ProcessDataReceive;
var
   lsTempStr: String;
   liTempBytes: Integer;
   liRecvBytes: Integer;
begin
   {Is the file open?}
   if ibFileOpen then
   begin
      {Set the data socket into synchronous mode}
      WSAAsyncSelect(isDataSocket, handle, 0, 0);
      {Receive the first set of data}
      liRecvBytes := recv(isDataSocket, TransferBuffer, 8192, 0);
      {Did you receive anything?}
      if (liRecvBytes >= 0) then
      begin
         {As long as you continue to receive data, loop}
         while (liRecvBytes > 0) do
         begin
            {Write what you have received to the file}
            for liTempBytes := 0 to (liRecvBytes - 1) do
               Write(TransferFile, Byte(TransferBuffer[liTempBytes]));
            {Clear the data transfer buffer}
            ClearBuffer(TransferBuffer, 8192);
            {Receive the next set of data}
            liRecvBytes := recv(isDataSocket, TransferBuffer, 8192, 0);
         end;
      end;
      {Close the data socket connection}
      CloseDataConnection;
      iiDataConnStatus := FTPD_NOOP;
   end;
end;

procedure TFTPServe.ProcessDataConnect;
var
   lsReplyMsg, lsRetnCode: String;
   lbProceed: boolean;
   liAddrStart, liAddrEnd: Integer;
begin
   {What is your data socket connection status?}
   case iiDataConnStatus of
      {We are sending a file}
      FTPD_SEND:
      begin
         {Open the file and send it}
         OpenSendFile;
         ProcessDataSend;
      end;
      {You are receiving a file}
      FTPD_RECV:
      begin
         {Open the file target, clear the transfer buffer, and receive
         the file}
         OpenReceiveFile;
         ClearBuffer(TransferBuffer, 512);
         ProcessDataReceive;
      end;
   end;
end;

procedure TFTPServe.ProcessListenAccept;
begin
   {Are there no current connections?}
   if (iiNumConnected = 0) then
   begin
      {Yes, accept this connection and send a greeting}
      isConnectionSocket := NetSocket.AcceptConnectingSocket(isListenSocket,
                         handle, UWM_SOCKETEVENT);
      if (isConnectionSocket <> INVALID_SOCKET) then
      begin
         {Increment the number of current connections}
         Inc(iiNumConnected);
         iiFTPStatus := FTP_CONNECT;
         {Get the client address}
         isClientAddr := NetSocket.GetSockAddr;
         {Clear the display}
         mleReceived.Clear;
         mleSent.Clear;
      end;
   end
   else
   begin
      {You have too many connections already, so accept the connection
       with the rejection socket}
      isRejectSocket := NetSocket.AcceptConnectingSocket(isListenSocket,
                         handle, UWM_REJECTSOCKETEVENT);
   end;
end;

{This procedure sends a rejection message to a client application, then 
disconnects the client}
procedure TFTPServe.SendConnectReject;
begin
   {Has the rejection message been sent yet?}
   if (not ibRejectSent) then
   begin
      {Send the rejection message}
      NetSocket.SocketSend(isRejectSocket, handle, UWM_REJECTSOCKETEVENT,
          '421 Unable to open connection at this time', SKT_NOOP);
      ibRejectSent := TRUE;
   end
   else
   begin
      {Close the connection}
      NetSocket.CloseSocketConnection(isRejectSocket);
      ibRejectSent := FALSE;
   end;
end;

procedure TFTPServe.SendMsg(aSendSock:_SOCKET; iMsg, iNewState, aiEvent: Integer; asAdditionalMsg: String);
var
   lsMsgStr: String;
begin
   {Which message needs to be sent}
   case iMsg of
      {Default messages for each FTP response code}
      110: lsMsgStr := '110 Restart Marker ';
      120: lsMsgStr := '120 Service not ready';
      125: lsMsgStr := '125 Data connection already open, starting transfer';
      150: lsMsgStr := '150 File OK, About to open data connection';
      200: lsMsgStr := '200 ';
      202: lsMsgStr := '202 Command not needed at this site';
      211: lsMsgStr := '211 ';
      212: lsMsgStr := '212 ';
      213: lsMsgStr := '213 ';
      214: lsMsgStr := '214 ';
      215: lsMsgStr := '215 WIN32';
      220: lsMsgStr := '220 Welcome to Delphi FTP v1.0';
      221: lsMsgStr := '221 Closing connection, Goodbye';
      225: lsMsgStr := '225 Data connection open, no transfer in progress';
      226: lsMsgStr := '226 Closing data connection, File transfer successful';
      227: lsMsgStr := '227 Entering passive Mode - ';
      230: lsMsgStr := '230 ';
      250: lsMsgStr := '250 Command complete';
      257: lsMsgStr := '257 ';
      331: lsMsgStr := '331 User name OK, Need password';
      332: lsMsgStr := '332 Need account for login';
      350: lsMsgStr := '350 Action pending, continue';
      421: lsMsgStr := '421 Service not available';
      425: lsMsgStr := '425 Can not open the data connection';
      426: lsMsgStr := '426 Connection closed, transfer aborted';
      450: lsMsgStr := '450 File not available';
      451: lsMsgStr := '451 Local error, action aborted';
      452: lsMsgStr := '452 Insufficient storage space';
      500: lsMsgStr := '500 Syntax error, command unrecognized';
      501: lsMsgStr := '501 Syntax error in arguments';
      502: lsMsgStr := '502 Command not available';
      503: lsMsgStr := '503 Bad command sequence';
      504: lsMsgStr := '504 User not logged in';
      530: lsMsgStr := '530 Parameter not implemented for command';
      532: lsMsgStr := '532 Account needed to store files';
      550: lsMsgStr := '550 File unavailable';
      551: lsMsgStr := '551 Action aborted, page type unknown';
      552: lsMsgStr := '552 Action aborted, storage allocation exceeded';
      553: lsMsgStr := '553 File name not allowed';
   end;
   {Send the default message with any additional message to the client}
   NetSocket.SocketSend(aSendSock, handle, aiEvent, lsMsgStr + asAdditionalMsg, SKT_NOOP);
   {Display the message for the user to see}
   mleSent.Lines.Add(lsMsgStr + asAdditionalMsg);
   iiFTPStatus := iNewState;
end;

{This procedure sends the appropriate message and response code for
commands that the client has sent}
procedure TFTPServe.SendContextMsg;
begin
   {Which command did you received?}
   case iiFTPStatus of
      {Send the appropriate response}
      FTP_NOOP:;
      FTP_CONNECT: SendMsg(isConnectionSocket, 220, FTP_NOOP,
                                               UWM_SOCKETEVENT, isSendMsg);
      FTP_USER: SendMsg(isConnectionSocket, 230, FTP_NOOP,
                                               UWM_SOCKETEVENT, isSendMsg);
      FTP_PASS: SendMsg(isConnectionSocket, 230, FTP_NOOP,
                                               UWM_SOCKETEVENT, isSendMsg);
      FTP_CWD: SendMsg(isConnectionSocket, 550, FTP_NOOP,
                                               UWM_SOCKETEVENT, isSendMsg);
      FTP_CDUP: SendMsg(isConnectionSocket, 550, FTP_NOOP,
                                               UWM_SOCKETEVENT, isSendMsg);
      FTP_REIN: SendMsg(isConnectionSocket, 502, FTP_NOOP,
                                               UWM_SOCKETEVENT, isSendMsg);
      FTP_QUIT: SendMsg(isConnectionSocket, 221, FTP_NOOP,
                                               UWM_SOCKETEVENT, isSendMsg);
      FTP_PORT: SendMsg(isConnectionSocket, 200, FTP_NOOP,
                                               UWM_SOCKETEVENT, isSendMsg);
      FTP_PASV: SendMsg(isConnectionSocket, 502, FTP_NOOP,
                                               UWM_SOCKETEVENT, isSendMsg);
      FTP_TYPE: SendMsg(isConnectionSocket, 200, FTP_NOOP,
                                               UWM_SOCKETEVENT, isSendMsg);
      FTP_STRU: SendMsg(isConnectionSocket, 502, FTP_NOOP,
                                               UWM_SOCKETEVENT, isSendMsg);
      FTP_MODE: SendMsg(isConnectionSocket, 200, FTP_NOOP,
                                               UWM_SOCKETEVENT, isSendMsg);
      FTP_RETR: SendMsg(isConnectionSocket, 502, FTP_NOOP,
                                               UWM_SOCKETEVENT, isSendMsg);
      FTP_STOR: SendMsg(isConnectionSocket, 502, FTP_NOOP,
                                               UWM_SOCKETEVENT, isSendMsg);
      FTP_STOU: SendMsg(isConnectionSocket, 502, FTP_NOOP,
                                               UWM_SOCKETEVENT, isSendMsg);
      FTP_APPE: SendMsg(isConnectionSocket, 502, FTP_NOOP,
                                               UWM_SOCKETEVENT, isSendMsg);
      FTP_REST: SendMsg(isConnectionSocket, 502, FTP_NOOP,
                                               UWM_SOCKETEVENT, isSendMsg);
      FTP_RNFR: SendMsg(isConnectionSocket, 502, FTP_NOOP,
                                               UWM_SOCKETEVENT, isSendMsg);
      FTP_RNTO: SendMsg(isConnectionSocket, 502, FTP_NOOP,
                                               UWM_SOCKETEVENT, isSendMsg);
      FTP_ABOR: SendMsg(isConnectionSocket, 502, FTP_NOOP,
                                               UWM_SOCKETEVENT, isSendMsg);
      FTP_DELE: SendMsg(isConnectionSocket, 550, FTP_NOOP,
                                               UWM_SOCKETEVENT, isSendMsg);
      FTP_RMD: SendMsg(isConnectionSocket, 550, FTP_NOOP,
                                               UWM_SOCKETEVENT, isSendMsg);
      FTP_MKD: SendMsg(isConnectionSocket, 550, FTP_NOOP,
                                               UWM_SOCKETEVENT, isSendMsg);
      FTP_PWD: SendMsg(isConnectionSocket, 257, FTP_NOOP,
                                               UWM_SOCKETEVENT, isSendMsg);
      FTP_LIST: SendMsg(isConnectionSocket, 226, FTP_NOOP,
                                               UWM_SOCKETEVENT, isSendMsg);
      FTP_NLST: SendMsg(isConnectionSocket, 226, FTP_NOOP,
                                               UWM_SOCKETEVENT, isSendMsg);
      FTP_OK: SendMsg(isConnectionSocket, 200, FTP_NOOP,
                                               UWM_SOCKETEVENT, isSendMsg);
      FTP_SYST: SendMsg(isConnectionSocket, 215, FTP_NOOP,
                                               UWM_SOCKETEVENT, isSendMsg);
      FTP_UNKWN: SendMsg(isConnectionSocket, 500, FTP_NOOP,
                                               UWM_SOCKETEVENT, isSendMsg);
   end;
   isSendMsg := '';
end;

procedure TFTPServe.ValidateLogin(asUserName: String);
begin
   isSendMsg := 'User Login ' + asUserName + ' OK';
   SendContextMsg;
end;

procedure TFTPServe.ValidatePassword(asPassword: String);
begin
   isSendMsg := 'User Login OK';
   SendContextMsg;
end;

procedure TFTPServe.SendWorkingDirectory;
begin
   {Get the current directory from the TFileListBox control}
   isSendMsg := flbCurDir.Directory;
   {Send the appropriate message to the client}
   SendContextMsg;
end;

procedure TFTPServe.CheckModeRequest(asMode: String);
begin
   {Is the requested mode "Stream"?}
   if (asMode = 'S') then
   begin
      {Yes, send a success response message}
      isSendMsg := 'Stream Mode Enabled';
      SendContextMsg;
   end
   else
   begin
      {No, send an error response message}
      isSendMsg := 'Mode ' + asMode + ' not supported';
      SendMsg(isConnectionSocket, 502, FTP_NOOP, UWM_SOCKETEVENT, isSendMsg);
   end;
end;

procedure TFTPServe.CheckTypeRequest(asType: String);
begin
   {Is the requested type "ASCII" or "IMAGE"?}
   if ((asType = 'A') or (asType = 'I')) then
   begin
      {Yes, send a success response message}
      if (asType = 'A') then
         isSendMsg := 'ASCII Type Enabled'
      else
         isSendMsg := 'Image Type Enabled';
      SendContextMsg;
   end
   else
   begin
      {No, send an error response message}
      isSendMsg := 'Type ' + asType + ' not supported';
      SendMsg(isConnectionSocket, 502, FTP_NOOP, UWM_SOCKETEVENT, isSendMsg);
   end;
end;

procedure TFTPServe.CheckPortRequest(asPort: String);
begin
   {Check the specified port value}
   isClientAddr := asPort;
   {Send a response, acknowledging the new port value}
   isSendMsg := 'Client Data Port Now ' + asPort;
   SendContextMsg;
end;

procedure TFTPServe.SendDirectoryList;
begin
   {Send the "preparing to connect data socket" message}
   SendMsg(isConnectionSocket, 150, FTP_NOOP,UWM_SOCKETEVENT, isSendMsg);
   {Save the TFileListBox control's contents to a temporary file}
   isFileName := '\tmp.lst';
   flbCurDir.Items.SaveToFile(isFileName);
   {Create the data socket}
   isDataSocket := NetSocket.CreateDataSocket(isConnectionSocket,
                   handle, UWM_DATASOCKETEVENT);
   iiDataConnStatus := FTPD_SEND;
   {Open the data socket connection so that you can send the file}
   OpenDataConnection(isClientAddr);
end;

procedure TFTPServe.SendRecvFile(asFileName: String);
begin
   {Send the "preparing to connect data socket" message}
   SendMsg(isConnectionSocket, 150, FTP_NOOP,UWM_SOCKETEVENT, isSendMsg);
   {Place the received file name into the transfer file name variable}
   isFileName := asFileName;
   {Create the data socket}
   isDataSocket := NetSocket.CreateDataSocket(isConnectionSocket,
                   handle, UWM_DATASOCKETEVENT);
   iiDataConnStatus := FTPD_SEND;
   {Open the data socket connection so that the file can be sent}
   OpenDataConnection(isClientAddr);
end;

procedure TFTPServe.ReceiveStoreFile(asFileName: String);
begin
   {Send the "preparing to connect data socket" message}
   SendMsg(isConnectionSocket, 150, FTP_NOOP,UWM_SOCKETEVENT, isSendMsg);
   {Place the received file name into the transfer file name variable}
   isFileName := asFileName;
   {Create the data socket}
   isDataSocket := NetSocket.CreateDataSocket(isConnectionSocket,
                   handle, UWM_DATASOCKETEVENT);
   iiDataConnStatus := FTPD_RECV;
   {Open the data socket connection so that the file can be received}
   OpenDataConnection(isClientAddr);
end;

procedure TFTPServe.CheckCommandMsg(asCommand, asArguments: String);
var
   lbCmdFound: Boolean;
begin
   lbCmdFound := FALSE;
   {Did you receive a NOOP command?}
   if (asCommand = 'NOOP') then
   begin
      lbCmdFound := TRUE;
      iiFTPStatus := FTP_OK;
      SendContextMsg;
   end;
   {Did you receive a USER command?}
   if (asCommand = 'USER') then
   begin
      lbCmdFound := TRUE;
      iiFTPStatus := FTP_USER;
      {Validate the login ID}
      ValidateLogin(asArguments);
   end;
   {Did you receive a PASS command?}
   if (asCommand = 'PASS') then
   begin
      lbCmdFound := TRUE;
      iiFTPStatus := FTP_PASS;
      {Validate the Password}
      ValidatePassword(asArguments);
   end;
   {Did you receive a CWD command?}
   if (asCommand = 'CWD') then
   begin
      lbCmdFound := TRUE;
      iiFTPStatus := FTP_CWD;
      SendContextMsg;
   end;
   {Did you receive a CDUP command?}
   if (asCommand = 'CDUP') then
   begin
      lbCmdFound := TRUE;
      iiFTPStatus := FTP_CDUP;
      SendContextMsg;
   end;
   {Did you receive a REIN command?}
   if (asCommand = 'REIN') then
   begin
      lbCmdFound := TRUE;
      iiFTPStatus := FTP_REIN;
      SendContextMsg;
   end;
   {Did you receive a QUIT command?}
   if (asCommand = 'QUIT') then
   begin
      lbCmdFound := TRUE;
      iiFTPStatus := FTP_QUIT;
      SendContextMsg;
      {Close the connection}
      NetSocket.CloseSocketConnection(isConnectionSocket);
      Dec(iiNumConnected);
      ibConnected := FALSE;
   end;
   {Did you receive a PORT command?}
   if (asCommand = 'PORT') then
   begin
      lbCmdFound := TRUE;
      iiFTPStatus := FTP_PORT;
      {Check the port request}
      CheckPortRequest(asArguments);
   end;
   {Did you receive a PASV command?}
   if (asCommand = 'PASV') then
   begin
      lbCmdFound := TRUE;
      iiFTPStatus := FTP_PASV;
      SendContextMsg;
   end;
   {Did you receive a TYPE command?}
   if (asCommand = 'TYPE') then
   begin
      lbCmdFound := TRUE;
      iiFTPStatus := FTP_TYPE;
      {Check the type request}
      CheckTypeRequest(asArguments);
   end;
   {Did you receive a STRU command?}
   if (asCommand = 'STRU') then
   begin
      lbCmdFound := TRUE;
      iiFTPStatus := FTP_STRU;
      SendContextMsg;
   end;
   {Did you receive a MODE command?}
   if (asCommand = 'MODE') then
   begin
      lbCmdFound := TRUE;
      iiFTPStatus := FTP_MODE;
      {Check the mode request}
      CheckModeRequest(asArguments);
   end;
   {Did you receive a RETR command?}
   if (asCommand = 'RETR') then
   begin
      lbCmdFound := TRUE;
      iiFTPStatus := FTP_RETR;
      {Send the requested file}
      SendRecvFile(asArguments);
   end;
   {Did you receive a STOR command?}
   if (asCommand = 'STOR') then
   begin
      lbCmdFound := TRUE;
      iiFTPStatus := FTP_STOR;
      {Receive the file being sent}
      ReceiveStoreFile(asArguments);
   end;
   {Did you receive a STOU command?}
   if (asCommand = 'STOU') then
   begin
      lbCmdFound := TRUE;
      iiFTPStatus := FTP_STOU;
      SendContextMsg;
   end;
   {Did you receive a APPE command?}
   if (asCommand = 'APPE') then
   begin
      lbCmdFound := TRUE;
      iiFTPStatus := FTP_APPE;
      SendContextMsg;
   end;
   {Did you receive a REST command?}
   if (asCommand = 'REST') then
   begin
      lbCmdFound := TRUE;
      iiFTPStatus := FTP_REST;
      SendContextMsg;
   end;
   {Did you receive a RNFR command?}
   if (asCommand = 'RNFR') then
   begin
      lbCmdFound := TRUE;
      iiFTPStatus := FTP_RNFR;
      SendContextMsg;
   end;
   {Did you receive a RNTO command?}
   if (asCommand = 'RNTO') then
   begin
      lbCmdFound := TRUE;
      iiFTPStatus := FTP_RNTO;
      SendContextMsg;
   end;
   {Did you receive a ABOR command?}
   if (asCommand = 'ABOR') then
   begin
      lbCmdFound := TRUE;
      iiFTPStatus := FTP_ABOR;
      SendContextMsg;
   end;
   {Did you receive a DELE command?}
   if (asCommand = 'DELE') then
   begin
      lbCmdFound := TRUE;
      iiFTPStatus := FTP_DELE;
      SendContextMsg;
   end;
   {Did you receive a RMD command?}
   if (asCommand = 'RMD') then
   begin
      lbCmdFound := TRUE;
      iiFTPStatus := FTP_RMD;
      SendContextMsg;
   end;
   {Did you receive a MKD command?}
   if (asCommand = 'MKD') then
   begin
      lbCmdFound := TRUE;
      iiFTPStatus := FTP_MKD;
      SendContextMsg;
   end;
   {Did you receive a PWD command?}
   if ((asCommand = 'PWD') or (asCommand = 'XPWD')) then
   begin
      lbCmdFound := TRUE;
      iiFTPStatus := FTP_PWD;
      {Send the current directory name}
      SendWorkingDirectory;
   end;
   {Did you receive a LIST command?}
   if (asCommand = 'LIST') then
   begin
      lbCmdFound := TRUE;
      iiFTPStatus := FTP_LIST;
      {Send the directory listing}
      SendDirectoryList;
   end;
   {Did you receive an NLST command?}
   if (asCommand = 'NLST') then
   begin
      lbCmdFound := TRUE;
      iiFTPStatus := FTP_NLST;
      {Send the directory listing}
      SendDirectoryList;
   end;
   {Did you receive a SYST command?}
   if (asCommand = 'SYST') then
   begin
      lbCmdFound := TRUE;
      iiFTPStatus := FTP_SYST;
      SendContextMsg;
   end;
   {Did you recognize the command?}
   if (not lbCmdFound) then
   begin
      iiFTPStatus := FTP_UNKWN;
      SendContextMsg;
   end;
end;

procedure TFTPServe.ReadCommandMsg;
var
   lsCommand: String;
   lsMessage: String;
   liCmdLength: Integer;
   liMsgLength: Integer;
begin
   {Receive the incoming data}
   liMsgLength := recv(isConnectionSocket, ReceiveBuffer, 8192, 0);
   {Did you receive anything?}
   if (liMsgLength = SOCKET_ERROR) and
         (WSAGetLastError <> WSAEWOULDBLOCK) then
      ShowMessage(NetSocket.SocketError)
   else
   begin
      {Get the first line}
      if NetSocket.GetLine(TempBuffer, ReceiveBuffer) then
      begin
         {Add the received line to the display}
         mleReceived.Lines.Add(StrPas(TempBuffer));
         {Is there anything in the hold buffer that you need to add
         in front of this line?}
         if (Length(isHoldBuffer) > 0) then
            lsMessage := isHoldBuffer + StrPas(TempBuffer)
         else
            lsMessage := StrPas(TempBuffer);
         {Reset the hold buffer}
         isHoldBuffer := '';
         {Find the end of the command}
         liCmdLength := Pos(' ', lsMessage);
         if (liCmdLength = 0) then
         begin
            {The entire line is the command}
            liCmdLength := Length(lsMessage);
            lsCommand := lsMessage;
            lsMessage := '';
         end
         else
         begin
            {Separate the command from the arguments}
            lsCommand := Copy(lsMessage, 1, (liCmdLength - 1));
            lsMessage := Copy(lsMessage, (liCmdLength + 1), Length(lsMessage));
         end;
         {Send the command and its arguments to the CheckCommandMsg procedure}
         CheckCommandMsg(lsCommand, lsMessage);
      end
      else
      begin
         {You have an incomplete line. Place it in the hold buffer and
         wait for the rest of the command line to arrive}
         if (Length(isHoldBuffer) > 0) then
            isHoldBuffer := isHoldBuffer + StrPas(TempBuffer)
         else
            isHoldBuffer := StrPas(TempBuffer);
      end;
   end;
end;

procedure TFTPServe.UWMSocketEvent(var Msg: TMessage);
begin
   {What socket event was received? Call the appropriate procedure}
   case Msg.LParamLo of
      FD_READ: ReadCommandMsg;
      FD_WRITE: SendContextMsg;
      FD_ACCEPT:;
      FD_CONNECT:;
      FD_CLOSE:;
   end;
end;

procedure TFTPServe.UWMDataSocketEvent(var Msg: TMessage);
begin
   {What socket event was received? Call the appropriate procedure}
   case Msg.LParamLo of
      FD_READ: ProcessDataReceive;
      FD_WRITE: ProcessDataSend;
      FD_OOB:;
      FD_ACCEPT: AcceptDataConnection;
      FD_CONNECT: ProcessDataConnect;
      FD_CLOSE: CloseDataConnection;
   end;
end;

procedure TFTPServe.UWMListenSocketEvent(var Msg: TMessage);
begin
   {What socket event was received? Call the appropriate procedure}
   case Msg.LParamLo of
      FD_READ:;
      FD_WRITE:;
      FD_OOB:;
      FD_ACCEPT: ProcessListenAccept;
      FD_CONNECT:;
      FD_CLOSE:;
   end;
end;

procedure TFTPServe.UWMRejectSocketEvent(var Msg: TMessage);
begin
   {What socket event was received? Call the appropriate procedure}
   case Msg.LParamLo of
      FD_READ:;
      FD_WRITE: SendConnectReject;
      FD_ACCEPT:;
      FD_CONNECT:;
      FD_CLOSE:;
   end;
end;

end.
