unit Socket;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, TSocketc, StdCtrls, ExtCtrls, NWinSock;

{ If Using the Delphi 32 Winsock.pas }
type
     _Socket = TSocket;
     fd_set = TFDSet;
     Pfd_set = PFDSet;
     timeval = TTimeVal;
     hostent = THostEnt;
     netent = TNetEnt;
     servent = TServEnt;
     protoent = TProtoEnt;
     in_addr = TInAddr;
     Pin_addr = PInAddr;
     Sockaddr_in = TSockAddrIn;
     PSockaddr_in = PSockAddrIn;
     WSADATA = TWSAData;
     LPWSADATA = PWSAData;
     sockaddr = TSockAddr;
     sockproto = TSockProto;
     linger = TLinger;
     LPSOCKADDR = ^sockaddr;
     LPSOCKADDR_IN = ^sockaddr_in;
     LPLINGER = ^linger;
     LPIN_ADDR = ^in_addr;
     LPFD_SET = ^fd_set;
     LPHOSTENT = ^hostent;
     LPSERVENT = ^servent;
     LPPROTOENT = ^protoent;
     LPTIMEVAL = ^timeval;
const
     FIONREAD = (IOC_OUT or ((4 and IOCPARM_MASK) SHL 16) or
           (102 SHL 8) or 127);
{ End of Delphi 32 Winsock.pas compatability type definitions }
const
  SKT_NOOP = 0;
  SKT_GETHOSTADDR = 1;
  SKT_CONNECTTOHOST = 2;
  SKT_SENDPACKET = 3;
  SKT_RECVPACKET = 4;
  UWM_SOCKETEVENT = WM_USER + 500;
 { DATE_TIME_REFERENCE_POINT: TDateTime = 2208988800;}

type
  TForm1 = class(TForm)
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    SocketDescription: TLabel;
    SystemState: TLabel;
    MaxNumSockets: TLabel;
    UDPPacketSize: TLabel;
    Label5: TLabel;
    HostName: TEdit;
    Label6: TLabel;
    HostResponse: TMemo;
    cbConnect: TButton;
    cbClose: TButton;
    Label7: TLabel;
    HostAddress: TEdit;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure ProcessHostLookup;
    procedure ProcessHostConnect;
    procedure ProcessHostSend;
    procedure ProcessHostReceive;
    procedure cbCloseClick(Sender: TObject);
    procedure cbConnectClick(Sender: TObject);
  private
    { Private declarations }
    isConnectionSocket: _SOCKET;
    procedure UWMSocketEvent(var Msg: TMessage); message UWM_SOCKETEVENT;
  public
    { Public declarations }
    NetSocket: _TSocket;
    ilInetTime: LongInt;
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}

function InTime(INetTime: LongInt): PChar; stdcall; far; external 'intime32';

procedure TForm1.FormCreate(Sender: TObject);
begin
     {Create the NetSocket class}
     NetSocket := _TSocket.Create;
     {Initialize the WinSock system}
     if not NetSocket.InitializeSocket then
     begin
        Close;
     end
     else
     begin
        {Retrieve and display the WinSock interface parameters}
        SocketDescription.Caption := NetSocket.GetSocketDescription;
        SystemState.Caption := NetSocket.GetSocketSystemState;
        MaxNumSockets.Caption := IntToStr(NetSocket.GetMaxNumSockets);
        UDPPacketSize.Caption := IntToStr(NetSocket.GetUDPPacketSize);
     end;
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
     {Cleanup the WinSock interface and system}
     NetSocket.ShutdownSocket;
     NetSocket.Free;
end;

procedure TForm1.ProcessHostLookup;
begin
   {We have received the host address, now we need to connect to it.}
   HostResponse.Lines.Add('Connecting To Host');
   if (NetSocket.ProcessHostAddress = TRUE) then
   begin
      {Display the host TCP/IP address}
      hostAddress.text := NetSocket.GetHostAddressString;
      {Connect to the host computer}
      isConnectionSocket := NetSocket.OpenSocketConnection(Handle, UWM_SOCKETEVENT, SKT_CONNECTTOHOST, IPPORT_TIMESERVER);
   end
   else
      HostResponse.Lines.Add('Cannot Find Host Address');
end;

procedure TForm1.ProcessHostConnect;
var
   lsErrorString: String;
begin
   {We have connected to the other computer, so update the display}
   HostResponse.Lines.Add('Sending To Host');
   NetSocket.SetSocketState(SKT_SENDPACKET);
   ProcessHostSend;
end;

procedure TForm1.ProcessHostSend;
var
   liBytesReceived: Integer;
begin
   {We don't really need to send anything}
   HostResponse.Lines.Add('Receiving From Host');
   NetSocket.SetSocketState(SKT_RECVPACKET);
end;

procedure TForm1.ProcessHostReceive;
var
   lsRetnStr: PChar;
   llINetTime, llIntelTime: LongInt;
begin
   {Receive the current system time from the host computer}
   recv(isConnectionSocket, ilINetTime, sizeof(LongInt), 0);
   if (ilINetTime <> 0) then
   begin
      {Convert it into a string for display purposes}
      llIntelTime := ntohl(ilINetTime);
      lsRetnStr := InTime(llIntelTime);
      lsRetnStr[(StrLen(lsRetnStr) - 1)] := #0;
      HostResponse.Lines.Add(StrPas(lsRetnStr));
      NetSocket.SetSocketState(SKT_NOOP);
      {Close the connection}
      NetSocket.CloseSocketConnection(isConnectionSocket);
   end;
end;

procedure TForm1.UWMSocketEvent(var Msg: TMessage);
begin
     {What is the current state of the socket? Call the appropriate procedure}
     case NetSocket.GetSocketState of
          SKT_GETHOSTADDR: ProcessHostLookup;
          SKT_CONNECTTOHOST: ProcessHostConnect;
          SKT_SENDPACKET: ProcessHostSend;
          SKT_RECVPACKET: ProcessHostReceive;
     end;
end;

procedure TForm1.cbCloseClick(Sender: TObject);
begin
     {Close the application}
     Close;
end;

procedure TForm1.cbConnectClick(Sender: TObject);
begin
   {Has the user entered a host name to connect to?}
   if Length(HostName.text) > 0 then
   begin
      {Clear the display and look-up the host address}
      HostResponse.Clear;
      HostResponse.Lines.Add('Looking Up Host Address');
      NetSocket.GetHostAddress(Handle, UWM_SOCKETEVENT, HostName.text, SKT_GETHOSTADDR);
   end;
end;

end.
