unit wserve;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  TSocketc, NWinSock, StdCtrls;

{ If Using the Delphi 32 Winsock.pas }
type
     _Socket = TSocket;
     fd_set = TFDSet;
     Pfd_set = PFDSet;
     timeval = TTimeVal;
     hostent = THostEnt;
     netent = TNetEnt;
     servent = TServEnt;
     protoent = TProtoEnt;
     in_addr = TInAddr;
     Pin_addr = PInAddr;
     Sockaddr_in = TSockAddrIn;
     PSockaddr_in = PSockAddrIn;
     WSADATA = TWSAData;
     LPWSADATA = PWSAData;
     sockaddr = TSockAddr;
     sockproto = TSockProto;
     linger = TLinger;
     LPSOCKADDR = ^sockaddr;
     LPSOCKADDR_IN = ^sockaddr_in;
     LPLINGER = ^linger;
     LPIN_ADDR = ^in_addr;
     LPFD_SET = ^fd_set;
     LPHOSTENT = ^hostent;
     LPSERVENT = ^servent;
     LPPROTOENT = ^protoent;
     LPTIMEVAL = ^timeval;
{ End of Delphi 32 Winsock.pas compatability type definitions }
const
  SKT_NOOP = 0;
  SKT_GETHOSTADDR = 1;
  SKT_CONNECTTOHOST = 2;
  SKT_SENDPACKET = 3;
  SKT_RECVPACKET = 4;
  HTTP_NOOP = 0;
  HTTP_CONNECT = 1;
  HTTP_GET = 2;
  HTTP_HEAD = 3;
  HTTP_POST = 4;
  HTTP_UNKWN = 29;
  UWM_SOCKETEVENT = WM_USER + 500;
  UWM_LISTENSOCKETEVENT = WM_USER + 501;
  UWM_REJECTSOCKETEVENT = WM_USER + 502;
type
  TWbServe = class(TForm)
    cbClose: TButton;
    mleReceived: TMemo;
    mleSent: TMemo;
    Label1: TLabel;
    Label2: TLabel;
    sleDefDir: TEdit;
    sleDefPage: TEdit;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure cbCloseClick(Sender: TObject);
    procedure ClearBuffer(DirtyBuffer: Pchar; BufLength: Integer);
    procedure OpenSendFile;
    procedure ProcessListenAccept;
    procedure SendConnectReject;
    procedure CloseConnection;
    procedure ProcessDataSend;
    function FileSizeHeader: String;
    function FileTypeHeader: String;
    function FileHeader: String;
    procedure SendFullMsg(aSendSock:_SOCKET; iMsg, iNewState, aiEvent: Integer;
                                             asAdditionalMsg: String);
    procedure SendContextMsg;
    procedure ConvertFileNameToDos;
    procedure SendFullObject;
    procedure SendShortObject;
    procedure SendFullHeader;
    procedure SendShortHeader;
    procedure CheckCommandMsg(asCommand, asArguments: String);
    procedure ReadCommandMsg;
  private
    { Private declarations }
    {Primary connection socket}
    isConnectionSocket: _SOCKET;
    {Socket for listening for incoming connections}
    isListenSocket: _SOCKET;
    {Socket for quickly rejecting an incoming connection}
    isRejectSocket: _SOCKET;
    {A boolean flag to indicate whether the connection socket is connected
    to another computer}
    ibConnected: Boolean;
    {A flag used to indicate the state of the HTTP conversation}
    iiHTTPStatus: Integer;
    {The hold buffer for use when receiving a long message from the client
    computer}
    isHoldBuffer: String;
    {The number of current connections}
    iiNumConnected: Integer;
    {A flag to indicate whether the rejection has been sent}
    ibRejectSent: Boolean;
    {A string to hold the message to be sent to the client}
    isSendMsg: String;
    {A flag to indicate whether the requested file is open}
    ibFileOpen: boolean;
    {The name of the requested file}
    isFileName: String;
    {The size of the requested file}
    ilFileSize: LongInt;
    {The amount of data in the send buffer}
    iiBufferAmt: Integer;
    {The address if the client}
    isClientAddr: String;
    {A flag to indicate whether the request is a full (HTTP 1.0) or short 
    (HTTP 0.9) request}
    ibIsFullRequest: Boolean;
    {The socket event handlers}
    procedure UWMSocketEvent(var Msg: TMessage); message UWM_SOCKETEVENT;
    procedure UWMListenSocketEvent(var Msg: TMessage); message UWM_LISTENSOCKETEVENT;
    procedure UWMRejectSocketEvent(var Msg: TMessage); message UWM_REJECTSOCKETEVENT;
  public
    { Public declarations }
    NetSocket: _TSocket;
    ReceiveBuffer: array[0..8192] of Char;
    TempBuffer: array[0..8192] of Char;
    TransferBuffer: array[0..8192] of Char;
    TransferFile: File of Byte;
  end;

var
  WbServe: TWbServe;

implementation

{$R *.DFM}

procedure TWbServe.FormCreate(Sender: TObject);
begin
   {Initialize all of the appropriate flags}
   ibConnected := FALSE;
   ibRejectSent := FALSE;
   ibIsFullRequest := FALSE;
   iiHTTPStatus := HTTP_NOOP;
   iiNumConnected := 0;
   {Create the NetSocket class}
   NetSocket := _TSocket.Create;
   {Initialize the WinSock system}
   if not NetSocket.InitializeSocket then
      Close
   else
   begin
      {Create a listening socket on TCP service port 80}
      isListenSocket := NetSocket.CreateListenSocket(handle,
                        UWM_LISTENSOCKETEVENT, 80);
      if (isListenSocket = INVALID_SOCKET) then
         Close;
   end;
end;

procedure TWbServe.FormDestroy(Sender: TObject);
begin
     {Clean up before shutting down}
     NetSocket.ShutdownSocket;
     NetSocket.Free;
end;

procedure TWbServe.cbCloseClick(Sender: TObject);
begin
   {Closing down, close any open sockets}
   NetSocket.CloseSocketConnection(isListenSocket);
   if ibConnected then
      NetSocket.CloseSocketConnection(isConnectionSocket);
   Close;
end;

{This procedure cleans the buffer passed to it, filling it with null characters}
procedure TWbServe.ClearBuffer(DirtyBuffer: Pchar; BufLength: Integer);
var
   liCurPos: Integer;
   lcNullChar: Char;
begin
   liCurPos := 0;
   lcNullChar := Char(0);
   while (liCurPos < BufLength) do
   begin
      DirtyBuffer^ := lcNullChar;
      Inc(liCurPos);
      Inc(DirtyBuffer);
   end;
end;

procedure TWbServe.OpenSendFile;
begin
   {Set the FILE OPEN flag to TRUE}
   ibFileOpen := TRUE;
   {Assign the file name to the file handle}
   AssignFile(TransferFile, isFileName);
   {Reset and open the file for reading}
   Reset(TransferFile);
   {Read the file size into the file size variable}
   ilFileSize := FileSize(TransferFile);
end;

procedure TWbServe.ProcessListenAccept;
begin
   {Are there no current connections?}
   if (iiNumConnected = 0) then
   begin
      {Yes, accept this connection and send a greeting}
      isConnectionSocket := NetSocket.AcceptConnectingSocket(isListenSocket,
                         handle, UWM_SOCKETEVENT);
      if (isConnectionSocket <> INVALID_SOCKET) then
      begin
         {Increment the number of current connections}
         Inc(iiNumConnected);
         iiHTTPStatus := HTTP_CONNECT;
         {Get the client address}
         isClientAddr := NetSocket.GetSockAddr;
         {Clear the display}
         mleReceived.Clear;
         mleSent.Clear;
      end;
   end
   else
   begin
      {You have too many connections already, so accept the connection
       with the rejection socket}
      isRejectSocket := NetSocket.AcceptConnectingSocket(isListenSocket,
                         handle, UWM_REJECTSOCKETEVENT);
   end;
end;

{This procedure sends a rejection message to a client application, then 
disconnects the client}
procedure TWbServe.SendConnectReject;
begin
   {Has the rejection message been sent yet?}
   if (not ibRejectSent) then
   begin
      {Send the rejection message}
      NetSocket.SocketSend(isRejectSocket, handle, UWM_REJECTSOCKETEVENT,
          'HTTP/1.0 503 Unable to open connection at this time' + #13#10, SKT_NOOP);
      ibRejectSent := TRUE;
   end
   else
   begin
      {Close the connection}
      NetSocket.CloseSocketConnection(isRejectSocket);
      ibRejectSent := FALSE;
   end;
end;

procedure TWbServe.CloseConnection;
begin
  {Before closing the socket connection, shut it down}
  shutdown(isConnectionSocket, 2);
  {Close the socket connection and destroy the socket}
  NetSocket.CloseSocketConnection(isConnectionSocket);
  {If the requested file is open, close it}
  if ibFileOpen then
  begin
     CloseFile(TransferFile);
     ibFileOpen := FALSE;
  end;
  {Decrement the number of current connections, so that more incoming
  connection requests can be accepted}
  Dec(iiNumConnected);
end;

procedure TWbServe.ProcessDataSend;
var
   liCurByteCount: Integer;
   lcCurByte: Byte;
   lpTransferDataPointer: PChar;
begin
   {Is the requested file open?}
   if ibFileOpen then
   begin
      {Initialize the byte counter}
      liCurByteCount := 0;
      {Are you at the end of the file?}
      if (not Eof(TransferFile)) then
      begin
         {No, clear the transfer buffer}
         ClearBuffer(TransferBuffer, 8192);
         {Read 4,096 bytes from the file into the transfer buffer}
         while ((liCurByteCount < 4096) and (not Eof(TransferFile))) do
         begin
            Read(TransferFile, lcCurByte);
            TransferBuffer[liCurByteCount] := Char(lcCurByte);
            Inc(liCurByteCount);
         end;
         {Send what you've read from the file}
         NetSocket.SocketSendBuffer(isConnectionSocket, handle, UWM_SOCKETEVENT,
                  TransferBuffer, liCurByteCount, SKT_SENDPACKET);
      end
      else
         {You've sent the entire file, so close the connection}
         CloseConnection;
   end;
end;

function TWbServe.FileSizeHeader: String;
begin
   {Build the file size message header line}
   FileSizeHeader := 'Content-Length: ' + IntToStr(ilFileSize) + #13#10;
end;

function TWbServe.FileTypeHeader: String;
var
   lsFileExt: String;
   lsFileType: String;
   liPosPeriod: Integer;
begin
   {Specify a default file type of HTML text}
   lsFileType := 'text/html';
   {Locate the extension in the file name}
   liPosPeriod := Pos('.', isFileName);
   if (liPosPeriod > 0) then
   begin
      {Is it a .GIF or JPEG file?}
      lsFileExt := Copy(isFileName, (liPosPeriod + 1), 3);
      if ((lsFileExt = 'gif') or (lsFileExt = 'GIF')) then
         lsFileType := 'image/gif';
      if ((lsFileExt = 'jpg') or (lsFileExt = 'JPG')) then
         lsFileType := 'image/jpeg';
   end;
   {Build the file content-type header line}
   FileTypeHeader := 'Content-Type: ' + lsFileType + #13#10;
end;

function TWbServe.FileHeader: String;
var
   lsHeaderStr: String;
begin
   {Put all three of these lines together into a message header section}
   lsHeaderStr := 'Allow: GET, HEAD' + #13#10;
   lsHeaderStr := lsHeaderStr + FileSizeHeader + FileTypeHeader;
   FileHeader := lsHeaderStr;
end;

procedure TWbServe.SendFullMsg(aSendSock:_SOCKET; iMsg, iNewState, aiEvent: Integer; asAdditionalMsg: String);
var
   lsMsgStr: String;
begin
   {Initialize the message with the specified response code}
   case iMsg of
      200: lsMsgStr := 'HTTP/1.0 200 OK' + #13#10;
      201: lsMsgStr := 'HTTP/1.0 201 Created' + #13#10;
      202: lsMsgStr := 'HTTP/1.0 202 Accepted' + #13#10;
      204: lsMsgStr := 'HTTP/1.0 204 No Content' + #13#10;
      301: lsMsgStr := 'HTTP/1.0 301 Moved Permanently' + #13#10;
      302: lsMsgStr := 'HTTP/1.0 302 Moved Temporarily' + #13#10;
      304: lsMsgStr := 'HTTP/1.0 304 Not Modified' + #13#10;
      400: lsMsgStr := 'HTTP/1.0 400 Bad Request' + #13#10;
      401: lsMsgStr := 'HTTP/1.0 401 Unauthorized' + #13#10;
      403: lsMsgStr := 'HTTP/1.0 403 Forbidden' + #13#10;
      404: lsMsgStr := 'HTTP/1.0 404 Not Found' + #13#10;
      500: lsMsgStr := 'HTTP/1.0 500 Internal Server Error' + #13#10;
      501: lsMsgStr := 'HTTP/1.0 501 Not Implemented' + #13#10;
      502: lsMsgStr := 'HTTP/1.0 502 Bad Gateway' + #13#10;
      503: lsMsgStr := 'HTTP/1.0 503 Service Unavailable' + #13#10;
   end;
   {Add the current date and time}
   lsMsgStr := lsMsgStr + 'Date: ' +
               FormatDateTime('ddd, d mmm yyyy hh:mm:ss "GMT"', Now());
   {Send what you have so far, then add them to the display}
   NetSocket.SocketSend(aSendSock, handle, aiEvent, lsMsgStr, SKT_NOOP);
   mleSent.Lines.Add(lsMsgStr);
   {Build two more header lines, telling the client what server software
   is servicing this request}
   lsMsgStr := 'MIME-Version: 1.0' + #13#10;
   lsMsgStr := lsMsgStr + 'Server: Delphi Web Server v1.0' + #13#10;
   {Send these two lines, plus what was passed in as the additional
   header information lines, and add them to the display}
   NetSocket.SocketSend(aSendSock, handle, aiEvent, lsMsgStr + asAdditionalMsg, SKT_NOOP);
   mleSent.Lines.Add(lsMsgStr + asAdditionalMsg);
   {Reset the current command status flag to the specified state value}
   iiHTTPStatus := iNewState;
end;

procedure TWbServe.SendContextMsg;
begin
   {What is the requested command?}
   case iiHTTPStatus of
      {NOOP, GET, and HEAD, don't do anything}
      HTTP_NOOP:;
      HTTP_GET:;
      HTTP_HEAD:;
      {POST and Unknown, send 401 and 400 responses accordingly}
      HTTP_POST: SendFullMsg(isConnectionSocket, 401, HTTP_NOOP, UWM_SOCKETEVENT, '');
      HTTP_UNKWN: SendFullMsg(isConnectionSocket, 400, HTTP_NOOP, UWM_SOCKETEVENT, '');
   end;
end;

procedure TWbServe.ConvertFileNameToDos;
var
   liSlashPos: Integer;
begin
   {Find the first slash character}
   liSlashPos := Pos('/', isFileName);
   {Loop as long as there are more slash characters}
   while (liSlashPos > 0) do
   begin
      {Copy the file name, replacing the slash character with the 
      backslash character}
      isFileName := Copy(isFileName, 1, (liSlashPos - 1)) + '\' +
                    Copy(isFileName, (liSlashPos + 1), Length(isFileName));
      liSlashPos := Pos('/', isFileName);
   end;
end;

procedure TWbServe.SendFullObject;
begin
   {Does the requested file exist?}
   if FileExists(isFileName) then
   begin
      {Yes, open the file and send the full header message, passing the
      FileHeader function as the additional message string parameter to
      the SendFullMsg procedure}
      OpenSendFile;
      SendFullMsg(isConnectionSocket, 200, HTTP_NOOP, UWM_SOCKETEVENT, FileHeader);
      {Now, send the file}
      ProcessDataSend;
   end
   else
   begin
      {No, send a full message informing the client that the requested
      file does not exist on this server}
      SendFullMsg(isConnectionSocket, 404, HTTP_NOOP, UWM_SOCKETEVENT, isFileName);
      {Close the connection}
      CloseConnection;
   end;
end;

procedure TWbServe.SendShortObject;
var
   isShortObjectMsg: String;
begin
   {Does the requested file exist?}
   if FileExists(isFileName) then
   begin
      {Yes, open and send it}
      OpenSendFile;
      ProcessDataSend;
   end
   else
   begin
      {No, build an HTML document stating that the requested file
      was not found}
      isShortObjectMsg := '<TITLE>Not Found</TITLE>';
      {Send the message to the client}
      NetSocket.SocketSend(isConnectionSocket, handle, UWM_SOCKETEVENT,
                                            isShortObjectMsg, SKT_NOOP);
      {Add the message to the display}
      mleSent.Lines.Add(isShortObjectMsg);
      {Close the connection}
      CloseConnection;
   end;
end;

procedure TWbServe.SendFullHeader;
begin
   {Does the requested file exist?}
   if FileExists(isFileName) then
   begin
      {Yes, open the file and send the full header message, passing the
      FileHeader function as the additional message string parameter to
      the SendFullMsg procedure}
      OpenSendFile;
      SendFullMsg(isConnectionSocket, 200, HTTP_NOOP, UWM_SOCKETEVENT, FileHeader);
      {Close the connection}
      CloseConnection;
   end
   else
   begin
      {No, send a full message informing the client that the requested
      file does not exist on this server}
      SendFullMsg(isConnectionSocket, 404, HTTP_NOOP, UWM_SOCKETEVENT, isFileName);
      {Close the connection}
      CloseConnection;
   end;
end;

procedure TWbServe.SendShortHeader;
var
   isShortHeaderMsg: String;
begin
   {Build the response document informing the client that the HEAD
   command is not allowed in the HTTP 0.9 protocol}
   isShortHeaderMsg := '<TITLE>Operation Not Allowed</TITLE>';
   {Send the message to the client}
   NetSocket.SocketSend(isConnectionSocket, handle, UWM_SOCKETEVENT,
                                            isShortHeaderMsg, SKT_NOOP);
   {Add the message to the display}
   mleSent.Lines.Add(isShortHeaderMsg);
   {Close the connection}
   CloseConnection
end;

procedure TWbServe.CheckCommandMsg(asCommand, asArguments: String);
var
   lbCmdFound: Boolean;
   liSpacePos: Integer;
begin
   {Initialize the command found flag to FALSE}
   lbCmdFound := FALSE;
   {Is the HTTP 1.0 indicator included?}
   if (Pos('HTTP/1.0', asArguments) > 0) then
   begin
      {Yes, mark this request as a full (HTTP 1.0) request}
      ibIsFullRequest := TRUE;
      {Now trim the HTTP 1.0 indicator from the file name}
      liSpacePos := Pos(' ', asArguments);
      if (liSpacePos > 0) then
         isFileName := Copy(asArguments, 1, (liSpacePos - 1));
   end
   else
   begin
      {No HTTP 1.0 indicator, so you must treat this request
       as an HTTP 0.9 request}
      ibIsFullRequest := FALSE;
      isFileName := asArguments;
   end;
   {Was the default page requested?}
   if (isFileName = '/') then
      isFileName := sleDefPage.Text
   else
      {No, add the root directory to the requested file name}
      isFileName := sleDefDir.Text + isFileName;
   {Call the ConvertFileNameToDos procedure to convert the file name into
   the DOS format so that you can use it}
   ConvertFileNameToDos;
   {Did the client send a GET command?}
   if (asCommand = 'GET') then
   begin
      {Yes, is it a full or short request (HTTP 1.0 or 0.9)?}
      lbCmdFound := TRUE;
      iiHTTPStatus := HTTP_GET;
      if (ibIsFullRequest) then
         SendFullObject
      else
         SendShortObject;
   end;
   {Did the client send a HEAD command?}
   if (asCommand = 'HEAD') then
   begin
      {Yes, is it a full or short request (HTTP 1.0 or 0.9)?}
      lbCmdFound := TRUE;
      iiHTTPStatus := HTTP_HEAD;
      if (ibIsFullRequest) then
         SendFullHeader
      else
         SendShortHeader;
      CloseConnection;
   end;
   {Did the client send a POST command?}
   if (asCommand = 'POST') then
   begin
      {Yes, send the client a NOT SUPPORTED response}
      lbCmdFound := TRUE;
      iiHTTPStatus := HTTP_POST;
      SendContextMsg;
      {Close the connection}
      CloseConnection;
   end;
   {Did you find a matching command?}
   if (not lbCmdFound) then
   begin
      {No, send a COMMAND UNKNOWN response}
      iiHTTPStatus := HTTP_UNKWN;
      SendContextMsg;
      {Close the connection}
      CloseConnection;
   end;
end;

procedure TWbServe.ReadCommandMsg;
var
   lsCommand: String;
   lsMessage: String;
   liCmdLength: Integer;
   liMsgLength: Integer;
begin
   {Receive the request from the client, then check for any errors}
   liMsgLength := recv(isConnectionSocket, ReceiveBuffer, 8192, 0);
   if (liMsgLength = SOCKET_ERROR) and
         (WSAGetLastError <> WSAEWOULDBLOCK) then
      ShowMessage(NetSocket.SocketError)
   else
   begin
      {Terminate the received request string}
      ReceiveBuffer[liMsgLength] := Char(0);
      {Get the first (request) line}
      if NetSocket.GetLine(TempBuffer, ReceiveBuffer) then
      begin
         {Add the request line to the display for the user to see}
         mleReceived.Lines.Add(StrPas(TempBuffer));
         {Is there anything in the hold buffer?}
         if (Length(isHoldBuffer) > 0) then
            {If so, add it in front of your new line}
            lsMessage := isHoldBuffer + StrPas(TempBuffer)
         else
            lsMessage := StrPas(TempBuffer);
         {Empty the hold buffer}
         isHoldBuffer := '';
         {Find the space character to determine the end of the command}
         liCmdLength := Pos(' ', lsMessage);
         if (liCmdLength = 0) then
         begin
            {If there is no space character, it is not the command
            line of the header}
            liCmdLength := Length(lsMessage);
            lsCommand := lsMessage;
            lsMessage := '';
         end
         else
         begin
            {Separate the command from the rest of the request}
            lsCommand := Copy(lsMessage, 1, (liCmdLength - 1));
            lsMessage := Copy(lsMessage, (liCmdLength + 1), Length(lsMessage));
         end;
         {Pass the command and parameter to the CheckCommandMsg procedure}
         CheckCommandMsg(lsCommand, lsMessage);
      end;
      {Any additional lines are of no concern in this application, so
      pass them directly to the display}
      while NetSocket.GetLine(TempBuffer, ReceiveBuffer) do
      begin
         mleReceived.Lines.Add(StrPas(TempBuffer));
      end;
      {Is there any left over? If so, place it into the hold buffer.}
      if (Length(StrPas(TempBuffer)) > 0) then
      begin
         if (Length(isHoldBuffer) > 0) then
            isHoldBuffer := isHoldBuffer + StrPas(TempBuffer)
         else
            isHoldBuffer := StrPas(TempBuffer);
      end;
   end;
end;

procedure TWbServe.UWMSocketEvent(var Msg: TMessage);
begin
   {What socket event was received? Call the appropriate procedure}
   case Msg.LParamLo of
      FD_READ: ReadCommandMsg;
      FD_WRITE: ProcessDataSend;
      FD_ACCEPT:;
      FD_CONNECT:;
      FD_CLOSE:;
   end;
end;

procedure TWbServe.UWMListenSocketEvent(var Msg: TMessage);
begin
   {What socket event was received? Call the appropriate procedure}
   case Msg.LParamLo of
      FD_READ:;
      FD_WRITE:;
      FD_OOB:;
      FD_ACCEPT: ProcessListenAccept;
      FD_CONNECT:;
      FD_CLOSE:;
   end;
end;

procedure TWbServe.UWMRejectSocketEvent(var Msg: TMessage);
begin
   {What socket event was received? Call the appropriate procedure}
   case Msg.LParamLo of
      FD_READ:;
      FD_WRITE: SendConnectReject;
      FD_ACCEPT:;
      FD_CONNECT:;
      FD_CLOSE:;
   end;
end;

end.
