unit trobot;

interface

uses
  WinTypes, WinProcs, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Tsocketc, NWinsock, StdCtrls;

{ If Using the Delphi 32 Winsock.pas }
type
     _Socket = TSocket;
     fd_set = TFDSet;
     Pfd_set = PFDSet;
     timeval = TTimeVal;
     hostent = THostEnt;
     netent = TNetEnt;
     servent = TServEnt;
     protoent = TProtoEnt;
     in_addr = TInAddr;
     Pin_addr = PInAddr;
     Sockaddr_in = TSockAddrIn;
     PSockaddr_in = PSockAddrIn;
     WSADATA = TWSAData;
     LPWSADATA = PWSAData;
     sockaddr = TSockAddr;
     sockproto = TSockProto;
     linger = TLinger;
     LPSOCKADDR = ^sockaddr;
     LPSOCKADDR_IN = ^sockaddr_in;
     LPLINGER = ^linger;
     LPIN_ADDR = ^in_addr;
     LPFD_SET = ^fd_set;
     LPHOSTENT = ^hostent;
     LPSERVENT = ^servent;
     LPPROTOENT = ^protoent;
     LPTIMEVAL = ^timeval;
{ End of Delphi 32 Winsock.pas compatability type definitions }
const
  SKT_NOOP = 0;
  SKT_GETHOSTADDR = 1;
  SKT_CONNECTTOHOST = 2;
  SKT_SENDPACKET = 3;
  SKT_RECVPACKET = 4;
{Constant declarations to represent HTTP commands and actions to be
taken. These constants keep track of the command last issued,
so that you can direct the server's responses to the appropriate 
procedure}
  HTTP_NOOP = 0;
  HTTP_CONNECT = 1;
  HTTP_GET = 2;
  HTTP_HEAD = 3;
  HTTP_POST = 4;
  HTTP_LINKCONNECT = 5;
  HTTP_LINKGET = 6;
  HTTP_LINKHEAD = 7;
  HTTP_LINKPOST = 8;
  HTTP_IMGCONNECT = 9;
  HTTP_IMGGET = 10;
  HTTP_IMGHEAD = 11;
  HTTP_IMGPOST = 12;
{User-defined event declarations, for socket events. The first is 
used for the primary socket, the second for address lookup notifications}
  UWM_SOCKETEVENT = WM_USER + 500;
  UWM_ADDRSOCKETEVENT = WM_USER + 501;
type
  TWebRobot = class(TForm)
    Label1: TLabel;
    sleURL: TEdit;
    cbConnect: TButton;
    Label2: TLabel;
    lbImages: TListBox;
    Label3: TLabel;
    lbLinks: TListBox;
    cbClose: TButton;
    cbTest: TButton;
    Label4: TLabel;
    stStatus: TLabel;
    Label5: TLabel;
    stState: TLabel;
    mleHTML: TMemo;
    Label6: TLabel;
    stURL: TLabel;
    procedure cbCloseClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure ProcessHostLookup;
    procedure ProcessHostConnect;
    procedure ClearBuffer(DirtyBuffer: Pchar; BufLength: Integer);
    procedure ProcessHostWelcome;
    procedure ProcessHostSend;
    function GetHTMLReference(TargetString, SourceString: PChar): Boolean;
    function GetRestHTMLReference(TargetString, SourceString: PChar): Boolean;
    procedure ProcessHostGet;
    procedure ProcessHostLinkGet;
    procedure ProcessHostReceive;
    procedure cbConnectClick(Sender: TObject);
    procedure cbTestClick(Sender: TObject);
  private
    { Private declarations }
    {The primary socket to be used in the connection with the server}
    isConnectionSocket: _SOCKET;
    {The secondary socket to be used in retrieving image files}
    isImageSocket: _SOCKET;
    {A flag to indicate whether the connection has been made}
    ibConnected: Boolean;
    {The command state indicator}
    iiHTTPStatus: Integer;
    {The hold buffer for incomplete message lines}
    isHoldBuffer: String;
    {The link test state indicator}
    iiTestStatus: Integer;
    {The number of the link being tested}
    iiTestNumber: Integer;
    {The port to use (if not the default)}
    iiAltPort: Integer;
    {The socket event-handler declarations}
    procedure UWMSocketEvent(var Msg: TMessage); message UWM_SOCKETEVENT;
    procedure UWMAddrSocketEvent(var Msg: TMessage); message UWM_ADDRSOCKETEVENT;
  public
    { Public declarations }
    {The TSocket class declaration}
    NetSocket: _TSocket;
    {The receive and temporary buffers}
    ReceiveBuffer: array[0..8192] of Char;
    TempBuffer: array[0..8192] of Char;
  end;

var
  WebRobot: TWebRobot;

implementation

{$R *.DFM}

procedure TWebRobot.cbCloseClick(Sender: TObject);
begin
   Close;
end;

procedure TWebRobot.FormCreate(Sender: TObject);
begin
   ibConnected := FALSE;
   iiHTTPStatus := HTTP_NOOP;
   iiTestStatus := HTTP_NOOP;
   {Create the NetSocket class}
   NetSocket := _TSocket.Create;
   {Initialize the WinSock system}
   if not NetSocket.InitializeSocket then
      Close;
end;

procedure TWebRobot.FormDestroy(Sender: TObject);
begin
     {Cleanup the WinSock interface and system}
     NetSocket.ShutdownSocket;
     NetSocket.Free;
end;

procedure TWebRobot.ProcessHostLookup;
begin
   {Update the display to reflect that you are attempting to connect
   to the server}
   stStatus.Caption := 'Connecting To Host';
   stStatus.Update;
   {Check to see whether you  need to connect 
    to an alternative service port}
   if (iiAltPort = 0) then
      iiAltPort := 80;
   {Call ProcessHostAddress (first written in Chapter 6) to reformat
   the received address so that you can connect}
   if (NetSocket.ProcessHostAddress = TRUE) then
   begin
      {Call OpenSocketConnection (First developed in Chapter 6) to
      create a new socket and open a connection to the host, this
      triggers the socket event when the connection has been
      established}
      if (iiTestStatus = HTTP_IMGCONNECT) then
         isImageSocket := NetSocket.OpenSocketConnection(Handle, UWM_SOCKETEVENT, SKT_CONNECTTOHOST, iiAltPort)
      else
         isConnectionSocket := NetSocket.OpenSocketConnection(Handle, UWM_SOCKETEVENT, SKT_CONNECTTOHOST, iiAltPort);
   end
   else
   begin
      {The host address was not found, so inform the user}
      stStatus.Caption := 'Cannot Find Host Address';
      stStatus.Update;
   end;
end;

procedure TWebRobot.ProcessHostConnect;
begin
   {Update the display to reflect that you are connected}
   stStatus.Caption := 'Receiving Welcome From Host';
   stStatus.Update;
   {Update the command state flags}
   iiHTTPStatus := HTTP_CONNECT;
   stState.Caption := 'HTTP_CONNECT';
   stState.Update;
   {Call ProcessHostWelcome to send the file or page request}
   ProcesshostWelcome;
end;

procedure TWebRobot.ClearBuffer(DirtyBuffer: Pchar; BufLength: Integer);
var
   liCurPos: Integer;
   lcNullChar: Char;
begin
   {Initialize the position pointer to 0}
   liCurPos := 0;
   {Create a null character}
   lcNullChar := Char(0);
   {Loop through the entire buffer passed in, setting each character
   equal to the null character}
   while (liCurPos < BufLength) do
   begin
      DirtyBuffer^ := lcNullChar;
      Inc(liCurPos);
      Inc(DirtyBuffer);
   end;
end;

procedure TWebRobot.ProcessHostWelcome;
var
   lsWelcomeMsg, lsRetnCode: String;
   lsResource: String;
   liURIStart, liURIEnd: Integer;
   liNumTrys, liAmtReceived: LongInt;
begin
   {You are connected, there is no response code to check}
   ibConnected := TRUE;
   {Set the command status flag to NOOP}
   iiHTTPStatus := HTTP_NOOP;
   {Display the response message, none in this case}
   stStatus.Caption := lsWelcomeMsg;
   stStatus.Update;
   {Are you connected?}
   if ibConnected then
   begin
      {What is your test status?, NOOP (getting the initial page), IMAGE
      (testing image URLs), or LINK (test Web links)}
      case iiTestStatus of
         { NOOP (getting the initial page)}
         HTTP_NOOP:
         begin
            {Set the command status flag to indicate that you sent a GET}
            iiHTTPStatus := HTTP_GET;
            stState.Caption := 'HTTP_GET';
            stState.Update;
            {Parse the user-entered URL;you want the file name}
            lsResource := sleURL.Text;
            liURIStart := Pos(':/', lsResource);
            if (liURIStart > 0) then
               lsResource := Copy(lsResource, (liURIStart + 3), Length(lsResource));
            liURIEnd := Pos('/', lsResource);
            if (liURIEnd = Length(lsResource)) then
               lsResource := '/index.htm'
            else
               lsResource := Copy(lsResource, liURIEnd, Length(lsResource));
            {Double-check the connection}
            if NetSocket.IsSocketConnected(isConnectionSocket) then
            begin
               {Update the display to reflect what you are doing}
               stStatus.Caption := 'Requesting Web Page';
               stStatus.Update;
               {Send the GET command to the server}
               NetSocket.SocketSend(isConnectionSocket, handle, UWM_SOCKETEVENT,
                              'GET ' + lsResource, SKT_SENDPACKET);
               {Clear the receive buffer}
               ClearBuffer(ReceiveBuffer, 8192);
            end
            else
            begin
               {You are not connected, so tell the user}
               stStatus.Caption := 'Not Connected';
               stStatus.Update;
            end;
         end;
         {LINK (test Web links)}
         HTTP_LINKCONNECT:
         begin
            {Set the command status flag to indicate that you sent a GET
            on a link test}
            iiHTTPStatus := HTTP_LINKGET;
            stState.Caption := 'HTTP_LINKGET';
            stState.Update;
            {Parse the link URL; you want the file name}
            lsResource := lbLinks.Items[iiTestNumber];
            liURIStart := Pos(':/', lsResource);
            if (liURIStart > 0) then
            begin
               lsResource := Copy(lsResource, (liURIStart + 3), Length(lsResource));
               liURIEnd := Pos('/', lsResource);
               if (liURIEnd > 0) then
               begin
                  if (liURIEnd = Length(lsResource)) then
                     lsResource := '/index.html'
                  else
                     lsResource := Copy(lsResource, liURIEnd, Length(lsResource));
               end
               else
                  lsResource := '/index.html'
            end;
            {Double-check the connection}
            if NetSocket.IsSocketConnected(isConnectionSocket) then
            begin
               {Update the display to reflect what you are doing}
               stStatus.Caption := 'Requesting Web Page';
               stStatus.Update;
               {Is it a CGI request? You don't test those.}
               if (Pos('cgi', lsResource) = 0) then
               begin
                  {Send the GET command to the server}
                  NetSocket.SocketSend(isConnectionSocket, handle, UWM_SOCKETEVENT,
                              'GET ' + lsResource, SKT_SENDPACKET);
                  {Clear the receive buffer}
                  ClearBuffer(ReceiveBuffer, 8192);
                  {You are running in synchronous mode now. You need 
                  to loop several times to ensure that the connection
                  has sufficient time to retrieve the link page}
                  liNumTrys := 0;
                  WSAAsyncSelect(isConnectionSocket, handle, 0, 0);
                  while ((recv(isConnectionSocket, ReceiveBuffer, 8192, 0) < 1) and
                        (liNumTrys < 1000000)) do
                     Inc(liNumTrys);
                  {If you didn't receive anything, place a 'Not Found'
                  page in the receive buffer}
                  if ((liNumTrys = 1000000) and (StrLen(ReceiveBuffer) = 0)) then
                     StrPCopy(ReceiveBuffer, '<TITLE>Not Found</TITLE>');
               end;
            end
            else
            begin
               {You are not connected, so tell the user.
                Place a 'Not Found' page in the receive buffer}
               stStatus.Caption := 'Not Connected';
               stStatus.Update;
               StrPCopy(ReceiveBuffer, '<TITLE>Not Found</TITLE>');
            end;
            {Call the ProcessHostLinkGet procedure to analyze the 
            received page}
            ProcessHostLinkGet;
         end;
         {IMAGE (testing image URLs)}
         HTTP_IMGCONNECT:
         begin
            {Set the command status flag to indicate that you sent a GET
            on an image link test}
            iiHTTPStatus := HTTP_IMGGET;
            stState.Caption := 'HTTP_IMGGET';
            stState.Update;
            {Parse the link URL; you want the file name}
            lsResource := lbImages.Items[iiTestNumber];
            liURIStart := Pos(':/', lsResource);
            if (liURIStart > 0) then
            begin
               lsResource := Copy(lsResource, (liURIStart + 3), Length(lsResource));
               liURIEnd := Pos('/', lsResource);
               if (liURIEnd = Length(lsResource)) then
                  lsResource := 'default.htm'
               else
                  lsResource := Copy(lsResource, (liURIEnd + 1), Length(lsResource));
            end;
            {Double-check the connection}
            if NetSocket.IsSocketConnected(isImageSocket) then
            begin
               {Update the display to reflect what you are doing}
               stStatus.Caption := 'Requesting Image';
               stStatus.Update;
               {Send the GET command to the server}
               NetSocket.SocketSend(isImageSocket, handle, UWM_SOCKETEVENT,
                              'GET ' + lsResource, SKT_SENDPACKET);
               {Clear the receive buffer}
               ClearBuffer(ReceiveBuffer, 8192);
               {You are running in synchronous mode now. You must
                  loop several times to ensure that you receive the 
                  entire image}
               WSAAsyncSelect(isConnectionSocket, handle, 0, 0);
               while (recv(isImageSocket, ReceiveBuffer, 8192, 0) < 1) do;
            end
            else
            begin
               {You are not connected, so tell the user.
                Place a 'Not Found' page in the receive buffer}
               stStatus.Caption := 'Not Connected';
               stStatus.Update;
               StrPCopy(ReceiveBuffer, '<TITLE>Not Found</TITLE>');
            end;
            {Call the ProcessHostLinkGet procedure to analyze the 
            received page}
            ProcessHostLinkGet
         end;
      end;
   end;
end;

procedure TWebRobot.ProcessHostSend;
begin
   if (ibConnected) then
   begin
         if (iiHTTPStatus <> HTTP_NOOP) then
         begin
            ClearBuffer(ReceiveBuffer, 8192);
         end;
   end;
end;

function TWebRobot.GetHTMLReference(TargetString, SourceString: PChar): Boolean;
var
   liCurPos: Integer;
   liCurTargetPos: Integer;
   liCarRetnPos: Integer;
   liCopyPos: Integer;
begin
   {Initialize the current position marker to 0}
   liCurPos := 0;
   {Scan the SourceString buffer, looking for the beginning of an HTML
   tag (marked by the left angle character <)}
   while ((SourceString[liCurPos] <> '<') and
         (SourceString[liCurPos] <> Char(0))) do
      Inc(liCurPos);
   {Initialize the target string position marker to 0}
   liCurTargetPos := 0;
   {Copy characters from the source string to the target string until
   either the end of the tag or the end of the source string is found}
   while ((SourceString[liCurPos] <> '>') and
         (SourceString[liCurPos] <> Char(0))) do
   begin
      if (SourceString[liCurPos] <> '<') then
      begin
         TargetString[liCurTargetPos] := SourceString[liCurPos];
         Inc(liCurPos);
         Inc(liCurTargetPos);
      end
      else
         Inc(liCurPos);
   end;
   {Mark the end of the tag with a null character in the target string}
   TargetString[liCurTargetPos] := Char(0);
   {Move the remaining text in the source string forward, so that the
   first character after the current HTML tag is the first character
   in the string, return TRUE if a complete HTML tag was found, return
   FALSE if an incomplete HTML tag was found}
   if (SourceString[liCurPos] = '>') then
   begin
      liCopyPos := 0;
      liCurPos := (liCurPos + 1);
      while (SourceString[liCurPos] <> Char(0)) do
      begin
         SourceString[liCopyPos] := SourceString[liCurPos];
         Inc(liCurPos);
         Inc(liCopyPos);
      end;
      SourceString[liCopyPos] := Char(0);
      GetHTMLReference := TRUE;
   end
   else
   begin
      SourceString[0] := Char(0);
      GetHTMLReference := FALSE;
   end;
end;

{This function does the same thing as the previous function, only it 
assumes that the first character is in the middle of an HTML tag and
only looks for the end of the tag}
function TWebRobot.GetRestHTMLReference(TargetString, SourceString: PChar): Boolean;
var
   liCurPos: Integer;
   liCurTargetPos: Integer;
   liCarRetnPos: Integer;
   liCopyPos: Integer;
begin
   {Initialize the current position marker to 0}
   {Initialize the target string position marker to 0}
   liCurPos := 0;
   liCurTargetPos := 0;
   {Copy characters from the source string to the target string until
   either the end of the tag or the end of the source string is found}
   while ((SourceString[liCurPos] <> '>') and
         (SourceString[liCurPos] <> Char(0))) do
   begin
      TargetString[liCurTargetPos] := SourceString[liCurPos];
      Inc(liCurPos);
      Inc(liCurTargetPos);
   end;
   {Mark the end of the tag with a null character in the target string}
   TargetString[liCurTargetPos] := Char(0);
   {Move the remaining text in the source string forward, so that the
   first character after the current HTML tag is the first character
   in the string, return TRUE if a complete HTML tag was found, return
   FALSE if an incomplete HTML tag was found}
   if (SourceString[liCurPos] = '>') then
   begin
      liCopyPos := 0;
      liCurPos := (liCurPos + 1);
      while (SourceString[liCurPos] <> Char(0)) do
      begin
         SourceString[liCopyPos] := SourceString[liCurPos];
         Inc(liCurPos);
         Inc(liCopyPos);
      end;
      SourceString[liCopyPos] := Char(0);
      GetRestHTMLReference := TRUE;
   end
   else
   begin
      SourceString[0] := Char(0);
      GetRestHTMLReference := FALSE;
   end;
end;

procedure TWebRobot.ProcessHostGet;
var
   lbComplete: Boolean;
   lsCurRef: String;
   liAddressStart, liAddressEnd: Integer;
begin
   {Are you connected?}
   if (ibConnected) then
   begin
      {Convert the received string to a Pascal string}
      lsCurRef := StrPas(ReceiveBuffer);
      {Did you get a NOT FOUND response?}
      if (Copy(lsCurRef, 1, 16) = '<TITLE>Not Found') then
      begin
         {Inform the user and shut down the connection}
         stStatus.Caption := 'URL Not Found';
         stStatus.Update;
         shutdown(isConnectionSocket, 0);
         NetSocket.CloseSocketConnection(isConnectionSocket);
         ibConnected := FALSE;
      end
      else
      begin
         {You did not receive a NOT FOUND, so begin parsing it to
         extract the tags in which you are interested}
         lbComplete := FALSE;
         {Is there anything in the hold buffer?}
         if (Length(isHoldBuffer) > 0) then
         begin
            {Yes, it must be an incomplete tag, so get the rest
            of the tag}
            if GetRestHTMLReference(TempBuffer, ReceiveBuffer) then
            begin
               {Add the two tag parts together}
               lsCurRef := isHoldBuffer + StrPas(TempBuffer);
               {Empty the hold buffer}
               isHoldBuffer := '';
               {Do you have either the BODY or HTML end tag? If so, you
               have reached the end of the page}
               if (lsCurRef = '/BODY') then
                  lbComplete := TRUE;
               if (lsCurRef = '/HTML') then
                  lbComplete := TRUE;
               {Did you find a hyperlink?}
               if (Copy(lsCurRef, 1, 6) = 'A HREF') then
               begin
                  {Yes, extract the URL and add it to the display}
                  liAddressStart := Pos('=', lsCurRef);
                  if (liAddressStart > 0) then
                  begin
                     lsCurRef := Copy(lsCurRef, (liAddressStart + 1), Length(lsCurRef));
                     if (Pos('"', lsCurRef) = 1) then
                        lsCurRef := Copy(lsCurRef, 2, Length(lsCurRef));
                     if (Pos('"', lsCurRef) > 0) then
                        lsCurRef := Copy(lsCurRef, 1, (Length(lsCurRef) - 1));
                  end;
                  lbLinks.Items.Add(lsCurRef);
               end;
               {Did you find an image?}
               if (Copy(lsCurRef, 1, 3) = 'IMG') then
               begin
                  {Yes, extract the URL and add it to the display}
                  liAddressStart := Pos('SRC=', lsCurRef);
                  if (liAddressStart > 0) then
                  begin
                     lsCurRef := Copy(lsCurRef, (liAddressStart + 5), Length(lsCurRef));
                     liAddressEnd := Pos('"', lsCurRef);
                     if (liAddressEnd > 0) then
                        lsCurRef := Copy(lsCurRef, 1, (liAddressEnd - 1))
                     else
                        lsCurRef := Copy(lsCurRef, 1, (Length(lsCurRef) - 1));
                  end;
                  lbImages.Items.Add(lsCurRef);
               end;
            end;
         end;
         {Clear the Temp Buffer}
         ClearBuffer(TempBuffer, StrLen(TempBuffer));
         {Loop as long as you find additional tags and haven't
         reached the end of the page}
         while (GetHTMLReference(TempBuffer, ReceiveBuffer) and
               (not lbComplete)) do
         begin
            {Convert the extracted tag to a Pascal string}
            lsCurRef := StrPas(TempBuffer);
            {Do you have the HTML end tag? If so, you
            have reached the end of the page}
            if (lsCurRef = '/HTML') then
               lbComplete := TRUE;
            {Did you find a hyperlink?}
            if (Copy(lsCurRef, 1, 6) = 'A HREF') then
            begin
               {Yes, extract the URL and add it to the display}
               liAddressStart := Pos('=', lsCurRef);
               if (liAddressStart > 0) then
               begin
                  lsCurRef := Copy(lsCurRef, (liAddressStart + 1), Length(lsCurRef));
                  if (Pos('"', lsCurRef) = 1) then
                     lsCurRef := Copy(lsCurRef, 2, Length(lsCurRef));
                  if (Pos('"', lsCurRef) > 0) then
                     lsCurRef := Copy(lsCurRef, 1, (Length(lsCurRef) - 1));
               end;
               lbLinks.Items.Add(lsCurRef);
            end;
            {Did you find an image?}
            if (Copy(lsCurRef, 1, 3) = 'IMG') then
            begin
               {Yes, extract the URL and add it to the display}
              liAddressStart := Pos('SRC=', lsCurRef);
              if (liAddressStart > 0) then
              begin
                 lsCurRef := Copy(lsCurRef, (liAddressStart + 5), Length(lsCurRef));
                 liAddressEnd := Pos('"', lsCurRef);
                 if (liAddressEnd > 0) then
                    lsCurRef := Copy(lsCurRef, 1, (liAddressEnd - 1))
                 else
                    lsCurRef := Copy(lsCurRef, 1, (Length(lsCurRef) - 1));
              end;
              lbImages.Items.Add(lsCurRef);
            end;
            {Clear the Temp Buffer}
            ClearBuffer(TempBuffer, StrLen(TempBuffer));
         end;
         {Have you reached the end of the page?}
         if (not lbComplete) then
         begin
            {No, place the remaining partial tag into the hold buffer}
            if (StrLen(TempBuffer) > 0) then
               isHoldBuffer := StrPas(TempBuffer);
            {Clear the receive buffer}
            ClearBuffer(ReceiveBuffer, 8192);
         end
         else
         begin
            {Yes, inform the user, shut down the connection, and change
            the default button to the TEST button}
            stStatus.Caption := 'Finnished Retrieving';
            stStatus.Update;
            shutdown(isConnectionSocket, 0);
            NetSocket.CloseSocketConnection(isConnectionSocket);
            cbConnect.Default := FALSE;
            cbTest.Default := TRUE;
            ibConnected := FALSE;
         end;
      end;
   end;
end;

procedure TWebRobot.ProcessHostLinkGet;
var
   lbComplete: Boolean;
   lsCurRef: String;
   lsHostAddress: String;
   liAddressStart, liAddressEnd: Integer;
   liNameStart, liNameEnd: Integer;
begin
   {Are you connected?}
   if (ibConnected) then
   begin
      {Initialize the complete flag to FALSE}
      lbComplete := FALSE;
      ibConnected := FALSE;
      {Are you testing a link or an image?}
      case iiHTTPStatus of
         {You are testing a link}
         HTTP_LINKGET:
         begin
            {Check the initial HTML to verify that you have a valid page}
            lsCurRef := StrPas(ReceiveBuffer);
            if (((Copy(lsCurRef, 1, 14) <> '<TITLE>Invalid') and
               (Copy(lsCurRef, 1, 16) <> '<TITLE>Not Found')) and
               (Pos('Not Found', lsCurRef) = 0)) then
            begin
               {You have a valid page, so remove this link 
                from the list box}
               lbLinks.Items.Delete(iiTestNumber);
               lbLinks.Refresh;
               {Decrement the test index number, so that once you 
               increment, you return to the same line in the list 
               box}
               Dec(iiTestNumber);
               {Display to the user the page's first few lines}
               mleHTML.Clear;
               mleHTML.Lines.Add(Copy(lsCurRef, 1, 200));
               mleHTML.Refresh;
            end;
            {Are there any untested links remaining?}
            if (lbLinks.Items.Count > (iiTestNumber + 1)) then
            begin
               {Increment the test index number}
               Inc(iiTestNumber);
               {Get the next link URL and parse it to get the host 
               address}
               iiTestStatus := HTTP_LINKCONNECT;
               lsHostAddress := lbLinks.Items[iiTestNumber];
               stURL.Caption := lsHostAddress;
               {Is it a complete URL?}
               if (Pos('://', lsHostAddress) = 0) then
               begin
                  {No, get the user-entered URL}
                  lsHostAddress := sleURL.text;
               end;
            end
            else
               {No more links to test}
               lbComplete := TRUE;
            {Turn off all asynchronous behavior}
            WSAAsyncSelect(isConnectionSocket, handle, 0, 0);
            {Continue receiving data until no more is received}
            while (recv(isConnectionSocket, ReceiveBuffer, 8192, 0) > 0) do;
            {Shut down the connection}
            shutdown(isConnectionSocket, 0);
            NetSocket.CloseSocketConnection(isConnectionSocket);
            isConnectionSocket := INVALID_SOCKET;
         end;
         {You are testing an image}
         HTTP_IMGGET:
         begin
            {Check for the beginnings of an HTML tag to see whether
             you are receiving a page rather than an image}
            if (ReceiveBuffer[0] <> '<') then
            begin
               {You have a valid image; 
                remove this URL from the list box}
               lbImages.Items.Delete(iiTestNumber);
               lbImages.Refresh;
               {Decrement the test index number, so that after you 
               increment, you return to the same line in the list 
               box}
               Dec(iiTestNumber);
               {Display to the user the first few bytes}
               lsCurRef := StrPas(ReceiveBuffer);
               mleHTML.Clear;
               mleHTML.Lines.Add(lsCurRef);
               mleHTML.Refresh;
            end;
            {Are there any untested images remaining?}
            if (lbImages.Items.Count > (iiTestNumber + 1)) then
            begin
               {Increment the test index number}
               Inc(iiTestNumber);
               {Get the next image URL and parse it to get the host 
               address}
               iiTestStatus := HTTP_IMGCONNECT;
               lsHostAddress := lbImages.Items[iiTestNumber];
               stURL.Caption := lsHostAddress;
               {Is it a complete URL?}
               if (Pos('://', lsHostAddress) = 0) then
               begin
                  {No, get the user-entered URL}
                  lsHostAddress := sleURL.text;
               end;
            end
            else
            begin
               {No more images to test, are there any links?}
               if (lbLinks.Items.Count > 0) then
               begin
                  {Yes, get the first link URL and parse it to get the 
                  host address}
                  iiTestStatus := HTTP_LINKCONNECT;
                  iiTestNumber := 0;
                  lsHostAddress := lbLinks.Items[0];
                  stURL.Caption := lsHostAddress;
                  {Is it a complete URL?}
                  if (Pos('://', lsHostAddress) = 0) then
                  begin
                     {No, get the user-ntered URL}
                     lsHostAddress := sleURL.text;
                  end;
               end
               else
                  {No more links to test}
                  lbComplete := TRUE;
            end;
            {Turn off all asynchronous behavior}
            WSAAsyncSelect(isImageSocket, handle, 0, 0);
            {Continue receiving data until no more is received}
            while (recv(isImageSocket, ReceiveBuffer, 8192, 0) > 0) do;
            {Shut down the connection}
            shutdown(isImageSocket, 0);
            NetSocket.CloseSocketConnection(isImageSocket);
            isImageSocket := INVALID_SOCKET;
         end;
      end;
      {Have you tested all links and images?}
      if (not lbComplete) then
      begin
         {No, parse the host name from the next to be tested}
         if (Length(lsHostAddress) > 0) then
         begin
            liNameStart := Pos(':/', lsHostAddress);
            if (liNameStart > 0) then
            begin
               lsHostAddress := Copy(lsHostAddress, (liNameStart + 3),
                                              Length(lsHostAddress));
               liNameEnd := Pos('/', lsHostAddress);
               if (liNameEnd > 0) then
                  lsHostAddress := Copy(lsHostAddress, 1, (liNameEnd - 1));
               liNameEnd := Pos(':', lsHostAddress);
               iiAltPort := 0;
               if (liNameEnd > 0) then
               begin
                  iiAltPort := StrToInt(Copy(lsHostAddress, (liNameEnd + 1), Length(lsHostAddress)));
                  lsHostAddress := Copy(lsHostAddress, 1, (liNameEnd - 1));
               end;
            end;
            {Set the status display and command status flag to indicate 
            that you are looking up the host address. }
            stStatus.Caption := 'Looking Up Host Address';
            stStatus.Update;
            iiHTTPStatus := HTTP_CONNECT;
            stState.Caption := 'HTTP_CONNECT';
            stState.Update;
            {Use the GetHostAddress procedure to look up the address of 
            the server. Chapter 6 covers this procedure.}
            NetSocket.GetHostAddress(Handle, UWM_ADDRSOCKETEVENT, lsHostAddress,
                                       SKT_GETHOSTADDR);
         end;
      end
      else
      begin
         {You have tested all images and links, so tell the user}
         stStatus.Caption := 'Finished Testing';
         stStatus.Update;
         iiTestStatus := HTTP_NOOP;
         cbTest.Default := FALSE;
         cbConnect.Default := TRUE;
      end;
   end;
end;

procedure TWebRobot.ProcessHostReceive;
var
   lsRetnStr: String;
   liCarRetnPos: Integer;
begin
   try
      {Receive the incoming data}
      liCarRetnPos := recv(isConnectionSocket, ReceiveBuffer, 8192, 0);
      ReceiveBuffer[liCarRetnPos] := Char(0);
      {You have received an FD_READ event, so check whether you
      received anything}
      if (ReceiveBuffer[0] = Char(0)) then
      begin
         {Clear the receive buffer}
         ClearBuffer(ReceiveBuffer, 8192);
      end
      else
      begin
         {You received data, so check which command you issued and
         call the appropriate procedure}
         case iiHTTPStatus of
            HTTP_NOOP:;
            HTTP_CONNECT: ProcessHostWelcome;
            HTTP_GET: ProcessHostGet;
            HTTP_LINKGET: ProcessHostLinkGet;
            HTTP_IMGGET: ProcessHostLinkGet;
         end;
      end;
   except
      {Clear the receive buffer}
      ClearBuffer(ReceiveBuffer, 8192);
   end;
end;

procedure TWebRobot.UWMSocketEvent(var Msg: TMessage);
begin
   case Msg.LParamLo of
      FD_READ: ProcessHostReceive;
      FD_WRITE: ProcessHostSend;
      FD_CONNECT: ProcessHostConnect;
      FD_CLOSE: ;
   end;
end;

procedure TWebRobot.UWMAddrSocketEvent(var Msg: TMessage);
begin
   if (Msg.lParamHi = 0) then
      ProcessHostLookup;
end;

procedure TWebRobot.cbConnectClick(Sender: TObject);
var
   lsHostAddress: String;
   liNameStart, liNameEnd: Integer;
begin
   {Make sure the URL TEdit control is not blank.}
   if (Length(sleURL.Text) > 0) then
   begin
      {Parse the URL to extract the host name}
      lsHostAddress := sleURL.text;
      stURL.Caption := lsHostAddress;
      liNameStart := Pos(':/', lsHostAddress);
      if (liNameStart > 0) then
      begin
         lsHostAddress := Copy(lsHostAddress, (liNameStart + 3),
                                              Length(lsHostAddress));
         liNameEnd := Pos('/', lsHostAddress);
         if (liNameEnd > 0) then
            lsHostAddress := Copy(lsHostAddress, 1, (liNameEnd - 1));
         {Check for an alternative TCP service port}
         liNameEnd := Pos(':', lsHostAddress);
         iiAltPort := 0;
         if (liNameEnd > 0) then
         begin
            iiAltPort := StrToInt(Copy(lsHostAddress, (liNameEnd + 1), Length(lsHostAddress)));
            lsHostAddress := Copy(lsHostAddress, 1, (liNameEnd - 1));
         end;
      end;
      {Clear the image and link list boxes}
      lbImages.Clear;
      lbLinks.Clear;
      {Set the status display and command status flag to indicate that
      the host address is being looked up}
      stStatus.Caption := 'Looking Up Host Address';
      stStatus.Update;
      iiHTTPStatus := HTTP_CONNECT;
      stState.Caption := 'HTTP_CONNECT';
      stState.Update;
      {Use the GetHostAddress procedure to look up the address of the
      server. Chapter 6 covers this procedure.}
      NetSocket.GetHostAddress(Handle, UWM_ADDRSOCKETEVENT, lsHostAddress,
                                       SKT_GETHOSTADDR);
   end;
end;

procedure TWebRobot.cbTestClick(Sender: TObject);
var
   lsHostAddress: String;
   liNameStart, liNameEnd: Integer;
begin
   {Initialize the number of links tested}
   iiTestNumber := 0;
   {Are there any images to test?}
   if (lbImages.Items.Count > 0) then
   begin
      {Yes, get the first image URL and parse it to get the host address}
      iiTestStatus := HTTP_IMGCONNECT;
      lsHostAddress := lbImages.Items[0];
      stURL.Caption := lsHostAddress;
      {Is it a complete URL?}
      if (Pos('://', lsHostAddress) = 0) then
      begin
         {No, get the user-entered URL}
         lsHostAddress := sleURL.text;
      end;
   end
   else
   begin
      {There are no images. Are there any hyperlinks to test?}
      if (lbLinks.Items.Count > 0) then
      begin
         {Yes, get the first link URL and parse it to get the host 
         address}
         iiTestStatus := HTTP_LINKCONNECT;
         lsHostAddress := lbLinks.Items[0];
         stURL.Caption := lsHostAddress;
         {Is it a complete URL?}
         if (Pos('://', lsHostAddress) = 0) then
         begin
            {No, get the user-entered URL}
            lsHostAddress := sleURL.text;
         end;
      end;
   end;
   {Do you have an URL to connect to?}
   if (Length(lsHostAddress) > 0) then
   begin
      {Yes, parse the host name and TCP port}
      liNameStart := Pos(':/', lsHostAddress);
      if (liNameStart > 0) then
      begin
         lsHostAddress := Copy(lsHostAddress, (liNameStart + 3),
                                              Length(lsHostAddress));
         liNameEnd := Pos('/', lsHostAddress);
         if (liNameEnd > 0) then
            lsHostAddress := Copy(lsHostAddress, 1, (liNameEnd - 1));
         liNameEnd := Pos(':', lsHostAddress);
         iiAltPort := 0;
         if (liNameEnd > 0) then
         begin
            iiAltPort := StrToInt(Copy(lsHostAddress, (liNameEnd + 1), Length(lsHostAddress)));
            lsHostAddress := Copy(lsHostAddress, 1, (liNameEnd - 1));
         end;
      end;
      {Set the status display and command status flag to indicate that
      the host address is being looked up}
      stStatus.Caption := 'Looking Up Host Address';
      stStatus.Update;
      iiHTTPStatus := HTTP_CONNECT;
      stState.Caption := 'HTTP_CONNECT';
      stState.Update;
      {Use the GetHostAddress procedure to look up the address of the
      server. Chapter 6 covers this procedure.}
      NetSocket.GetHostAddress(Handle, UWM_ADDRSOCKETEVENT, lsHostAddress,
                                       SKT_GETHOSTADDR);
   end;
end;

end.
