unit ChatForm;

(*
  Limitations to the Chat demo:
    * You can not send a file to your self.
    * Selecitng Setup will cause the Chat server & client to disconnect.
    
*)
interface

uses
  WinTypes, WinProcs, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Tabnotbk, ExtCtrls, dWinsock, BuffSock, Buttons, dwsUtils,
  Chatcfg;

const
  ConnectText = '&Connect to Chat Server [%s]';
  

type
  TChatWindow = class(TForm)
    ChatClient: TTextClient;
    ChatServer: TTextServer;
    SendFileServer: TBinaryServer;
    GetFileClient: TBinaryClient;
    Panel2: TPanel;
    pnlMessage: TPanel;
    DnldMessage: TPanel;
    Panel3: TPanel;
    bExit: TButton;
    cbServer: TCheckBox;
    bSetup: TButton;
    bChat: TSpeedButton;
    mChat: TMemo;
    mText: TMemo;
    SendFileDlg: TOpenDialog;
    bSendFile: TSpeedButton;
    cbAcceptFiles: TCheckBox;
    procedure ChatServerInfo(Sender: TObject; icode: TSockInfo;
      Socket: TSocketBase);
    procedure ChatServerReject(Sender: TObject; Socket: TSocketBase);
    procedure ChatServerAccept(Sender: TObject; Socket: TSocketBase);
    procedure ChatServerDisconnect(Sender: TObject; Socket: TSocketBase);
    procedure ChatServerLineReceived(socket: TLineSocketBase;
      const line: string; complete: Boolean);
    procedure ChatClientLineReceived(socket: TLineSocketBase;
      const line: string; complete: Boolean);
    procedure ChatClientConnect(Sender: TObject; Socket: TSocketBase);
    procedure ChatClientDisconnect(Sender: TObject; Socket: TSocketBase);
    procedure FormCreate(Sender: TObject);
    procedure mTextKeyUp(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure GetFileClientFileReceived(socket: TBufferedSocketBase);
    procedure GetFileClientConnect(Sender: TObject; Socket: TSocketBase);
    procedure SendFileServerFileSent(socket: TBufferedSocketBase);
    procedure SendFileServerRead(Sender: TObject; Socket: TSocketBase);
    procedure GetFileClientRead(Sender: TObject; Socket: TSocketBase);
    procedure bExitClick(Sender: TObject);
    procedure bSetupClick(Sender: TObject);
    procedure bChatClick(Sender: TObject);
    procedure cbServerClick(Sender: TObject);
    procedure bSendFileClick(Sender: TObject);
    procedure GetFileClientDisconnect(Sender: TObject;
      Socket: TSocketBase);
  private
    { Private declarations }
    procedure ConnectionMessage( Socket: TSocketBase; Info: string );
    procedure SetItUp;
  public
    { Public declarations }
    ReceiveFile: string;
    ReceiveSize: longint;
    SendFile:    string;
    SendSize:    Longint;
    IncomingPath: String;
    UserName: string;
  end;

var
  ChatWindow: TChatWindow;

implementation

{$R *.DFM}

procedure TChatWindow.ChatServerInfo(Sender: TObject; icode: TSockInfo;
  Socket: TSocketBase);
begin
  pnlMessage.Caption := format('%s: %s', [(Sender as TComponent).Name,
                                         SockInfoToString( iCode )]);
end;

procedure TChatWindow.ConnectionMessage( Socket: TSocketBase; Info: string );
begin
  mChat.Lines.Add( format('Server: Connection from %s %s',
                       [Socket.RemoteAddress, Info]));
end;

procedure TChatWindow.ChatServerReject(Sender: TObject; Socket: TSocketBase);
begin
  ConnectionMessage(Socket, 'Rejected');
  (Socket as TLineSocket).WriteLine('Server: Conneciton Rejected');
end;

procedure TChatWindow.ChatServerAccept(Sender: TObject; Socket: TSocketBase);
begin
  ConnectionMessage(Socket, 'Accpeted');
  (Socket as TLineSocket).WriteLine('Server: Conneciton Accepted');
end;

procedure TChatWindow.ChatServerDisconnect(Sender: TObject; Socket: TSocketBase);
begin
  ConnectionMessage(Socket, 'Disconnected');
end;

procedure TChatWindow.ChatServerLineReceived(socket: TLineSocketBase;
  const line: string; complete: Boolean);
var
  I: integer;
begin
  { Delphi 1 needs to handle longer strings. }
  for I := 0 to ChatServer.ClientCount -1 do { Send to ALL Clients }
    (ChatServer.Client[I] as TLineSocket).WriteLine( Line );
end;

procedure TChatWindow.ChatClientLineReceived(socket: TLineSocketBase;
          const line: string; complete: Boolean);
var
  Temp: string;
begin
  { Delphi 1 needs to handle strings longer than 255 chars. }

  if Line[1] = #255 then
  begin { we have an incomming file }
    Temp := line;
    system.Delete(temp, 1, 1);                        { Remove #255 }
    ReceiveFile := IncomingPath +
                   StrTokenExt(', ','""', Temp);      { File name   }
    ReceiveSize := StrToInt(StrToken(', ', Temp));    { Size        }
    GetFileClient.Close;
    GetFileClient.Address := StrToken(', ', Temp);    { IP          }
    GetFileClient.Port    := Temp;                    { Port        }
    GetFileClient.Open(TBufferedSocket);  { only if I'm not sending the file }
  end
  else
    mChat.Lines.Add( Line );  { Just some text }
end;

procedure TChatWindow.ChatClientConnect(Sender: TObject; Socket: TSocketBase);
begin
  if Not Socket.Connected then
  begin
    mChat.Lines.Add('Server: Connect Failed (Server not Present)');
    ChatClientDisconnect(Sender, Socket);
  end;
end;

procedure TChatWindow.ChatClientDisconnect(Sender: TObject; Socket: TSocketBase);
begin
  bChat.Down := False;
end;

procedure TChatWindow.FormCreate(Sender: TObject);
var
  Done: boolean;
begin
  mText.lines.Clear;
  mChat.Lines.Clear;
  UserName := ChatClient.Address;
  IncomingPath := ExtractFilepath(Application.Exename);
  ShowMessage('This is a multi-user chat demo. To use it best, run it on more than'#13+
              'one PC with a least one copy in server mode. Then connect all of the'#13+
              'other copies to that server''s address. You will then be able to chat'#13+
              'to each other much like IRC');
  Done := False;
  repeat
    try
      SendFileServer.Listen(TBufferedSocket);
      Done := True;
    except
      SendFileServer.Port := IntToStr(SendFileServer.GetPort + 1);
    end;
  until Done;
  SetItUp;
end;

procedure TChatWindow.mTextKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key = vk_Return then
    while MText.Lines.Count > 0 do
    begin
      (ChatClient.Conn as TLineSocket).WriteLine(format('%s: %s',
                                   [ UserName, MText.Lines[0]]));
      MText.Lines.Delete(0);
    end;

end;

procedure TChatWindow.GetFileClientFileReceived(socket: TBufferedSocketBase);
begin
  DnldMessage.Caption := format('Received %s: %d bytes', [ReceiveFile, ReceiveSize]);
  mChat.Lines.Add('Server: '+DnldMessage.Caption);
end;

procedure TChatWindow.GetFileClientConnect(Sender: TObject;
  Socket: TSocketBase);
Var
  C: char;
begin
  if (bSendFile.Down) and           { Check to see if we are the sender }
     (ExtractFileName(ReceiveFile) = SendFile) then
    C := #253                           { Refused by Sender (we already have the file) }
  else
  if (cbAcceptFiles.Checked) then       { Want to accept a file }
  begin
    { Get ready to Receive the file }
    (Socket as TBufferedSocket).RecvFile(ReceiveFile, ReceiveSize);
    mChat.Lines.Add(format('Server: Recieving %s: %d bytes', [ReceiveFile, ReceiveSize]));
    C := #255;                          { Accept the file  }
  end
  else
    C := #254;                          { refuse the file }
  (Socket as TBufferedSocket).WriteTobuffer(C, 1);
  (Socket as TBufferedSocket).SendBuffer;
end;

procedure TChatWindow.SendFileServerFileSent(socket: TBufferedSocketBase);
begin
  DnldMessage.Caption := format('%s Received %s: %d bytes',
                         [Socket.RemoteAddress,SendFile, SendSize]);
  mChat.Lines.Add('Server: '+DnldMessage.Caption);
  bSendFile.Tag := bSendFile.Tag - 1;  { Track how many we are sending }
  bSendFile.Down := bSendFile.Tag > 0; { Bring the button up if we are done }
end;

procedure TChatWindow.SendFileServerRead(Sender: TObject; Socket: TSocketBase);
Var
  C: char;
begin
  (Socket as TBufferedSocket).ReadFromBuffer(C, 1);
  (Socket as TBufferedSocket).ClearReadBuffer;
  case  C of
    #255: begin
            bSendFile.Tag := bSendFile.Tag + 1;  { Track how many we are sending }
            (Socket as TBufferedSocket).SendFile(SendFileDlg.FileName);
            (Socket as TBufferedSocket).SendBuffer;
          end;
    #254,
    #253: begin
            Socket.Close;
            if C = #254 then  { Notify sender that remote refused }
            begin
              DnldMessage.Caption := format('%s Refused: %s',
                                     [Socket.RemoteAddress,SendFile]);
              mChat.Lines.Add('Server: '+DnldMessage.Caption);
            end;
          end;
  end;
end;

procedure TChatWindow.GetFileClientRead(Sender: TObject; Socket: TSocketBase);
begin
  if (Socket as TBufferedSocket).ReceivingStream then  { show progress }
  begin
    DnldMessage.Caption := format('%s: %d bytes left',
          [ReceiveFile, ReceiveSize - (Socket as TBufferedSocket).RecvStreamPos]);
    Application.processMessages;
  end
  else
    Socket.Close;
end;

procedure TChatWindow.bExitClick(Sender: TObject);
begin
  Close;
end;

procedure TChatWindow.bSetupClick(Sender: TObject);
begin
  if ((not ChatClient.Connected) and (ChatServer.ClientCount = 0)) or
     (MessageDlg('Connections will be lost during setup,'#13'Loose connections and continue with setup?',
                  mtConfirmation, [mbYes, mbNo], 0) = mrYes) then
  begin
    with TSetupDlg.Create(self) do
    try
      if ChatClient.Connected then
      begin
        ChatClient.Close;
        bChat.Down := False;
      end;
      ChatServer.CloseDown;
      { Transfer Defaults }
      eHost.Text     := ChatClient.Address;
      ename.Text     := UserName;
      eIncoming.text := IncomingPath;
      mHosts.Lines.Assign(ChatServer.RestrictedHosts);
      cbAccept.Checked     := ChatServer.Restriction = hrAccept;
      cbReverseDNS.Checked := ChatServer.AllowOnlyIfReverseDNS;

      if ShowModal = mrOK then
      begin
        ChatClient.Address := eHost.Text;
        UserName           := ename.Text;

        IncomingPath       := eIncoming.text;
        if IncomingPath[Length(IncomingPath)] <> '\' then
          IncomingPath := IncomingPath + '\';

        ChatServer.RestrictedHosts.Assign(mHosts.Lines);
        if cbAccept.Checked then
          ChatServer.Restriction := hrAccept
        else
          ChatServer.Restriction := hrDeny;
        ChatServer.AllowOnlyIfReverseDNS := cbReverseDNS.Checked;
        SetItUp;
      end;
    finally
      Free
    end;
  end;
end;

procedure TChatWindow.SetItUp;
begin
  PnlMessage.Caption := format('Press Chat to connect to %s', [ChatClient.Address]);
  bChat.hint := PnlMessage.Caption;
  DNldMessage.Caption := Format('Downloads dir: %s', [incomingpath]);
  if cbServer.Checked then
    cbServerClick(Self);
  bChat.Down := ChatClient.Connected;
end;

procedure TChatWindow.bChatClick(Sender: TObject);
begin
  ChatClient.Close; { In case it's open }
  if bChat.Down then
    ChatClient.Open(TLineSocket);
end;

procedure TChatWindow.cbServerClick(Sender: TObject);
begin
  if cbServer.Checked then
  try
    ChatServer.Listen(TLineSocket)
  except
    cbServer.Checked := False;
  end
  else
    ChatServer.CloseDown;
end;

procedure TChatWindow.bSendFileClick(Sender: TObject);
var
  FileHeader: string;
  SR: TSearchRec;
  Sending: boolean;
begin
  Sending := False;
  try
    if GetFileClient.Connected then
    begin
      ShowMessage('You are currently receiving a file,'#13+
                  'Please wait until the files has been received');
      exit;
    end;

    if not bSendFile.Down then
    begin
      if (MessageDlg('File Transmision in progress'#13'Stop Transfer and send a new file??',
                    mtConfirmation, [mbYes, mbNo], 0) = mrNo) then
      begin
        Sending := True;
        Exit;
      end
      else
        SendFileServer.CloseDown;
    end;

    if ChatClient.Connected then
    begin
      if SendFileDlg.Execute then
      begin
        if FindFirst(SendFileDlg.FileName, faAnyFile, SR) = 0 then
        begin
          Sending := True;
          SendFile := ExtractFilename(SendFileDlg.FileName);
          SendSize := SR.Size;
          FindClose(SR);
          Fileheader := format(#255'"%s",%d,%s,%s',
                        [SendFile, SendSize, ChatClient.Conn.LocalAddress,
                         SendFileServer.Port]);
          (ChatClient.Conn as TLineSocket).WriteLine(FileHeader);
        end;
      end;
    end
    else
    begin
      ShowMessage('You must be connected (chatting) first');
    end;
  finally
    bSendFile.Down := Sending;
  end;
end;


procedure TChatWindow.GetFileClientDisconnect(Sender: TObject;
  Socket: TSocketBase);
begin
  Socket.Close;
end;

end.

