(*****************************************************************************
 * dWinsock component for Borland Delphi.
 *
 * (C) 1995 by Ulf Sderberg, ulfs@sysinno.se
 *             Marc Palmer,   marc@anyware.co.uk
 *             Keith Hawes,   khawes@ccmail.com
 *
 *  buffsock.pas - classes derived from tSocket to handle buffered and line
 *                 oriented IO.
 *
 * $Revision:   1.13  $
 *
 * -- History --
 * $Log:   //SUPPORT_SYSTEMS/PVCS/source/Buffsock._pas  $
 * 
 *    Rev 1.13   23 Jun 1997 15:38:08   khawes
 * Changes for 2.51
 * 
 * 
 *    Rev 1.12   04 Jun 1997 09:18:34   khawes
 * Version 2.51
 * 
 * 
 *    Rev 1.11   04 Jun 1997 09:13:56   khawes
 * Call DoDisconnect on a ConnResetByPeer
 * Call SendBuffer in DefSendStream
 * Change Marc's email Address
 * 
 * 
 *    Rev 1.10   19 May 1997 16:09:16   khawes
 * Name Change TCustom*Socket to TBasicCustom*Socket
 * Added SendBuffer call to SendFile
 * 
 * 
 *    Rev 1.9   25 Apr 1997 16:12:28   khawes
 * Added Override of Close to TBufferedSocket to kill any active streams.
 *
 * 
 *    Rev 1.8   25 Apr 1997 15:13:44   khawes
 * Features & Chages for Beta of 2.50
 * 
 * 
 *    Rev 1.7   26 Mar 1997 16:00:34   khawes
 * More bug fixes from Marc.
 * 
 * 
 *    Rev 1.6   26 Mar 1997 15:45:56   khawes
 * More BuffSock changes.  FileSend & Recieve is complete
 * 
 * 
 *    Rev 1.5   14 Mar 1997 16:17:32   khawes
 * Added new TBinaryClient/Server and TTextCient/Server Components
 * 
 * 
 *    Rev 1.4   28 Jan 1997 09:48:00   khawes
 * MP: Misc changes & Bug fixes
 * Changed to allow Close to be called even if the socket is already closed
 * Abort pending async functions on a close.
 *
 *
 *    Rev 1.3   11 Nov 1996 14:47:08   khawes
 *
 *    Rev 1.2   21 Oct 1996 16:34:52   khawes
 * * No more GPFs when setting read or write buffer size
 * * No (or less) duplicate OnReadBufferXXX and OnWriteBufferXXX events
 * * Multiple lines received in one burst are now handled correctly in
 *   TLineSocket
 * * ClearXXXBuffer will now work correctly.
 * 
 * Many thanks to Peter Guletsky and Billy McDaniel for their patience
 * and support in finding these and other bugs in BuffSock's debut
 * release.
 * 
 * 
 *    Rev 1.1   07 Aug 1996 08:52:50   KHAWES
 * Changed uses statment to remofe ifdef win32 and use jsut dWinsock instead.
 * Ran through NoTabs to switch tabs to spaces.
 * 
 * 
 *    Rev 1.0   05 Aug 1996 11:40:52   KHAWES
 * classes derived from tSocket to handle buffered and line oriented IO.
 *
 *****************************************************************************)

unit buffsock;

interface

uses SysUtils, Classes, WSAPI, dWinsock, dwsError;

const

  BS_ERR_BlockSizeTooBig = 'Block size is larger than buffer size.';
  BS_ERR_PullUpAmountTooBig = 'Cannot pull up buffer - amount to pull up larger than total buffer.';
  BS_ERR_BufferSizeTooSmall = 'Cannot change buffer size - data will be lost.';
  BS_ERR_BlockSizeTooSmall = 'Cannot read block/record - size is zero.';
  BS_ERR_InvalidBufferSize = 'Cannot change buffer size - data will be lost.';
  BS_ERR_ReadBufferOverFlow = 'Read buffer full, user did not flush';
  BS_ERR_WriteBufferOverFlow = 'Write buffer full, user did not flush';
  BS_ERR_PascalString = 'Attempt to read more than 255 chars into a Delphi 1.0x Pascal string.';
  BS_ERR_IllegalMethod = 'Recv or Send cannot be called on a buffered socket.';
  BS_ERR_BuffFullNoLineChars = 'Receiving line that is too big.';

  charLF : string[1] = #10;
  charCR : string[1] = #13;
  charCRLF : string[2] = #13#10;

type
  EBufferedSocketException = class(Exception);

  TBufferedSocketBase = class;

  TBufferedSocketClass = class of TSocketBase;

  TBufferedSocketNotifyEvent = procedure ( socket : TBufferedSocketBase ) of object;

  TLineSocketType = (lstCR, lstLF, lstCRLF);

  TBufferedSocketBase = class(TStreamSocket)
  private
    FOnWriteBufferEmpty, FOnWriteBufferFull,
    FOnReadBufferEmpty, FOnReadBufferFull,
    FOnWriteBufferChange, FOnReadBufferChange,
    FOnFileSent, FOnFileReceived : TBufferedSocketNotifyEvent;
    FDestroySendStream, FDestroyRecvStream,
    FSendingStream, FReceivingStream : Boolean;

    procedure DoRead; override;
    procedure DoWrite; override;

    procedure AllocReadBuf;
    procedure AllocWriteBuf;

    procedure FreeReadBuf;
    procedure FreeWriteBuf;

    procedure PullUpReadBuffer( amount : u_int);
    procedure PullUpWriteBuffer( amount : u_int);

    { These function raise exceptions if we are going to lose data
      as a result of the action. }
    procedure SetReadBufSize( sz : u_int);
    procedure SetWriteBufSize( sz : u_int);
    procedure SetRecordSize( sz : u_int);

    function GetReadBufferBytesFree : u_int;
    function GetWriteBufferBytesFree : u_int;

    procedure KillSendStream;
    procedure SendStreamData;
    procedure KillRecvStream;
    procedure RecvStreamData;
  protected
    FReadBuffer : Pointer;     { allocate as necessary }
    FWriteBuffer : Pointer;    { allocate as necessary }
    FReadBufSize : u_int;  { size of current alloc'd buffers }
    FWriteBufSize : u_int;  { size of current alloc'd buffers }
    FReadBufPos : u_int;  { offset of current alloc'd buffers }
    FWriteBufPos : u_int;  { offset of current alloc'd buffers }
    FRecordSize : u_int;
    FRecvStream, FSendStream : TStream;
    FRecvStreamSize : Longint;

    procedure DoReadBufferEmpty; virtual; { call OnReadxxxx event }
    procedure DoReadBufferFull; virtual; { call OnReadxxxx event }
    procedure DoWriteBufferEmpty; virtual; { call OnReadxxxx event }
    procedure DoWriteBufferFull; virtual; { call OnReadxxxx event }
    procedure DoWriteBufferChange; virtual; { call OnReadxxxx event }
    procedure DoReadBufferChange; virtual; { call OnReadxxxx event }
    procedure DoFileSent; virtual; { call OnReadxxxx event }
    procedure DoFileReceived; virtual; { call OnReadxxxx event }

    { this function will read from the in buffer up to bufsize,
      or up to FBytesIn_InBuffer, whichever is the lower.
      Should return number of bytes read - i.e. 0 means nothing to
      read}
    function DefReadFromBuffer( var destbuf; bufsize : u_int ) : u_int; virtual;

    { read data from buffer and write it to a stream }
    function DefReadFromBufferToStream( AStream : TStream;
      count : u_int ) : u_int; virtual;

    { Attempt to read 'blocksize' bytes. If < blocksize bytes in read buffer,
     return FALSE. Exception raised if bufsize > FReadBufSize }
    function DefReadBlockFromBuffer( var destbuf; bufsize : u_int ) : Boolean; virtual;
    { Attempt to read FRecordSize bytes. If < FRecordSize bytes in read buffer,
     return FALSE. Exception raised if FRecordSize > FReadBufSize }
    function DefReadRecordFromBuffer( var destbuf ) : Boolean; virtual;

    { this function will write to the out buffer up to bufsize,
      or up to (FBufSize - FBytesIn_InBuffer), whichever is the lower.
      Should return number of bytes written - i.e. 0 means there
      is no buffer space - the data does not actually get sent here,
      that is done by a call to SendBuffer in your OnWrite handler. }
    function DefWriteToBuffer( var srcbuf; bufsize : u_int ) : u_int; virtual;

    { read data from stream and write to buffer }
    function DefWriteToBufferFromStream( AStream : TStream;
      count : u_int ) : u_int; virtual;

    { Attempt to write 'blocksize' bytes. If < blocksize space in write buffer,
     return FALSE. Exception raised if bufsize > FWriteBufSize }
    function DefWriteBlockToBuffer( var srcbuf; bufsize : u_int ) : Boolean; virtual;
    { Attempt to write FRecordSize bytes. If < FRecordSize bytes free in write buffer,
      return FALSE. Exception raised if FRecordSize > FWriteBufSize }
    function DefWriteRecordToBuffer( var srcbuf ) : Boolean; virtual;

{ LINE BASED FUNCTIONS }
  { Returns length of line in buffer starting at char index.
    Returns High(u_int) if no line in buffer.
    This is used to do quick checks for the end of a line just being
    received. }
    function  DefGetLineLengthFrom( startidx : u_int;
      const linechars : string) : u_int;

  { Returns length of first line in buffer. Returns High(u_int) if no
    line in buffer }
    function  DefGetLineLength( const linechars : string) : u_int;

  { Return TRUE if data contains CR, LF or CRLF according to FOptions }
    function  DefIsLineInBuffer( const linechars : string) : Boolean;

  { Return TRUE if data contains CR, LF or CRLF according to FOptions,
    and if TRUE gives the user the line as a O-Pascal string.
    Limited to 256 char lines under Delphi 1 - use DefStrReadLine instead. }
    function  DefReadLine(var l : string; const linechars : string;
      withlinechars : Boolean) : Boolean; virtual;

  { Writes the contents of the string to the buffer, and adds
    the correct CR, LF or CRLF as per the FOptions.
    Return TRUE if data can fit into the buffer.
    Limited to 256 char lines under Delphi 1 - use DefStrReadLine instead. }
    function  DefWriteLine(const l : string;
      const linechars : string) : Boolean; virtual;

  { Return TRUE if data contains CR, LF or CRLF according to FOptions,
    and if TRUE copies the line to the PChar array, up to BUFSIZE chars.
    WARNING: If your buffer is too small, the rest of the line will be lost.}
    function  DefStrReadLine(l : PChar; bufsize : u_int;
      const linechars : string; withlinechars : Boolean ) : Boolean; virtual;

  { Writes the contents of the PChar to the buffer, and adds
    the correct CR, LF or CRLF as per the FOptions.
    Return TRUE if data can fit into the buffer. }
    function  DefStrWriteLine(l : PChar;
      const linechars : string) : Boolean; virtual;

  { For Delphi 1 users with lines > 255 chars }
    function  DefRebuildLongLine( dest : PChar; buflen : Integer;
      const firstpart : string; const linechars : string;
      withlinechars : Boolean) : Boolean;

  { File Transfer functions }
    procedure DefSendFile( const fn : string);
    procedure DefRecvFile( const fn : string; size : Longint);
    procedure DefSendStream( s : TStream; DestroyOnComplete : Boolean);
    procedure DefRecvStream( s : TStream; size : Longint;
      DestroyOnComplete : Boolean);

    procedure DefCancelRecvStream;
    procedure DefCancelSendStream;

    function GetSendStreamSize : Longint;
    function GetSendStreamPos : Longint;
    function GetRecvStreamSize : Longint;
    function GetRecvStreamPos : Longint;

    property SendingStream : Boolean read FSendingStream;
    property SendStreamSize : Longint read GetSendStreamSize;
    property SendStreamPos : Longint read GetSendStreamPos;
    property ReceivingStream : Boolean read FReceivingStream;
    property RecvStreamSize : Longint read GetRecvStreamSize;
    property RecvStreamPos : Longint read GetRecvStreamPos;

    property ReadBufferBytesUsed : u_int read FReadBufPos;
    property ReadBufferBytesFree : u_int read GetReadBufferBytesFree;
    property WriteBufferBytesUsed : u_int read FWriteBufPos;
    property WriteBufferBytesFree : u_int read GetWriteBufferBytesFree;
    property ReadBufferSize : u_int read FReadBufSize write SetReadBufSize;
    property WriteBufferSize : u_int read FWriteBufSize write SetWriteBufSize;
    property RecordSize : u_int read FRecordSize write SetRecordSize;

    property OnReadBufferChange : TBufferedSocketNotifyEvent
      read FOnReadBufferChange write FOnReadBufferChange;
    property OnReadBufferEmpty : TBufferedSocketNotifyEvent
      read FOnReadBufferEmpty write FOnReadBufferEmpty;
    property OnReadBufferFull : TBufferedSocketNotifyEvent
      read FOnReadBufferFull write FOnReadBufferFull;

    property OnWriteBufferChange : TBufferedSocketNotifyEvent
      read FOnWriteBufferChange write FOnWriteBufferChange;
    property OnWriteBufferFull : TBufferedSocketNotifyEvent
      read FOnWriteBufferFull write FOnWriteBufferFull;
    property OnWriteBufferEmpty : TBufferedSocketNotifyEvent
      read FOnWriteBufferEmpty write FOnWriteBufferEmpty;
    property OnFileSent : TBufferedSocketNotifyEvent
      read FOnFileSent write FOnFileSent;
    property OnFileReceived : TBufferedSocketNotifyEvent
      read FOnFileReceived write FOnFileReceived;

  public

    constructor Create(AParent : TSockCtrl); override; { default buf sizes etc. }
    destructor Destroy; override; { free bufs etc. }

    procedure ClearReadBuffer;
    procedure ClearWriteBuffer;

    function Send(var buf; cnt : integer) : integer; override;
    function Recv(var buf; cnt : integer) : integer; override;
    procedure Close; override;

    { This function should be called in your OnWrite handler,
      it should try to call TStreamSocket.Send for the number of
      bytes in the out buffer, to send as much as possible. The
      remaining bytes in the Out_Buffer must be pulled-up to the top.}
    procedure SendBuffer; virtual;

    { This function should be called in your OnRead handler,
      it should try to call TStreamSocket.Recv for the number of
      bytes free in the IN buffer, to receive as much as possible.}
    procedure RecvBuffer; virtual;

    function PeekReadBuffer( var destbuf; bufsize : u_int ) : u_int;
    function PeekWriteBuffer( var destbuf; bufsize : u_int ) : u_int;

  end;

  TBufferedSocket = class (TBufferedSocketBase)
  public

    function ReadFromBuffer( var destbuf; bufsize : u_int ) : u_int; virtual;
    function ReadFromBufferToStream( AStream : TStream;
      count : u_int ) : u_int; virtual;
    { Attempt to read 'blocksize' bytes. If < blocksize bytes in read buffer,
     return FALSE. Exception raised if bufsize > FReadBufSize }
    function ReadBlockFromBuffer( var destbuf; bufsize : u_int ) : Boolean; virtual;
    { Attempt to read FRecordSize bytes. If < FRecordSize bytes in read buffer,
     return FALSE. Exception raised if FRecordSize > FReadBufSize }
    function ReadRecordFromBuffer( var destbuf ) : Boolean; virtual;

    { this function will write to the out buffer up to bufsize,
      or up to (FBufSize - FBytesIn_InBuffer), whichever is the lower.
      Should return number of bytes written - i.e. 0 means there
      is no buffer space - the data does not actually get sent here,
      that is done by a call to SendBuffer in your OnWrite handler. }
    function WriteToBuffer( var srcbuf; bufsize : u_int ) : u_int; virtual;
    function WriteToBufferFromStream( AStream : TStream;
      count : u_int ) : u_int; virtual;
    { Attempt to write 'blocksize' bytes. If < blocksize space in write buffer,
     return FALSE. Exception raised if bufsize > FWriteBufSize }
    function WriteBlockToBuffer( var srcbuf; bufsize : u_int ) : Boolean; virtual;
    { Attempt to write FRecordSize bytes. If < FRecordSize bytes free in write buffer,
      return FALSE. Exception raised if FRecordSize > FWriteBufSize }
    function WriteRecordToBuffer( var srcbuf ) : Boolean; virtual;

    procedure SendFile( const fn : string);
    procedure RecvFile( const fn : string; size : Longint);
    procedure SendStream( s : TStream; DestroyOnComplete : Boolean);
    procedure RecvStream( s : TStream; size : Longint;
      DestroyOnComplete : Boolean);
    procedure CancelRecvStream;
    procedure CancelSendStream;

    property SendingStream;
    property SendStreamSize;
    property SendStreamPos;
    property ReceivingStream;
    property RecvStreamSize;
    property RecvStreamPos;

    property ReadBufferBytesUsed;
    property ReadBufferBytesFree;
    property WriteBufferBytesUsed;
    property WriteBufferBytesFree;
    property ReadBufferSize;
    property WriteBufferSize;
    property RecordSize;

    property OnReadBufferChange;
    property OnReadBufferEmpty;
    property OnReadBufferFull;

    property OnWriteBufferChange;
    property OnWriteBufferFull;
    property OnWriteBufferEmpty;
  end;

  TLineSocketOptions = (  lsoAddLineChars,
                          lsoAutoFlushAfterWrite,
                          lsoReturnLineChars,
                          lsoLineReceivedEvents);
  TLineSocketOptionsSet = set of TLineSocketOptions;

  TLineSocketBase = class;
  TLineReceivedEvent =
    procedure ( socket : TLineSocketBase; const line : string;
                complete : Boolean) of object;

  TLineSocketBase = class (TBufferedSocketBase)
  private
    function LineReceived : Boolean;

  protected
    FLineType : TLineSocketType;
    FOnLineReceived : TLineReceivedEvent;
    FLineOptions : TLineSocketOptionsSet;

    function GetLineChars : string;

    procedure DoReadBufferFull; override; { call OnReadxxxx event }

    { Override to implement lsoAutoFlushAfterWrite }
    function  DefWriteLine(const l : string;
      const linechars : string) : Boolean; override;
    function  DefStrWriteLine(l : PChar;
      const linechars : string) : Boolean; override;

    procedure DoOnLineReceived( const line : string; complete : Boolean); virtual;

    property LineOptions : TLineSocketOptionsSet
      read FLineOptions write FLineOptions;
    property LineType : TLineSocketType
      read FLineType write FLineType;
    property OnLineReceived : TLineReceivedEvent
      read FOnLineReceived write FOnLineReceived;

  public
    constructor Create(AParent : TSockCtrl); override; { default buf sizes etc. }

    procedure RecvBuffer; override;
  end;

  TLineSocket = class (TLineSocketBase)
  public
    function  GetLineLength : u_int;
    function  IsLineInBuffer : Boolean;
    function  ReadLine(var l : string) : Boolean;
    function  WriteLine(const l : string) : Boolean;
    function  StrReadLine(l : PChar; bufsize : u_int) : Boolean;
    function  StrWriteLine(l : PChar) : Boolean;

    { For Delphi 1 users with lines > 255 chars }
    function RebuildLongLine( dest : PChar; buflen : Integer;
      const firstpart : string) : Boolean;

    property LineType;

    property LineOptions;
    property OnLineReceived;

    property ReadBufferBytesUsed;
    property ReadBufferBytesFree;
    property WriteBufferBytesUsed;
    property WriteBufferBytesFree;
    property ReadBufferSize;
    property WriteBufferSize;

    property OnReadBufferChange;
    property OnReadBufferEmpty;
    property OnReadBufferFull;

    property OnWriteBufferChange;
    property OnWriteBufferFull;
    property OnWriteBufferEmpty;
  end;

  TBufferedClientBase = class(TCustomClientSocket)
  private
    FOnWriteBufferEmpty,
    FOnReadBufferEmpty,
    FOnWriteBufferFull,
    FOnReadBufferFull,
    FOnWriteBufferChange,
    FOnReadBufferChange,
    FOnFileSent,
    FOnFileReceived:      TBufferedSocketNotifyEvent;

    FReadBufSize : u_int;
    FWriteBufSize : u_int;
    FRecordSize : u_int;

  protected
    property OnWriteBufferEmpty: TBufferedSocketNotifyEvent
             read FOnWriteBufferEmpty write FOnWriteBufferEmpty;
    property OnReadBufferEmpty: TBufferedSocketNotifyEvent
             read FOnReadBufferEmpty write FOnReadBufferEmpty;
    property OnWriteBufferFull: TBufferedSocketNotifyEvent
             read FOnWriteBufferFull write FOnWriteBufferFull;
    property OnReadBufferFull: TBufferedSocketNotifyEvent
             read FOnReadBufferFull write FOnReadBufferFull;
    property OnWriteBufferChange: TBufferedSocketNotifyEvent
             read FOnWriteBufferChange write FOnWriteBufferChange;
    property OnReadBufferChange: TBufferedSocketNotifyEvent
             read FOnReadBufferChange write FOnReadBufferChange;
    property OnFileSent: TBufferedSocketNotifyEvent
             read FOnFileSent write FOnFileSent;
    property OnFileReceived: TBufferedSocketNotifyEvent
             read FOnFileReceived write FOnFileReceived;

    property ReadBufferSize : u_int read FReadBufSize write FReadBufSize;
    property WriteBufferSize : u_int read FWriteBufSize write FWriteBufSize;
    property RecordSize : u_int read FRecordSize write FRecordSize;

    function CreateConn: TSocketBase; override;

  public
    { Public declarations }
    constructor Create(AOwner : TComponent); override;

  published
    { Published declarations }
  end;

  TBufferedServerBase = class(TCustomBasicServerSocket)  { Base }
  private
    FOnWriteBufferEmpty,
    FOnReadBufferEmpty,
    FOnWriteBufferFull,
    FOnReadBufferFull,
    FOnWriteBufferChange,
    FOnReadBufferChange,
    FOnFileSent,
    FOnFileReceived:      TBufferedSocketNotifyEvent;

    FReadBufSize : u_int;
    FWriteBufSize : u_int;
    FRecordSize : u_int;

  protected
    property OnWriteBufferEmpty: TBufferedSocketNotifyEvent
             read FOnWriteBufferEmpty write FOnWriteBufferEmpty;
    property OnReadBufferEmpty: TBufferedSocketNotifyEvent
             read FOnReadBufferEmpty write FOnReadBufferEmpty;
    property OnWriteBufferFull: TBufferedSocketNotifyEvent
             read FOnWriteBufferFull write FOnWriteBufferFull;
    property OnReadBufferFull: TBufferedSocketNotifyEvent
             read FOnReadBufferFull write FOnReadBufferFull;
    property OnWriteBufferChange: TBufferedSocketNotifyEvent
             read FOnWriteBufferChange write FOnWriteBufferChange;
    property OnReadBufferChange: TBufferedSocketNotifyEvent
             read FOnReadBufferChange write FOnReadBufferChange;
    property OnFileSent: TBufferedSocketNotifyEvent
             read FOnFileSent write FOnFileSent;
    property OnFileReceived: TBufferedSocketNotifyEvent
             read FOnFileReceived write FOnFileReceived;

    property ReadBufferSize : u_int read FReadBufSize write FReadBufSize;
    property WriteBufferSize : u_int read FWriteBufSize write FWriteBufSize;
    property RecordSize : u_int read FRecordSize write FRecordSize;

    function CreateConn: TSocketBase; override;

  public
    { Public declarations }
    constructor Create(AOwner : TComponent); override;

  published
    { Published declarations }
  end;

  TTextClientBase = class(TBufferedClientBase)
  private
    FOnLineReceived : TLineReceivedEvent;
    FLineOptions : TLineSocketOptionsSet;
    FLineType : TLineSocketType;
  protected
    property OnLineReceived : TLineReceivedEvent
             read FOnLineReceived write FOnLineReceived;
    property LineOptions : TLineSocketOptionsSet
             read FLineOptions write FLineOptions;
    property LineType : TLineSocketType
             read FLineType write FLineType;

    function CreateConn: TSocketBase; override;
  public
    constructor Create(AOwner : TComponent); override;
  end;

  TTextServerBase = class(TBufferedServerBase)
  private
    FOnLineReceived : TLineReceivedEvent;
    FLineOptions : TLineSocketOptionsSet;
    FLineType : TLineSocketType;
  protected
    property OnLineReceived : TLineReceivedEvent
             read FOnLineReceived write FOnLineReceived;
    property LineOptions : TLineSocketOptionsSet
             read FLineOptions write FLineOptions;
    property LineType : TLineSocketType
             read FLineType write FLineType;

    function CreateConn: TSocketBase; override;
  public
    constructor Create(AOwner : TComponent); override;
  end;


  TBinaryClient = class(TBufferedClientBase)
    published
      { Standard Properties }
      property Address;
      property Port;
      property Protocol;

      property TimeOut;
      property Options;

      property OnInfo;
      property OnConnect;
      property OnTimeOut;
      property OnDisconnect;
      property OnRead;
      property OnWrite;
      property OnAsyncComplete;

      { Buffered Properties }
      property OnWriteBufferEmpty;
      property OnReadBufferEmpty;
      property OnWriteBufferFull;
      property OnReadBufferFull;
      property OnWriteBufferChange;
      property OnReadBufferChange;
      property OnFileSent;
      property OnFileReceived;

      property ReadBufferSize;
      property WriteBufferSize;
      property RecordSize;

  end;

  TBinaryServer = class(TBufferedServerBase)
    published
      { Standard Properties }
      property Address;
      property Port;
      property Protocol;

      property MaxConnections;
      property ClientOptions;
      property QueueLength;

      property OnInfo;
      property OnAccept;
      property OnReject;
      property OnDisconnect;
      property OnRead;
      property OnWrite;
      property OnAsyncComplete;

      { Buffered Properties }
      property OnWriteBufferEmpty;
      property OnReadBufferEmpty;
      property OnWriteBufferFull;
      property OnReadBufferFull;
      property OnWriteBufferChange;
      property OnReadBufferChange;
      property OnFileSent;
      property OnFileReceived;

      property ReadBufferSize;
      property WriteBufferSize;
      property RecordSize;

      property RestrictedHosts;
      property Restriction;
      property AllowOnlyIfReverseDNS;
  end;

  TTextClient = class(TTextClientBase)
    published
      { Standard Properties }
      property Address;
      property Port;
      property Protocol;

      property TimeOut;
      property Options;

      property OnInfo;
      property OnConnect;
      property OnTimeOut;
      property OnDisconnect;
      property OnRead;
      property OnWrite;
      property OnAsyncComplete;

      { Buffered Properties }
      property OnWriteBufferEmpty;
      property OnReadBufferEmpty;
      property OnWriteBufferFull;
      property OnReadBufferFull;
      property OnWriteBufferChange;
      property OnReadBufferChange;
      property OnFileSent;
      property OnFileReceived;
      property OnLineReceived;

      property ReadBufferSize;
      property WriteBufferSize;
      property RecordSize;
      property LineType;
      property LineOptions;

  end;

  TTextServer = class(TTextServerBase)
    published
      { Standard Properties }
      property Address;
      property Port;
      property Protocol;

      property MaxConnections;
      property ClientOptions;
      property QueueLength;

      property OnInfo;
      property OnAccept;
      property OnReject;
      property OnDisconnect;
      property OnRead;
      property OnWrite;
      property OnAsyncComplete;

      { Buffered Properties }
      property OnWriteBufferEmpty;
      property OnReadBufferEmpty;
      property OnWriteBufferFull;
      property OnReadBufferFull;
      property OnWriteBufferChange;
      property OnReadBufferChange;
      property OnFileSent;
      property OnFileReceived;
      property OnLineReceived;

      property ReadBufferSize;
      property WriteBufferSize;
      property RecordSize;
      property LineType;
      property LineOptions;

      property RestrictedHosts;
      property Restriction;
      property AllowOnlyIfReverseDNS;
  end;

implementation

{$ifndef Win32}

procedure SetLength(var s : string; l : Byte);
begin
  s[0] := Char(Ord(l));
end;

{$endif}

{
TBufferedSocketBase

This is to be used as a base class for general (binary) buffered sockets.

Provides methods for reading and writing to and from the buffer and
several events that occur when the buffer changes.
}

constructor TBufferedSocketBase.Create(AParent : TSockCtrl);
begin
  inherited Create(AParent);

  FReadBuffer := nil;
  FWriteBuffer := nil;
  FOnWriteBufferEmpty := nil;
  FOnWriteBufferFull := nil;
  FOnReadBufferEmpty := nil;
  FOnReadBufferFull := nil;
  FReadBufSize := 1024;
  FWriteBufSize := 1024;
  FReadBufPos := 0;
  FWriteBufPos := 0;
  FRecordSize := 16;

  AllocReadBuf;
  AllocWriteBuf;
end;

destructor TBufferedSocketBase.Destroy;
begin
  FreeReadBuf;
  FreeWriteBuf;
  inherited Destroy;
end;

function TBufferedSocketBase.Send(var buf; cnt : integer) : integer;
begin
  raise EBufferedSocketException.Create( BS_ERR_IllegalMethod );
end;

function TBufferedSocketBase.Recv(var buf; cnt : integer) : integer;
begin
  raise EBufferedSocketException.Create( BS_ERR_IllegalMethod );
end;

procedure TBufferedSocketBase.Close;
begin
  inherited Close;
  KillSendStream;
  KillRecvStream;
end;

procedure TBufferedSocketBase.AllocReadBuf;
begin
  FreeReadBuf;

  if FReadBufSize < 1 then
    raise EBufferedSocketException.Create( BS_ERR_InvalidBufferSize );

  FReadBuffer := AllocMem( FReadBufSize);
end;

procedure TBufferedSocketBase.AllocWriteBuf;
begin
  FreeWriteBuf;

  if FWriteBufSize < 1 then
    raise EBufferedSocketException.Create( BS_ERR_InvalidBufferSize );

  FWriteBuffer := AllocMem( FWriteBufSize);
end;

procedure TBufferedSocketBase.FreeReadBuf;
begin
  if Assigned(FReadBuffer) then
  begin
{$ifdef Win32}
    FreeMem( FReadBuffer );
{$else}
    FreeMem( FReadBuffer, FReadBufSize );
{$endif}
    FReadBuffer := nil;
  end;
end;

procedure TBufferedSocketBase.FreeWriteBuf;
begin
  if Assigned(FWriteBuffer) then
  begin
{$ifdef Win32}
    FreeMem( FWriteBuffer );
{$else}
    FreeMem( FWriteBuffer, FWriteBufSize );
{$endif}
    FWriteBuffer := nil;
  end;
end;

{ These function raise exceptions if we are going to lose data
  as a result of the action. }
procedure TBufferedSocketBase.SetReadBufSize( sz : u_int);
var
  newbuf : PChar;
begin
  if (sz < FReadBufPos) then
    raise EBufferedSocketException.Create( BS_ERR_BufferSizeTooSmall);

  if (sz = FReadBufSize) then Exit; { nothing to do }

  newbuf := AllocMem(sz);

  if FReadBufPos > 0 then
    Move(FReadBuffer^, newbuf^, FReadBufPos);

  FreeReadBuf;
  FReadBufSize := sz;
  FReadBuffer := newbuf;
end;

procedure TBufferedSocketBase.SetWriteBufSize( sz : u_int);
var
  newbuf : PChar;
begin
  if (sz < FWriteBufPos) then
    raise EBufferedSocketException.Create( BS_ERR_BufferSizeTooSmall);

  if (sz = FWriteBufSize) then Exit; { nothing to do }

  newbuf := AllocMem(sz);

  if FWriteBufPos > 0 then
    Move(FWriteBuffer^, newbuf^, FWriteBufPos);

  FreeWriteBuf;
  FWriteBufSize := sz;
  FWriteBuffer := newbuf;
end;

procedure TBufferedSocketBase.SetRecordSize( sz : u_int);
begin
  if (sz > FWriteBufSize) or (sz > FReadBufSize) then
    raise EBufferedSocketException.Create( BS_ERR_BufferSizeTooSmall);

  FRecordSize := sz;
end;

{ 'amount' is the number of bytes at the beginning of the buffer that
  should now be eliminated by copying the data above down. }
procedure TBufferedSocketBase.PullUpReadBuffer( amount : u_int);
begin
  if (amount > FReadBufSize) then
    raise EBufferedSocketException.Create( BS_ERR_PullUpAmountTooBig);

  if amount < FReadBufSize then
    Move( (@(PChar(FReadBuffer)[amount]))^, FReadBuffer^, FReadBufSize-amount);

  Dec(FReadBufPos, amount);
  if FReadBufPos = 0 then DoReadBufferEmpty;
end;

procedure TBufferedSocketBase.PullUpWriteBuffer( amount : u_int);
begin
  if (amount > FWriteBufSize) then
    raise EBufferedSocketException.Create( BS_ERR_PullUpAmountTooBig);

  Move( (@(PChar(FWriteBuffer)[amount]))^, FWriteBuffer^, FWriteBufSize-amount);
  Dec(FWriteBufPos, amount);
  if FWriteBufPos = 0 then DoWriteBufferEmpty;
end;

procedure TBufferedSocketBase.DoReadBufferEmpty; { call OnReadxxxx event }
begin
  if Assigned(FOnReadBufferEmpty) then FOnReadBufferEmpty(Self);
end;

procedure TBufferedSocketBase.DoReadBufferFull; { call OnReadxxxx event }
begin
  if Assigned(FOnReadBufferFull) then FOnReadBufferFull(Self);
  if FReadBufPos >= FReadBufSize then
    raise EBufferedSocketException.Create(BS_ERR_ReadBufferOverFlow);
end;

procedure TBufferedSocketBase.DoWriteBufferChange;{ call OnReadxxxx event }
begin
  if Assigned(FOnWriteBufferChange) then FOnWriteBufferChange(Self);
end;

procedure TBufferedSocketBase.DoReadBufferChange; { call OnReadxxxx event }
begin
  if Assigned(FOnReadBufferChange) then FOnReadBufferChange(Self);
end;

procedure TBufferedSocketBase.DoWriteBufferEmpty;{ call OnReadxxxx event }
begin
{ Check if file send has finished and notify app. This way we are sure
  that all of the data has been SENT, not just written to the buffer,
  because we are past EOF and the write buffer is empty, so we are done! }
  if SendingStream then
    if FSendStream.Size <= FSendStream.Position then
    begin
      KillSendStream;
      DoFileSent;
    end;

  if Assigned(FOnWriteBufferEmpty) then FOnWriteBufferEmpty(Self);

end;

procedure TBufferedSocketBase.DoWriteBufferFull;  { call OnReadxxxx event }
begin
  if Assigned(FOnWriteBufferFull) then FOnWriteBufferFull(Self);
{ Commented this out 14/7/96 - doesn't look right.
  if FWriteBufPos >= FWriteBufSize-1 then
    raise EBufferedSocketException.Create(BS_ERR_WriteBufferOverFlow);}
end;

procedure TBufferedSocketBase.ClearReadBuffer;
begin
  if Assigned(FReadBuffer) then
  begin
    FillChar( FReadBuffer^, FReadBufSize, #0);
    FReadBufPos := 0;
    DoReadBufferEmpty;
  end;
end;

procedure TBufferedSocketBase.ClearWriteBuffer;
begin
  if Assigned(FWriteBuffer) then
  begin
    FillChar( FWriteBuffer^, FWriteBufSize, #0);
    FWriteBufPos := 0;
    DoWriteBufferEmpty;
  end;
end;

function TBufferedSocketBase.GetReadBufferBytesFree : u_int;
begin
  Result := FReadBufSize - FReadBufPos;
end;

function TBufferedSocketBase.GetWriteBufferBytesFree : u_int;
begin
  Result := FWriteBufSize - FWriteBufPos;
end;

{ This function is called by DoWrite.
  It tries to call TStreamSocket.Send for the number of
  bytes in the out buffer, to send as much as possible. The
  remaining bytes in the Out_Buffer must be pulled-up to the top.
  In the OnWriteBufferChange event you can reload the buffer with data. }
procedure TBufferedSocketBase.SendBuffer;
var
  bsent : u_int;
begin
  if FWriteBufPos = 0 then Exit; { Don't send if nothing in buffer! }

  try
    repeat
      bsent := inherited Send( PChar(FWriteBuffer)^, FWriteBufPos);
      if bsent > 0 then
      begin
        PullUpWriteBuffer(bsent);

    { Hmmm. Is there a potential problem here? If the WriteBufferChange method
      calls SendBuffer to make sure more data gets sent, could this cause
      too much recursion and kill the stack?

      If so, maybe we should try to send some more data if we can after
      calling DoWritebufferChange. Then we would have to let users know
      that they must never call SendBuffer from within WriteBufferChange.
      Maybe we could block it with a flag }
       DoWriteBufferChange;

       if FSendingStream then
         SendStreamData;

      end;
    until bsent <= 0;
  except
    on E: ESockError do
    begin
      if LastError = WSAEConnReset then
        DoDisconnect
      else
        raise;
    end
    else
      raise;
  end;
end;

{ This function should be called in your OnRead handler,
  it should try to call TStreamSocket.Recv for the number of
  bytes free in the IN buffer, to receive as much as possible.}
procedure TBufferedSocketBase.RecvBuffer;
var
  bread : u_int;
begin
  try
    repeat
      bread := inherited Recv( PChar(FReadBuffer)[FReadBufPos],
        GetReadBufferBytesFree);
      if bread > 0 then
      begin
        Inc(FReadBufPos, bread);

        DoReadBufferChange;

        if FReceivingStream then
          RecvStreamData;

        if FReadBufPos = FReadBufSize then DoReadBufferFull;
      end;
    until (bread <= 0) or (Not Connected);
  except
    on E: ESockError do
    begin
      if LastError = WSAEConnReset then
        DoDisconnect
      else
        raise;
    end
    else
      raise;
  end;
end;

{ this function will read from the in buffer up to bufsize,
  or up to FBytesIn_InBuffer, whichever is the lower.
  Should return number of bytes read - i.e. 0 means nothing to
  read}
function TBufferedSocketBase.DefReadFromBuffer( var destbuf; bufsize : u_int ) : u_int;
var
  bytestoget : u_int;
begin
  bytestoget := bufsize;
  if bytestoget > FReadBufPos then bytestoget := FReadBufPos;
  if bytestoget > 0 then
  begin
    Move( FReadBuffer^, destbuf, bytestoget);
    PullUpReadBuffer(bytestoget);
  end;
  Result := bytestoget;
end;

{ read data from buffer and write it to a stream }
function TBufferedSocketBase.DefReadFromBufferToStream( AStream : TStream;
  count : u_int ) : u_int;
var
  bytestoget : u_int;
begin
  bytestoget := count;
  if bytestoget > FReadBufPos then bytestoget := FReadBufPos;
  if bytestoget > 0 then
  begin
    AStream.Write( FReadBuffer^, bytestoget);
    PullUpReadBuffer(bytestoget);
  end;
  Result := bytestoget;
end;

{ Attempt to read 'blocksize' bytes. If < blocksize bytes in read buffer,
 return FALSE. Exception raised if bufsize > FReadBufSize }
function TBufferedSocketBase.DefReadBlockFromBuffer( var destbuf; bufsize : u_int ) : Boolean;
var
  bytestoget : u_int;
begin
  Result := False;
  bytestoget := bufsize;
{ We are reading blocks, so fail if there is not enough data }
  if bytestoget > FReadBufPos then Exit;

  if bufsize > FReadBufSize then
    raise EBufferedSocketException.Create( BS_ERR_BlockSizeTooBig);

  if bytestoget > 0 then
  begin
    Move( FReadBuffer^, destbuf, bytestoget);
    PullUpReadBuffer(bytestoget);
    Result := True;
  end else raise EBufferedSocketException.Create( BS_ERR_BLOCKSizeTooSmall);
end;

{ Attempt to read FRecordSize bytes. If < FRecordSize bytes in read buffer,
 return FALSE. Exception raised if FRecordSize > FReadBufSize }
function TBufferedSocketBase.DefReadRecordFromBuffer( var destbuf ) : Boolean;
begin
  Result := DefReadBlockFromBuffer( destbuf, FRecordSize);
end;

{ this function will write to the out buffer up to bufsize,
  or up to (FBufSize - FBytesIn_InBuffer), whichever is the lower.
  Should return number of bytes written - i.e. 0 means there
  is no buffer space - the data does not actually get sent here,
  that is done by a call to SendBuffer in your OnWrite handler. }
function TBufferedSocketBase.DefWriteToBuffer( var srcbuf; bufsize : u_int ) : u_int;
var
  bytestoput, bytesfree : u_int;
begin
  bytestoput := bufsize;
  bytesfree := GetWriteBufferBytesFree;

  if bufsize > FWriteBufSize then
    raise EBufferedSocketException.Create( BS_ERR_WriteBufferOverflow);

  if bytestoput > bytesfree then bytestoput := bytesfree;

  if bytestoput > 0 then
  begin
    Move( srcbuf, (@(PChar(FWriteBuffer)[FWriteBufPos]))^, bytestoput);
    Inc(FWriteBufPos, bytestoput);
    if FWriteBufPos = FWriteBufSize then DoWriteBufferFull;
  end;
  Result := bytestoput;
end;

{ read data from stream and write to buffer }
function TBufferedSocketBase.DefWriteToBufferFromStream( AStream : TStream;
  count : u_int ) : u_int;
var
  bytestoput, bytesfree : u_int;
begin
  bytestoput := count;
  bytesfree := GetWriteBufferBytesFree;

  if count > FWriteBufSize then
    raise EBufferedSocketException.Create( BS_ERR_WriteBufferOverflow);

  if bytestoput > bytesfree then bytestoput := bytesfree;

  if bytestoput > 0 then
  begin
    AStream.Read( (@(PChar(FWriteBuffer)[FWriteBufPos]))^, bytestoput);
    Inc(FWriteBufPos, bytestoput);
    if FWriteBufPos = FWriteBufSize then DoWriteBufferFull;
  end;
  Result := bytestoput;
end;

{ Attempt to write 'blocksize' bytes. If < blocksize space in write buffer,
 return FALSE. Exception raised if bufsize > FWriteBufSize }
function TBufferedSocketBase.DefWriteBlockToBuffer( var srcbuf; bufsize : u_int ) : Boolean;
var
  bytesfree : u_int;
begin
{  Result := False;   Value Never used  }     
  bytesfree := GetWriteBufferBytesFree;
  if bufsize > bytesfree then
    raise EBufferedSocketException.Create( BS_ERR_BlockSizeTooBig);
{ We are writing blocks, so fail if there is not enough data }

  if bufsize > 0 then
  begin
    Move( srcbuf, (@(PChar(FWriteBuffer)[FWriteBufPos]))^, bufsize);
    Inc(FWriteBufPos, bufsize);
    if FWriteBufPos = FWriteBufSize then DoWriteBufferFull;
    Result := True;
  end else Result := True; {raise EBufferedSocketException.Create( BS_ERR_BLOCKSizeTooSmall);}
 end;

{ Attempt to write FRecordSize bytes. If < FRecordSize bytes free in write buffer,
  return FALSE. Exception raised if FRecordSize > FWriteBufSize }
function TBufferedSocketBase.DefWriteRecordToBuffer( var srcbuf ) : Boolean;
begin
  Result := DefWriteBlockToBuffer(srcbuf, FRecordSize);
end;

function TBufferedSocketBase.PeekReadBuffer( var destbuf; bufsize : u_int ) : u_int;
var
  bytestoget : u_int;
begin
  bytestoget := bufsize;
  if bytestoget > FReadBufPos then bytestoget := FReadBufPos;
  if bytestoget > 0 then
  begin
    Move( FReadBuffer^, destbuf, bytestoget);
  end;
  Result := bytestoget;
end;

function TBufferedSocketBase.PeekWriteBuffer( var destbuf; bufsize : u_int ) : u_int;
var
  bytestoget : u_int;
begin
  bytestoget := bufsize;
  if bytestoget > FWriteBufPos then bytestoget := FWriteBufPos;
  if bytestoget > 0 then
  begin
    Move( FWriteBuffer^, destbuf, bytestoget);
  end;
  Result := bytestoget;
end;

procedure TBufferedSocketBase.DoRead;
begin
  RecvBuffer;
  inherited DoRead;
end;

procedure TBufferedSocketBase.DoWrite;
begin
  SendBuffer;
  inherited DoWrite;
end;

{ Return length of the line in the buffer INCLUDING CR/LFs }
function  TBufferedSocketBase.DefGetLineLengthFrom( startidx : u_int;
  const linechars : string) : u_int;
var
  idx : longint;

  { search 'scanme' char buff for occurence of 'target' char buff.
    Why not StrPos? Well this will not stop scanning 'scanme' if it
    encounters #0 - which might happen if the user is sending Pascal strings! }
  function NotNullTermStrPos( const target : string; scanme : PChar;
    start, stop : Integer) : Integer;
  var
    I, TargetPos, TargetLen : Integer;
  begin
    TargetPos := 1;
    Result := -1;
    TargetLen := Length(target);
    for I := start to stop do
    begin
      if target[TargetPos] = scanme[I] then
      begin
        if TargetPos = 1 then Result := I; { record length up to here }
        Inc(TargetPos);
      end else
      begin
        TargetPos := 1; { start again }
        Result := -1;
      end;

      { Have we matched the whole string? }
      if TargetPos > TargetLen then
        Exit;
    end;
  end;

begin
  Result := High(u_int);

  if (not Assigned(FReadBuffer)) or (startidx >= FReadBufPos) then Exit;

  idx := NotNullTermStrPos( linechars, PChar(FReadBuffer),
    startidx, FReadBufPos-1);

  { Now return the length including the line chars! }
  if idx <> -1 then
    Result := idx+Length(linechars) { idx is the position of the first line char }
  else
    Result := High(u_int);
end;

function  TBufferedSocketBase.DefGetLineLength(const linechars : string) : u_int;
begin
  Result := DefGetLineLengthFrom(0, linechars);
end;

function  TBufferedSocketBase.DefIsLineInBuffer( const linechars : string) : Boolean;
begin
  Result := DefGetLineLength(linechars) <> High(u_int);
end;

function  TBufferedSocketBase.DefReadLine(var l : string;
  const linechars : string; withlinechars : Boolean) : Boolean;
var
  len : u_int;
begin
  Result := False;
  len := DefGetLineLength(linechars);
  if len = High(u_int) then Exit;

{$ifdef Win32}
  SetLength(l, len);
  DefReadFromBuffer( PChar(l)^, len);
{$else}
 { Check that we have a line less than 255 chars }
  if len > 255 then
  begin
  { Line is longer than 255 chars so read 255 chars and raise exception }
    DefReadFromBuffer( (@l[1])^, 255);
    SetLength(l, 255);
    raise EBufferedSocketException.Create(BS_ERR_PascalString);
  end else
  begin
  { Line can be read OK into Delphi 1 pascal string }
    DefReadFromBuffer( (@l[1])^, len);
    SetLength(l, len);
  end;
{$endif}

  if not (withlinechars) then
    SetLength(l, len-Length(linechars));

  Result := True;
end;

function  TBufferedSocketBase.DefWriteLine(const l : string;
  const linechars : string) : Boolean;
var
  len, lcharlen : u_int;
begin
  Result := False;

  len := Length(l);
  lcharlen := Length(linechars);

{ Make sure whole line, including line chars will fit before we even try! }
  if GetWriteBufferBytesFree < (lcharlen + len) then Exit;
{$ifdef Win32}
  Result := DefWriteBlockToBuffer( PChar(l)^, len );
{$else}
  Result := DefWriteBlockToBuffer( (@l[1])^, len );
{$endif}

  if not Result then Exit; { data write failed, forget about line chars! }

  if lcharlen <> 0 then
  begin
{$ifdef Win32}
    Result := DefWriteBlockToBuffer( PChar(linechars)^, lcharlen );
{$else}
    Result := DefWriteBlockToBuffer( (@linechars[1])^, lcharlen );
{$endif}
  end;
end;

function  TBufferedSocketBase.DefStrReadLine(l : PChar; bufsize : u_int;
  const linechars : string; withlinechars : Boolean) : Boolean;
var
  len : u_int;
begin
  Result := False;
  len := DefGetLineLength( linechars);
  if (len = High(u_int)) or (len > bufsize-1) then Exit;

  if not (withlinechars) then
    Dec(len, Length(linechars));

  DefReadFromBuffer( PChar(l)^, len);
  l[len-1] := #0;
end;

function  TBufferedSocketBase.DefStrWriteLine(l : PChar;
  const linechars : string) : Boolean;
var
  len, lcharlen : u_int;
begin
  Result := False;

  len := StrLen(l);
  lcharlen := Length(linechars);

{ Make sure whole line, including line chars will fit before we even try! }
  if GetWriteBufferBytesFree < (lcharlen + len) then Exit;

  Result := DefWriteBlockToBuffer( l^, len );

  if not Result then Exit; { data write failed, forget about line chars! }

  if lcharlen <> 0 then
  begin
{$ifdef Win32}
    Result := DefWriteBlockToBuffer( PChar(linechars)^, lcharlen );
{$else}
    Result := DefWriteBlockToBuffer( (@linechars[1])^, lcharlen );
{$endif}
  end;
end;

function TBufferedSocketBase.DefRebuildLongLine( dest : PChar; buflen : Integer;
  const firstpart, linechars : string; withlinechars : boolean) : Boolean;
var
  space, count : Integer;
begin
  Result := False;
  StrPLCopy( dest, firstpart, buflen);
  count := Length(firstpart);
  space := buflen-count;
  if space > 0 then Result := DefStrReadLine( dest+count, space,
    linechars, withlinechars);
end;

{ File/Stream Transfer functions }

function TBufferedSocketBase.GetSendStreamSize : Longint;
begin
  if Assigned(FSendStream) then
    Result := FSendStream.Size
  else raise
    Exception.Create( 'Cannot get stream size - not sending any stream');
end;

function TBufferedSocketBase.GetSendStreamPos : Longint;
begin
  if Assigned(FSendStream) then
    Result := FSendStream.Position
  else raise
    Exception.Create( 'Cannot get stream position - not sending any stream');
end;

function TBufferedSocketBase.GetRecvStreamSize : Longint;
begin
  if Assigned(FRecvStream) then
    Result := FRecvStream.Size
  else raise
    Exception.Create( 'Cannot get stream size - not receiving any stream');
end;

function TBufferedSocketBase.GetRecvStreamPos : Longint;
begin
  if Assigned(FRecvStream) then
    Result := FRecvStream.Position
  else raise
    Exception.Create( 'Cannot get stream position - not receiving any stream');
end;

procedure TBufferedSocketBase.DefSendStream( s : TStream;
  DestroyOnComplete : Boolean);
begin
  if FSendingStream then
    raise Exception.Create( 'Already sending a stream.');

  KillSendStream;
  FDestroySendStream := DestroyOnComplete;
  FSendingStream := True;
  FSendStream := s;
  FSendStream.Seek( 0, soFromBeginning );
  SendStreamData; { KH April 20 - Prime Write buffer }
  SendBuffer;
end;

procedure TBufferedSocketBase.DefRecvStream( s : TStream; size : Longint;
  DestroyOnComplete : Boolean);
begin
  if FReceivingStream then
    raise Exception.Create( 'Already receiving a stream.');

  KillRecvStream;

  FReceivingStream := True;
  FDestroyRecvStream := DestroyOnComplete;
  FRecvStreamSize := size;
  FRecvStream := s;
  RecvStreamData; { in case there is some in the buffer now! }
end;

procedure TBufferedSocketBase.DefCancelRecvStream;
begin
  KillRecvStream;
end;

procedure TBufferedSocketBase.DefCancelSendStream;
begin
  KillSendStream;
end;

procedure TBufferedSocketBase.KillSendStream;
begin
  if Assigned(FSendStream) and FDestroySendStream then FSendStream.Free;
  FSendStream := nil;
  FSendingStream := False;
end;

procedure TBufferedSocketBase.KillRecvStream;
begin
  if Assigned(FRecvStream) and FDestroyRecvStream then
    FRecvStream.Free;
  FRecvStream := nil;
  FReceivingStream := False;
end;

procedure TBufferedSocketBase.DefSendFile( const fn : string);
begin
  if FSendingStream then
    raise Exception.Create( 'Already sending a stream.');

  DefSendStream(TFileStream.Create( fn, fmOpenRead + fmShareDenyWrite), True);
end;

procedure TBufferedSocketBase.SendStreamData;
var
  btosend : u_int;
begin
  btosend := FSendStream.Size - FSendStream.Position;
  if btosend > WriteBufferBytesFree then
    btosend := WriteBufferBytesFree;

  DefWriteToBufferFromStream( FSendStream, btosend);
end;

procedure TBufferedSocketBase.DefRecvFile( const fn : string; size : Longint);
begin
  if FReceivingStream then
    raise Exception.Create( 'Already receiving a stream.');

  DefRecvStream(TFileStream.Create( fn, fmCreate), size, True);
end;

procedure TBufferedSocketBase.RecvStreamData;
var
  btoget : u_int;
begin
  btoget := FRecvStreamSize - FRecvStream.Position;
  if btoget > ReadBufferBytesUsed then
    btoget := ReadBufferBytesUsed;

  DefReadFromBufferToStream( FRecvStream, btoget);

  if FRecvStreamSize = FRecvStream.Position then
  begin
    KillRecvStream;
    DoFileReceived;
  end;
end;

procedure TBufferedSocketBase.DoFileSent;
begin
  if Assigned( FOnFileSent) then FOnFileSent(Self);
end;

procedure TBufferedSocketBase.DoFileReceived;
begin
  if Assigned( FOnFileReceived) then FOnFileReceived(Self);
end;

{
TBufferedSocket

This is to be used as a general (binary) buffered socket.

Provides methods for reading and writing to and from the buffer and
several events that occur when the buffer changes.
}

function TBufferedSocket.ReadFromBuffer( var destbuf; bufsize : u_int ) : u_int;
begin
  Result := DefReadFromBuffer( destbuf, bufsize);
end;

function TBufferedSocket.ReadFromBufferToStream( AStream : TStream;
  count : u_int ) : u_int;
begin
  Result := DefReadFromBufferToStream( AStream, count);
end;

{ Attempt to read 'blocksize' bytes. If < blocksize bytes in read buffer,
 return FALSE. Exception raised if bufsize > FReadBufSize }
function TBufferedSocket.ReadBlockFromBuffer( var destbuf; bufsize : u_int ) : Boolean;
begin
  Result := DefReadBlockFromBuffer( destbuf, bufsize);
end;

{ Attempt to read FRecordSize bytes. If < FRecordSize bytes in read buffer,
 return FALSE. Exception raised if FRecordSize > FReadBufSize }
function TBufferedSocket.ReadRecordFromBuffer( var destbuf ) : Boolean;
begin
  Result := DefReadRecordFromBuffer( destbuf );
end;

{ this function will write to the out buffer up to bufsize,
  or up to (FBufSize - FBytesIn_InBuffer), whichever is the lower.
  Should return number of bytes written - i.e. 0 means there
  is no buffer space - the data does not actually get sent here,
  that is done by a call to SendBuffer in your OnWrite handler. }
function TBufferedSocket.WriteToBuffer( var srcbuf; bufsize : u_int ) : u_int;
begin
  Result := DefWriteToBuffer( srcbuf, bufsize);
end;

function TBufferedSocket.WriteToBufferFromStream( AStream : TStream;
  count : u_int ) : u_int;
begin
  Result := DefWriteToBufferFromStream( AStream, count);
end;

{ Attempt to write 'blocksize' bytes. If < blocksize space in write buffer,
 return FALSE. Exception raised if bufsize > FWriteBufSize }
function TBufferedSocket.WriteBlockToBuffer( var srcbuf; bufsize : u_int ) : Boolean;
begin
  Result := DefWriteBlockToBuffer( srcbuf, bufsize);
end;

{ Attempt to write FRecordSize bytes. If < FRecordSize bytes free in write buffer,
  return FALSE. Exception raised if FRecordSize > FWriteBufSize }
function TBufferedSocket.WriteRecordToBuffer( var srcbuf ) : Boolean;
begin
  Result := DefWriteRecordToBuffer(srcbuf);
end;

{ File Transfer functions }
procedure TBufferedSocket.SendStream( s : TStream; DestroyOnComplete : Boolean);
begin
  DefSendStream( s, DestroyOnComplete);
end;

procedure TBufferedSocket.RecvStream( s : TStream; size : Longint;
  DestroyOnComplete : Boolean);
begin
  DefRecvStream( s, size, DestroyOnComplete);
end;

procedure TBufferedSocket.CancelRecvStream;
begin
  DefCancelRecvStream;
end;

procedure TBufferedSocket.CancelSendStream;
begin
  DefCancelSendStream;
end;

procedure TBufferedSocket.SendFile( const fn : string);
begin
  DefSendFile(fn);
end;

procedure TBufferedSocket.RecvFile( const fn : string; size : Longint);
begin
  DefRecvFile(fn, size);
end;


{
TLineSocketBase

This is to be used as a base class for any Line-based sockets.

Provides methods for read and writing lines and an event for notifying
the user when a line is received.
}

constructor TLineSocketBase.Create(AParent : TSockCtrl);
begin
  inherited Create(AParent);

  FLineOptions := [lsoAddLineChars, lsoLineReceivedEvents, lsoAutoFlushAfterWrite ];
  FLineType := lstCRLF;
end;

procedure TLineSocketBase.DoReadBufferFull; { call OnReadxxxx event }
begin
  while LineReceived do;

  if ReadBufferBytesFree <= 0 then
    inherited DoReadBufferFull;
end;

function TLineSocketBase.LineReceived : Boolean;
var
  newline : string;
  fullline : Boolean;
begin
  Result := False;
  if lsoLineReceivedEvents in FLineOptions then
  begin
  { DefReadLine will raise an exception if the line is too long }
    fullline := True;
    try
      Result := DefReadLine(newline, GetLineChars, lsoReturnLinechars in FLineOptions);
    except
      on e : EBufferedSocketException do
        if e.Message = BS_ERR_PascalString then
        begin
          fullline := False;
          Result := True;
        end else raise;
    end;

    if Result then
      DoOnLineReceived( newline, fullline );
  end;
end;

procedure TLineSocketBase.DoOnLineReceived( const line : string; complete : Boolean);
begin
  if Assigned(FOnLineReceived) then
    FOnLineReceived(Self, line, complete);
end;

procedure TLineSocketBase.RecvBuffer;
begin
  repeat
    inherited RecvBuffer;
    while LineReceived do;
  until not ((lsoLineReceivedEvents in FLineOptions)  {if AUTOREAD only}
    and Connected and (InCount > 0));   {while smth is out there}
end;

function TLineSocketBase.GetLineChars : string;
begin
  Result := '';
  case FLineType of
    lstCR :   Result := charCR;
    lstLF :   Result := charLF;
    lstCRLF : Result := charCRLF;
  end;
end;

{ Override to implement lsoAutoFlushAfterWrite }
function  TLineSocketBase.DefWriteLine(const l : string;
  const linechars : string) : Boolean;
begin
  Result := inherited DefWriteLine( l, linechars);
  if Result and (lsoAutoFlushAfterWrite in FLineOptions) then
    SendBuffer;
end;

{ Override to implement lsoAutoFlushAfterWrite }
function  TLineSocketBase.DefStrWriteLine(l : PChar;
  const linechars : string) : Boolean;
begin
  Result := inherited DefStrWriteLine( l, linechars);
  if Result and (lsoAutoFlushAfterWrite in FLineOptions) then
    SendBuffer;
end;

{
TLineSocket

This is to be used as a standard Line-based socket.

Provides methods for read and writing lines and an event for notifying
the user when a line is received.
}

function  TLineSocket.GetLineLength : u_int;
begin
  Result := DefGetLineLength( GetLineChars);
end;

function  TLineSocket.IsLineInBuffer : Boolean;
begin
  Result := DefIsLineInBuffer( GetLineChars);
end;

function  TLineSocket.ReadLine(var l : string) : Boolean;
begin
  Result := DefReadLine( l, GetLineChars, lsoReturnLineChars in FLineOptions );
end;

function  TLineSocket.WriteLine(const l : string) : Boolean;
begin
  if lsoAddLineChars in FLineOptions then
    Result := DefWriteLine( l, GetLineChars )
  else
    Result := DefWriteLine( l, '');
end;

function  TLineSocket.StrReadLine(l : PChar; bufsize : u_int) : Boolean;
begin
  Result := DefStrReadLine(l, bufsize, GetLineChars,
    lsoReturnLineChars in FLineOptions);
end;

function  TLineSocket.StrWriteLine(l : PChar) : Boolean;
begin
  if lsoAddLineChars in FLineOptions then
    Result := DefStrWriteLine(l, GetLineChars)
  else
    Result := DefStrWriteLine(l, '');
end;

{ For Delphi 1 users with lines > 255 chars }
function TLineSocket.RebuildLongLine( dest : PChar; buflen : Integer;
  const firstpart : string) : Boolean;
begin
  Result := DefRebuildLongLine( dest, buflen, firstpart, GetLineChars,
    lsoReturnLinechars in FLineOptions);
end;

{
TBuffered*Socket

This is to be used as base for Buffered Socket Client Controls.
}

constructor TBufferedClientBase.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
  FReadBufSize  := 1024;
  FWriteBufSize := 1024;
  FRecordSize   :=   16;
end;

function TBufferedClientBase.CreateConn: TSocketBase;
begin
  Result := inherited CreateConn;
  (Result as TBufferedSocketBase).OnWriteBufferEmpty  := FOnWriteBufferEmpty;
  (Result as TBufferedSocketBase).OnReadBufferEmpty   := FOnReadBufferEmpty;
  (Result as TBufferedSocketBase).OnWriteBufferFull   := FOnWriteBufferFull;
  (Result as TBufferedSocketBase).OnReadBufferFull    := FOnReadBufferFull;
  (Result as TBufferedSocketBase).OnWriteBufferChange := FOnWriteBufferChange;
  (Result as TBufferedSocketBase).OnReadBufferChange  := FOnReadBufferChange;
  (Result as TBufferedSocketBase).OnFileSent          := FOnFileSent;
  (Result as TBufferedSocketBase).OnFileReceived      := FOnFileReceived;

  (Result as TBufferedSocketBase).ReadBufferSize      := FReadBufSize;
  (Result as TBufferedSocketBase).WriteBufferSize     := FWriteBufSize;
  (Result as TBufferedSocketBase).RecordSize          := FRecordSize;
end;

constructor TBufferedServerBase.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
  FReadBufSize  := 1024;
  FWriteBufSize := 1024;
  FRecordSize   :=   16;
end;

function TBufferedServerBase.CreateConn: TSocketBase;
begin
  Result := inherited CreateConn;
  (Result as TBufferedSocketBase).OnWriteBufferEmpty  := FOnWriteBufferEmpty;
  (Result as TBufferedSocketBase).OnReadBufferEmpty   := FOnReadBufferEmpty;
  (Result as TBufferedSocketBase).OnWriteBufferFull   := FOnWriteBufferFull;
  (Result as TBufferedSocketBase).OnReadBufferFull    := FOnReadBufferFull;
  (Result as TBufferedSocketBase).OnWriteBufferChange := FOnWriteBufferChange;
  (Result as TBufferedSocketBase).OnReadBufferChange  := FOnReadBufferChange;
  (Result as TBufferedSocketBase).OnFileSent          := FOnFileSent;
  (Result as TBufferedSocketBase).OnFileReceived      := FOnFileReceived;

  (Result as TBufferedSocketBase).ReadBufferSize      := FReadBufSize;
  (Result as TBufferedSocketBase).WriteBufferSize     := FWriteBufSize;
  (Result as TBufferedSocketBase).RecordSize          := FRecordSize;
end;

constructor TTextClientBase.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
  FLineOptions := [lsoAddLineChars, lsoLineReceivedEvents, lsoAutoFlushAfterWrite ];
  FLineType := lstCRLF;
end;

function TTextClientBase.CreateConn: TSocketBase;
begin
  Result := inherited CreateConn;
  (Result as TLineSocketBase).OnLineReceived    := FOnLineReceived;

  (Result as TLineSocketBase).LineOptions           := FLineOptions;
  (Result as TLineSocketBase).LineType          := FLineType;
end;

constructor TTextServerBase.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
  FLineOptions := [lsoAddLineChars, lsoLineReceivedEvents, lsoAutoFlushAfterWrite ];
  FLineType := lstCRLF;
end;

function TTextServerBase.CreateConn: TSocketBase;
begin
  Result := inherited CreateConn;
  (Result as TLineSocketBase).OnLineReceived    := FOnLineReceived;

  (Result as TLineSocketBase).LineOptions           := FLineOptions;
  (Result as TLineSocketBase).LineType          := FLineType;
end;

end.
