(*****************************************************************************
 * dWinsock component for Borland Delphi.
 *
 * (C) 1995 by Ulf Sderberg, ulfs@sysinno.se
 *             Marc Palmer,   marc@anyware.co.uk
 *             Keith Hawes,   khawes@ccmail.com
 *
 *  dWinsock.pas - Winsock Client & Server components for Delpi
 *
 * $Revision:   1.39  $
 *
 * -- History --
 * $Log:   //SUPPORT_SYSTEMS/PVCS/Source/dWinSock._pas  $
 * 
 *    Rev 1.39   23 Jun 1997 15:38:46   khawes
 * Changes for 2.51
 * 
 * 
 *    Rev 1.33   27 May 1997 10:21:54   khawes
 * Remove // comments
 * 
 *
 *    Rev 1.32   27 May 1997 09:38:38   khawes
 * 
 *    Rev 1.31   27 May 1997 09:37:18   khawes
 * Changes for C++ Builder support
 * 
 * 
 *    Rev 1.30   23 May 1997 17:06:04   khawes
 * Added :
 * I.Free  { KH May 23, 97 Fix Memory leak }
 * To free the object after an async request.
 * 
 * 
 *    Rev 1.29   19 May 1997 16:11:32   khawes
 * Name Change TCustom*Socket to TBasicCustom*Socket
 * Added call to Close in TSocketBase.DoDisconnect
 * 
 *
 *    Rev 1.28   08 May 1997 16:08:10   khawes
 * Added ForceTimeout to TBaseSocket.Destroy
 * 
 * 
 *    Rev 1.27   08 May 1997 16:06:22   khawes
 * Removed TSocketBase.Close's abstract declarations and consolidated the 
 * Closes for Stream and Datagram sockets.
 * 
 * 
 *    Rev 1.26   29 Apr 1997 13:01:42   khawes
 * Call ForceTimeout on Close
 * 
 * 
 *    Rev 1.25   25 Apr 1997 15:18:22   khawes
 * Moved StripChar to dwsUtils
 * Added siDisconnected
 *
 * 
 *    Rev 1.24   26 Mar 1997 15:45:58   khawes
 * More BuffSock changes.  FileSend & Recieve is complete
 * 
 * 
 *    Rev 1.23   19 Mar 1997 12:32:36   khawes
 * Type 
 *   THostRestriction = (hrAccept, hrDeny);
 * 
 * TSockCtrl
 *   protected function CreateConn: TSocketBase; virtual;
 * 
 * TCustomBasicServerSocket 
 *   Private
 *     FRestrictedHosts : TStrings;
 *     FRestriction     : THostRestriction;
 *     FDNSLookup       : Boolean;
 *   protected
 *     procedure SetRestrictedHosts(Value: TStrings);
 *     function  IsHostInRestrictedList(Host: shortstring): boolean; virtual;
 *   public
 *     property RestrictedHosts: TStrings read FRestrictedHosts write 
 * SetRestrictedHosts;
 *     property Restriction : THostRestriction read FRestriction write FRestriction;
 *     property AllowOnlyIfReverseDNS: boolean read FDNSLookup write 
 * FDNSLookup;
 * 
 * TCustomServer
 *   Published
 *     property RestrictedHosts;
 *     property Restriction;
 *     property AllowOnlyIfReverseDNS;
 * 
 *
 *    Rev 1.22   21 Feb 1997 11:32:20   khawes
 * KH: Raise an exception in TSockCtrl.GetConn if FSocketClass is nil.
 * KH: Moved SetSocketClass from TBasicClientSocket to TSockCtrl.
 * KH: Set FStarted to False after a successful WSACleanup.
 * KH: DoReject virtual method in TCustomBasicServerSocket.
 * KH: Handle NT same machine bug: FD_READ before FD_CONNECT.
 * 
 * 
 *    Rev 1.21   19 Feb 1997 09:01:18   khawes
 * KH: Fix Access intermitant vialoation bug on Socket Close by removing the
 * correct socket
 * KH: Stopped TStreamSocket.Send from sending nothing. 
 * 
 * 
 *    Rev 1.20   05 Feb 1997 08:42:22   khawes
 * Fixed possible uninitilized Tag vatriable
 *
 *
 *    Rev 1.19   28 Jan 1997 09:48:02   khawes
 * MP: Misc changes & Bug fixes
 * Changed to allow Close to be called even if the socket is already closed
 * Abort pending async functions on a close.
 * 
 *
 *    Rev 1.17   23 Oct 1996 12:48:26   khawes
 * Fix Bug with changing the offset for registry variables.  Also moved Resources
 * to not conflice with Async Pro by Default.
 * 
 * 
 *    Rev 1.16   06 Sep 1996 08:43:40   KHAWES
 * MP: Changed the definition of socket type
 * 
 *
 *    Rev 1.15   28 Aug 1996 17:26:18   KHAWES
 * (KH) Added a new state called ssListening and a property called Listening to 
 * determine if a socket is listening.
 * 
 * 
 *    Rev 1.14   21 Aug 1996 08:40:52   KHAWES
 * Fix error message Offset bug.
 * 
 * 
 *    Rev 1.13   12 Aug 1996 11:26:34   KHAWES
 * Added SendToPort to the TDataGramSocket to allow sending to a specified 
 * host and now Port.
 * 
 * 
 *    Rev 1.12   07 Aug 1996 09:18:26   KHAWES
 * Fix bug in timer chain insert
 * Fix bug in notification of siNoSuchSocket.
 *
 * 
 *    Rev 1.11   07 Aug 1996 08:50:20   KHAWES
 * Changed Unit name to always be dWinsock.  Removed Register procedure.
 * Use dwsReg16 or dwsReg32 to install into the VCL
 * 
 *
 *    Rev 1.10   05 Aug 1996 11:16:50   KHAWES
 * (MP)
 * Added SocketType & Peek methods.
 *
 *
 *    Rev 1.9   05 Aug 1996 09:52:14   KHAWES
 * (KH)
 * Change to Connected property.  Internaly change this to a state flag to track
 * ssNotConnected, ssConnected and ssClosing. 
 * 
 * Used the new FState to only call ReadFlush once during closing of a socket.
 * 
 * 
 *    Rev 1.8   01 Jul 1996 09:08:18   KHAWES
 * Created CallParentOnXXX for all Events to replace the OnAssigned(...) then ...
 * Created DoXXX to use instead of OnAssigned to allow overriding of events by
 * sockets to make creation of Protocol socktes easier.
 * (Marc)
 *
 * Moved Old DoXXX methods from Public to Protected
 * (Keith)
 * 
 *
 *    Rev 1.7   21 Jun 1996 16:24:30   KHAWES
 * Added ForceTimeout to force dWinsock lookup function to terminate on the
 * next timer tick.
 *
 * Added a WSACancelAsyncRequest to WaitFor so that when a timeout occurs
 * we cancel the pending request.  (This is cleaner than just ignoring the result
 * when in finally comes in)
 *
 *
 *    Rev 1.6   21 Jun 1996 15:33:02   KHAWES
 * Change while loop for 200ms delay to Sleep(200)
 * 
 *
 *
 *    Rev 1.5   07 Jun 1996 17:00:58   KHAWES
 * Not adding Socket to the list until we know it is valid.  Freeing the socket if it is
 * not.
 *
 * Added new Event OnReject to TCustomServer.
 *  It is called when the Max number of connections is reached.
 *  After the call to OnReject the socket is automatically closed.
 *
 * 
 *    Rev 1.4   29 May 1996 17:40:56   KHAWES
 * Added storage to the Async Tracking objects to hold the parameters being
 * passed to the async call so that the values will be valid for the life of the call.
 * Fixes problem with NT3.51 SP4
 * 
 * 
 *    Rev 1.3.1.0   29 May 1996 17:02:48   KHAWES
 * Added storage to the Async Tracking objects to hold the parameters being
 * passed to the async call so that the values will be valid for the life of the call.  
 * Fixes problem with NT3.51 SP4
 *
 * 
 *    Rev 1.3   20 May 1996 16:20:24   khawes
 * Changed:
 *   function  TrackTask( InfoType: TAsyncInfoClass ): TAsyncInfo;
 * to:
 *   function  TrackTask( Info: TAsyncInfo ): THandle;
 * 
 * Changes to TAsync*Info classes:
 *  Added buffers to keep parameters valid for durration of Async call
 *  Added contrsuctors to create & initiate Async Call
 *  Moved AsyncStartError from TSockCtrl to TAsyncInfo
 * 
 * Normalized Spelling of WinsockAPI
 *
 * 
 *    Rev 1.2   24 Apr 1996 09:18:18   khawes
 * Remove Time-Bomb
 * 
 *
 *    Rev 1.1   29 Mar 1996 14:12:52   khawes
 * Testing $Log...
 *
 *
 *    Rev 1.0   29 Mar 1996 14:10:52   khawes
 * First checkin
 *
 *****************************************************************************)
unit dWinsock;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Forms, DWSError, wsAPI;

{ Stop C++ builder from including winsock.h }
const _WINSOCKAPI_ = 1;

{$IfNDef WIN32}
type ShortString = string;
{$EndIf}

var
  T: pointer;
const
  CM_SOCKMSG      = WM_USER+1111;
  CM_SOCKCLOSE    = CM_SOCKMSG+1;
  { KH - Added for "false blocking" and non-blocking }
  CM_AsyncDone    = CM_SockClose+1;
  CM_AsyncBlock   = CM_AsyncDone+1;
  CM_ReleaseBlock = CM_AsyncBlock+1;

type
  TAsyncOptionsType = ( csoRead, csoWrite, csoOOB );
  TAsyncOptions = set of TAsyncOptionsType;

  THostRestriction = (hrAccept, hrDeny);

  { Socket info codes }

  { MP 20/04/95 added siInactive }
  {    25/04/95 added siConnected, siClosed, siTimedOut }
  { KH 950719 added siAccept }
  { KH 950821 added siNoSuchSocket & Documentation}
  { KH 97 April 24 added siDisconnected }
  TSockInfo = (  siInactive,      { not used yet - obsolete?                  }
                 siLookUp,        { A Lookup is being performed               }
                 siConnect,       { A Connection is taking place              }
                 siConnected,     { Connection was sucessful                  }
                 siListen,        { Listining for connecitons                 }
                 siRecv,          { Recieving Data                            }
                 siSend,          { Sending Data                              }
                 siClosed,        { Connection Closed                         }
                 siTimedOut,      { Timout Occured                            }
                 siAccept,        { Connection Accepted by server             }
                 siNoSuchSocket,  { Message arrived from a socket that has    }
                                  {   not been accepted by the server.        }
                                  {   *Socket will be NIL*                    }
                 siReject,        { Connection Rejected by server             }
                 siDisconnected); { Connection was disconnected               }

  TSockCtrl     = class;  { Forward declaration }
  TSocketBase   = class;  { Forward declaration }
  TSocketState = ( ssNotConnected, ssConnected, ssClosing, ssListening);

  { KH 950905 }
  { KH - Socket message buffering records }
  { KH 951219  - Change record to hold a TMessage }
  { KH 951222  - Add SockCtrl Field to track which control gets the message}
  PSockMsg = ^TSockMsg;
  TSockMsg = record
    Message: TMessage;
    SockCtrl: TSockCtrl;
    next: PSockMsg;
  end;

  { KH - Last Async Task -- Used during synchronous Lookup functions }
  TLastTask = record
    handle: Thandle;
    err,
    buflen: word;
  end;

  { KH 950905 Result objects for AsyncLookup*By* }

  { KH May 13, 1996: Add Buffers for any info passed to async functins }

  TAsyncInfo = class(TObject)
  private
    FTaskHandle: Thandle;
    FAsyncBuf: TAsyncBuf;
    FError: word;
  public
    procedure AsyncStartError( handle: hWnd );
    property TaskHandle: Thandle read FTaskHandle;
    property Error: word read FError;
  end;

  THostSendBuffer = array[0..64] of char;
  TAsyncHostInfo = class(TAsyncInfo)
  private
    FBuffer: THostSendBuffer;
    Fsa: in_Addr;

    function GetName : ShortString;
    function GetAddress: ShortString;
    function GetAddr: in_addr;
    function GetInfo : HostEnt;            { US May 25, 1997: C++Builder workaround }
  public
	{ US May 25, 1997: Merged CreateByAddress and CreateByHost into one
      constructor because of C++Builder ambiguities }
    constructor Create(handle: hWnd; byAddr : boolean; AAddrOrHost: ShortString );
    property Name: ShortString read GetName;
    property Address: ShortString read GetAddress;
    property Addr: in_addr read GetAddr;
    property Info: HostEnt read GetInfo; { US May 25, 1997: C++Builder WA }
    property PassedBuffer: THostSendBuffer read FBuffer;
  end;

  TAsyncServiceInfo = class(TAsyncInfo)
  private
    fService:  array[0..64] of char;
    fProtocol: array[0..32] of char;
    function GetName: ShortString;
    function GetPort: Word;
    function GetPortStr: ShortString;
    function GetProtocol: ShortString;
    function GetInfo : ServEnt;            { US May 25, 1997: C++Builder WA }
  public
    constructor CreateByService(handle: hWnd; AService, AProtocol: ShortString);
    constructor CreateByPort(handle: hWnd; APort: u_int; AProtocol: ShortString);
    property Name: ShortString read GetName;
    property Port: Word read GetPort;
    property PortStr: ShortString read GetPortStr;
    property Protocol: ShortString read GetProtocol;
    property Info: ServEnt read GetInfo;

  end;

  TAsyncProtocolInfo = class(TAsyncInfo)
  private
    fName: array[0..64] of char;

    function getName: ShortString;
    function GetNumber: Word;
    function GetNumberStr: ShortString;
    function GetInfo : ProtoEnt;            { US May 25, 1997: C++Builder WA }
  public
    constructor CreateByName(handle: hWnd; AName: ShortString);
    constructor CreateByNumber(handle: hWnd; ANumber: integer);
    property Name: ShortString read GetName;
    property Number: Word read GetNumber;
    property NumberStr: ShortString read GetNumberStr;
    property Info: ProtoEnt read GetInfo;  { US May 25, 1997: was FAsyncBuf.Protocol }
  end;

  { Define notification events for socket controls. }
  TSockInfoEvent = procedure (Sender : TObject; icode : TSockInfo;
                              Socket: TSocketBase) of object;
  TSocketEvent = procedure (Sender : TObject; Socket : TSocketBase) of object;
  TSockeTAsyncEvent = procedure (Sender : TObject; AsyncInfo: TAsyncInfo)
                                of object;

{ MP 951129 Changes for UDP  - change TSocket to TSocketBase and introduced
        TStreamSocket and TDatagramSocket  }

  { TSocketBase -- socket api wrapper class. }
  TSocketBase = class(TObject)
  private
    function SError(err, IgnoreErr, MsgId: Integer): integer;
  protected
    FLastError  : integer;      { Last return code form a winsock.dll call }
    FSocket     : TSock;        { socket id }
    FParent     : TSockCtrl;    { socket owner }
    FAddr       : sockaddr_in;  { host address }
    FState      : TSocketstate;
    FBytesSent  : integer;      { bytes sent by last SendBuf call }
  { New user defined Tag property }
    FTag        : Longint;
    FSocketType : Integer;
  { New Unique Name property }
    FName       : String;

{MP Added 30 June}
    procedure DoRead; virtual;
    procedure DoWrite; virtual;
    procedure DoAccept; virtual;
    procedure DoConnect; virtual;
    procedure DoDisconnect; virtual;
    procedure DoInfo(icode : TSockInfo); virtual;
    procedure DoTimeout; virtual;
    procedure DoReject; virtual;

    procedure CallParentOnRead;
    procedure CallParentOnWrite;
    procedure CallParentOnAccept;
    procedure CallParentOnConnect;
    procedure CallParentOnDisconnect;
    procedure CallParentOnInfo(icode : TSockInfo);
    procedure CallParentOnTimeout;
    procedure CallParentOnReject;

    function  GetName: string ; virtual;
    function  GetConnected : Boolean; virtual; { Leave this virtual! }
    function  GetAState( index: integer ): Boolean;
    procedure SetSocketType( t : Integer); virtual;
    property  SocketType : Integer read FSocketType write SetSocketType;

  public

    constructor Create(AParent : TSockCtrl); virtual;
    destructor Destroy; override; {KH 950718}

    function Send(var buf; cnt : integer) : integer; virtual; abstract;
    function Recv(var buf; cnt : integer) : integer; virtual; abstract;
    function Peek(var buf; cnt : integer) : integer;

    procedure FillSocket(var name, addr, service : ShortString; var port : Word);

    function LocalAddress : ShortString;
    function LocalPort : Word;

    function RemoteHost : ShortString;
    function RemoteAddress : ShortString;
    function RemotePort : Word;

    procedure SetOptions; virtual;
    procedure Listen(var name, addr, service : ShortString; port : Word;
              nqlen : integer); virtual; abstract;
    procedure Open(var name, addr, service : ShortString; port : Word;
              opts : TAsyncOptions); virtual; abstract;
    procedure Close; virtual;

    procedure ReadFlush;

    function InCount : integer;

    property BytesSent : integer read FBytesSent;
    property LastError: integer read FLastError;
    property SocketID: TSock read FSocket;
    property Connected: boolean read GetConnected;
    property Listening: boolean index integer(ssListening) read getAState;
    property State: TSocketState read FState;       { KH Aug 5, 96 - More info than just Connected }
    property Parent: TSockCtrl read FParent;
    property Addr: sockaddr_in read Faddr;
    property Name: string read GetName write FName;

    property Tag : Longint read FTag write FTag default 0;
  end;

  TStreamSocket = class(TSocketBase)
  private
    function RecvText : ShortString;
    procedure SendText(const s : ShortString);
    procedure SetSocketType( t : Integer); override;
  public

    constructor Create(AParent : TSockCtrl); override;
    destructor Destroy; override; {KH 950718}

    function Send(var buf; cnt : integer) : integer; override;
    function Recv(var buf; cnt : integer) : integer; override;

    procedure Listen(var name, addr, service : ShortString; port : Word;
              nqlen : integer); override;
    procedure Open(var name, addr, service : ShortString; port : Word;
              opts : TAsyncOptions); override;
    procedure Close; override;
    property SocketType;
    property Text : ShortString read RecvText write SendText;
  end;

  TDatagramSocket = class(TSocketBase)
  private
    function RecvText : ShortString;
    procedure SendText(const s : ShortString);

    function GetBroadcast : Boolean;
    procedure SetBroadcast( val : Boolean);

{This proc is used internally to make sure we always have a valid FSocket,
        because with a datagram you do not have to call Open, the socket
  may not be valid when you call SendTo or RecvFrom }
    procedure CreateSocket;
    procedure SetSocketType( t : Integer); override;
  public

    constructor Create(AParent : TSockCtrl); override;
    destructor Destroy; override; {KH 950718}

    { Send or receive dtaagram to/from perr denoted by Address or Host property }
    function Send(var buf; cnt : integer) : integer; override;
    function Recv(var buf; cnt : integer) : integer; override;
    { Datagram specific, ignores Address and Host properties }
    function SendTo( const host : shortstring; var buf; cnt : integer) : integer; virtual;
    function SendToPort( const host : shortstring; port: word; var buf; cnt : integer) : integer; virtual;
    function RecvFrom( var buf; cnt : integer;
                       var fromaddr : SockAddr;
                       var fromlen : Integer) : integer; virtual;

    procedure Listen( var name, addr, service : ShortString; port : Word;
                      nqlen : integer); override;
    procedure Open( var name, addr, service : ShortString; port : Word;
                    opts : TAsyncOptions); override;
    procedure Close; override;
    property SocketType;
    property Text : ShortString read RecvText write SendText;
    property Broadcast : Boolean read GetBroadcast write SetBroadcast;
  end;

  TSocketClass = class of TSocketBase;

  TSocketList = class (TList)
  protected
    function GetSocket( Index : Integer ) : TSocketBase;
  public
    property Sockets[ Index : Integer ] : TSocketBase read GetSocket;
  end;

  TAsyncDataList = class (Tlist)
  protected
    function GetTask( TaskHandle : tHandle ) : TAsyncInfo;
  public
    property Tasks[ TaskHandle : tHandle ] : TAsyncInfo read GetTask;
  end;


  { TSockCtrl -- socket control component base class. }
  TSockCtrl = class(TComponent)
  private
    { US 950509 }
    FHWnd         : HWnd;

    { MP 14 Jan 96 - Moved FSocketClass from TCustomBasicServerSocket }
    FSocketClass    : TSocketClass;

    { Event handler references }
    FOnInfo       : TSockInfoEvent;
    { MP 21/4/95 Moved from TClient+TSocket and 2 new properties added }
    FOnDisconnect    : TSocketEvent;
    FOnRead          : TSocketEvent;
    FOnWrite         : TSocketEvent;
    FOnTimeOut       : TSocketEvent;
    { KH 950905 Added FOnAsyncComplete }
    FOnAsyncComplete : TSockeTAsyncEvent;

    fTimedOut     : Boolean;
    fAsyncData    : TAsyncDataList;

    { MP 25/4/95 New fields to handle timeout + timer event chains }
    FTimerChainParent, FTimerChainChild : TSockCtrl;
    FTimeOutRemaining   :  Integer;
    FTimeOutActive      :  Boolean;

    { Design time connection info }
    FHost         : ShortString;
    FAddress      : ShortString;
    FService      : ShortString;
    FPort         : Word;
    FUserAddress  : ShortString;
    FUserPort     : ShortString;

    FConn         : TSocketBase;        { Run time connection info }
    FTimeOut      : integer;        { timeout length in seconds }

    { MP 951129 Added Protocol property for UDP/TCP handling. This has been
        treated in the same way as the Address and Port properties are now. This
      is because the Winsock 2 API allows for other protocols that we do not
      yet know about (i.e. ATM) and each protocol will have a unique ID number.
      This approach will allow us to handle that when the time comes. }
    FUserProtocol : ShortString;
    FProtocolID   : Integer;
    FProtocolName : ShortString;

    { Access functions }
    procedure SetProtocol(const p : ShortString);
    procedure SetAddress(const a : ShortString);
    procedure SetPort(const port : ShortString);
{ MP 25/4/95 }
    procedure SetTimeOut( p : Integer);

    { Returns the WinSock.DLL description }
    function GetDescription : ShortString;
    function GetMaxDatagramSize : Word;
    function SAsyncError(TaskHandle: THandle; MsgId: integer): THandle;

  protected
    { Protected declarations }
    { KH 950905 }
    LastTask: TLastTask;

    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;
    procedure CBSockClose(ASocket : TSocketBase); virtual;

    { US 950509 }
    procedure WndProc(var Message : TMessage);
    procedure OnSockMsg(var Message : TMessage); virtual; abstract;
    procedure OnAsyncDoneMsg( Var Message: TMessage); virtual;

    { KH 950905 }
    function WaitFor(ATask: THandle): integer; virtual;

    { MP 14 Jan 95 .... added member function to make sure FConn is not NIL }
    { KH 14 Mar 96 .... Moved from private to protected                     }
    function GetConn : TSocketBase;
    { KH 14 Mar 96 All sockets need to be created through here! }
    function CreateConn: TSocketBase; virtual;

    { MP 25/4/95 }
    procedure TimerEvent( Sender : TObject);
    procedure UseTimer;
    procedure ReleaseTimer;

    { KH 951222 to remove this controls messages from the message queue }
    procedure ClearQueuedMessages;

    { MP 25/4/95  New properties }
    property OnTimeOut : TSocketEvent read FOnTimeOut write FOnTimeOut;
    property TimeOut : Integer read FTimeOut write SetTimeOut;

    { MP 21/4/95  Moved these props from client+server to TSockctrl
      KH 950825 Moved to protected Section }
    property OnDisconnect : TSocketEvent read FOnDisconnect write FOnDisconnect;
    property OnRead : TSocketEvent read FOnRead write FOnRead;
    property OnWrite : TSocketEvent read FOnWrite write FOnWrite;
    property OnAsyncComplete : TSockeTAsyncEvent read FOnAsyncComplete
                                                 write FOnAsyncComplete;

    { DoXXX procedures } { Moved to Protected July 1, 96 }
    procedure DoDisconnect(var Socket: TSocketBase); virtual;
    procedure DoInfo(Socket: TSocketBase; icode : TSockInfo); virtual;
    procedure DoRead(var Socket: TSocketBase); virtual;
    procedure DoWrite(var Socket: TSocketBase); virtual;
    procedure DoTimeOut(var Socket: TSocketBase); virtual;

    function  TrackTask( Info: TAsyncInfo ): THandle;

    property  Protocol : ShortString read FUserProtocol write SetProtocol;

  public
    { Public declarations }
{    class function LocalHost : ShortString; Converted to a function }
    procedure ForceTimeout; { KH 960621 Add a way to stop an Open, or Lookup* }

    { KH 950905 name changes for Lookup* Functions }
    { US 950826 changed var to const }

    { Lookup Functions }
    function LookupHostByAddr(const a : ShortString) : ShortString;
    function LookupAddrByName(const name : ShortString) : in_addr;
    function LookupAddrByNameStr(const name : ShortString) : ShortString;
    function LookupPortByService(const service : ShortString) : Word;
    function LookupServiceByPort(port : Word) : ShortString;

    { Async Lookup Funcitons.
      Returns a TaskHandle that can be used to match the function call with
      the correct OnAsyncComplete event. }
    function AsyncLookupHostByAddr(const a: ShortString): tHandle;
    function AsyncLookupHostByName(const name: ShortString): tHandle;
    function AsyncLookupServiceByService(const service: ShortString): tHandle;
    function AsyncLookupServicebyPort(const port: Word): thandle;

    { MP 27 Jan 96   New functions to access variables now hidden by new
    property handling for Address, Port and Protocol. }
    function GetHost : ShortString;
    function GetAddress : ShortString;
    function GetPort : Integer;
    function GetService : ShortString;
    function GetProtocolName : ShortString;
    function GetProtocolID : Integer;

    { MP 14 Jan 96 added }
    procedure SetSocketClass( sc : TSocketClass );

    property Handle : HWND read FHWnd;  { US 950509 }
                { MP 14 Jan 96 .... added member function to make sure FConn is not NIL }
    property Conn : TSocketBase read GetConn;
    property Description : ShortString read GetDescription;
    property MaxDatagramSize : Word read GetMaxDatagramSize;
    property Address : ShortString read FUserAddress write SetAddress;
    property Port : ShortString read FUserPort write SetPort;
    property OnInfo : TSockInfoEvent read FOnInfo write FOnInfo;

  published
    { Published declarations }
  end;

  { Definition of the TCustomClientSocket component class }
  TCustomClientSocket = class(TSockCtrl)
  private
    FOnConnect    : TSocketEvent;
    FOptions      : TAsyncOptions;

  protected
    { Protected declarations }
    procedure OnSockMsg(var Message : TMessage); override;
    { KH 950825 DoXXX }
    procedure DoConnect(var Socket: TSocketBase); virtual;
    procedure DoDisconnect(var Socket: TSocketBase); override;


  public
    { Public declarations }
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;

    procedure Open(ASocketClass : TSocketClass);
    procedure Close;
    function Connected : boolean;

    {property Host : ShortString read FHost write SetHost;  }
    property Options : TAsyncOptions read FOptions write FOptions
                                            default [csoRead, csoWrite];
    property OnConnect : TSocketEvent read FOnConnect write FOnConnect;
    property OnTimeOut;
    property TimeOut;
    { KH 950825 Added }
    property OnDisconnect;
    property OnRead;
    property OnWrite;
    property OnAsyncComplete;

    property Protocol;
  published
    { Published declarations }
  end;

  { Definition of the TCustomBasicServerSocket component class }
  TCustomBasicServerSocket = class(TSockCtrl)
  private
    { Event handler references }
    FOnAccept       : TSocketEvent;
    FOnReject       : TSocketEvent;

    FMaxConns       : Integer;
    FConns          : TSocketList;

    { MP 20/4/95 }
    FOptions        : TAsyncOptions;

    { MP 18 Jan 96 }
    FQueueLength    : Integer;

    { KH 18 MAR 97 }
    FRestrictedHosts : TStrings;
    FRestriction     : THostRestriction;
    FDNSLookup       : Boolean;

    function GetClient(cid : integer) : TSocketBase;
    function GetClientCount : integer;

  protected
    { Protected declarations }
    procedure OnSockMsg(var Message : TMessage); override;
    procedure CBSockClose(ASocket : TSocketBase); override;
    procedure DoAccept(var Socket: TSocketBase); virtual;
    procedure DoDisconnect(var Socket: TSocketBase); override;
    procedure DoReject(var Socket: TSocketBase); virtual;
    procedure SetRestrictedHosts(Value: TStrings);
    function  IsHostInRestrictedList(Host: shortstring): boolean; virtual;

  public
    { Public declarations }
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;

    procedure Listen( ASocketClass : TSocketClass);
    procedure Close;

    { MP 20/04/95 added CloseDown declaration. Used CloseDown to avoid
    confusion with winsock's Shutdown }
    procedure CloseDown; { close server and all connections }
    { Return client socket }

    property Client[cid : integer] : TSocketBase read GetClient; default;
    property ClientCount : Integer read GetClientCount;
    property RestrictedHosts: TStrings read FRestrictedHosts write SetRestrictedHosts;
    property Restriction : THostRestriction read FRestriction write FRestriction;
    property AllowOnlyIfReverseDNS: boolean read FDNSLookup write FDNSLookup;

    property OnAccept : TSocketEvent read FOnAccept write FOnAccept;
    property OnReject : TSocketEvent read FOnReject write FOnReject;
    { MP 25/4/95 New property }
    property MaxConnections : Integer read FMaxConns write FMaxConns default 16;
    property ClientOptions : TAsyncOptions read FOptions write FOptions
                              default [csoRead, csoWrite];
    { KH 950825 Added }
    property OnDisconnect;
    property OnRead;
    property OnWrite;
    property OnAsyncComplete;
    property Protocol;
    property QueueLength : Integer read FQueueLength write FQueueLength default 5;
  published
    { Published declarations }
  end;

{Definition of TBasicClientSocket - The actual component }
  TBasicClientSocket = class(TCustomClientSocket)
  published
    property Address;
    property Port;
    property Protocol;

    property TimeOut;
    property Options;

    property OnInfo;
    property OnConnect;
    property OnTimeOut;
    property OnDisconnect;
    property OnRead;
    property OnWrite;
    property OnAsyncComplete;
  end;

{Definition of TBasicServerSocket - The actual component }
  TBasicServerSocket = class(TCustomBasicServerSocket)
  published
    property Address;
    property Port;
    property Protocol;

    property MaxConnections;
    property ClientOptions;
    property QueueLength;

    property RestrictedHosts;
    property Restriction;
    property AllowOnlyIfReverseDNS;

    property OnInfo;
    property OnAccept;
    property OnReject;
    property OnDisconnect;
    property OnRead;
    property OnWrite;
    property OnAsyncComplete;
  end;

function  ErrorStr(e: integer): ShortString;
procedure StartUp;
function SockInfoToString( si : TSockInfo ) : string;
function LocalHost : ShortString;

var
  WinsockAPI: TWinsockAPI ;

implementation

uses ExtCtrls, dWsUtils;

var
  TimerUserCount : Integer;
  TimerChainRoot : TSockCtrl;

var
  nUsers        : integer;
  nWSErr        : integer;
  myVerReqd     : word;
  myWSAData     : WSADATA;
  Timer         : TTimer;
  BlockingTasks : integer;
  MsgQueue      : PSockMsg;

function Started: boolean;
begin
  if WinsockAPI = nil then
    WinsockAPI := TWinsockAPI.Create;
  Result := WinsockAPI.Started;
end;

function SockInfoToString( si : TSockInfo ) : string;
const
  SockInfoStrings : array[low(TSockInfo)..High(TSockInfo)] of string =
              ('Inactive',
               'Looking-up a host name',
               'Connecting',
               'Connection',
               'Listening',
               'Receiving Data',
               'Sending Data',
               'Closed',
               'Connection timed out',
               'Accepting connection',
               'Error: No such socket',
               'Rejecting Connection',
               'Disconnected');
begin
  Result := SockInfoStrings[ si ];
end;

function  ErrorStr(e: integer): ShortString;
begin
  if e = 0 then
    Result := ''
  else
    Result := LoadStr(e+WSAOffset);
  if (e <> 0) and (Result = '') then
    Result := format('Unknown Error %d', [e]);
end;

{ KH extract message forming code to use for both TSocketBase.sError and
  TSockCtrl.sAsyncError                                                }

procedure SockException(Err, Msgid: integer);
var
  NPos: integer;
  SPos: integer;
  Msg,
  temp: ShortString;
begin
  Msg := ErrorStr(MsgId);
  temp := Uppercase(Msg);
  NPos := pos('%D', temp);
  SPos := pos('%S', temp);
  if Msg = '' then
    temp := ErrorStr(err)
  else if (NPos = 0) and (SPos = 0) then
    temp := msg
  else if (NPos < SPos) and (NPos <> 0) then
    temp := format(msg, [err, ErrorStr(err)])
  else
    temp := format(msg, [ErrorStr(err), err]);
  raise ESockError.Create(temp);
end;

function MakeAsyncMask( Options : TAsyncOptions) : Longint;
begin
  Result := 0;

  if csoRead in Options then
    Result := FD_READ;

  if csoWrite in Options then
    Result := Result or FD_WRITE;

  if csoOOB in Options then
    Result := Result or FD_OOB;
end;

{ StartUp -- See if a Windows Socket DLL is present on the system. }
procedure StartUp;
begin
  if Started then
    exit;
  BlockingTasks:= 0;
  nUsers := 0;
  myVerReqd:=$0101;
  { myVerReqd:=$0707; KH 27-FEB-96  Removed - How did it work before?}
  nWSErr := WinsockAPI.WSAStartup(myVerReqd, myWSAData);
  if nWSErr <> 0 then
    SockException(nWSErr, DWSErrStart);
end;

{ CleanUp -- Tell Windows Socket DLL we don't need its services any longer. }
procedure CleanUp; far;
var
  temp: PSockMsg;
begin
  if WinsockAPI <> nil then
  begin
    if Started then
    begin
      if WinsockAPI.WSAisBlocking then
        WinsockAPI.WSACancelBlockingCall;
      nWSErr := WinsockAPI.WSACleanup;
    end;
  end;
  while MsgQueue <> nil do
  begin
    temp := MsgQueue;
    MsgQueue := MsgQueue^.Next;
    Dispose(temp);
  end;
  BlockingTasks := 0;
  if WinsockAPI <> nil then
  begin
    {$ifdef win32}
    { This is the fix to the access violation bug   }
    { in NT 3.51                                    }
    { Pause 200 milliseconds before unloading the   }
    { winsock dll to let is finish it's processing  }
    { KH 6/5/96                                     }
    sleep(200);
    {$endif}
    WinsockAPI.free;
    WinsockAPI := nil;
  end;
end;

{--------------------------------------------------------------
  TSocketList implementation
 --------------------------------------------------------------}
function TSocketList.GetSocket( Index : Integer ) : TSocketBase;
begin
  Result := Items[Index];
end;

{--------------------------------------------------------------
  TASsyncDataList implementation
 --------------------------------------------------------------}
function TAsyncDataList.GetTask( TaskHandle : tHandle ) : TAsyncInfo;
var
  I: integer;
begin
  Result := nil;
  for I := 0 to Count-1 do
  begin
    if Items[I] <> nil then
      if TAsyncInfo(Items[I]).TaskHandle = TaskHandle then
      begin
        Result := Items[I];
        break;
      end;
  end;
end;

{--------------------------------------------------------------
  TAsync*Info implementation
 --------------------------------------------------------------}
 function SafeStrPas( S: pChar): ShortString;
 begin
   if S = nil then
     Result := ''
   else
     Result := StrPas(S);
 end;


procedure TAsyncInfo.AsyncStartError( handle: hWnd );
begin
  PostMessage( handle, CM_AsyncDone, Taskhandle,
               WinsockAPI.WSAMakeAsyncReply( 0, WinsockAPI.WSAGetlastError))
end;

{ US May 25, 1997: merged CreateByAddress and CreateByHost into Create }
constructor TAsyncHostInfo.Create( handle: hWnd; byAddr : boolean; AAddrOrHost: ShortString );
begin
  inherited create;

  StrPcopy(FBuffer, AAddrOrHost);

  if byAddr then
    begin
      { Code from old CreateByAddress }
      Fsa.s_addr := WinsockAPI.inet_addr(FBuffer);
      if Fsa.s_addr = 0 then
        raise ESockError.Create(ErrorStr(DWSErrBasAddress));

      FTaskHandle := WinsockAPI.WSAAsyncGetHostByAddr( handle, CM_AsyncDone,
                     pChar(@Fsa.s_addr), 4, PF_INET, @FAsyncBuf, SizeOf(FAsyncBuf));
    end
  else
    begin
      { Code from old CreateByHost }
      FTaskHandle := WinsockAPI.WSAAsyncGetHostByName( handle, CM_AsyncDone,
                     FBuffer, @FAsyncBuf, SizeOf(FAsyncBuf));
    end;

  if TaskHandle = 0 then
    AsyncStartError(Handle);
end;

function TAsyncHostInfo.GetName: ShortString;
begin
  Result := SafeStrPas(FAsyncBuf.Host.h_name);
end;

function TAsyncHostInfo.GetAddress: ShortString;
var
  sa: in_addr;
begin
  sa := GetAddr;
  Result := SafeStrPas(WinsockAPI.inet_ntoa(sa));
end;

function TAsyncHostInfo.GetAddr: in_addr;
var
  pa: Pchar;
begin
  pa := FAsyncBuf.Host.h_addr_list^;
  Result.S_un_b.s_b1 := byte(pa[0]);
  Result.S_un_b.s_b2 := byte(pa[1]);
  Result.S_un_b.s_b3 := byte(pa[2]);
  Result.S_un_b.s_b4 := byte(pa[3]);
end;

{ US May 25, 1997: C++Builder workaround }
function TAsyncHostInfo.GetInfo : HostEnt;
begin
  Result := FAsyncBuf.Host;
end;

constructor TAsyncServiceInfo.CreateByService(handle: hWnd; AService, AProtocol: ShortString);
begin
  inherited create;

  StrPcopy(FService, AService);
  StrPCopy(FProtocol, AProtocol);

  FTaskHandle := WinsockAPI.WSAAsyncGetServByName( handle, CM_AsyncDone,
                 FService, FProtocol, @FAsyncBuf, SizeOf(FAsyncBuf));
  if TaskHandle = 0 then
    AsyncStartError(Handle);
end;

constructor TAsyncServiceInfo.CreateByPort(handle: hWnd; APort: u_int; AProtocol: ShortString);
begin
  inherited create;

  StrPCopy(FProtocol, AProtocol);

  FTaskHandle := WinsockAPI.WSAAsyncGetServByPort( handle, CM_AsyncDone,
                 WinsockAPI.htons(port), FProtocol, @FAsyncBuf, SizeOf(FAsyncBuf));
  if TaskHandle = 0 then
    AsyncStartError(Handle);
end;

function TAsyncServiceInfo.GetName: ShortString;
begin
  Result := SafeStrPas(FAsyncBuf.Service.s_name);
end;

function TAsyncServiceInfo.GetPort: Word;
begin
  Result := WinsockAPI.ntohs(FAsyncBuf.Service.s_Port);
end;

function TAsyncServiceInfo.GetPortStr: ShortString;
begin
  Result := IntToStr(getPort);
end;

{ US May 25, 1997: C++Builder workaround }
function TAsyncServiceInfo.GetInfo : ServEnt;
begin
  Result := FAsyncBuf.Service;
end;

function TAsyncServiceInfo.GetProtocol: ShortString;
begin
  Result := SafeStrPas(FAsyncBuf.Service.s_Proto);
end;

constructor TAsyncProtocolInfo.CreateByName(handle: hWnd; AName: ShortString);
begin
  inherited create;

  StrPCopy(FName, AName);
  FTaskHandle := WinsockAPI.WSAAsyncGetProtoByName( handle, CM_AsyncDone,
                 Fname, @FAsyncBuf, SizeOf(FAsyncBuf));
  if TaskHandle = 0 then
    AsyncStartError(handle);
end;

constructor TAsyncProtocolInfo.CreateByNumber(handle: hWnd; ANumber: integer);
begin
  inherited create;

  FTaskHandle := WinsockAPI.WSAAsyncGetProtoByNumber( handle, CM_AsyncDone,
                 ANumber, @FAsyncBuf, SizeOf(FAsyncBuf));
  if TaskHandle = 0 then
    AsyncStartError(handle);
end;

function TAsyncProtocolInfo.getName: ShortString;
begin
  Result := SafeStrPas(FAsyncBuf.Protocol.p_name);
end;

function TAsyncProtocolInfo.GetNumber: Word;
begin
  Result := WinsockAPI.ntohs(FAsyncBuf.Protocol.p_proto);
end;

function TAsyncProtocolInfo.GetNumberStr: ShortString;
begin
  Result := InttoStr(GetNumber);
end;

{ US May 25, 1997: C++Builder workaround }
function TAsyncProtocolInfo.GetInfo : ProtoEnt;
begin
  Result := FAsyncBuf.Protocol;
end;


{--------------------------------------------------------------
  TSocketBase implementation
 --------------------------------------------------------------}

constructor TSocketBase.Create(AParent : TSockCtrl);
begin
  inherited Create;
  FParent := AParent;
  FSocket := INVALID_SOCKET;
  FSocketType := SOCK_STREAM;
  FAddr.sin_family := PF_INET;
  FAddr.sin_addr.s_addr := INADDR_ANY;
  FAddr.sin_port := 0;
  FState := ssNotConnected;
  FBytesSent := 0;
  FName := '';
end;

destructor TSocketBase.Destroy;
begin
  if FSocket <> INVALID_SOCKET  then
  begin
    FParent.ForceTimeout;
    WinsockAPI.CloseSocket(FSocket);
  end;
  inherited Destroy;
end;

{ LocalAddress -- get local address }
function TSocketBase.LocalAddress : ShortString;
var
  sa : sockaddr_in;
  nl : integer;
begin
  Result := '';
  if FSocket = INVALID_SOCKET then exit;
  nl := SizeOf(sa);
  try { KH 950821 Change in error handling }
    sError(WinsockAPI.getsockname(FSocket, sa, nl), 0, DWSErrGetSockName);
    Result := StrPas(WinsockAPI.inet_ntoa(sa.sin_addr));
  except
    on ESockError do ;
  end;
end;

{ LocalPort -- get local port number }
function TSocketBase.LocalPort : Word;
var
  sa : sockaddr_in;
  nl : integer;
begin
  Result := 0;
  if FSocket = INVALID_SOCKET then exit;
  nl := SizeOf(sa);
  try { KH 950821 Change in error handling }
    sError(WinsockAPI.getsockname(FSocket, sa, nl), 0, DWSErrGetSockName);
    Result := WinsockAPI.ntohs(sa.sin_port);
  except
    on ESockError do ;
  end;
end;

{ RemoteHost -- get name of connected remote host }
function TSocketBase.RemoteHost : ShortString;
var
  sa  : sockaddr_in;
  nl  : integer;
  buf : pAsyncBuf;
  ATask: Thandle;
begin
  Result := '';
  if not Connected then exit;
  nl := sizeof(sa);
  { Get connection address info }
  sError(WinsockAPI.getpeername(FSocket, sa, nl), 0, DWSErrGetPeerName);
  FAddr := sa;
  { Do a reverse lookup to get the host name }
  new(buf);
  try
    FParent.UseTimer;
    ATask := FParent.SAsyncError(WinsockAPI.WSAAsyncGetHostByAddr( FParent.Handle,
                     CM_AsyncBlock, PChar(@FAddr.sin_addr.s_addr), 4, PF_INET,
                     buf, SizeOf(Buf^)), DWSErrAsyncGetHostByAddr);
    if FParent.WaitFor(ATask) = 0 then
      Result := StrPas(Buf^.Host.h_name);
  finally
    dispose(buf);
    FParent.ReleaseTimer;
  end;
end;

{ RemoteAddress -- get address of connected remote host }
function TSocketBase.RemoteAddress : ShortString;
var
  sa : sockaddr_in;
  nl : integer;
begin
  Result := '?';
  if (FSocket = INVALID_SOCKET) or (not Connected) then exit;
  nl := SizeOf(sa);
  sError(WinsockAPI.getpeername(FSocket, sa, nl), 0, DWSErrGetPeerName);
  Result := StrPas(WinsockAPI.inet_ntoa(sa.sin_addr));
end;

{ RemotePort -- get remote port number }
function TSocketBase.RemotePort : word;
var
  sa : sockaddr_in;
  nl : integer;
begin
  Result := 0;
  if (FSocket = INVALID_SOCKET) or (not Connected) then exit;
  nl := SizeOf(sa);
  try { KH 950821 Change in error handling }
    sError(WinsockAPI.getpeername(FSocket, sa, nl), 0, DWSErrGetPeerName);
    Result := WinsockAPI.ntohs(sa.sin_port)
  except
    on ESockError do Result := 0;
  end;
end;


{ FillSocket -- fill in address and port fields in socket struct }
procedure TSocketBase.FillSocket(var name, addr, service : ShortString;
          var port : Word);
var
  s : array [1..32] of char;
begin
  { Fill in address field }
  if name <> '' then            { Host name given }
    begin
      FAddr.sin_addr := FParent.LookupAddrbyName(name);        {KH 950712}
      addr := StrPas(WinsockAPI.inet_ntoa(FAddr.sin_addr));
    end
  else if addr <> '' then       { IP address given }
    begin
      FAddr.sin_addr.s_addr := INADDR_ANY;
      if addr <> '0.0.0.0' then { beware of Trumpet bug! }
        begin
          StrPCopy(@s, addr);
          FAddr.sin_addr.s_addr := WinsockAPI.inet_addr(@s);
        end;
    end
  else                          { Neither name or address given }
    raise ESockError.Create(ErrorStr(DWSErrNoAddress));

  { Fill in port number field }
  if service <> '' then
    port := FParent.LookupPortByService(service); {KH 950712}
  FAddr.sin_port := WinsockAPI.htons(port);
end;

{ SetOptions -- set socket options }
procedure TSocketBase.SetOptions;
begin
  {Place Holder}
end;

procedure TSocketbase.Close;
var
  wasconn : Boolean;
begin
  if (not Started) or (FSocket = INVALID_SOCKET) then exit;

  FParent.ForceTimeout;
  wasconn := (State = ssConnected);
  FState := ssClosing;
  if wasconn then
  begin
    ReadFlush;
  end;

  WinsockAPI.closesocket(FSocket);

  CallParentOnInfo(siClosed);
  FSocket := INVALID_SOCKET;

  FBytesSent := 0;
  PostMessage(FParent.Handle, CM_SOCKCLOSE, 0, Longint(self));
end;

function TSocketBase.SError(err, IgnoreErr, Msgid: integer): integer;
begin
  result := err;
  if err = SOCKET_ERROR then
  begin
    err := WinsockAPI.WSAGetLastError;
    FLastError := err;
    if err = IgnoreErr then { Ignore a specific error for both speed and  }
    begin;                  { leaving break on exceptions on for debugging }
      Result := 0;          { really helps for ignoring WsaeWouldBlock    }
      exit;
    end;
    SockException(err, MsgId);
  end
  else
    FLastError := 0;
end;


{ ReadFlush -- Forces all data to be sent }
procedure TSocketBase.ReadFlush;
var
  temp: TSocketBase;
begin
  if not assigned(parent.FOnRead) then
    exit;
  while InCount > 0 do
  begin
    temp := self;
    parent.DoRead(temp);
  end;
end;

{ InCount -- Get # of bytes in receive buffer }
function TSocketBase.InCount : integer;
const
  FIONREAD = $40000000 or ((longint(4)) shl 16) or (ord('f') shl 8) or 127;
var
  n : longint;
begin
  sError(WinsockAPI.ioctlsocket(FSocket, FIONREAD, n), 0, DWSErrInCount);
  Result := n and $ffff;
end;

procedure TSocketBase.DoRead;
begin
  CallParentOnRead;
end;

procedure TSocketBase.DoWrite;
begin
  CallParentOnWrite;
end;

procedure TSocketBase.DoAccept;
begin
  CallParentOnAccept;
end;

procedure TSocketBase.DoConnect;
begin
  CallParentOnConnect;
end;

procedure TSocketBase.DoDisconnect;
begin
  CallParentOnDisconnect;
  Close; { KH May 19, 1997 - Close the socket }
end;

procedure TSocketBase.DoInfo(icode : TSockInfo);
begin
  CallParentOnInfo(icode);
end;

procedure TSocketBase.DoTimeout;
begin
  CallParentOnTimeout;
end;

procedure TSocketBase.DoReject;
begin
  CallParentOnReject;
end;

procedure TSocketBase.CallParentOnRead;
begin
  if Assigned(FParent.FOnRead) then
    FParent.FOnRead(FParent, Self);
end;

procedure TSocketBase.CallParentOnWrite;
begin
  if Assigned(FParent.FOnWrite) then
    FParent.FOnWrite(FParent, Self);
end;

procedure TSocketBase.CallParentOnAccept;
begin
  if Assigned( (FParent as TCustomBasicServerSocket).FOnAccept) then
    (FParent as TCustomBasicServerSocket).FOnAccept(FParent, Self);
end;

procedure TSocketBase.CallParentOnConnect;
begin
  if Assigned( (FParent as TCustomClientSocket).FOnConnect) then
    (FParent as TCustomClientSocket).FOnConnect(FParent, Self);
end;

procedure TSocketBase.CallParentOnDisconnect;
begin
  CallparentOnInfo(siDisconnected);  { KH April 24, 97}
  if Assigned(FParent.FOnDisconnect) then
    FParent.FOnDisconnect(FParent, Self);
end;

procedure TSocketBase.CallParentOnInfo(icode : TSockInfo);
begin
  case icode of
    siListen: FState := ssListening;
    siClosed: FState := ssNotConnected;
  end;
  if Assigned(FParent.FOnInfo) then
    FParent.FOnInfo(FParent, icode, Self);
end;

procedure TSocketBase.CallParentOnTimeout;
begin
  if Assigned(FParent.FOnTimeout) then
    FParent.FOnTimeout(FParent, Self);
end;

procedure TSocketBase.CallParentOnReject;
begin
  if Assigned( (FParent as TCustomBasicServerSocket).FOnReject) then
    (FParent as TCustomBasicServerSocket).FOnReject(FParent, Self);
end;

function TSocketBase.GetName: string;
const
  UniqueSocketNum: integer = 0;
begin
  if FName = '' then
  begin
    FName := format('socket%d', [UniqueSocketNum]);
    Inc(UniqueSocketNum);
  end;
  Result := Fname;
end;

function TSocketBase.GetConnected : Boolean;
begin
  Result := FState in [ssConnected];
end;

function TSocketBase.GetAState(index: integer) : Boolean;
begin
  Result := FState = TSocketState(index);
end;

{ this should be overriden in derived classes to prevent bad assignments! }
procedure TSocketBase.SetSocketType( t : Integer);
begin
  FSocketType := t;
end;

{ Request that passed buffer be filled with peeked data. }
function TSocketBase.Peek(var buf; cnt : integer) : integer;
begin
{  FParent.DoInfo(Self, siRecv);} {KH 950719}
  if (FSocket = INVALID_SOCKET) or (not Connected) then
    raise ESockError.Create(ErrorStr(DWSErrSockNotOpen));

  try { KH 950821 Change in error handling }
    Result := sError(WinSockAPI.recv(FSocket, buf, cnt, MSG_PEEK),
            WsaeWouldBlock, DWSErrPeek);
  except
    on ESockError do
    begin
{      Close;   KH June 3, 1997 }
      raise;
    end;
  end;
end;

{--------------------------------------------------------------
  TDatagramSocket implementation
 --------------------------------------------------------------}

constructor TDatagramSocket.Create(AParent : TSockCtrl);
begin
  inherited Create(AParent);
  FParent := AParent;
  FSocketType := SOCK_DGRAM;
  FAddr.sin_family := PF_INET;
  FAddr.sin_addr.s_addr := INADDR_ANY;
  FAddr.sin_port := 0;
  FState := ssNotConnected;
  FBytesSent := 0;
{ MP - Do this here 'cos a UDP socket always needs FSocket to be valid in case
  send is called }
  if (not Started) then
    SockException(WSANotInitialised, 0);

  CreateSocket;
end;

destructor TDatagramSocket.Destroy;
begin
  inherited Destroy;
end;

procedure TDatagramSocket.CreateSocket;
begin
  if FSocket <> INVALID_SOCKET then Exit;

  FSocket := WinsockAPI.socket(PF_INET, FSocketType, IPPROTO_IP);
  if FSocket = INVALID_SOCKET then
    sError(Socket_Error, 0, DWSErrCreate);
end;

procedure TDatagramSocket.SetBroadcast( val : Boolean);
var
  buf, buflen, err : Integer;
begin
  buflen := SizeOf(buf);
  if val then buf := 1 else buf := 0;
  err := WinsockAPI.setsockopt( FSocket, Integer(SOL_SOCKET),
                           SO_BROADCAST, PChar(@buf), buflen);
  if err <> 0 then SockException(err, 0);
end;

function TDatagramSocket.GetBroadcast : Boolean;
var
        buf, buflen, err : Integer;
begin
  buflen := SizeOf(buf);
  err := WinsockAPI.getsockopt( FSocket, Integer(SOL_SOCKET),
                           SO_BROADCAST, PChar(@buf), buflen);
  if err = 0 then
  begin
    Result := buf <> 0;
  end
  else
  begin
    Result := False;
    SockException(err, 0);
  end;
end;

{ Open a connection. }
procedure TDatagramSocket.Open( var name, addr, service : ShortString; port : Word;
                                opts : TAsyncOptions);
begin
  if (not Started) then
    SockException(WSANotInitialised, 0);

  if Connected then
    SockException(WSAEIsConn, 0);

  CreateSocket;

  FParent.DoInfo(Self, siLookUp);

  FillSocket(name, addr, service, port);

  SetOptions;

  WinsockAPI.WSAAsyncSelect( FSocket, FParent.Handle, CM_SOCKMSG,
                             MakeAsyncMask(opts) or FD_CONNECT or FD_CLOSE);

  { MP 25/4/95 }
  try
    {FParent.UseTimer; } { No timeouts in UDP ? }
    FParent.DoInfo(Self, siConnect);
    try { KH 950821 Change in error handling }
      sError(WinsockAPI.connect(FSocket, FAddr, SizeOf(FAddr)),
             WsaeWouldBlock, DWSErrConnect);
      { No real conneciton is made with UDP, so no FD_CONNECT is received,
        this means we must set this to true here. }
      FState := ssConnected;
    except
      on ESockError do
      begin
        if FSocket <> INVALID_SOCKET then
          Close;
        raise;
      end;
    end;
  except
    {FParent.ReleaseTimer; } { Ditto...no timeouts in UDP? }
    raise;
  end;
end;

{ This Close does something strange. Because we always need a valid FSocket,
        it closes and then re-creates the socket. }
procedure TDatagramSocket.Close;
begin
  inherited CLose;
  CreateSocket;
end;

function TDatagramSocket.RecvText : ShortString;
{$ifndef win32}
var
  n : integer;
{$endif}
begin
  {$ifdef win32}
    SetLength(Result, 255);
    SetLength(Result, Recv(pChar(@Result[1])^, 255));
  {$else}
    n := Recv(PChar(@Result[1])^, 255);
    Result[0] := char(n);
  {$endif}
end;

procedure TDatagramSocket.SendText(const s : ShortString);
begin
  FBytesSent := Send(PChar(@s[1])^, Length(s));
end;

{ Send contents of passed buffer.
  For Datagrams, a call to Send can only be done on a CONNECTED
  datagram socket. Call Open before using Send or Recv
}

function TDatagramSocket.Send(var buf; cnt : integer) : integer;
begin
  Result := 0;
  if not Connected then SockException(0, DWSErrSockNotOpen);
  { Check here to see if cnt > iMaxUdpDg returned from WSAStartup and raise
        exception if it is. }
  FParent.DoInfo(Self, siSend); {KH 950719}
  if (FSocket = INVALID_SOCKET) then exit;

  if (cnt > FParent.MaxDatagramSize) and 
     (FParent.MaxDatagramSize <> 0) then
    raise ESockError.Create(ErrorStr(DWSErrDatagramTooBig));


  try { KH 950821 Change in error handling }
    result := SError( WinsockAPI.send(FSocket, buf, cnt, 0),
                      WsaeWouldBlock, DWSErrSend);
  except
    on ESockError do
    begin
{      close;   KH June 3, 1997 }
      raise;
    end;
  end;
end;

{ Send contents of passed buffer.
  SendTo can *only* be called on UNCONNECTED datagram sockets
}
function TDatagramSocket.SendTo(const host : shortstring; var buf; cnt : integer) : integer;
begin
  Result := SendToPort(Host, FParent.FPort, buf, cnt);
end;

function TDatagramSocket.SendToPort(const host : shortstring; Port: word; var buf; cnt : integer) : integer;
var
  name, addr, dummy : shortstring;
begin
  { Check here to see if cnt > iMaxUdpDg returned from WSAStartup and raise
        exception if it is. }
  FParent.DoInfo(Self, siSend); {KH 950719}

  if (cnt > FParent.MaxDatagramSize) and
     (FParent.MaxDatagramSize <> 0) then
    raise ESockError.Create(ErrorStr(DWSErrDatagramTooBig));

  if WinsockAPI.IsInternetAddress( host ) then
  begin
    name := '';
    addr := host;
  end
  else
  begin
    name := host;
    addr := '';
  end;

  dummy := '';
  FillSocket( name, addr, dummy, Port);
  try { KH 950821 Change in error handling }
    result := SError(WinsockAPI.sendto( FSocket, buf, cnt, 0, FAddr,
                                   SizeOf(FAddr)), WsaeWouldBlock, DWSErrSend);
  except
    on ESockError do
    begin
{      close;   KH June 3, 1997 }
      raise;
    end;
  end;
end;

{ Request that passed buffer be filled with received data.
        For Datagrams, a call to Recv can only be done on a CONNECTED
  datagram socket. Call Open before using Send or Recv
}
function TDatagramSocket.Recv( var buf; cnt : integer) : integer;
begin
  CallParentOnInfo(siRecv); {KH 950719}
  if (FSocket = INVALID_SOCKET) then
    raise ESockError.Create(ErrorStr(DWSErrSockNotOpen));

  try { KH 950821 Change in error handling }
    Result := sError( WinsockAPI.recv( FSocket, buf, cnt, 0),
                      WsaeWouldBlock, DWSErrRecv);
  except
    on ESockError do
    begin
{      Close;   KH June 3, 1997 }
      raise;
    end;
  end;
end;

{ Request that passed buffer be filled with received data. }
function TDatagramSocket.RecvFrom( var buf; cnt : integer;
                                   var fromaddr : SockAddr;
                                   var fromlen : Integer) : integer;
begin
  CallParentOnInfo(siRecv); {KH 950719}

  try { KH 950821 Change in error handling }
    Result := sError( WinsockAPI.recvfrom( FSocket, buf, cnt, 0, fromaddr, fromlen),
                      WsaeWouldBlock, DWSErrRecv);
  except
    on ESockError do
    begin
{      Close;   KH June 3, 1997 }
      raise;
    end;
  end;
end;

{ Listen -- wait for incoming connection. }
procedure TDatagramSocket.Listen( var name, addr, service : ShortString; port : Word;
                                  nqlen : integer);
begin
  if (not Started) then
    SockException(WSANOTINITIALISED, 0);

  FSocket := WinsockAPI.socket(PF_INET, FSocketType, IPPROTO_IP);
  if FSocket = INVALID_SOCKET then
    sError(Socket_Error, 0, DWSErrCreate);

  FillSocket(name, addr, service, port);

  SetOptions;

  try { KH 950821 Change in error handling }
    sError(WinsockAPI.bind(FSocket, FAddr, SizeOf(FAddr)), 0, DWSErrBind);
    FParent.FPort := LocalPort; { in case port number was 0 }
    FParent.FUserPort := IntToStr(LocalPort);
  except
    on ESockError do
    begin
      Close;
      raise;
    end;
  end;

  WinsockAPI.WSAAsyncSelect(FSocket, FParent.Handle, CM_SOCKMSG, FD_READ or FD_WRITE);
  (*
  { 950816 MP replaced q with nqlen}
  try { KH 950821 Change in error handling }
    sError(WinsockAPI.listen(FSocket, nqlen), 0, DWSErrListen);
    FParent.DoInfo(Self, siListen);
  except
    on ESockError do
    begin
      if FSocket <> INVALID_SOCKET then
        Close;
      raise;
    end;
  end;
  *)
end;


{ Make sure only UDP style socket types assigned }
procedure TDatagramSocket.SetSocketType( t : Integer);
begin
  FSocketType := t;
end;

{--------------------------------------------------------------
  TStreamSocket implementation
 --------------------------------------------------------------}

constructor TStreamSocket.Create(AParent : TSockCtrl);
begin
  inherited Create(AParent);
  FParent := AParent;
  FSocket := INVALID_SOCKET;
  FAddr.sin_family := PF_INET;
  FAddr.sin_addr.s_addr := INADDR_ANY;
  FAddr.sin_port := 0;
  FState := ssNotConnected;
  FBytesSent := 0;
end;

destructor TStreamSocket.Destroy;
begin
  inherited Destroy;
end;

{ Listen -- wait for incoming connection. }
procedure TStreamSocket.Listen(var name, addr, service : ShortString; port : Word;
                        nqlen : integer);
begin
  if (not Started) then
    SockException(WSANOTINITIALISED, 0);

  FSocket := WinsockAPI.socket(PF_INET, FSocketType, IPPROTO_IP);
  if FSocket = INVALID_SOCKET then
    sError(Socket_Error, 0, DWSErrCreate);

  FillSocket(name, addr, service, port);

  SetOptions;

  try { KH 950821 Change in error handling }
    sError(WinsockAPI.bind(FSocket, FAddr, SizeOf(FAddr)), 0, DWSErrBind);
    FParent.FPort := LocalPort; { in case port number was 0 }
    FParent.FUserPort := IntToStr(LocalPort);
  except
    on ESockError do
    begin
      Close;
      raise;
    end;
  end;

  WinsockAPI.WSAAsyncSelect(FSocket, FParent.Handle, CM_SOCKMSG, FD_ACCEPT or FD_CLOSE);

  { 950816 MP replaced q with nqlen}
  try { KH 950821 Change in error handling }
    sError(WinsockAPI.listen(FSocket, nqlen), 0, DWSErrListen);
    CallParentOnInfo(siListen);
  except
    on ESockError do
    begin
      if FSocket <> INVALID_SOCKET then
        Close;
      raise;
    end;
  end;
end;

{ Open a connection. }
procedure TStreamSocket.Open(var name, addr, service : ShortString; port : Word;
                          opts : TAsyncOptions);
begin
  if (not Started) then
    SockException(WSANotInitialised, 0);

  if Connected then
    SockException(WSAEIsConn, 0);

  FSocket := WinsockAPI.socket(PF_INET, FSocketType, IPPROTO_IP);
  if FSocket = INVALID_SOCKET then
    sError(Socket_Error, 0, DWSErrCreate);

  CallParentOnInfo(siLookUp);

  FillSocket(name, addr, service, port);

  SetOptions;

  WinsockAPI.WSAAsyncSelect(FSocket, FParent.Handle, CM_SOCKMSG,
             MakeAsyncMask(opts) or FD_CONNECT or FD_CLOSE);

  { MP 25/4/95 }
  try
    FParent.UseTimer; { start timeout check }
    CallParentOnInfo(siConnect);
    try { KH 950821 Change in error handling }
      sError(WinsockAPI.connect(FSocket, FAddr, SizeOf(FAddr)),
             WsaeWouldBlock, DWSErrConnect);
    except
      on ESockError do
      begin
        if FSocket <> INVALID_SOCKET then
          Close;
        raise;
      end;
    end;
  except
    FParent.ReleaseTimer;
    raise;
  end;
end;

procedure TStreamSocket.Close;
begin
  inherited Close;
  { MP 951127 }
  FParent.ReleaseTimer;
end;

function TStreamSocket.RecvText : ShortString;
{$ifndef win32}
var
  n : integer;
{$endif}
begin
  {$ifdef win32}
    SetLength(Result, 255);
    SetLength(Result, Recv(pChar(@Result[1])^, 255));
  {$else}
    n := Recv(PChar(@Result[1])^, 255);
    Result[0] := char(n);
  {$endif}
end;

procedure TStreamSocket.SendText(const s : ShortString);
begin
  FBytesSent := Send(PChar(@s[1])^, Length(s));
end;

{ Send contents of passed buffer. }
function TStreamSocket.Send(var buf; cnt : integer) : integer;
begin
  Result := 0;
  if cnt = 0 then  { Added to prevent empty messages and unneeded }
    exit;          { OnInfo events as suggested by MJW            }
  CallParentOnInfo(siSend); {KH 950719}
  if (FSocket = INVALID_SOCKET) or (not Connected) then exit;
  try { KH 950821 Change in error handling }
    result := SError(WinsockAPI.send(FSocket, buf, cnt, 0), WsaeWouldBlock,
                     DWSErrSend);
  except
    on ESockError do
    begin
{      close;   KH June 3, 1997 }
      raise;
    end;
  end;
end;

{ Request that passed buffer be filled with received data. }
function TStreamSocket.Recv(var buf; cnt : integer) : integer;
begin
  CallParentOnInfo(siRecv); {KH 950719}
  if (FSocket = INVALID_SOCKET) or (not (State in [ssConnected, ssClosing])) then
    raise ESockError.Create(ErrorStr(DWSErrSockNotOpen));

  try { KH 950821 Change in error handling }
    Result := sError(WinsockAPI.recv(FSocket, buf, cnt, 0), WsaeWouldBlock,
                     DWSErrRecv);
  except
    on ESockError do
    begin
{      Close;   KH June 3, 1997 }
      raise;
    end;
  end;
end;

{ Make sure only TCP style socket types assigned }
procedure TStreamSocket.SetSocketType( t : Integer);
begin
  case t of
    SOCK_STREAM : FSocketType := t;
  else
    raise ESockError.Create( ErrorStr( DWSErrInvalidStreamType) );
  end;
end;

{--------------------------------------------------------------
  TSockCtrl implementation
 --------------------------------------------------------------}

{ Create -- initalization }
constructor TSockCtrl.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);

  { The control should be visible at design time only.
    At run time, check if the WINSOCK has been started. }
  if not (csDesigning in ComponentState) then
  begin
    { US 950509 }
    FHWnd := AllocateHWnd(WndProc);
    StartUp;
  end;

  FHost := '';
  FAddress := '0.0.0.0';
  FUserAddress := '';
  FUserPort := '';

  FService := '';
  FPort := 0;

  FillChar(LastTask, SizeOf(LastTask), 0);
  fTimedOut := False;

  Protocol := 'tcp';
  FConn := nil;
  FSocketClass := TStreamSocket;

  inc(nUsers);
end;


{ Destroy -- destruction }
destructor TSockCtrl.Destroy;

begin
  ReleaseTimer;
  FConn.Free;
  fAsyncData.Free;
  { US 950509 }
  if not (csDesigning in ComponentState) then
    DeallocateHWnd(FHwnd);
  Dec(nUsers);
  if nUsers <= 0 then
    CleanUp;
  inherited Destroy;
end;

procedure TSockCtrl.CBSockClose(ASocket : TSocketBase);
begin
  ClearQueuedMessages;
end;

{ US 950509: WndProc -- trap socket messages. }
procedure TSockCtrl.WndProc(var Message : TMessage);
{ KH 20-NOV-95  Added Call to Applicaiton.HandleException(self) }
var
  temp: PSockMsg;

  procedure NewMsg(Var Message: TMessage);
  begin
    if MsgQueue = nil then
    begin
      MsgQueue := new(PSockMsg);
      temp := MsgQueue;
    end
    else
    begin
      temp := MsgQueue;
      while temp^.Next <> nil do
        temp := temp^.Next;
      temp^.Next := new(PSockMsg);
      temp := temp^.Next;
    end;
    temp^.Message := Message;  { KH 951219 }
    temp^.SockCtrl := Self;    { KH 951226 -- Save Socket Control }
    temp^.Next := nil;
  end;

begin
  try   { Makes it easier to always call the default windProc }
    if (Application.Terminated) and (Message.Msg <> cm_SockClose) then
    begin           {KH 960207 - Why bother we're closing}
      exit;
    end;
    with Message do
    begin
      try
        case Msg of
          CM_SOCKMSG   : if (BlockingTasks = 0) and (MsgQueue = nil) then
                           OnSockMsg(Message)                          { KH 950905 }
                         else
                           NewMsg(Message);                            { KH 950905 }
          CM_SOCKCLOSE : CBSockClose(TSocketBase(Message.lParam));
          CM_AsyncDone : OnAsyncDoneMsg(Message);                      { KH 950905 }
          CM_AsyncBlock: with LastTask do                              { KH 950905 }
                           begin
                             handle := Thandle(wParam);
                             err    := WinsockAPI.WSAGeTAsyncError( lParam );
                             buflen := WinsockAPI.WSAGeTAsyncBuflen( lParam );
                           end;
          CM_ReleaseBlock: {};          { Stop call to DefWindowProc } { KH 950905 }
          end;
      except
        Application.HandleException(Self);
      end;
    end;

    if (BlockingTasks = 0) and (MsgQueue <> nil) then { KH 951219 }
    begin
      temp := MsgQueue;
      MsgQueue := MsgQueue^.Next;
      if temp^.SockCtrl <> nil then  { Ignore msgs for sockets closed after msg was queued}
        Temp^.SockCtrl.OnSockMsg(Temp^.Message);  { KH 951222 Send msg to owning control }
      Dispose(temp);
      if MsgQueue <> nil then
        PostMessage(Handle, cm_ReleaseBlock, 0, 0);
    end;
  finally
    with message do
      Result := DefWindowProc(FHWnd, Msg, wParam, lParam);
  end;
end;

procedure TSockCtrl.OnAsyncDoneMsg( Var Message: TMessage);
var
  I: TAsyncInfo;
begin
  if fAsyncData <> nil then
  begin
    I := fAsyncData.Tasks[Message.wParam];
    if I = nil then
      exit; { Can't find Task so lets ignore it }
    {MP 951127 fix to release timer }
    ReleaseTimer;

    I.FError := WinsockAPI.WSAGeTAsyncError(Message.lParam);
    if Assigned(FOnAsyncComplete) then
    begin
      FOnAsyncComplete(Self, I);
    end;
    fAsyncData.Delete(fAsyncData.IndexOf(I));
    I.Free;  { KH May 23, 97 Fix Memory leak }
    if fAsyncData.Count = 0 then
    begin
      fAsyncData.Free;
      fAsyncData := nil;
    end;
  end;
end;

function TSockCtrl.WaitFor(ATask: THandle): integer;
var
  giveup : Boolean;
begin
  Inc(BlockingTasks);
  FillChar(LastTask, SizeOf(LastTask), 0);  { Dec 18  Clear LastTask Before Starting }
  giveup := False;
  while (LastTask.Handle <> ATask) and (Not FTimedOut) and (not giveup) do
  begin
    Application.ProcessMessages;
    if Assigned(FConn) then
      giveup := ssClosing = FConn.FState;

    if Application.Terminated then giveup := True;

    if giveup then
    begin
      FTimeOutRemaining := 0;
      WinsockAPI.WSACancelAsyncRequest( ATask);
    end;
  end;
  Dec(BlockingTasks);
  if BlockingTasks >= 0 then
  begin
    BlockingTasks := 0;
    PostMessage(Handle, CM_ReleaseBlock, 0, 0);
  end;
  if (LastTask.Handle <> ATask) and (FTimedOut) then
  begin
    lastTask.err := WSAETimedOut;
    WinsockAPI.WSACancelAsyncRequest(ATask);   { KH June 21, 96 Cancel call if we timed out }
  end;
  Result := LastTask.err;
end;

{ MP 25/4/95  Handle the time out timer events
              This gets a bit tricky, because we don't want to keep
              wasting CPU time if we have already timed out, so we release
              the timer if we time out. This can only be done once the
              other components in the chain have been called.
}
procedure TSockCtrl.TimerEvent( Sender : TObject );
begin
  if Assigned(FTimerChainChild) then
    FTimerChainChild.TimerEvent(Sender);
  if FTimeOutRemaining > 0 then
    Dec(FTimeOutRemaining);
  if FTimeOutRemaining = 0 then
    begin
      ReleaseTimer; { do this NOW in case event handler takes too long! }
      DoInfo(Conn, siTimedOut);
      { MP This should actually pass the actual socket in the case of a server }
      DoTimeOut(FConn);
    end;
end;

{ DoDisconnect -- Call the OnDisconnect event handler if any }
procedure TSockCtrl.DoDisconnect(var Socket: TSocketBase);
begin
  Socket.DoDisconnect;
end;

{ DoInfo -- call the OnInfo event handler if any. }
procedure TSockCtrl.DoInfo(Socket: TSocketBase; icode : TSockInfo);
begin
  Socket.DoInfo(icode);
end;

{ DoRead -- Call the OnRead event handler if any }
procedure TSockCtrl.DoRead(var Socket: TSocketBase);
begin
  Socket.DoRead;
end;

{ DoWrite -- Call the OnWrite event handler if any }
procedure TSockCtrl.DoWrite(var Socket: TSocketBase);
begin
  Socket.DoWrite;
end;

{ DoTimeOut -- Call the OnTimeOut event handler if any }
procedure TSockCtrl.DoTimeOut(var Socket: TSocketBase);
begin
  FTimedOut := True;
  Socket.DoTimeOut;
end;

{ GetDescription -- return description of WinSock implementation }
function TSockCtrl.GetDescription : ShortString;
begin
  Result := StrPas(myWSAdata.szDescription);
end;

function TSockCtrl.GetMaxDatagramSize : Word;
begin
  Result := myWSAdata.iMaxUdpDg;
end;

{ KH, Feb 20, 1997:  added check for nil and raise an exception if it is. }
function TSockCtrl.GetConn : TSocketBase;
begin
  if not Assigned(FConn) then
    FConn := CreateConn;
  Result := FConn;
end;

function TSockCtrl.CreateConn: TSocketBase;
begin
  Result := nil;
  if FSocketClass <> nil then
    Result := FSocketClass.Create(Self)
  else
    SockException(0, DWSErrNoClass);
end;

function TSockCtrl.SAsyncError(TaskHandle: tHandle; MsgId: integer): THandle;
begin
  Result := TaskHandle;
  if TaskHandle = 0 then
    SockException(WinsockAPI.WSAGetLastError, MsgId);
end;

{ LocalHost -- return name of local host }
{class} function {TSockCtrl.} LocalHost : ShortString;
var
  sh : array [0..255] of char;
begin
  if not Started then
    begin
      Result := '';
      Exit;
    end;
  if WinsockAPI.gethostname(sh, 255) = 0 then
    Result := StrPas(sh)
  else
    Result := '';
end;
{ KH 960621 Add a way to stop an Open, or Lookup* }
procedure TSockCtrl.ForceTimeout;
begin
  FTimeOutRemaining := 0;
end;

{ Set protocol }
procedure TSockCtrl.SetProtocol(const p : ShortString);
var
  protoint : Integer;

  procedure ProtocolIsInt;
  begin
    FProtocolID := protoint;
    FProtocolName := '';
    case protoint of
      IPPROTO_IP : FProtocolName := 'IP';
      IPPROTO_ICMP : FProtocolName := 'ICMP';
      IPPROTO_GGP : FProtocolName := 'GGP';
      IPPROTO_TCP : FProtocolName := 'TCP';
      IPPROTO_PUP : FProtocolName := 'PUP';
      IPPROTO_UDP : FProtocolName := 'UDP';
      IPPROTO_IDP : FProtocolName := 'IDP';
      IPPROTO_ND  : FProtocolName := 'ND';

      IPPROTO_RAW : FProtocolName := 'RAW';
    end;
  end;

  procedure ProtocolIsStr;
  var
    up : ShortString;
  begin
    FProtocolName := p;
    FProtocolID := -1;
    up := UpperCase(p);
    if up = 'IP' then FProtocolID := IPPROTO_IP;
    if up = 'ICMP' then FProtocolID := IPPROTO_ICMP;
    if up = 'GGP' then FProtocolID := IPPROTO_GGP;
    if up = 'TCP' then FProtocolID := IPPROTO_TCP;
    if up = 'PUP' then FProtocolID := IPPROTO_PUP;
    if up = 'UDP' then FProtocolID := IPPROTO_UDP;
    if up = 'IDP' then FProtocolID := IPPROTO_IDP;
    if up = 'ND' then FProtocolID := IPPROTO_ND;
    if up = 'RAW' then FProtocolID := IPPROTO_RAW;
  end;

begin
{ We should really trap changes here if the socket is already open.
  It will not be good to change from TCP to UDP once you have
  opened the connection! Should raise an exception. }
  FUserProtocol := p;

  protoint := StrToIntDef( p, -1);
  if protoint <> -1 then ProtocolIsInt
  else ProtocolIsStr;
end;

{ Set host address }
procedure TSockCtrl.SetAddress(const a : ShortString);
begin
  started;  { Called to initilize WinsockAPI object if necessary }
  FUserAddress := a;
  if not (csDesigning in ComponentState) then
  begin
    if WinsockAPI.IsInternetAddress(a) then
    begin
      FAddress := a;
      FHost := '';
    end
    else
    begin
      FHost := a;
      FAddress := '';
    end;
  end;
end;

{ Set port number }
procedure TSockCtrl.SetPort(const port : ShortString);
var
  p : Word;
begin
  started;  { Called to initilize WinsockAPI object if necessary }
  FUserPort := port;
  if WinsockAPI.IsPortNumber(port) then
  begin
    p := StrToInt(port);
    FPort := p;
    if Not Started then
      FService := ''
    else
    begin
      try
        FService := LookupServiceByPort(p);
      except
        on ESockError do FService := '';
      end;
    end;
  end
  else
  begin
    FService := port;
    if Not Started then
      FPort := 0
    else
    begin
      try
        FPort := LookupPortByService(port);
      except
        on ESockError do FPort := 0;
      end;
    end;
  end;

end;

function TSockCtrl.GetHost : ShortString;
begin
  Result := FHost;
end;

function TSockCtrl.GetAddress : ShortString;
begin
  Result := FAddress;
end;

function TSockCtrl.GetPort : Integer;
begin
  Result := FPort;
end;

function TSockCtrl.GetService : ShortString;
begin
  Result := FService;
end;

function TSockCtrl.GetProtocolName : ShortString;
begin
  Result := FProtocolName;
end;

function TSockCtrl.GetProtocolID : Integer;
begin
  Result := FProtocolID;
end;

{ MP 15 Jan 97 improved }
procedure TSockCtrl.SetSocketClass( sc : TSocketClass );
var
  tag : Longint;
begin
  tag := 0;
  if sc <> FSocketClass then
  begin
    if Assigned(FConn) then
    begin
      tag := FConn.Tag;
      FConn.Free;
      FConn := nil;
    end;
    FSocketClass := sc;
  end;
  GetConn.Tag := tag; { create it }
end;


{ MP 25/4/95 }
{ Set time out delay }
procedure TSockCtrl.SetTimeOut( p : Integer);
begin
  if p < 0 then p := 0; { trap negatives }
  FTimeOut := p;
end;

{ there is one global timer, and the different controls chain the calls
to the OnTimer event. }
procedure TSockCtrl.UseTimer;
begin
  if (csDesigning in ComponentState) then
    Exit;

  if (FTimeOut = 0) or (not Assigned(FOnTimeOut)) then exit;

  if not Assigned(Timer) then
  begin
    Timer := TTimer.Create(Application); { MP 28 Jan 96 Changed from (Self) to fix GPF }
    Timer.Interval := 1000;
  end;

  { Add ourselfes to the top of the timer chain }
  FTimerChainParent := nil;                   { we will be the root we have no parent }
  FTimerChainChild := TimerChainRoot;         { Make the current root our child       }
  if (TimerChainRoot <> nil) then             { Setup prefious root's parent to       }
    TimerChainRoot.FTimerChainParent := Self; { point to us (the new parent)          }
  TimerChainRoot := Self;                     { Make us the root of the chain         }

  Timer.OnTimer := TimerEvent;
  FTimeOutActive := True;
  FTimeOutRemaining := FTimeOut;
  Inc(TimerUserCount);
  FTimedOut := False;
  { MP 28 Jan 96 - if the timer was switched off we need to switch it on again.}
  if TimerUserCount = 1  then Timer.Enabled := True;
end;

procedure TSockCtrl.ReleaseTimer;
begin
  if (csDesigning in ComponentState) then Exit;

  { US 950502 + removed lots of if FTimeOutActive from other places }
  if not FTimeOutActive then Exit;

  { remove ourselves from the chain }
  if Assigned(FTimerChainParent) then
    { reinstate previous handler }
    FTimerChainParent.FTimerChainChild := FTimerChainChild
  else
  begin
    if Assigned( FTimerChainChild) then
      Timer.OnTimer := FTimerChainChild.TimerEvent
    else
      Timer.OnTimer := nil;
    TimerChainRoot := FTimerChainChild;
  end;

  if Assigned(FTimerChainChild) then
    FTimerChainChild.FTimerChainParent := FTimerChainParent;

  Dec(TimerUserCount);
  FTimeOutActive := False;
  if TimerUserCount = 0 then
  begin
    { We disable the timer to save windows resources. If we call UseTimer it
      will be enabled again. }
    Timer.Enabled := False;
                      Timer.OnTimer := nil; { MP 28 Jan 96 added to make sure no bad stuff happens }
    { MP 28 Jan 96 Removed as part of fix that makes Application own
      the timer rather than the control. As the Application owns it,
      it is automatically destroyed when the app exits.
    Timer.Free;
    Timer := nil;}
  end;

{ MP added 951125 to ensure we don't remove ourselves twice if ReleaseTiemr called
        twice }
  FTimerChainParent := nil;
  FTimerChainChild := nil;
end;

procedure TSockCtrl.ClearQueuedMessages;
var
  Temp: PSockMsg;
begin
  Temp := MsgQueue;
  while temp <> nil do
  begin
    if temp^.SockCtrl = self then
      temp^.SockCtrl := nil;
    temp := temp^.next;
  end;
end;

{ Reverse -- try to do a reverse lookup }
{ US 950826 changed 'var a' to 'const a' }
function TSockCtrl.LookupHostByAddr(const a : ShortString) : ShortString;
var
  Buf: pAsyncBuf;
  s    : array[0..31] of char;
  sa  : in_addr;
  ATask: tHandle;
begin
  Result := '';
  StrPCopy(@s, a);  { US 950826 added @ in front of 1st arg }
  sa.s_addr := WinsockAPI.inet_addr(s);
  if sa.s_addr = 0 then
    raise ESockError.Create(ErrorStr(DWSErrBasAddress));

  new(buf);
  try
    UseTimer;
    ATask := SAsyncError(WinsockAPI.WSAAsyncGetHostByAddr(handle, CM_AsyncBlock,
                         PChar(@sa.s_addr), 4, PF_INET,
                         buf, SizeOf(Buf^)),
                         DWSErrAsyncGetHostbyAddr);
    if WaitFor(ATask) = 0 then
      Result := StrPas(buf^.Host.h_name)
    else
      SockException(LastTask.err, DWSErrAsyncHostToAddr);
  finally
    Dispose(buf);
    ReleaseTimer;
  end;
end;


{ LookupName -- try to look up host name }
function TSockCtrl.LookupAddrByName(const name : ShortString) : in_addr;
var
  buf   : pAsyncBuf;
  pa    : PChar;
  sz    : array [0..64] of char;
  sa    : in_addr;
  ATask : THandle;
begin
  StrPCopy(sz, name);
  New(buf);
  try
    UseTimer;
    ATask := sAsyncError( WinsockAPI.WSAAsyncGetHostByName( handle, CM_AsyncBlock,
                                     SZ, buf, SizeOf(buf^) ),
                                     DWSErrAsyncGetHostByName);
    if WaitFor(ATask) = 0 then
    begin
     { US 950518 fixed h_addr bug }
      pa := buf^.Host.h_addr_list^;
      sa.S_un_b.s_b1:=byte(pa[0]);
      sa.S_un_b.s_b2:=byte(pa[1]);
      sa.S_un_b.s_b3:=byte(pa[2]);
      sa.S_un_b.s_b4:=byte(pa[3]);
      Result := sa;
    end
    else
      SockException(LastTask.err,  DWSErrAsyncAddrToHost);
  finally
    Dispose(buf);
    ReleaseTimer;
  end;
end;

function TSockCtrl.LookupAddrByNameStr(const name : ShortString): ShortString;
var
  sa: in_addr;
begin
  sa := LookupAddrByName(name);
  Result := StrPas(WinsockAPI.inet_ntoa(sa));
end;

{ LookupService -- try to lookup service name }
{ KH 950905 }
function TSockCtrl.LookupPortByService(const service : ShortString) : Word;
var
  buf : pAsyncBuf;
  proto : array [0..32] of char;
  name : array [0..64] of char;
  ATask : THandle;
begin
  Result := 0;
  StrPCopy(proto, FProtocolName);
  StrPCopy(name, service);
  New(buf);
  try
    UseTimer;
    ATask := sAsyncError( WinsockAPI.WSAAsyncGetServByName( handle, CM_AsyncBlock,
                                     name, proto, buf, SizeOf(buf^)),
                                     DWSErrAsyncGetServbyName );
    if WaitFor(ATask) = 0 then
      Result := WinsockAPI.ntohs(buf^.service.s_port);
  (* KH  -- no exceptions, Port & Service properties cross reference each other
    else
      SockException(LastTask.err, 'Error %d "%s" finding port for service ' +
                                  name);
  *)
  finally
    Dispose(buf);
    ReleaseTimer;
  end;
end;

{ MP 18/7/95 LookupPort -- try to lookup port and return service name }
function TSockCtrl.LookupServiceByPort( port : Word) : ShortString;
var
  Buf: pAsyncBuf;
  proto  : array [1..32] of char;
  ATask: THandle;
begin
  Result := '';
  StrPCopy(@proto, FProtocolName);
  New(buf);
  try
    UseTimer;
    ATask := SAsyncError(WinsockAPI.WSAAsyncGetServByPort(Handle, CM_AsyncBlock,
                         WinsockAPI.htons(port), @proto, buf, SizeOF(Buf^)),
                         DWSErrAsyncGetServByPort);
    if WaitFor(ATask) = 0 then
      Result := Strpas(buf^.Service.s_name);
  (* KH  -- no exceptions, Port & Service properties cross reference each other
    else
      SockException(LastTask.err, 'Error %d "%s" finding service for port' +
                                  name);
  *)
  finally
    Dispose(buf);
    ReleaseTimer;
  end;
end;

function  TSockCtrl.TrackTask( Info: TAsyncInfo ): THandle;
begin
  Result := Info.TaskHandle;
  if fAsyncData = nil then
    fAsyncData := TAsyncDataList.Create;
  fAsyncData.Add(Info);
end;

{ US May 25, 1997: uses new constructor with boolean param }
function TSockCtrl.AsyncLookupHostByAddr(const a: ShortString): tHandle;
begin
  Result := TrackTask( TAsyncHostInfo.Create(Handle, true, A) );
end;

{ US May 25, 1997: uses new constructor with boolean param }
function TSockCtrl.AsyncLookupHostByName(const name: ShortString): tHandle;
begin
  Result := TrackTask( TAsyncHostInfo.Create(Handle, false, name) );
end;

function TSockCtrl.AsyncLookupServiceByService(const service: ShortString): tHandle;
begin
  Result := TrackTask( TAsyncServiceInfo.CreateByService(handle, Service, fProtocolName ));
end;

function TSockCtrl.AsyncLookupServicebyPort(const port: Word): thandle;
begin
  Result := TrackTask( TAsyncServiceInfo.CreatebyPort(handle, Port, fProtocolName ));
end;

{--------------------------------------------------------------
  TCustomClientSocket implementation.
 --------------------------------------------------------------}

constructor TCustomClientSocket.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
  FOptions := [ csoRead, csoWrite ];
end;

destructor TCustomClientSocket.Destroy;
begin
  inherited Destroy;
end;

procedure TCustomClientSocket.Open(ASocketClass : TSocketClass);
begin
  if Connected then
    raise ESockError.Create(ErrorStr(DWSErrSockAlreadyOpen));

  SetSocketClass( ASocketClass ); { creates the new socket! }

  FConn.Open(FHost, FAddress, FService, FPort, FOptions);
end;

procedure TCustomClientSocket.Close;
begin
  { MP 16/1/97 }
  if (FConn = nil) then Exit;
{    raise ESockError.Create(ErrorStr(DWSErrSockNotOpen));}
  ReleaseTimer;
  FConn.Close;
end;

function TCustomClientSocket.Connected : boolean;
begin
  Result := false;
  if FConn <> nil then
    Result := FConn.Connected;
end;

procedure TCustomClientSocket.DoConnect(var Socket: TSocketBase);
begin
{ MP added 27/11/95  to stop Connected showing TRUE when connect failed,
        and to make TimeOuts get called}
  { MP 951127 moved this up here - must always free the timeout }
  ReleaseTimer;
  if Socket.LastError = 0 then
  begin
    Socket.FState := ssConnected;
    { MP 950425 Let app know connection is made }
    DoInfo(Socket, siConnected);
  end;

  Socket.DoConnect;
end;

procedure TCustomClientSocket.DoDisconnect(var Socket: TSocketBase);
begin
  if FConn.Connected then
  begin
    { MP 20/4/95 }
    ReleaseTimer;
  end;
  inherited DoDisconnect(Socket);
 { US 950502 user must call xxx.Close method in OnDisconnect event  }
 { KH May 19, 1997 - Socket is now closed in inherited DoDisconnect }
end;

{ OnSockMsg -- handle CM_SOCKMSG }
procedure TCustomClientSocket.OnSockMsg(var Message : TMessage);
var
  sock : TSock;
  evt, err : word;
begin
  sock := TSock(Message.wParam);
  evt := WinsockAPI.WSAGetSelectEvent(Message.lParam);
  err := WinsockAPI.WSAGetSelectError(Message.lParam);
  if (Sock <> Conn.SocketID) then
    exit; { Stray Message Ignore it }

  { US 950826 added following line }
  FConn.FLastError := err;

  case evt of
    FD_CONNECT: if not FConn.Connected then      { NT same machine bug, FD_READ may happen before }
                  DoConnect(FConn);              { FD_CONNECT, so don't call twice, see below     }
    FD_CLOSE:   DoDisconnect(FConn);
    FD_OOB: ;
    FD_READ:    begin
                  if not FConn.Connected then    { NT same machine bug, If FD_READ comes first    }
                    DoConnect(FConn);            { Force a connect - We must be connected         }
                  DoRead(FConn);                 { Do the read as normal                          }
                end;
    FD_WRITE:   DoWrite(FConn);
  end;
end;
{--------------------------------------------------------------
  TCustomBasicServerSocket functions
 --------------------------------------------------------------}

constructor TCustomBasicServerSocket.Create(AOwner : TComponent);
begin
  FConns := nil;
  inherited Create( AOwner );
  {FConn := TStreamSocket.Create( Self ); }
  FConns := TSocketList.Create;
  FMaxConns := 16;
  FQueueLength := 5;
  FOptions := [ csoRead, csoWrite ];
  FRestriction := hrDeny;
  FRestrictedHosts := TStringList.Create;
  FDNSLookup := False;
end;

destructor TCustomBasicServerSocket.Destroy;
var
  i : integer;
begin
  if FConns <> nil then
  begin
    for i := 0 to FConns.Count-1 do
      FConns.Sockets[i].Free;
    FConns.Free;
  end;
  inherited Destroy;
end;

function TCustomBasicServerSocket.GetClient(cid : integer) : TSocketBase;
begin
  Result := FConns[cid];
end;

function TCustomBasicServerSocket.GetClientCount : integer;
begin
  Result := FConns.Count;
end;

procedure TCustomBasicServerSocket.Close;
begin
  { US 950502 }
  if (FConn = nil) then Exit;
  ReleaseTimer;
  FConn.Close;
end;

{ MP 20/04/95 CloseDown added. Closes all connection sockets and then closes
  the server socket. Useful for shutting down entire server without destroying
  the actual server object }
procedure TCustomBasicServerSocket.CloseDown;
var
  i : Integer;
begin
  for i := 0 to FConns.Count-1 do
    FConns.Sockets[i].Close;
  FConns.Clear;  { KH 6/5/96 - Clear the client list }
  { MP 20/4/95 }
  Close;

  { US 950502 }
  ReleaseTimer;
end;

{ US 950427: CBSockClose }
procedure TCustomBasicServerSocket.CBSockClose(ASocket : TSocketBase);
var
  i : integer;

begin
  inherited cbSockClose(ASocket);
  if ASocket = FConn then Exit;  { Server's socket will NOT be in the list }
  for i := 0 to FConns.Count-1 do
    if FConns.Sockets[i] = ASocket then  { Removed ".FSocket" as suggested by MJW }
    begin
      FConns.Sockets[i].Free;
      FConns.Delete(i);
      FConns.Pack;  { ok, not particularly efficient }
      Break; { KH 950713 Why Keep going we just removed it }
    end;
end;

procedure TCustomBasicServerSocket.DoAccept(var Socket: TSocketBase);
var
  nl         : integer;
  reject     : Boolean;
  IP         : string;
  Aport      : u_int;
begin
  { KH 6/5/96 - Get Socket handle first and check to see if it is valid before
    assig it to the connectio list }

  Socket := CreateConn;      { Create a new Socket           }
  nl := sizeof(sockaddr_in);
  Socket.FSocket := WinsockAPI.accept(FConn.FSocket, Socket.FAddr, nl);

  if Socket.FSocket = SOCKET_ERROR then { Unable to Accept! }
  begin
    WinsockAPI.WSAGetLastError;         { Clear the error                      }
    Socket.Free;                        { Free the Socket                      }
    Socket := nil;                      { Set it to nil to avoid any confusion }
    Exit;                               { Leave                                }
  end;

  FConns.Add(Socket);                   { Add it to the Conneciton list        }

  WinsockAPI.WSAAsyncSelect(Socket.FSocket, Handle, CM_SOCKMSG,
                            MakeAsyncMask(FOptions) or FD_CLOSE);
  Socket.FState := ssConnected;

  { Check to see if the new conneciton needs to be rejected }
  { Reject if:  Too Manu current connections                }
  {             Restriction Host list tells us to           }

  Reject := FConns.Count > FMaxConns; { Too many connections? }
  if not Reject then
  begin
    WinsockAPI.CrackSockAddr_in(Socket.Addr, IP, APort);
    try
      if AllowOnlyIfReverseDNS then
        Reject := (LookupHostByAddr(IP) = '');
    except
      Reject := true;
    end;
    if Not Reject then
      Reject := not (IsHostInRestrictedList(IP) xor (FRestriction = hrDeny));
                                    {  In List  xor   Deny?  Result Reject }
                                    {  Y              Y         F      T   }
                                    {  Y              N         T      F   }
                                    {  N              Y         T      F   }
                                    {  N              N         F      T   }
  end;

  if Reject then               { Reject }
  begin
    DoInfo(Socket, siReject);
    DoReject(Socket);
  end
  else                         { We can accept this connection }
  begin
    DoInfo(Socket, siAccept);
    Socket.DoAccept;
  end;
end;

procedure TCustomBasicServerSocket.DoDisconnect(var Socket: TSocketBase);
begin
  { MP 18/4/95 changed this from NOT FConns[ to FConns[
    I think the logic was slightly erroneous }
  if Socket.Connected then
  begin
   { MP 25/4/95 }
   ReleaseTimer;
   inherited DoDisconnect(Socket);
   {Info(siClosed); KH 950719}
 end;
 { US 950502 user must call xxx.Close method in OnDisconnect event  }
 { KH May 19, 1997 - Socket is now closed in inherited DoDisconnect }
end;

procedure TCustomBasicServerSocket.DoReject(var Socket: TSocketBase);
begin
  Socket.DoReject;
  Socket.Close;
end;

procedure TCustomBasicServerSocket.SetRestrictedHosts(Value: TStrings);
begin
  FRestrictedHosts.Assign(Value);
end;

function  TCustomBasicServerSocket.IsHostInRestrictedList(Host: shortstring): boolean;
var
  I: integer;
  IsIP: Boolean;
  TempAddr,
  TempHost: String;
  Part: String;
begin
  Result := False;
  I := 0;
  while (I < FRestrictedHosts.Count) do { Check for entry in RestrictedHost list }
  begin
    if Result = true then
      Exit;

    IsIP := True; { Assume we have an IP address }
    Result := True;
    TempHost := Host;
    TempAddr := FRestrictedHosts[I];
    while (IsIP) and (TempAddr <> TempHost) and (TempAddr <> '') and (TempHost <> '') do
    begin                                   { check for wild cards }
      Part := StrToken('.', TempAddr);
      isIP := IsIP and ((Part = '*') or (StrToIntDef(Part, -1) > -1));
      if (Part <> StrToken('.', TempHost)) and (Part <> '*') then
        Result := False; { can't exit. need to check if it is not an IP address }
    end;

    if (not IsIP) then
    begin
      Result := True;
      TempHost := LookupHostByAddr(Host);
      TempAddr := FRestrictedHosts[I];
      while (TempAddr <> TempHost) and (TempAddr <> '') and (TempHost <> '') do
      begin                                   { check for wild cards }
        Part := StrToken('.', TempAddr);
        if (Part <> StrToken('.', TempHost)) and (Part <> '*') then
        begin
          Result := False;
        end;
      end;
    end;
    Inc(I);
  end;
end;


{ OnSockMsg -- handle CM_SOCKMSG from WINSOCK }
{ KH 950821 Remove use of cid and switch to Socket and check on existance of
{           Socket and issue an info(siNoSuchSocket) if it does no exist. }
{           Cleaned up processing to avoide extra calls                   }
procedure TCustomBasicServerSocket.OnSockMsg(var Message : TMessage);

  function FindConn(var Socket: TSocketBase; sock: TSock): boolean;
  var
    i : integer;
  begin
    Result := False;
    for i := 0 to FConns.Count-1 do
      if FConns.Sockets[i].FSocket = sock then
      begin
        Socket := FConns.Sockets[i];
        Result := True;
        Exit;
      end;
  end;

var
  sock  : TSock;
  evt : word;
  err : word;
  Socket: TSocketBase;

begin
  sock := TSock(Message.wParam);
  evt := WinsockAPI.WSAGetSelectEvent(Message.lParam);
  err := WinsockAPI.WSAGetSelectError(Message.lParam);

  if evt = FD_Accept then
    DoAccept(Socket)
  else
  begin
    if FSocketClass.InheritsFrom(TStreamSocket) then
    begin
      if (not FindConn(Socket, Sock)) then
      begin
        if assigned(FOnInfo) then
          FOnInfo(self, siNoSuchSocket, nil );
        Exit;
      end;
    end
    else if FSocketClass.InheritsFrom(TDatagramSocket) then
      Socket := Conn;
    { US 950826 added following line }
    Socket.FLastError := err;
    case evt of
      FD_CLOSE: DoDisconnect(Socket);
      FD_READ:  DoRead(Socket);
      FD_WRITE: DoWrite(Socket);
      FD_OOB: ;
    end;
  end;
end;

procedure TCustomBasicServerSocket.Listen(ASocketClass : TSocketClass);
begin
  FSocketClass := ASocketClass;
  { MP moved from .Create }
  FConn := FSocketClass.Create( Self );
  FConn.Listen(FHost, FAddress, FService, FPort, FQueueLength);
end;

{--------------------------------------------------------------
  Unit initialization code.
 --------------------------------------------------------------}

initialization
  TimerUserCount := 0;
  TimerChainRoot := nil;
  Timer := nil;
  MsgQueue := nil;
  WinsockAPI := TWinsockAPI.Create;
  {$ifndef win32}
  AddExitProc(CleanUp);
  {$else}
finalization
  Cleanup;
  {$endif}

end.
