unit sockfact;

interface

uses sysUtils, Classes, dWinsock, TypInfo;

type
	ESocketFactory = Exception;

	TSockCtrlClass = class of TSockCtrl;

  TSocketFactoryNotifyEvent =
  	procedure (Sender : TObject; ASocket : TSockCtrl) of object;

	TSocketFactory = class (TComponent)
  protected
		FInstancesSoFar : Longint;
  	FTemplate : TSockCtrl;
  	FSockets : TList;
    FMaxSockets : Integer;
    FOnSocketCreated, FOnSocketDestroyed : TSocketFactoryNotifyEvent;

  	function GetCount : Integer;
    function GetSocket( i : Integer) : TSockCtrl;

		{ watch for deletions }
		procedure Notification(AComponent: TComponent;
    	Operation: TOperation); override;
  public
		constructor Create( AOwner : TComponent); override;
    destructor Destroy; override;

  	function NewSocket : TSockCtrl;
		procedure DeleteSocket( s : TSockCtrl);

    procedure DeleteAll;

    function HasSocket( s : TSockCtrl) : Boolean;

  	property Count : Integer read GetCount;
  	property Sockets[ i : Integer] : TSockCtrl read GetSocket;
  published
  	property TemplateSocket : TSockCtrl read FTemplate write FTemplate;
  	property MaxSockets : Integer read FMaxSockets write FMaxSockets;

    property OnSocketCreated : TSocketFactoryNotifyEvent
    	read FOnSocketCreated write FOnSocketCreated;
    property OnSocketDestroyed : TSocketFactoryNotifyEvent
    	read FOnSocketDestroyed write FOnSocketDestroyed;
  end;

  procedure CopyProps( Source, Dest : TObject; Filter : TTypeKinds;
  	const ExcludeProps : array of string);
  procedure CopyProperties( Source, Dest : TObject;
		const ExcludeProps : array of string);
  procedure CopyEvents( Source, Dest : TObject;
  	const ExcludeProps : array of string);
  procedure GetPropsAndValues( o : TObject; Filter : TTypeKinds; l : TStrings);
  procedure GetProperties( o : TObject; l : TStrings);
  procedure GetEvents( o : TObject; l : TStrings);

implementation

uses WinTypes;

procedure CopyProps( Source, Dest : TObject; Filter : TTypeKinds;
	const ExcludeProps : array of string);
var
	propinf : PPropList;
  I, propcount : Integer;
  tinfo : PTypeInfo;

  ordval : Longint;
  strval : string;
  floatval : Extended;
  classval : TObject;
  methodval : TMethod;

  function Excl( const propn : string) : Boolean;
  var
  	J : Integer;
  begin
  	for J := Low(ExcludeProps) to High(ExcludeProps) do
    	if CompareText(ExcludeProps[J], propn) = 0 then
      begin
      	Result := True;
        Exit;
      end;
    Result := False;
  end;

begin
  tinfo := Source.ClassInfo;
	propcount := GetPropList( tinfo, Filter, nil);
  if propcount <> 0 then
  begin
  	GetMem( propinf, SizeOf(PPropInfo)*propcount);
    try
			propcount := GetPropList( tinfo, Filter, propinf);
      for I := 0 to propcount-1 do
      begin
      	if Excl(propinf[I]^.Name) then Continue;

        case propinf[I]^.PropType^.Kind of
          tkUnknown :
          	begin
            end;
          {$ifdef Win32}
          tkInteger, tkChar, tkEnumeration, tkSet, tkWChar :
					{$else}
          tkInteger, tkChar, tkEnumeration, tkSet :
        	{$endif}
          	begin
							ordval := GetOrdProp( Source, propinf[I]);
              SetOrdProp( Dest, propinf[I], ordval);
            end;
          tkFloat :
          	begin
							floatval := GetFloatProp( Source, propinf[I]);
              SetFloatProp( Dest, propinf[I], floatval);
            end;
          {$ifdef VER90}
          tkString, tkLString, tkLWString  :
					{$else}
          {$ifdef VER100}
          tkString, tkLString :
		    	{$else}
          tkString :
          {$endif}
          {$endif}
          	begin
							strval := GetStrProp( Source, propinf[I]);
              SetStrProp( Dest, propinf[I], strval);
            end;
          tkClass :
          	begin
            end;
          tkMethod :
          	begin
							methodval := GetMethodProp( Source, propinf[I]);
              SetMethodProp( Dest, propinf[I], methodval);
            end;
        end;
      end;
    finally
    	FreeMem(propinf, SizeOf(PPropInfo)*propcount);
    end;
  end;
end;

procedure CopyProperties( Source, Dest : TObject;
	const ExcludeProps : array of string);
begin
	CopyProps( Source, Dest, tkProperties, ExcludeProps);
end;

procedure CopyEvents( Source, Dest : TObject;
	const ExcludeProps : array of string);
begin
	CopyProps( Source, Dest, tkMethods, ExcludeProps);
end;

procedure GetPropsAndValues( o : TObject; Filter : TTypeKinds; l : TStrings);
type
	PPointer = ^Pointer;
var
	propinf : PPropList;
  I, propcount : Integer;
  tinfo : PTypeInfo;

  ordval : Longint;
  strval, valname : string;
  floatval : Extended;
  classval : TObject;
  methodval : TMethod;
  methpoint : Pointer;
	xobj : TObject;
begin
  tinfo := o.ClassInfo;
	propcount := GetPropList( tinfo, Filter, nil);
  l.Clear;
  if propcount <> 0 then
  begin
  	GetMem( propinf, SizeOf(PPropInfo)*propcount);
    try
			propcount := GetPropList( tinfo, Filter, propinf);
      for I := 0 to propcount-1 do
      begin
        valname := '[Unknown]';

        case propinf[I]^.PropType^.Kind of
          tkUnknown :
          	begin
            end;
          {$ifdef Win32}
          tkInteger, tkChar, tkEnumeration, tkSet, tkWChar :
					{$else}
          tkInteger, tkChar, tkEnumeration, tkSet :
        	{$endif}
          	begin
							ordval := GetOrdProp( o, propinf[I]);
              case propinf[I]^.PropType^.Kind of
	              tkInteger : valname := IntToStr(ordval);
                tkChar :
                    if not (Char(ordval) in [#33..#127]) then
                    	valname := Char(ordval)
                    else
                    	valname := Format('#%d', [ordval]);
                tkEnumeration :
                {$ifdef VER100}
                	valname := GetEnumName( propinf[I]^.PropType^, ordval);
                {$else}
                	valname := GetEnumName( propinf[I]^.PropType, ordval);
                {$endif}
                tkSet :
                	valname := Format('enum (%d)', [ordval])
              end;
            end;
          tkFloat :
          	begin
							floatval := GetFloatProp( o, propinf[I]);
              valname := FloatToStr(floatval);
            end;
          {$ifdef VER90}
          tkString, tkLString, tkLWString  :
					{$else}
          {$ifdef VER100}
          tkString, tkLString :
					{$else}
          tkString :
          {$endif}
          {$endif}
          	begin
							valname := GetStrProp( o, propinf[I]);
            end;
          tkClass :
          	begin
            	valname := 'Instance of '+
              {$ifdef VER100}
              	GetTypeData(propinf[I]^.PropType^)^.ClassType.ClassName;
              {$else}
              	GetTypeData(propinf[I]^.PropType)^.ClassType.ClassName;
              {$endif}
            end;
          tkMethod :
          	begin
							methodval := GetMethodProp( o, propinf[I]);
						  xobj := TObject(methodval.Data);
              valname :=
              	xobj.MethodName(methodval.Code);
            end;
        end;

      	l.Add( Format( '%s = %s', [propinf[I]^.Name, valname]));
      end;
    finally
    	FreeMem(propinf, SizeOf(PPropInfo)*propcount);
    end;
  end;
end;

procedure GetProperties( o : TObject; l : TStrings);
begin
	GetPropsAndValues( o, tkProperties, l);
end;

procedure GetEvents( o : TObject; l : TStrings);
begin
	GetPropsAndValues( o, tkMethods, l);
end;

constructor TSocketFactory.Create( AOwner : TComponent);
begin
	inherited Create(AOwner);
	FSockets := TList.Create;
end;

destructor TSocketFactory.Destroy;
begin
	{ We don't free the sockets here because we the form automatically kills
  	them! }
	FSockets.Free;
	inherited Destroy;
end;

function TSocketFactory.GetCount : Integer;
begin
	Result := FSockets.Count;
end;

function TSocketFactory.GetSocket( i : Integer) : TSockCtrl;
begin
	Result := TSockCtrl(FSockets[i]);
end;

{ watch for deletions }
procedure TSocketFactory.Notification(AComponent: TComponent;
  Operation: TOperation);
var
	idx : Integer;
begin
  inherited Notification( AComponent, Operation);

	if Operation = opRemove then
  begin
  	if AComponent = FTemplate then
    	FTemplate := nil
    else if FSockets.IndexOf(AComponent) <> -1 then
	  begin
  		FSockets.Remove(AComponent);
    	if Assigned( FOnSocketDestroyed) then
    		FOnSocketDestroyed( Self, AComponent as TSockCtrl);
  	end;
  end;
end;

function TSocketFactory.NewSocket : TSockCtrl;
begin
	if FTemplate <> nil then
  begin
  	if (FMaxSockets > 0) and (Count >= FMaxSockets) then
    	raise ESocketFactory.Create( 'Too many sockets!');

  	Result := TSockCtrlClass(FTemplate.ClassType).Create( Owner);
		Inc(FInstancesSoFar);
    FSockets.Add(Result);

    Result.Name := Format( '%sCopy%d', [FTemplate.Name, FInstancesSoFar]);

    CopyProperties( FTemplate, Result, ['Name']);
    CopyEvents( FTemplate, Result, ['']);

    if Assigned( FOnSocketCreated) then
    	FOnSocketCreated( Self, Result);
  end else raise ESocketFactory.Create( 'No template assigned!');
end;

procedure TSocketFactory.DeleteSocket( s : TSockCtrl);
begin
	if FSockets.IndexOf(s) <> -1 then
		s.Free
  else raise
  	ESocketFactory.Create( 'Socket factory does not own this socket!');
end;

procedure TSocketFactory.DeleteAll;
begin
  { Free all of them - Notification gets called so they are removed
    from the list automatically! You MUST NOT do it here! }
	while FSockets.Count > 0 do
  	TSockCtrl(FSockets[0]).Free;
end;

function TSocketFactory.HasSocket( s : TSockCtrl) : Boolean;
begin
	Result := FSockets.IndexOf(s) <> -1;
end;

end.
