(*****************************************************************************
 * dWinsock component for Borland Delphi.
 *
 * (C) 1995 by Ulf Sderberg, ulfs@sysinno.se
 *             Marc Palmer,   marc@cablenet.net
 *             Keith Hawes,   khawes@ccmail.com
 *
 *  wsAPI.pas - Delphi equivilant of winsock.h.
 *            - dWinsock's exception class.
 *            - dWinsock's tWinsockAPI class.
 *
 * $Revision:   1.9  $
 *
 * -- History --
 * $Log:   //SUPPORT_SYSTEMS/PVCS/Source/WSAPI._PAS  $
 * 
 *    Rev 1.9   27 May 1997 09:37:20   khawes
 * Changes for C++ Builder support
 * 
 * 
 *    Rev 1.8   19 May 1997 16:11:56   khawes
 * Name Change TCustom*Socket to TBasicCustom*Socket
 * 
 * 
 *    Rev 1.7   25 Apr 1997 15:19:52   khawes
 * Big Fix in CrackSock_addr
 * 
 * 
 *    Rev 1.6   07 Mar 1997 10:34:34   khawes
 * Added public properties:
 *   MaxDatagramSize : Word
 *   MaxSockets : Word 
 *   Version : Word 
 *   HighVersion : Word
 *   Description : string
 *   property SystemStatus : string
 *   property VendorInfo : Pointer
 * 
 * 
 *    Rev 1.5   21 Feb 1997 11:32:20   khawes
 * KH: Raise an exception in TSockCtrl.GetConn if FSocketClass is nil.
 * KH: Moved SetSocketClass from TBasicClientSocket to TSockCtrl.
 * KH: Set FStarted to False after a successful WSACleanup.
 * KH: DoReject virtual method in TCustomBasicServerSocket.
 * KH: Handle NT same machine bug: FD_READ before FD_CONNECT.
 * 
 * 
 *    Rev 1.4   05 Feb 1997 08:42:42   khawes
 * Free FDLL's memory on destroy
 * 
 * 
 *    Rev 1.3   29 May 1996 16:47:24   KHAWES
 * Changed Definition of SELECT
 * changed fd_set variables to be VAR to force a pointer
 *
 * 
 *    Rev 1.2   20 May 1996 16:14:36   khawes
 * Added additional ";" to function prototypes.  Removed usless comments.
 * 
 *
 *    Rev 1.1   24 Apr 1996 09:18:20   khawes
 * Remove Time-Bomb
 *
 *
 *    Rev 1.0   29 Mar 1996 10:00:50   khawes
 * Begin PVCS version controll
 *
 *
 *****************************************************************************)
{$A+}

unit Wsapi;

interface
{--------------------------------------------------------------
  Winsock.h file for Delphi Based on a conversion by Marc B. Manza
  of Center for Applied Large-Scale Computing as modified by
  Mike Caughran of Cedar Island Software in Nov 1994
  Winsock.inc, Winsock.if & Winsock.imp

  Converted to a single unit by Keith Hawes. For use with Delphi and
  dWinsock.pas (A Delphi Winsock VCL)

  (C) 1995 by Ulf Sderberg, ulfs@sysinno.se
              Marc Palmer,   marc@cablenet.net
              Keith Hawes,   khawes@ccmail.com

  -- History --
  1.42 10-Nov-95 - Keith
       * Changes for 32bit WSOCK32.DLL and Delphi32
         Works with Delphi32 Beta and windows 95.  But will not run under NT.
  2.00 Beta 3
       27-Jan-96 - Marc
       * Changed all sockaddr references in API functions to PSockaddr.
         (Makes no difference to existing code but is nevertheless correct).
       31-Jan-96 - Keith
       * Changed WSAData.iMaxSockets, and WSAData.iMaxUdpDg from shortInt to word
       * Changed wsock32 to wsock32.dll to allow running under NT.
  2.00 Beta 4
       06-Feb-96 - Keith
       * Begin Conversion of wsapi to tWinsockAPI an new object to allow
         dynamic loading of the Winsock DLL.
       * Add .DLL to DLLname if not extention exists Begause LoadLibrary
         requires an extention.
       * Changed u_int from Word to Cardinal (whoops, How did this work before?)
  2.00 Beta 5 (prerelease)
       02-Mar-96 - Keith
       * Moved ESockError to wsAPI.
       * Added an exception when trying to change the DLL name while it is
         loaded.
       * Made Port unsighed everywhere.
       07-Mar-96
       * Change to use function names instead of number to link to Winsock.dll
         funcitons.
       * Changed many of the API declarations to more closely match what is in
         Borlands translatin of winsock.h, winsock.pas.
       12-Mar-96
       * Added prerelease time bomb to wsaStartup.
       20-Mar-96
       * Moved ";" to end of line on prototypes.

---------------------------------------------------------------}

Uses WinTypes, WinProcs, dwsError, sysUtils, Classes;

type
  { dWinsock exception type }
  ESockError = class(Exception);

  u_int = cardinal;
  TSock = integer;

const
  FD_SETSIZE = 64;

type
  PFD_SET = ^FD_SET;
  fd_set = packed record
    fd_count : word;  {Borland has a integer here but the .h has a u_short --KH}
    fd_array : array[0..FD_SETSIZE-1] of TSock;
  end;

  ptimeval = ^timeval;
  timeval = record
    tv_sec, tv_usec : longint;
  end;

const
  IOCPARM_MASK   =   $007F;
  IOC_VOID       =   $20000000;
  IOC_OUT        =   $40000000;
  IOC_IN         =   $80000000;
  IOC_INOUT      =   (IOC_IN OR IOC_OUT);

  FIONREAD     = IOC_OUT or
    ((Longint(SizeOf(Longint)) and IOCPARM_MASK) shl 16) or
    (Longint(Byte('f')) shl 8) or 127;
  FIONBIO      = IOC_IN or
    ((Longint(SizeOf(Longint)) and IOCPARM_MASK) shl 16) or
    (Longint(Byte('f')) shl 8) or 126;
  FIOASYNC     = IOC_IN or
    ((Longint(SizeOf(Longint)) and IOCPARM_MASK) shl 16) or
    (Longint(Byte('f')) shl 8) or 125;


  MAXGETHOSTSTRUCT   =  1024;


type
  PPchar = ^PChar;


  PHostEnt = ^HostEnt;
  hostent = packed record
    h_name : PChar;
    h_aliases : PPChar;
    h_addrtype : SmallInt;
    h_length : SmallInt;
    h_addr_list : PPChar;
  end;

  netent = packed record
    n_name : PChar;
    n_aliases : ^PChar;
    n_addrtype : SmallInt;
    n_net : LongInt;
  end;

  PServEnt = ^ServEnt;
  servent = packed record
    s_name : PChar;
    s_aliases : ^PChar;
    s_port : Word;
    s_proto : PChar;
  end;

  PProtoEnt = ^ProtoEnt;
  protoent = packed record
    p_name : PChar;
    p_aliases : ^Pchar;
    p_proto : SmallInt;
  end;

PAsyncBuf = ^TAsyncBuf;
TAsyncBuf = packed record { KH Used in calls to WASAsyncXbyY calls }
  case integer of
    0: (Host:     Hostent);
    1: (Service:  ServEnt);
    2: (Protocol: ProtoEnt);
  255: (Reserved: array[1..MaxGetHostStruct] of byte);
end;

const
  IPPROTO_IP         =  0;
  IPPROTO_ICMP       =  1;
  IPPROTO_GGP        =  2;
  IPPROTO_TCP        =  6;
  IPPROTO_PUP        =  12;
  IPPROTO_UDP        =  17;
  IPPROTO_IDP        =  22;
  IPPROTO_ND         =  77;

  IPPROTO_RAW        =  255;
  IPPROTO_MAX        =  256;

  IPPORT_ECHO        =  7;
  IPPORT_DISCARD     =  9;
  IPPORT_SYSTAT      =  11;
  IPPORT_DAYTIME     =  13;
  IPPORT_NETSTAT     =  15;
  IPPORT_FTP         =  21;
  IPPORT_TELNET      =  23;
  IPPORT_SMTP        =  25;
  IPPORT_TIMESERVER  =  37;
  IPPORT_NAMESERVER  =  42;
  IPPORT_WHOIS       =  43;
  IPPORT_MTP         =  57;

  IPPORT_TFTP        =  69;
  IPPORT_RJE         =  77;
  IPPORT_FINGER      =  79;
  IPPORT_TTYLINK     =  87;
  IPPORT_SUPDUP      =  95;

  IPPORT_EXECSERVER  =  512;
  IPPORT_LOGINSERVER =  513;
  IPPORT_CMDSERVER   =  514;
  IPPORT_EFSSERVER   =  520;

  IPPORT_BIFFUDP     =  512;
  IPPORT_WHOSERVER   =  513;
  IPPORT_ROUTESERVER =  520;

  IPPORT_RESERVED    =  1024;

  IMPLINK_IP         =  155;
  IMPLINK_LOWEXPER   =  156;
  IMPLINK_HIGHEXPER  =  158;

Type
  SunB = record
    s_b1, s_b2, s_b3, s_b4 : Byte;
  end;

  SunW = record
    s_w1, s_w2 : Word;
  end;

  PIn_Addr = ^In_Addr;
  in_addr = record
    case integer of
      0 : (S_un_b : SunB);
      1 : (S_un_w : SunW);
      2 : (S_addr : LongInt);
    end;

type
  PSockAddr_In = ^SockAddr_In;
  sockaddr_in = record
    case integer of
     0:(sin_family : SmallInt;
        sin_port : Word;
        sin_addr : in_addr;
        sin_zero : array[0..7] of char);
     1:(sa_family: Word;
        sa_data: array[0..13] of Char);
  end;

const
  INADDR_ANY       = $00000000;
  INADDR_LOOPBACK  = $7F000001;
  INADDR_BROADCAST = $FFFFFFFF;
  INADDR_NONE      = $FFFFFFFF;

const
  WSADESCRIPTION_LEN     =   256;
  WSASYS_STATUS_LEN      =   128;

type
  LPWSADATA = ^WSAData;
  WSAData = record
    wVersion : word;
    wHighVersion : word;
    szDescription : array[0..WSADESCRIPTION_LEN] of char;
    szSystemStatus : array[0..WSASYS_STATUS_LEN] of char;
    iMaxSockets : word;
    iMaxUdpDg : word;
    lpVendorInfo : PChar;
  end;

const
  IP_OPTIONS          = 1;
  IP_MULTICAST_IF     = 2;
  IP_MULTICAST_TTL    = 3;
  IP_MULTICAST_LOOP   = 4;
  IP_ADD_MEMBERSHIP   = 5;
  IP_DROP_MEMBERSHIP  = 6;

  IP_DEFAULT_MULTICAST_TTL   = 1;
  IP_DEFAULT_MULTICAST_LOOP  = 1;
  IP_MAX_MEMBERSHIPS         = 20;

const
  INVALID_SOCKET    = TSock(NOT(0));
  SOCKET_ERROR      = -1;

  SOCK_STREAM    =   1;
  SOCK_DGRAM     =   2;
  SOCK_RAW       =   3;
  SOCK_RDM       =   4;
  SOCK_SEQPACKET =   5;

  SO_DEBUG       =   $0001;
  SO_ACCEPTCONN  =   $0002;
  SO_REUSEADDR   =   $0004;
  SO_KEEPALIVE   =   $0008;
  SO_DONTROUTE   =   $0010;
  SO_BROADCAST   =   $0020;
  SO_USELOOPBACK =   $0040;
  SO_LINGER      =   $0080;
  SO_OOBINLINE   =   $0100;

  SO_DONTLINGER  =   $ff7f;

  SO_SNDBUF      =   $1001;
  SO_RCVBUF      =   $1002;
  SO_SNDLOWAT    =   $1003;
  SO_RCVLOWAT    =   $1004;
  SO_SNDTIMEO    =   $1005;
  SO_RCVTIMEO    =   $1006;
  SO_ERROR       =   $1007;
  SO_TYPE        =   $1008;

  SO_CONNDATA     = $7000;
  SO_CONNOPT      = $7001;
  SO_DISCDATA     = $7002;
  SO_DISCOPT      = $7003;
  SO_CONNDATALEN  = $7004;
  SO_CONNOPTLEN   = $7005;
  SO_DISCDATALEN  = $7006;
  SO_DISCOPTLEN   = $7007;

  SO_OPENTYPE     = $7008;

  SO_SYNCHRONOUS_ALERT    = $10;
  SO_SYNCHRONOUS_NONALERT = $20;

  SO_MAXDG        = $7009;
  SO_MAXPATHDG    = $700A;

  TCP_NODELAY     = $0001;
  TCP_BSDURGENT   = $7000;

  AF_UNSPEC      =   0;
  AF_UNIX        =   1;
  AF_INET        =   2;
  AF_IMPLINK     =   3;
  AF_PUP         =   4;
  AF_CHAOS       =   5;
  AF_IPX         =   6;
  AF_NS          =   6;
  AF_NBS         =   7;
  AF_ISO         =   7;
  AF_OSI         =   7;
  AF_ECMA        =   8;
  AF_DATAKIT     =   9;
  AF_CCITT       =   10;
  AF_SNA         =   11;
  AF_DECnet      =   12;
  AF_DLI         =   13;
  AF_LAT         =   14;
  AF_HYLINK      =   15;
  AF_APPLETALK   =   16;
  AF_NETBIOS     =   17;
  AF_VOICEVIEW   =   18;

  AF_MAX         =   19;

type
  PSockAddr = ^SockAddr;
  sockaddr = sockaddr_in; {use the sa_* variant}

  sockproto = record
    sp_family : Word;
    sp_protocol : Word;
  end;

const
  PF_UNSPEC      =   AF_UNSPEC;
  PF_UNIX        =   AF_UNIX;
  PF_INET        =   AF_INET;
  PF_IMPLINK     =   AF_IMPLINK;
  PF_PUP         =   AF_PUP;
  PF_CHAOS       =   AF_CHAOS;
  PF_NS          =   AF_NS;
  PF_IPX         =   AF_IPX;
  PF_NBS         =   AF_NBS;
  PF_ISO         =   AF_ISO;
  PF_OSI         =   AF_OSI;
  PF_ECMA        =   AF_ECMA;
  PF_DATAKIT     =   AF_DATAKIT;
  PF_CCITT       =   AF_CCITT;
  PF_SNA         =   AF_SNA;
  PF_DECnet      =   AF_DECnet;
  PF_DLI         =   AF_DLI;
  PF_LAT         =   AF_LAT;
  PF_HYLINK      =   AF_HYLINK;
  PF_APPLETALK   =   AF_APPLETALK;
  PF_NETBIOS     =   AF_NETBIOS;
  PF_VOICEVIEW   =   AF_VOICEVIEW;
  PF_MAX         =   AF_MAX;

type
  PLinger = ^Linger;
  linger = record
    l_onoff : Word;
    l_linger : Word;
  end;

const
  SOL_SOCKET     =   $ffff;

  SOMAXCONN      =   5;

  MSG_OOB        =   $1;
  MSG_PEEK       =   $2;
  MSG_DONTROUTE  =   $4;

  MSG_MAXIOVLEN  =   16;

  MSG_PARTIAL    =   $8000;

  FD_READ            =  $01;
  FD_WRITE           =  $02;
  FD_OOB             =  $04;
  FD_ACCEPT          =  $08;
  FD_CONNECT         =  $10;
  FD_CLOSE           =  $20;


type  {Define Function prototypes for Winsock dll}

d_accept = function( s: TSock; var addr: Sockaddr; var addrlen: integer ): TSock;         {$ifdef win32} StdCall; {$endif}
d_bind = function ( s: TSock; var addr: sockaddr; namelen: Integer ): Integer;            {$ifdef win32} StdCall; {$endif}
d_closesocket = function ( s: TSock ): Integer;                                           {$ifdef win32} StdCall; {$endif}
d_connect = function ( s: TSock; var name: sockaddr; namelen: integer ): integer;         {$ifdef win32} StdCall; {$endif}
d_ioctlsocket = function ( s: TSock; cmd: longint; var arg: Longint ): Integer;           {$ifdef win32} StdCall; {$endif}
d_getpeername = function ( s: TSock; var name: sockaddr; var namelen: integer ): Integer; {$ifdef win32} StdCall; {$endif}
d_getsockname = function ( s: TSock; var name: sockaddr; var namelen: integer ): Integer; {$ifdef win32} StdCall; {$endif}
d_getsockopt = function ( s: TSock; level, optname: Integer; optval: PChar;
               var optlen: Integer ): Integer;                                            {$ifdef win32} StdCall; {$endif}
d_htonl = function ( hostlong: LongInt ): LongInt;                                        {$ifdef win32} StdCall; {$endif}
d_htons = function ( hostshort: Word ): Word;                                             {$ifdef win32} StdCall; {$endif}
d_inet_addr = function ( cp: PChar ): LongInt;                                            {$ifdef win32} StdCall; {$endif}
d_inet_ntoa = function ( inaddr: in_addr ): PChar;                                        {$ifdef win32} StdCall; {$endif}
d_listen = function ( s: TSock; backlog: Integer ): Integer;                              {$ifdef win32} StdCall; {$endif}
d_ntohl = function ( netlong: LongInt ): LongInt;                                         {$ifdef win32} StdCall; {$endif}
d_ntohs = function ( netshort: Word ): Word;                                              {$ifdef win32} StdCall; {$endif}
d_recv = function ( s: TSock; var buf; len, flags: Integer ): Integer;                    {$ifdef win32} StdCall; {$endif}
d_recvfrom = function ( s: TSock; var buf; len, flags: Integer;
             var from: Sockaddr; var fromlen: Integer ): Integer;                         {$ifdef win32} StdCall; {$endif}
d_select = function ( nfds: Integer; var readfds, writefds, exceptfds: fd_set;
           timeout: ptimeval ): longint;                                                  {$ifdef win32} StdCall; {$endif}
d_send = function ( s: TSock; var buf; len, flags: Integer ): Integer;                    {$ifdef win32} StdCall; {$endif}
d_sendto = function ( s: TSock; var buf; len, flags: Integer;
           var addrto: Sockaddr; tolen: Integer ): Integer;                               {$ifdef win32} StdCall; {$endif}
d_setsockopt = function ( s: TSock; level, optname: Integer; optval: PChar;
               optlen: Integer ): Integer;                                                {$ifdef win32} StdCall; {$endif}
d_shutdown = function ( s: TSock; how: Integer ): Integer;                                {$ifdef win32} StdCall; {$endif}
d_socket = function ( af, struc, protocol: Integer ): TSock;                             {$ifdef win32} StdCall; {$endif}
d_gethostbyaddr = function ( addr: Pointer; len, struc: Integer ): PHostEnt;             {$ifdef win32} StdCall; {$endif}
d_gethostbyname = function ( name: PChar ): PHostEnt;                                     {$ifdef win32} StdCall; {$endif}
d_gethostname = function ( name: PChar; len: Integer ): Integer;                          {$ifdef win32} StdCall; {$endif}
d_getservbyport = function ( port: u_int; proto: PChar ): PServEnt;                       {$ifdef win32} StdCall; {$endif}
d_getservbyname = function ( name, proto: PChar ): PServEnt;                              {$ifdef win32} StdCall; {$endif}
d_getprotobynumber = function ( proto: Integer ): PProtoEnt;                              {$ifdef win32} StdCall; {$endif}
d_getprotobyname = function ( name: PChar ): PProtoEnt;                                   {$ifdef win32} StdCall; {$endif}
d_WSAStartup = function ( wVersionRequired: word; var WSData: WSADATA ): Integer;         {$ifdef win32} StdCall; {$endif}
d_WSACleanup = function: Integer;                                                         {$ifdef win32} StdCall; {$endif}
d_WSASetLastError = procedure ( iError: Integer );                                        {$ifdef win32} StdCall; {$endif}
d_WSAGetLastError = function: Integer;                                                    {$ifdef win32} StdCall; {$endif}
d_WSAIsBlocking = function: BOOL;                                                         {$ifdef win32} StdCall; {$endif}
d_WSAUnhookBlockingHook = function: Integer;                                              {$ifdef win32} StdCall; {$endif}
d_WSASetBlockingHook = function ( lpBlockFunc: TFarProc ): TFarProc;                      {$ifdef win32} StdCall; {$endif}
d_WSACancelBlockingCall = function: Integer;                                              {$ifdef win32} StdCall; {$endif}
d_WSAAsyncGetServByName = function ( hWin: HWND; wMsg: u_int; name, proto: PChar;
                          buf: PAsyncBuf; buflen: Integer ): THandle;                     {$ifdef win32} StdCall; {$endif}
d_WSAAsyncGetServByPort = function (  hWin: HWND; wMsg, port: u_int; proto: PChar;
                          buf: PAsyncBuf; buflen: Integer ): THandle;                     {$ifdef win32} StdCall; {$endif}
d_WSAAsyncGetProtoByName = function ( hWin: HWND; wMsg: u_int; name: PChar;
                           buf: PAsyncBuf; buflen: Integer ): THandle;                    {$ifdef win32} StdCall; {$endif}
d_WSAAsyncGetProtoByNumber = function ( hWin: HWND; wMsg: u_int;
                             number: Integer; buf: PAsyncBuf;
                             buflen: Integer ): THandle;                                  {$ifdef win32} StdCall; {$endif}
d_WSAAsyncGetHostByName = function ( hWin: HWND; wMsg: u_int; name: PChar;
                          buf: PAsyncBuf; buflen: Integer ): THandle;                     {$ifdef win32} StdCall; {$endif}
d_WSAAsyncGetHostByAddr = function ( hWin: HWND; wMsg: u_int; addr: PChar;
                          len, struc: Integer; buf: PAsyncBuf;
                          buflen: Integer ): THandle;                                     {$ifdef win32} StdCall; {$endif}
d_WSACancelAsyncRequest =  function ( hAsyncTaskHandle: THandle ): Integer;               {$ifdef win32} StdCall; {$endif}
d_WSAAsyncSelect = function ( s: TSock; hWin: HWND; wMsg: u_int;
                   lEvent: longint ): Integer;                                            {$ifdef win32} StdCall; {$endif}

type tWinsockAPI = Class(TObject)
private
  hDLL:                      THandle;
  FDLL:                      pChar;
  fStarted:                  boolean;
  FMaxSockets:               Word;
  FMaxUDPDg:                 Word;
  FVersion:                  Word;
  FHighVersion:              Word;
  FDescription:              pChar;
  FSystemStatus:             pChar;
  FVendorInfo:               pointer;


  FAccept:                   d_accept;
  Fbind:                     d_bind;
  Fclosesocket:              d_closesocket;
  Fconnect:                  d_connect;
  Fioctlsocket:              d_ioctlsocket;
  Fgetpeername:              d_getpeername;
  Fgetsockname:              d_getsockname;
  Fgetsockopt:               d_getsockopt;
  Fhtonl:                    d_htonl;
  Fhtons:                    d_htons;
  Finet_addr:                d_inet_addr;
  Finet_ntoa:                d_inet_ntoa;
  Flisten:                   d_listen;
  Fntohl:                    d_ntohl;
  Fntohs:                    d_ntohs;
  Frecv:                     d_recv;
  Frecvfrom:                 d_recvfrom;
  Fselect:                   d_select;
  Fsend:                     d_send;
  Fsendto:                   d_sendto;
  Fsetsockopt:               d_setsockopt;
  Fshutdown:                 d_shutdown;
  Fsocket:                   d_socket;
  Fgethostbyaddr:            d_gethostbyaddr;
  Fgethostbyname:            d_gethostbyname;
  Fgethostname:              d_gethostname;
  Fgetservbyport:            d_getservbyport;
  Fgetservbyname:            d_getservbyname;
  Fgetprotobynumber:         d_getprotobynumber;
  Fgetprotobyname:           d_getprotobyname;
  FWSAStartup:               d_WSAStartup;
  FWSACleanup:               d_WSACleanup;
  FWSASetLastError:          d_WSASetLastError;
  FWSAGetLastError:          d_WSAGetLastError;
  FWSAIsBlocking:            d_WSAIsBlocking;
  FWSAUnhookBlockingHook:    d_WSAUnhookBlockingHook;
  FWSASetBlockingHook:       d_WSASetBlockingHook;
  FWSACancelBlockingCall:    d_WSACancelBlockingCall;
  FWSAAsyncGetServByName:    d_WSAAsyncGetServByName;
  FWSAAsyncGetServByPort:    d_WSAAsyncGetServByPort;
  FWSAAsyncGetProtoByName:   d_WSAAsyncGetProtoByName;
  FWSAAsyncGetProtoByNumber: d_WSAAsyncGetProtoByNumber;
  FWSAAsyncGetHostByName:    d_WSAAsyncGetHostByName;
  FWSAAsyncGetHostByAddr:    d_WSAAsyncGetHostByAddr;
  FWSACancelAsyncRequest:    d_WSACancelAsyncRequest;
  FWSAAsyncSelect:           d_WSAAsyncSelect;
protected
  function  GetDLL: string;
  procedure SetDll(Value: String);

  function  GetDescription: string;
  function  getSystemStatus: string;
  { Dynamic Load Enabling Function }
  function NeedDll(Proc: pChar): TFarProc;
  procedure NewPchar(var New: pChar; Value: pChar);
public
  property DLLName: string read GetDLL write SetDll;
  property Started: boolean read FStarted;

  { Added Mar 7 1997 KH}
  property MaxDatagramSize : Word read FMaxUDPDG;
  property MaxSockets : Word read FMaxSockets;
  property Version : Word read FVersion;
  property HighVersion : Word read FHighVersion;
  property Description : string read GetDescription;
  property SystemStatus : string read GetSystemStatus;
  property VendorInfo : Pointer read FVendorInfo;

  constructor create;
  destructor  destroy; override;
  { Delphi functions for C macros }
  function FIONBIO (x, y : SmallInt; t : LongInt) : LongInt;
  function WSAMakeAsyncReply ( Buflen, Error: Word ): LongInt;
  function WSAMakeSelectReply( Event, Error: Word ): LongInt;
  function WSAGetAsyncBuflen ( Param: LongInt ): Word;
  function WSAGetAsyncError  ( Param: LongInt ): Word;
  function WSAGetSelectEvent ( Param: LongInt ): Word;
  function WSAGetSelectError ( Param: LongInt ): Word;
  { Helpful Delphi Functions }
  function IsInternetAddress(const hostname : string) : Boolean;
  function IsPortNumber(const port: string) : Boolean;
  procedure CrackSockAddr_In( sa : sockaddr_in; var hostname : string; var port : u_int);

  { Delphi functions / Procedures to access version 1.1 of the Winsock API. }
  function accept ( s: TSock; var addr: Sockaddr; var addrlen: integer ): TSock;
  function bind ( s: TSock; var addr: sockaddr; namelen: Integer ): Integer;
  function closesocket ( s: TSock ): Integer;
  function connect ( s: TSock; var name: sockaddr; namelen: integer ): integer;
  function ioctlsocket ( s: TSock; cmd: longint; var arg: Longint ): Integer;
  function getpeername ( s: TSock; var name: sockaddr; var namelen: integer ): Integer;
  function getsockname ( s: TSock; var name: sockaddr; var namelen: integer ): Integer;
  function getsockopt ( s: TSock; level, optname: Integer; optval: PChar;
           var optlen: Integer ): Integer;
  function htonl ( hostlong: LongInt ): LongInt;
  function htons ( hostshort: Word ): Word;
  function inet_addr ( cp: PChar ): LongInt;
  function inet_ntoa ( inaddr: in_addr ): PChar;
  function listen ( s: TSock; backlog: Integer ): Integer;
  function ntohl ( netlong: LongInt ): LongInt;
  function ntohs ( netshort: Word ): Word;
  function recv ( s: TSock; var buf; len, flags: Integer ): Integer;
  function recvfrom ( s: TSock; var buf; len, flags: Integer;
           var from: Sockaddr; var fromlen: Integer ): Integer;
  function select ( nfds: Integer; var readfds, writefds, exceptfds: fd_set;
           timeout: ptimeval ): longint;
  function send ( s: TSock; var buf; len, flags: Integer ): Integer;
  function sendto ( s: TSock; var buf; len, flags: Integer;
           var addrto: Sockaddr; tolen: Integer ): Integer;
  function setsockopt ( s: TSock; level, optname: Integer; optval: PChar;
           optlen: Integer ): Integer;
  function shutdown ( s: TSock; how: Integer ): Integer;
  function socket ( af, struc, protocol: Integer ): TSock;
  function gethostbyaddr ( addr: Pointer; len, struc: Integer ): PHostEnt;
  function gethostbyname ( name: PChar ): PHostEnt;
  function gethostname ( name: PChar; len: Integer ): Integer;
  function getservbyport ( port: u_int; proto: PChar ): PServEnt;
  function getservbyname ( name, proto: PChar ): PServEnt;
  function getprotobynumber ( proto: Integer ): PProtoEnt;
  function getprotobyname ( name: PChar ): PProtoEnt;
  function WSAStartup ( wVersionRequired: word; var WSData: WSADATA ): Integer;
  function WSACleanup : Integer;
  procedure WSASetLastError ( iError: Integer );
  function WSAGetLastError : Integer;
  function WSAIsBlocking : BOOL;
  function WSAUnhookBlockingHook : Integer;
  function WSASetBlockingHook ( lpBlockFunc: TFarProc ): TFarProc;
  function WSACancelBlockingCall : Integer;
  function WSAAsyncGetServByName ( hWin: HWND; wMsg: u_int; name, proto: PChar;
           buf: PAsyncBuf; buflen: Integer ): THandle;
  function WSAAsyncGetServByPort ( hWin: HWND; wMsg, port: u_int; proto: PChar;
           buf: PAsyncBuf; buflen: Integer ): THandle;
  function WSAAsyncGetProtoByName ( hWin: HWND; wMsg: u_int; name: PChar;
           buf: PAsyncBuf; buflen: Integer ): THandle;
  function WSAAsyncGetProtoByNumber ( hWin: HWND; wMsg: u_int;
           number: Integer; buf: PAsyncBuf;
           buflen: Integer ): THandle;
  function WSAAsyncGetHostByName ( hWin: HWND; wMsg: u_int; name: PChar;
           buf: PAsyncBuf; buflen: Integer ): THandle;
  function WSAAsyncGetHostByAddr ( hWin: HWND; wMsg: u_int; addr: PChar;
           len, struc: Integer; buf: PAsyncBuf;
           buflen: Integer ): THandle;
  function WSACancelAsyncRequest ( hAsyncTaskHandle: THandle ): Integer;
  function WSAAsyncSelect ( s: TSock; hWin: HWND; wMsg: u_int;
           lEvent: longint ): Integer;
end;

implementation

constructor tWinsockAPI.create;
begin
  inherited create;
  fStarted := False;
  hDll := 0;
  FDLL := nil;
  FDescription := nil;
  FSystemStatus := nil;
end;

destructor  tWinsockAPI.destroy;
begin
  if FDLL <> nil then
    StrDispose( FDLL);
  if FDescription <> nil then
    StrDispose(FDescription);
  if FSystemStatus <> nil then
    StrDispose(FSystemStatus);
  if hDll > hInstance_Error then
  begin
    FreeLibrary(hDLL);
    hDLL := 0;
  end;
  inherited destroy;
end;

function tWinsockAPI.GetDLL: string;
begin
  if FDLL <> nil then
    Result := StrPas(FDLL)
  else
    Result := ''
end;

procedure tWinsockAPI.SetDLL(Value: string);
begin
  if hDLL <> 0 then
    raise ESockError.Create(LoadStr(DWSErrDLLAlreadyLoaded));
  if FDLL <> nil then
  begin
    if CompareText(Value, DLLName) = 0 then
      exit;
    StrDispose(FDLL);
  end;
  if Value = '' then
    FDLL := nil
  else
  begin
    if ExtractFileExt(Value) = '' then
      Value := ChangeFileExt(Value, '.DLL');  { Add .DLL if necessary }
    FDLL := StrAlloc(Length(Value)+1);
    StrPCopy( FDLL, Value );
  end;
end;

function  tWinsockAPI.GetDescription: string;
begin
  if FDescription <> nil then
    Result := StrPas(FDescription)
  else
    Result := ''
end;

function  tWinsockAPI.getSystemStatus: string;
begin
  if FSystemStatus <> nil then
    Result := StrPas(FSystemStatus)
  else
    Result := ''
end;

function tWinsockAPI.FIONBIO (x, y : SmallInt; t : LongInt) : LongInt;
begin
  FIONBIO := (IOC_IN OR ((IOCPARM_MASK shl 16) OR (x SHL 8) OR (Y)));
end;

function tWinsockAPI.WSAMakeAsyncReply( Buflen, Error: Word ): LongInt;
begin
  WSAMakeAsyncReply := MakeLong (Buflen, Error);
end;

function tWinsockAPI.WSAMakeSelectReply( Event, Error: Word ): LongInt;
begin
  result := MakeLong (Event, Error);
end;

function tWinsockAPI.WSAGetAsyncBuflen( Param: LongInt ): Word;
begin
  result := LOWORD(Param);
end;

function tWinsockAPI.WSAGetAsyncError( Param: LongInt ): Word;
begin
  result := HIWORD(Param);
end;

function tWinsockAPI.WSAGetSelectEvent( Param: LongInt ): Word;
begin
  result := LOWORD(Param);
end;

function tWinsockAPI.WSAGetSelectError( Param: LongInt ): Word;
begin
  result := HIWORD(Param);
end;

function tWinsockAPI.IsInternetAddress(const hostname : string) : Boolean;
var
  buf : array[0..100] of Char;
begin
  StrPLCopy( buf, hostname, SizeOf(buf));
  Result := inet_addr(buf) <> INADDR_NONE;
end;

function tWinsockAPI.IsPortNumber(const port: string) : Boolean;
begin
  Result := StrToIntDef(port, -1) <> -1;
end;

procedure tWinsockAPI.CrackSockAddr_In( sa : sockaddr_in; var hostname : string; var port : u_int);
begin
  hostname := StrPas( inet_ntoa( sockaddr_in(sa).sin_addr ) );
  port := ntohl(sockaddr_in(sa).sin_port);
end;

function tWinsockAPI.NeedDll(Proc: pChar): TFarProc;
{$ifdef win32}
  const WinsockDLL = 'WSOCK32.DLL';
{$else}
  const WinsockDLL = 'WINSOCK.DLL';
{$endif}
begin
  if hDLL < HInstance_Error then
  begin
    if FDLL = nil then
      DLLName := WinsockDLL;
    hDLL := LoadLibrary(FDLL);
    if hDLL < HInstance_Error then
      raise Exception.Create(format('Unable to load Library "%s" Error %d',
      [DLLName, hDLL]));
  end;
  Result := GetProcAddress(hDLL, Proc);
  if Result = nil then
    raise Exception.Create('Unable to locate function in DLL');
end;

procedure tWinsockAPI.NewPchar(var New: pChar; Value: pChar);
begin
  if New <> nil then
    StrDispose(New);
  if Value = nil then
    New := nil
  else
  begin
    New := StrAlloc(StrLen(Value)+1);
    StrCopy( New, Value );
  end;
end;

function TWinsockAPI.accept( s: TSock; var addr: Sockaddr; var addrlen: integer ): TSock;
begin
  if not Assigned(FAccept) then
    @FAccept := NeedDLL('accept');
  Result := FAccept(s, addr, addrlen );
end;

function TWinsockAPI.bind ( s: TSock; var addr: sockaddr; namelen: Integer ): Integer;
begin
  if not Assigned(FBind) then
    @FBind := NeedDLL('bind');
  Result := FBind(s, Addr, namelen );
end;

function tWinsockAPI.closesocket ( s: TSock ): Integer;
begin
  if not Assigned(Fclosesocket) then
    @Fclosesocket := NeedDLL('closesocket');
  Result := Fclosesocket( s );
end;

function tWinsockAPI.connect ( s: TSock; var name: sockaddr; namelen: integer ): integer;
begin
  if not assigned(Fconnect) then
    @Fconnect := NeedDLL('connect');
  Result := Fconnect( s, name, namelen );
end;

function tWinsockAPI.ioctlsocket ( s: TSock; cmd: longint; var arg: Longint ): Integer;
begin
  if not assigned(Fioctlsocket) then
    @Fioctlsocket := NeedDLL('ioctlsocket');
  Result := Fioctlsocket( s, cmd, arg );
end;

function tWinsockAPI.getpeername ( s: TSock; var name: sockaddr; var namelen: integer ): Integer;
begin
  if not assigned(Fgetpeername) then
    @Fgetpeername := NeedDLL('getpeername');
  Result := Fgetpeername( s, name, namelen );
end;

function tWinsockAPI.getsockname ( s: TSock; var name: sockaddr; var namelen: integer ): Integer;
begin
  if not assigned(Fgetsockname) then
    @Fgetsockname := NeedDLL('getsockname');
  Result := Fgetsockname( s, name, namelen );
end;

function tWinsockAPI.getsockopt ( s: TSock; level, optname: Integer; optval: PChar;
         var optlen: Integer ): Integer;
begin
  if not assigned(Fgetsockopt) then
    @Fgetsockopt := NeedDLL('getsockopt');
  Result := Fgetsockopt( s, level, optname, optval, optlen );
end;

function tWinsockAPI.htonl ( hostlong: LongInt ): LongInt;
begin
  if not assigned(Fhtonl) then
    @Fhtonl := NeedDLL('htonl');
  Result := Fhtonl( hostlong );
end;

function tWinsockAPI.htons ( hostshort: Word ): Word;
begin
  if not assigned(Fhtons) then
    @Fhtons := NeedDLL('htons');
  Result := Fhtons( hostshort );
end;

function tWinsockAPI.inet_addr ( cp: PChar ): LongInt;
begin
  if not assigned(Finet_addr) then
    @Finet_addr := NeedDLL('inet_addr');
  Result := Finet_addr( cp );
end;

function tWinsockAPI.inet_ntoa ( inaddr: in_addr ): PChar;
begin
  if not assigned(Finet_ntoa) then
    @Finet_ntoa := NeedDLL('inet_ntoa');
  Result := Finet_ntoa( inaddr );
end;

function tWinsockAPI.listen ( s: TSock; backlog: Integer ): Integer;
begin
  if not assigned(Flisten) then
    @Flisten := NeedDLL('listen');
  Result := Flisten( s, backlog );
end;

function tWinsockAPI.ntohl ( netlong: LongInt ): LongInt;
begin
  if not assigned(Fntohl) then
    @Fntohl := NeedDLL('ntohl');
  Result := Fntohl( netlong );
end;

function tWinsockAPI.ntohs ( netshort: Word ): Word;
begin
  if not assigned(Fntohs) then
    @Fntohs := NeedDLL('ntohs');
  Result := Fntohs( netshort );
end;

function tWinsockAPI.recv ( s: TSock; var buf; len, flags: Integer ): Integer;
begin
  if not assigned(Frecv) then
    @Frecv := NeedDLL('recv');
  Result := Frecv( s, buf, len, flags );
end;

function tWinsockAPI.recvfrom ( s: TSock; var buf; len, flags: Integer;
         var from: Sockaddr; var fromlen: Integer ): Integer;
begin
  if not assigned(Frecvfrom) then
    @Frecvfrom := NeedDLL('recvfrom');
  Result := Frecvfrom( s, buf, len, flags, from, fromlen );
end;

function tWinsockAPI.select ( nfds: Integer; var readfds, writefds, exceptfds: fd_set;
         timeout: ptimeval ): longint;
begin
  if not assigned(Fselect) then
    @Fselect := NeedDLL('select');
  Result := Fselect( nfds, readfds, writefds, exceptfds, timeout );
end;

function tWinsockAPI.send ( s: TSock; var buf; len, flags: Integer ): Integer;
begin
  if not assigned(Fsend) then
    @Fsend := NeedDLL('send');
  Result := Fsend( s, buf, len, flags );
end;

function tWinsockAPI.sendto ( s: TSock; var buf; len, flags: Integer;
         var addrto: Sockaddr; tolen: Integer ): Integer;
begin
  if not assigned(Fsendto) then
    @Fsendto := NeedDLL('sendto');
  Result := Fsendto( s, buf, len, flags, addrto, tolen );
end;

function tWinsockAPI.setsockopt ( s: TSock; level, optname: Integer; optval: PChar;
         optlen: Integer ): Integer;
begin
  if not assigned(Fsetsockopt) then
    @Fsetsockopt := NeedDLL('setsockopt');
  Result := Fsetsockopt( s, level, optname, optval, optlen );
end;

function tWinsockAPI.shutdown ( s: TSock; how: Integer ): Integer;
begin
  if not assigned(Fshutdown) then
    @Fshutdown := NeedDLL('shutdown');
  Result := Fshutdown( s, how );
end;

function tWinsockAPI.socket ( af, struc, protocol: Integer ): TSock;
begin
  if not assigned(Fsocket) then
    @Fsocket := NeedDLL('socket');
  Result := Fsocket( af, struc, protocol );
end;

function tWinsockAPI.gethostbyaddr ( addr: Pointer; len, struc: Integer ): PHostEnt;
begin
  if not assigned(Fgethostbyaddr) then
    @Fgethostbyaddr := NeedDLL('gethostbyaddr');
  Result := Fgethostbyaddr( addr, len, struc );
end;

function tWinsockAPI.gethostbyname ( name: PChar ): PHostEnt;
begin
  if not assigned(Fgethostbyname) then
    @Fgethostbyname := NeedDLL('gethostbyname');
  Result := Fgethostbyname( name );
end;

function tWinsockAPI.gethostname ( name: PChar; len: Integer ): Integer;
begin
  if not assigned(Fgethostname) then
    @Fgethostname := NeedDLL('gethostname');
  Result := Fgethostname( name, len );
end;

function tWinsockAPI.getservbyport ( port: u_int; proto: PChar ): PServEnt;
begin
  if not assigned(Fgetservbyport) then
    @Fgetservbyport := NeedDLL('getservbyport');
  Result := Fgetservbyport( port, proto );
end;

function tWinsockAPI.getservbyname ( name, proto: PChar ): PServEnt;
begin
  if not assigned(Fgetservbyname) then
    @Fgetservbyname := NeedDLL('getservbyname');
  Result := Fgetservbyname( name, proto );
end;

function tWinsockAPI.getprotobynumber ( proto: Integer ): PProtoEnt;
begin
  if not assigned(Fgetprotobynumber) then
    @Fgetprotobynumber := NeedDLL('getprotobynumber');
  Result := Fgetprotobynumber( proto );
end;

function tWinsockAPI.getprotobyname ( name: PChar ): PProtoEnt;
begin
  if not assigned(Fgetprotobyname) then
    @Fgetprotobyname := NeedDLL('getprotobyname');
  Result := Fgetprotobyname( name );
end;

function tWinsockAPI.WSAStartup ( wVersionRequired: word; var WSData: WSADATA ): Integer;
begin
  if not assigned(FWSAStartup) then
    @FWSAStartup := NeedDLL('WSAStartup');
  Result := FWSAStartup( wVersionRequired, WSData );
  FStarted := Result = 0; { Did we start sucessfully? }

  FMaxSockets   := WSData.iMaxSockets;
  FMaxUdpDg     := WSData.iMaxUdpDg;
  FVersion      := WSData.wVersion;
  FHighVersion  := WSData.wHighVersion;
  FVendorInfo   := WSData.lpVendorInfo;
  NewpChar(FDescription,  WSData.szDescription);
  NewpChar(FSystemStatus, WSData.szSystemStatus);

end;

function tWinsockAPI.WSACleanup : Integer;
begin
  if not assigned(FWSACleanup) then
    @FWSACleanup := NeedDLL('WSACleanup');
  Result := FWSACleanup;
  FStarted := Not (Result = 0); { We're not started anymore }
end;

procedure tWinsockAPI.WSASetLastError ( iError: Integer );
begin
  if not assigned(FWSASetLastError) then
    @FWSASetLastError := NeedDLL('WSASetLastError');
  FWSASetLastError( iError );
end;

function tWinsockAPI.WSAGetLastError : Integer;
begin
  if not assigned(FWSAGetLastError) then
    @FWSAGetLastError := NeedDLL('WSAGetLastError');
  Result := FWSAGetLastError;
end;

function tWinsockAPI.WSAIsBlocking : BOOL;
begin
  if not assigned(FWSAIsBlocking) then
    @FWSAIsBlocking := NeedDLL('WSAIsBlocking');
  Result := FWSAIsBlocking;
end;

function tWinsockAPI.WSAUnhookBlockingHook : Integer;
begin
  if not assigned(FWSAUnhookBlockingHook) then
    @FWSAUnhookBlockingHook := NeedDLL('WSAUnhookBlockingHook');
  Result := FWSAUnhookBlockingHook;
end;

function tWinsockAPI.WSASetBlockingHook ( lpBlockFunc: TFarProc ): TFarProc;
begin
  if not assigned(FWSASetBlockingHook) then
    @FWSASetBlockingHook := NeedDLL('WSASetBlockingHook');
  Result := FWSASetBlockingHook( lpBlockFunc );
end;

function tWinsockAPI.WSACancelBlockingCall : Integer;
begin
  if not assigned(FWSACancelBlockingCall) then
    @FWSACancelBlockingCall := NeedDLL('WSACancelBlockingCall');
  Result := FWSACancelBlockingCall;
end;

function tWinsockAPI.WSAAsyncGetServByName ( hWin: HWND; wMsg: u_int; name, proto: PChar;
         buf: PAsyncBuf; buflen: Integer ): THandle;
begin
  if not assigned(FWSAAsyncGetServByName) then
    @FWSAAsyncGetServByName := NeedDLL('WSAAsyncGetServByName');
  Result := FWSAAsyncGetServByName( hWin, wMsg, name, proto, buf, buflen );
end;

function tWinsockAPI.WSAAsyncGetServByPort ( hWin: HWND; wMsg, port: u_int; proto: PChar;
         buf: PAsyncBuf; buflen: Integer ): THandle;
begin
  if not assigned(FWSAAsyncGetServByPort) then
    @FWSAAsyncGetServByPort := NeedDLL('WSAAsyncGetServByPort');
  Result := FWSAAsyncGetServByPort( hWin, wMsg, port, proto, buf, buflen );
end;

function tWinsockAPI.WSAAsyncGetProtoByName ( hWin: HWND; wMsg: u_int; name: PChar;
         buf: PAsyncBuf; buflen: Integer ): THandle;
begin
  if not assigned(FWSAAsyncGetProtoByName) then
    @FWSAAsyncGetProtoByName := NeedDLL('WSAAsyncGetProtoByName');
  Result := FWSAAsyncGetProtoByName( hWin, wMsg, name, buf, buflen );
end;

function tWinsockAPI.WSAAsyncGetProtoByNumber ( hWin: HWND; wMsg: u_int;
         number: Integer; buf: PAsyncBuf;
         buflen: Integer ): THandle;
begin
  if not assigned(FWSAAsyncGetProtoByNumber) then
    @FWSAAsyncGetProtoByNumber := NeedDLL('WSAAsyncGetProtoByNumber');
  Result := FWSAAsyncGetProtoByNumber( hWin, wMsg, Number, buf, buflen );
end;

function tWinsockAPI.WSAAsyncGetHostByName ( hWin: HWND; wMsg: u_int; name: PChar;
         buf: PAsyncBuf; buflen: Integer ): THandle;
begin
  if not assigned(FWSAAsyncGetHostByName) then
    @FWSAAsyncGetHostByName := NeedDLL('WSAAsyncGetHostByName');
  Result := FWSAAsyncGetHostByName( hWin, wMsg, name, buf, buflen );
end;

function tWinsockAPI.WSAAsyncGetHostByAddr ( hWin: HWND; wMsg: u_int; addr: PChar;
         len, struc: Integer; buf: PAsyncBuf;
         buflen: Integer ): THandle;
begin
  if not assigned(FWSAAsyncGetHostByAddr) then
    @FWSAAsyncGetHostByAddr := NeedDLL('WSAAsyncGetHostByAddr');
  Result := FWSAAsyncGetHostByAddr( hWin, wMsg, addr, len, struc, buf, buflen );
end;

function tWinsockAPI.WSACancelAsyncRequest ( hAsyncTaskHandle: THandle ): Integer;
begin
  if not assigned(FWSACancelAsyncRequest) then
    @FWSACancelAsyncRequest := NeedDLL('WSACancelAsyncRequest');
  Result := FWSACancelAsyncRequest( hAsyncTaskHandle );
end;

function tWinsockAPI.WSAAsyncSelect ( s: TSock; hWin: HWND; wMsg: u_int;
         lEvent: longint ): Integer;
begin
  if not assigned(FWSAAsyncSelect) then
    @FWSAAsyncSelect := NeedDLL('WSAAsyncSelect');
  Result := FWSAAsyncSelect( s, hWin, wMsg, lEvent );
end;

end.

