{$J+}  { Use writable typed constants }
{$A-}
unit Dgram;
interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, StdCtrls, dWinsock, ExtCtrls, Buttons, Spin, Grids;

type
  TForm1 = class(TForm)
    Label1: TLabel;
    UDPLabel: TLabel;
    FromLabel: TLabel;
    Label4: TLabel;
    PortLabel: TLabel;
    Label6: TLabel;
    Grid: TStringGrid;
    Label7: TLabel;
    Bevel1: TBevel;
    SendOnCheckBox: TCheckBox;
    GroupBox1: TGroupBox;
    Label3: TLabel;
    IntervalSpinEdit: TSpinEdit;
    Label5: TLabel;
    HostEdit: TEdit;
    SingleHostRadioButton: TRadioButton;
    BroadcastRadioButton: TRadioButton;
    BitBtn1: TBitBtn;
    ReceiveOnCheckBox: TCheckBox;
    Timer1: TTimer;
    Label8: TLabel;
    PacketsReceivedLabel: TLabel;
    Label9: TLabel;
    PacketsSentLabel: TLabel;
    BroadcastAddressEdit: TEdit;
    ConnectCheckBox: TCheckBox;
    HelpButton: TBitBtn;
    ClientSocket1: TBasicClientSocket;
    ServerSocket1: TBasicServerSocket;
    procedure ClientSocket1Connect(Sender: TObject; Socket: TSocketBase);
    procedure ServerSocket1Read(Sender: TObject; Socket: TSocketBase);
    procedure FormCreate(Sender: TObject);
    procedure ServerSocket1Accept(Sender: TObject; Socket: TSocketBase);
    procedure ServerSocket1Info(Sender: TObject; icode: TSockInfo;
      Socket: TSocketBase);
    procedure Timer1Timer(Sender: TObject);
    procedure ReceiveOnCheckBoxClick(Sender: TObject);
    procedure SendOnCheckBoxClick(Sender: TObject);
    procedure IntervalSpinEditChange(Sender: TObject);
    procedure ConnectCheckBoxClick(Sender: TObject);
    procedure HelpButtonClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    procedure SendPrices;
    procedure ToggleReceive;
    procedure ToggleSend;
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}
uses WSApi;

const
  DEMO_PORT = '12000';

type
  TSharesRecord = record
     PacketId : Longint;
     Names : array[0..4] of array[0..4] of Char; { use Pchars not pascal strings }
     Prices : array[0..4] of Longint;  { not using reals because of byte orders }
  end;

var
  OldPrices : TSharesRecord;

procedure TForm1.ClientSocket1Connect(Sender: TObject;
  Socket: TSocketBase);
begin
{ if Clientsocket1.Connected then
    with Socket as TStreamSocket do Text := ('Hello!');   }
end;

procedure TForm1.ServerSocket1Read(Sender: TObject; Socket: TSocketBase);
const
  num_datagrams_so_far : Longint = 0;
var
  newprices : TSharesRecord;
  sz, fromlen : Integer;
  fromsa : sockaddr_in;
  fromname : string;
  fromport : u_int;
  C : Integer;
begin
  fromlen := SizeOf( fromsa );
  sz := (Socket as TDatagramSocket).RecvFrom( newprices, SizeOf(newprices),
                                              sockaddr(fromsa), fromlen);
  Grid.Cells[0, 0] := 'Packet ID';
  Grid.Cells[0,1] := IntToStr(newprices.PacketId);
  for C :=  0 to 4 do begin
    Grid.Cells[C+1, 0] := StrPas(newprices.Names[C]);
    Grid.Cells[C+1, 1] := IntToStr(newprices.Prices[C]);
  end;

  if sz = SOCKET_ERROR then begin
    ShowMessage( 'An error occurred calling recvfrom' );
    Exit;
  end;

  WinsockAPI.CrackSockAddr_In( fromsa, fromname, fromport );

  FromLabel.Caption := fromname;
  PortLabel.Caption := IntToStr(fromport);

  Inc( num_datagrams_so_far );
  PacketsReceivedLabel.Caption := IntToStr( num_datagrams_so_far );
end;

procedure TForm1.FormCreate(Sender: TObject);
var
   I : Integer;
begin
  Randomize;
  { Start timer at a sensible interval. }
  Timer1.Interval := IntervalSpinEdit.Value*1000;

  ServerSocket1.Port := DEMO_PORT;
  ServerSocket1.Listen( TDatagramSocket );

  ClientSocket1.SetSocketClass( TDatagramSocket);

  UDPLabel.Caption := IntToStr(ClientSocket1.MaxDatagramSize);

  { Seed prices with 'sensible' values }
  for I := 0 to 4 do
    OldPrices.Prices[I] := Random(400)+400;

  Timer1.Enabled := True;
end;

procedure TForm1.ServerSocket1Accept(Sender: TObject; Socket: TSocketBase);
begin
  { this code never happens because datagram server's don't
    accept connections... it's just here to prove it to you.  }
  ShowMessage('Accept');
end;

procedure TForm1.ServerSocket1Info(Sender: TObject; icode: TSockInfo;
  Socket: TSocketBase);
begin
{InfoLabel.Caption := SockInfoToString( icode );}
end;

procedure TForm1.SendPrices;
const
  num_datagrams_sent : Longint = 0;
  FluctuationRange = 20;
var
  PricestoSend : TSharesRecord;
  I : Integer;
  Delta, Rnd : Longint;

  procedure Limit( var price : Longint );
  begin
    if price < 0 then price := 0;
    if price > 2000 then price := 2000; { Stop the prices getting stupidly high }
  end;

begin
{ This is REALLY crap :)   }
  with PricesToSend do begin
    StrPCopy(Names[0], 'BORL');
    StrPCopy(Names[1], 'MSFT');
    StrPCopy(Names[2], 'DWS');
    StrPCopy(Names[3], 'ULFS');
    StrPCopy(Names[4], 'BT');
  end;

  { do a silly random price generator that adds
    +/- half of the fluctuation range }
  for I := 0 to 4 do begin
    PricesToSend.Prices[I] := OldPrices.Prices[I];
    Rnd := Random( FluctuationRange);
    Delta := Longint( Rnd-(FluctuationRange div 2) );
    Inc( PricesToSend.Prices[I], Delta);
    Limit( PricesToSend.Prices[I] );
  end;

  PricesToSend.PacketId := num_datagrams_sent;

  OldPrices := PricesToSend;

  if SingleHostRadioButton.Checked then begin
    with ClientSocket1.Conn as TDatagramSocket do begin
      Broadcast := False;
      if ConnectCheckBox.Checked then
        Send( PricesToSend, SizeOf(PricesToSend))
      else
        SendTo( HostEdit.Text, PricesToSend, SizeOf(PricesToSend));
    end;
  end else begin
    { do broadcast stuff }
    with ClientSocket1.Conn as TDatagramSocket do begin
      Broadcast := True;
      SendTo( BroadcastAddressEdit.Text, PricesToSend, SizeOf(PricesToSend));
    end;

  end;

  Inc(num_datagrams_sent);
  PacketsSentLabel.Caption := IntToStr(num_datagrams_sent);
end;

procedure TForm1.ToggleSend;
begin
  Timer1.Enabled := not Timer1.Enabled;
end;

procedure TForm1.ToggleReceive;
const
  Receiving : Boolean = True;
begin
  if Receiving then begin
    ServerSocket1.Close;
    Receiving := False;
  end else begin
    ServerSocket1.Listen( TDatagramSocket );
    Receiving := True;
  end;
end;

procedure TForm1.Timer1Timer(Sender: TObject);
begin
  SendPrices;
end;

procedure TForm1.ReceiveOnCheckBoxClick(Sender: TObject);
begin
  ToggleReceive;
end;

procedure TForm1.SendOnCheckBoxClick(Sender: TObject);
begin
  ToggleSend;
end;

procedure TForm1.IntervalSpinEditChange(Sender: TObject);
begin
{ Make sure user hasn't put invalid value into the Spin Edit }
  if StrToIntDef(IntervalSpinEdit.Text, -1) <> -1 then
    Timer1.Interval := IntervalSpinEdit.Value * 1000;
end;

procedure TForm1.ConnectCheckBoxClick(Sender: TObject);
begin
  Timer1.Enabled := False;

  if ConnectCheckBox.Checked then begin
    with ClientSocket1 do begin
      Address := HostEdit.Text;
      Port := DEMO_PORT;
      Open( TDatagramSocket );
    end;
  end else Clientsocket1.Close;

  Timer1.Enabled := True;
end;

procedure TForm1.HelpButtonClick(Sender: TObject);
begin
  Application.HelpContext( 1000 );
end;

end.
