unit nwck32dp;

{******************************************************************************}
{ Delphi 32-bit NWCK32DP.DLL Interface                                         }
{                                                                              }
{ Author       : Sean Stromberg (Imagicom)                                     }
{ Last Modified: March 16, 1996                                                }
{                                                                              }
{ See the NWCK32DP.DLL help for information on the routines.                   }
{                                                                              }
{ Copyright 1996 - Sapient Technology, Inc.                                    }
{******************************************************************************}

interface

uses
  SysUtils, WinTypes, Dialogs;

type
  TNWCKVersion = record
    MinorVersion, MajorVersion : Byte;
  end;

var
  WinSockVersion : Word;

const
  NWCK_DLL_NAME  = 'NWCK32DP.DLL';

{ General Purpose }
procedure NWCKInit; pascal;
function  NWCKVersion : DWORD; pascal;

{ Network Browsing }
function  NWCKGetServerNameDlg(Parent : HWND; Buffer : PChar) : integer; pascal;

{ WinSock API - Administrative }
function  NWCKStartWinSock(minVersion : Word) : Boolean; pascal;
function  NWCKEndWinSock : integer; pascal;
function  NWCKGetWinSockInfo(var Version : Word; Description, Status : PChar;
                             var MaxSock, MaxDGram : ShortInt; Vendor : PChar) : integer; pascal;
function  NWCKGetMyName(Name, IPAddress : PChar) : integer; pascal;
function  NWCKGetServerIP(Name, IPAddress : PChar) : integer; pascal;
function  NWCKLastWinSockError(var Error : Integer; ErrorString : PChar) : integer; pascal;

{ WinSock API - DataGram }
function  NWCKCreateUDPServer1(LocalPort : Integer; WindowHandle : HWND;
                               Message : UINT; var Handle : LongInt) : integer; pascal;
function  NWCKCreateUDPServer2(LocalPort : Integer; Thread : DWORD;
                               Message : UINT; var Handle : LongInt) : integer; pascal;
function  NWCKCreateUDPClient1(WindowHandle : HWND; Message : UINT;
                               var Handle : LongInt) : integer; pascal;
function  NWCKCreateUDPClient2(WindowHandle : HWND; Message : UINT;
                               var Handle : LongInt) : integer; pascal;
function  NWCKSetUDPBufferSize(Handle : LongInt; BufferSize : UINT) : integer; pascal;
function  NWCKUDPWriteToPort(Handle : LongInt; Data : Pointer; DataLength : integer;
                             RemoteName : PChar; RemotePort : integer) : integer; pascal;
function  NWCKUDPWriteToService(Handle : LongInt; Data : Pointer; DataLength : integer;
                                RemoteName, RemoteService : PChar) : integer; pascal;
function  NWCKUDPRead(Handle : LongInt; Data : Pointer; DataLength : integer;
                      ReplyAddress : PChar) : integer; pascal;
function  NWCKUDPReply(Handle : LongInt; Data : Pointer; DataLength : integer;
                       ReplyAddress : PChar) : integer; pascal;
function  NWCKLastUDPWinSockError(Handle : LongInt; var Error : integer;
                                  ErrorString : PChar) : integer; pascal;
function  NWCKCloseUDP(Handle : LongInt) : integer; pascal;

{ WinSock API - Streams }
const
  NWCK_STREAM_RECVDATA        =     15005;
  NWCK_STREAM_DATASENT        =     15006;
  NWCK_STREAM_WRITEERROR      =     15007;
  NWCK_STREAM_READERROR       =     15008;
  NWCK_STREAM_CONNECT         =     15009;
  NWCK_STREAM_ACCEPT          =     15010;
  NWCK_STREAM_SHUTDOWN        =     15011;

  NWCK_STREAM_WINSOCK_NOERROR =     0;
  NWCK_WIN32_ERROR            =     1;
  NWCK_WIN32_MEM_ERROR        =     2;
  NWCK_WIN32_PRG_ERROR        =     3;
  NWCK_STREAM_NOTSERVER       =     4;

function NWCKCreateStreamServerWindow(LocalPort : integer; WindowHandle : HWND;
                                      Message : UINT; var Handle : LongInt) : integer; pascal;
function NWCKCreateStreamServerThread(LocalPort : integer; ThreadID, Message : UINT;
                                      var Handle : LongInt) : integer; pascal;
function NWCKCreateStreamCommWindow(WindowHandle : HWND; Message : UINT;
                                    var Handle : LongInt) : integer; pascal;
function NWCKCreateStreamCommThread(ThreadID, Message : UINT;
                                    var Handle : LongInt) : integer; pascal;
function NWCKStreamAccept(OriginalHandle, CommHandle : LongInt) : integer; pascal;
function NWCKCreateStreamClientWindow(WindowHandle : HWND; Message : UINT;
                                      var Handle : LongInt) : integer; pascal;
function NWCKCreateStreamClientThread(ThreadID, Message : UINT; var Handle : LongInt) : integer; pascal;
function NWCKStreamConnectToPort(Handle : LongInt; RemoteName : LPSTR; RemotePort : integer) : integer; pascal;
function NWCKStreamConnectToService(Handle : LongInt; RemoteName, RemoteService : LPSTR) : integer; pascal;
function NWCKStreamWrite(Handle : LongInt; Data : Pointer; DataLength : integer) : integer; pascal;
function NWCKStreamRead(Handle : LongInt; Data : Pointer; DataLength : integer) : integer; pascal;
function NWCKSetStreamBufferSize(Handle : LongInt; BufferSize : UINT) : integer; pascal;
function NWCKLastStreamWinSockError(Handle : LongInt; var Error : integer; ErrorString : PChar) : integer; pascal;
function NWCKCloseStream(Handle : LongInt) : integer; pascal;

{ Internet API - Sending Mail }
function NWCKMIMEPrepStart(OutFile, From : PChar; var Handle : LongInt) : integer; pascal;
function NWCKMIMEPrepNewMesg(Handle : LongInt) : integer; pascal;
function NWCKMIMEPrepTO(Handle : LongInt; ToName : PChar) : integer; pascal;
function NWCKMIMEPrepCC(Handle : LongInt; CCName : PChar) : integer; pascal;
function NWCKMIMEPrepSubj(Handle : LongInt; Subject : PChar) : integer; pascal;
function NWCKMIMEPrepFile(Handle : LongInt; FileAttachment : PChar) : integer; pascal;
function NWCKMIMEPrepData(Handle : LongInt; Body : Pointer; DataLength : integer) : integer; pascal;
function NWCKMIMEPrepEndMesg(Handle : LongInt) : integer; pascal;
function NWCKMIMEPrepFinish(Handle : LongInt) : integer; pascal;

const
  SMTP_SHUTDOWN		 =    0;	// Just informs you that mail services have ended
  SMTP_MEMORY_ERR	 =    1;	// Low memory problems
  SMTP_READ_ERR		 =    2;	// Winsock read error
  SMTP_WRITE_ERR	 =    3;	// Winsock write error
  SMTP_NOCONN_ERR	 =    4;	// Not able to connect to SMTP server
  SMTP_RESPONSE_ERR	 =    5;	// SMTP returned an ERROR as a response
  SMTP_FILEOPEN_ERR	 =    6;	// General File I/O unable to open file
  SMTP_FILEREAD_ERR	 =    7;	// General File I/O read error
  SMTP_NOMSG_ERR	 =    8;	// There were no messages to send
  SMTP_WINCREAT_ERR	 =    9;	// Unable to create hidden window
  SMTP_NOSERVNAM_ERR	 =    10;	// Not a valid SMTP Server Name
  SMTP_SOCKCREAT_ERR	 =    11;	// Unable to create a stream socket
  SMTP_SERVDISCON_ERR	 =    12;	// Server disconnected us
  SMTP_NOSENDER_ERR	 =    13;	// Sender's E-Mail address not in file
  SMTP_FILEFORMAT_ERR	 =    14;	// The send file(MIME created) format is bad
  SMTP_MAILSENT_NUM	 =    15;	// The LPARAM is the message # just sent OK.
  SMTP_MAILRCPT_ERR	 =    16;	// There was a bad Recipient Name. The LPARAM = message#
  SMTP_CONNECT_OK	 =    17;	// Connected to SMTP Server - OK!
  SMTP_MAILSENT_OK	 =    18;	// All Mail was sent successfully

  function NWCKSMTPStartWindow(MyWindow : HWND; MyMessage : UINT; SendFile : PChar;
                               var Handle : LongInt) : integer; pascal;
  function NWCKSMTPStartThread(MyThread : UINT; MyMessage : UINT; SendFile : PChar;
                               var Handle : LongInt) : integer; pascal;
  function NWCKSMTPSendMail(Handle : LongInt; SMTPServerName : PChar) : integer; pascal;
  function NWCKSMTPClose(Handle : LongInt) : integer; pascal;

{ Internet API - Receiving Mail }
type
  TNWCKMIMEMSG = record
    MessageID       : array[1..64] of char;
    DateString      : array[1..40] of char;
    FromString      : array[1..80] of char;
    SubjectString   : array[1..256] of char;
    ToList          : array[1..32] of PChar;
    ToNumber        : integer;
    CCList          : array[1..32] of PChar;
    CCNumber        : integer;
    FileAttachments : array[1..8] of PChar;
    FileNumber      : integer;
    DataString      : PChar;
    DataSize        : integer;
  end;

  PNWCKMIMEMSG = ^TNWCKMIMEMSG;

const
  POP3_SHUTDOWN          =    0;
  POP3_MEMORY_ERR        =    1;
  POP3_READ_ERR          =    2;
  POP3_WRITE_ERR         =    3;
  POP3_NOCONN_ERR        =    4;
  POP3_RESPONSE_ERR      =    5;
  POP3_FILEOPEN_ERR      =    6;
  POP3_FILEREAD_ERR      =    7;
  POP3_FILEWRIT_ERR	 =    8;
  POP3_NOMSG_ERR	 =    9;
  POP3_WINCREAT_ERR	 =    10;
  POP3_NOSERVNAM_ERR	 =    11;
  POP3_SOCKCREAT_ERR	 =    12;
  POP3_SERVDISCON_ERR	 =    13;
  POP3_NOUSERNAM_ERR	 =    14;
  POP3_NOPWORD_ERR	 =    15;
  POP3_RTRVNG_MSG	 =    16;
  POP3_CONNECT_OK	 =    17;
  POP3_TOTALMAIL_MSG	 =    18;
  POP3_MAILRTRV_OK	 =    19;

function NWCKPOP3StartWindow(MyWindow : HWND; MyMessage : UINT; OutFile : PChar;
                             var Handle : LongInt) : integer; pascal;
function NWCKPOP3StartThread(MyThread, MyMessage : UINT; OutFile : PChar;
                             var Handle : LongInt) : integer; pascal;
function NWCKPOP3ISThereMail(Handle : LongInt; PopServer, User,
                             PassWord : PChar) : integer; pascal;
function NWCKPOP3PullMail(Handle : LongInt; PopServer, User, Password : PChar;
                          DeleteMessage : Boolean) : integer; pascal;
function NWCKPOP3NumMail(Handle : LongInt; var NumberMessages : integer) : integer; pascal;
function NWCKPOP3Close(Handle : LongInt) : integer; pascal;

function NWCKMIMEScanStart(InFile, Directory : PChar; var Handle : LongInt) : integer; pascal;
function NWCKMIMEScanNumMessages(Handle : LongInt; var NumberOfMessages : integer) : integer; pascal;
function NWCKMIMEScanGetFirst(Handle : LongInt) : PNWCKMIMEMSG; pascal;
function NWCKMIMEScanGetNext(Handle : LongInt) : PNWCKMIMEMSG; pascal;
function NWCKMIMEScanClose(Handle : LongInt) : integer; pascal;

implementation

{ General Purpose }
procedure NWCKInit;                    external NWCK_DLL_NAME name 'NWCKINIT';
function  NWCKVersion;                 external NWCK_DLL_NAME name 'NWCKVERSION';

{ Network Browsing }
function  NWCKGetServerNameDlg;        external NWCK_DLL_NAME name 'NWCKGETSERVERNAMEDLG';

{ WinSock API - Administrative }
function  NWCKStartWinSock;            external NWCK_DLL_NAME name 'NWCKSTARTWINSOCK';
function  NWCKEndWinSock;              external NWCK_DLL_NAME name 'NWCKENDWINSOCK';
function  NWCKGetWinSockInfo;          external NWCK_DLL_NAME name 'NWCKGETWINSOCKINFO';
function  NWCKGetMyName;               external NWCK_DLL_NAME name 'NWCKGETMYNAME';
function  NWCKGetServerIP;             external NWCK_DLL_NAME name 'NWCKGETSERVERIP';
function  NWCKLastWinSockError;        external NWCK_DLL_NAME name 'NWCKLASTWINSOCKERROR';

{ WinSock API - DataGram }
function  NWCKCreateUDPServer1;        external NWCK_DLL_NAME name 'NWCKCREATEUDPSERVER1';
function  NWCKCreateUDPServer2;        external NWCK_DLL_NAME name 'NWCKCREATEUDPSERVER2';
function  NWCKCreateUDPClient1;        external NWCK_DLL_NAME name 'NWCKCREATEUDPCLIENT1';
function  NWCKCreateUDPClient2;        external NWCK_DLL_NAME name 'NWCKCREATEUDPCLIENT2';
function  NWCKSetUDPBufferSize;        external NWCK_DLL_NAME name 'NWCKSETUDPBUFFERSIZE';
function  NWCKUDPWriteToPort;          external NWCK_DLL_NAME name 'NWCKUDPWRITETOPORT';
function  NWCKUDPWriteToService;       external NWCK_DLL_NAME name 'NWCKUDPWRITETOSERVICE';
function  NWCKUDPRead;                 external NWCK_DLL_NAME name 'NWCKUDPREAD';
function  NWCKUDPReply;                external NWCK_DLL_NAME name 'NWCKUDPREPLY';
function  NWCKLastUDPWinSockError;     external NWCK_DLL_NAME name 'NWCKLASTUDPWINSOCKERROR';
function  NWCKCloseUDP;                external NWCK_DLL_NAME name 'NWCKCLOSEUDP';

{ WinSock API - Streams }
function NWCKCreateStreamServerWindow; external NWCK_DLL_NAME name 'NWCKCREATESTREAMSERVERWINDOW';
function NWCKCreateStreamServerThread; external NWCK_DLL_NAME name 'NWCKCREATESTREAMSERVERTHREAD';
function NWCKCreateStreamCommWindow;   external NWCK_DLL_NAME name 'NWCKCREATESTREAMCOMMWINDOW';
function NWCKCreateStreamCommThread;   external NWCK_DLL_NAME name 'NWCKCREATESTREAMCOMMTHREAD';
function NWCKStreamAccept;             external NWCK_DLL_NAME name 'NWCKSTREAMACCEPT';
function NWCKCreateStreamClientWindow; external NWCK_DLL_NAME name 'NWCKCREATESTREAMCLIENTWINDOW';
function NWCKCreateStreamClientThread; external NWCK_DLL_NAME name 'NWCKCREATESTREAMCLIENTTHREAD';
function NWCKStreamConnectToPort;      external NWCK_DLL_NAME name 'NWCKSTREAMCONNECTTOPORT';
function NWCKStreamConnectToService;   external NWCK_DLL_NAME name 'NWCKSTREAMCONNECTTOSERVICE';
function NWCKStreamWrite;              external NWCK_DLL_NAME name 'NWCKSTREAMWRITE';
function NWCKStreamRead;               external NWCK_DLL_NAME name 'NWCKSTREAMREAD';
function NWCKSetStreamBufferSize;      external NWCK_DLL_NAME name 'NWCKSETSTREAMBUFFERSIZE';
function NWCKLastStreamWinSockError;   external NWCK_DLL_NAME name 'NWCKLASTSTREAMWINSOCKERROR';
function NWCKCloseStream;              external NWCK_DLL_NAME name 'NWCKCLOSESTREAM';

{ Internet API - Sending Mail }
function NWCKMIMEPrepStart;            external NWCK_DLL_NAME name 'NWCKMIMEPREPSTART';
function NWCKMIMEPrepNewMesg;          external NWCK_DLL_NAME name 'NWCKMIMEPREPNEWMESG';
function NWCKMIMEPrepTO;               external NWCK_DLL_NAME name 'NWCKMIMEPREPTO';
function NWCKMIMEPrepCC;               external NWCK_DLL_NAME name 'NWCKMIMEPREPCC';
function NWCKMIMEPrepSubj;             external NWCK_DLL_NAME name 'NWCKMIMEPREPSUBJ';
function NWCKMIMEPrepFile;             external NWCK_DLL_NAME name 'NWCKMIMEPREPFILE';
function NWCKMIMEPrepData;             external NWCK_DLL_NAME name 'NWCKMIMEPREPDATA';
function NWCKMIMEPrepEndMesg;          external NWCK_DLL_NAME name 'NWCKMIMEPREPENDMESG';
function NWCKMIMEPrepFinish;           external NWCK_DLL_NAME name 'NWCKMIMEPREPFINISH';
function NWCKSMTPStartWindow;          external NWCK_DLL_NAME name 'NWCKSMTPSTARTWINDOW'
function NWCKSMTPStartThread;          external NWCK_DLL_NAME name 'NWCKSMTPSTARTTHREAD'
function NWCKSMTPSendMail;             external NWCK_DLL_NAME name 'NWCKSMTPSENDMAIL'
function NWCKSMTPClose;                external NWCK_DLL_NAME name 'NWCKSMTPCLOSE'

{ Internet API - Receiving Mail }
function NWCKPOP3StartWindow;          external NWCK_DLL_NAME name 'NWCKPOP3STARTWINDOW';
function NWCKPOP3StartThread;          external NWCK_DLL_NAME name 'NWCKPOP3STARTTHREAD';
function NWCKPOP3IsThereMail;          external NWCK_DLL_NAME name 'NWCKPOP3ISTHEREMAIL';
function NWCKPOP3PullMail;             external NWCK_DLL_NAME name 'NWCKPOP3PULLMAIL';
function NWCKPOP3NumMail;              external NWCK_DLL_NAME name 'NWCKPOP3NUMMAIL';
function NWCKPOP3Close;                external NWCK_DLL_NAME name 'NWCKPOP3CLOSE';

function NWCKMIMEScanStart;            external NWCK_DLL_NAME name 'NWCKMIMESCANSTART'
function NWCKMIMEScanNumMessages;      external NWCK_DLL_NAME name 'NWCKMIMESCANNUMMESSAGES'
function NWCKMIMEScanGetFirst;         external NWCK_DLL_NAME name 'NWCKMIMESCANGETFIRST'
function NWCKMIMEScanGetNext;          external NWCK_DLL_NAME name 'NWCKMIMESCANGETNEXT'
function NWCKMIMEScanClose;            external NWCK_DLL_NAME name 'NWCKMIMESCANCLOSE'

procedure NWCK32_Exit; far;
begin
  NWCKEndWinSock;
end;

begin
  WinSockVersion := MAKEWORD(1, 1);
  AddExitProc(NWCK32_Exit);
  NWCKInit;
  NWCKStartWinSock(WinSockVersion);
end.
